; a simple wrapper for booting Proverb
; This file should probably be about 100 lines.

(cond-expand
  (guile
    (debug-enable 'debug)
    ; allow the stack to grow to the system limit
    (debug-set! stack 0)
    ; allow position recording
    (read-enable 'positions)
    ; use Scheme errors instead of SIGPIPE
    (sigaction SIGPIPE SIG_IGN)
    ; get these modules loaded while the universe is still normal
    (resolve-interface '(ice-9 regex))
    (resolve-interface '(ice-9 session))
    (resolve-interface '(ice-9 format))
    (resolve-interface '(ice-9 rw))
    (resolve-interface '(ice-9 rdelim))
    (resolve-interface '(ice-9 syncase))
    (resolve-interface '(ice-9 documentation))
    (resolve-interface '(ice-9 optargs))
    (resolve-interface '(ice-9 pretty-print))
    ; these split the monolith of guile into digestible pieces
    (load "studio/guile-modules.scm")
    (load "studio/guile-networking.scm")
    (load "studio/guile-filesystem.scm")
    (load "studio/guile-errors.scm")
    ; these provide R6RS libraries fashioned from Guile
    (load "studio/guile-rnrs.scm")
    (use-modules (rnrs))
    (load "studio/guile-rnrs-arithmetic-bitwise.scm")

    ; let us provide support dir to other modules
    (load "studio/guile-bootstrap.scm")
    (define cld (dirname (port-filename (current-load-port))))
    (eval `(,set! support-dir ,(string-append cld "/support/")) (resolve-module '(proverb bootstrap)))

    ; grab Proverb
    (load "studio/util.scm")

    (use-modules (proverb dialect))
    ;(define cld (current-load-dir))
    ;(eval `(,set! support-dir ,(string-append cld "/support/")) (resolve-module '(proverb filesystem)))

    (append-load-path! cld)
    (define bogo-forward #f)
    (load "studio/main.scm")
    (use-modules (proverb main))

    ; load up the default contents of the evaluation environment
    (load "studio/hexdump.scm")
    (load "studio/interactive.scm")
  )
  (chicken
    (use syntax-case)
    (set! load load-relative)
    ; dialect must be loaded before any other studio code
    (load "studio/studio-dialect.scm")
    (load "studio/util.scm")
  )
  (chez
    (load "studio/studio-dialect.scm")
    (load "studio/util.scm")
  )
  (sisc
    (set! max-stack-trace-depth 32)
    (load "studio/sisc-rnrs.scm")
    (load "studio/sisc-rnrs-arithmetic-bitwise.scm")
    (import threading)
    (import string-io)
    (import custom-io)
    (import file-manipulation)
    (import hashtable)
    (load "studio/sisc-studio-dialect.scm")
    (define studio-string-split string-split)
    ;(library (bogo bogo)
    ;  (export bogo1)
    ;  (import bogo)
    ;  (define bogo1 123)
    ;)
    (display "SISC BAR:\n")
    (load "studio/util.scm")
    (load "studio/main.scm")
  )
  (else
    (display "UNSUPPORTED DIALECT!\n")
  )
)

; and exit. Our driver will call main
