#define _FILE_OFFSET_BITS 64

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/mman.h>

/* include the guile header files */
#include <guile/gh.h>

/* This appears necessary on the mac */
char **environ=NULL;

/* A macro for publishing guile bindings */
#define PUBLISH(scheme_name,c_name,args) { gh_define(scheme_name,gh_new_procedure(scheme_name,c_name,args,0,0)); gh_eval_str("(define-public " scheme_name " " scheme_name ")"); }

#if SCM_MINOR_VERSION > 6
#	define STRINGBUF_TAG           scm_tc7_stringbuf
#	define STRING_TAG            scm_tc7_string
#define STRINGBUF_OUTLINE_CHARS(buf)   ((char *)SCM_CELL_WORD_1(buf))
#define STRINGBUF_OUTLINE_LENGTH(buf)  (SCM_CELL_WORD_2(buf))
#endif

static SCM the_guardian;

SCM
guile_mmap_file_wrapper(SCM filename,SCM len,SCM prot,SCM flags,SCM offset)
{
char *c_filename;
int fd;
SCM scm_retval;
void *addr;
off_t size;

  c_filename=SCM_CHARS(filename);
  fd=open(c_filename,O_RDONLY);
  if(fd == -1) {
    return SCM_BOOL_F;
  }
  
  addr=NULL;
  size=lseek(fd,0,SEEK_END);
  if(size==0) {
    return gh_str2scm(NULL,0);
  }

  //addr=mmap(NULL,(size+4095)&~(size_t)0xfff,PROT_READ|PROT_WRITE,MAP_PRIVATE,fd,0);
  addr=mmap(NULL,(size+4095)&~(size_t)0xfff,PROT_READ,MAP_SHARED,fd,0);

  close(fd);

  if(addr==(void *)-1) {
    return SCM_BOOL_F;
  }

#if SCM_MINOR_VERSION > 6
  {
    SCM stringbuf;
    SCM res;
    stringbuf = scm_double_cell (STRINGBUF_TAG, (scm_t_bits) addr,
                              (scm_t_bits) size, (scm_t_bits) 0);

    /* In guile 1.8, we guard the stringbuf, not the string */
    gh_call1(the_guardian,stringbuf);

    res = scm_double_cell (STRING_TAG, SCM_UNPACK(stringbuf),
                           (scm_t_bits)0, (scm_t_bits) size);
    scm_retval=res;
  }
#else
  {
  void *freeaddr;
  // an allocation actually happens for null length strings.
  scm_retval=gh_str2scm(NULL,0);
  freeaddr=SCM_CHARS(scm_retval);
  SCM_SET_CELL_WORD_1(scm_retval,addr);
  SCM_SET_STRING_LENGTH(scm_retval,size);
  // go free scheme's string
  scm_must_free(freeaddr);
  gh_call1(the_guardian,scm_retval);
  }
#endif

  return scm_retval;
}

static SCM
post_gc_mmap_reaper() {
#if SCM_MINOR_VERSION > 6
  SCM cursor;
  while((cursor=gh_call0(the_guardian)) != SCM_BOOL_F) {
    munmap(STRINGBUF_OUTLINE_CHARS(cursor),STRINGBUF_OUTLINE_LENGTH(cursor));
    // fake the length
    SCM_SET_CELL_WORD_2(cursor,0);
    // whip up a fake string to satisfy scheme's reaper
    SCM_SET_CELL_WORD_1(cursor,scm_must_malloc(0,"bogus"));
  }
  return SCM_BOOL_T;
#else
  SCM cursor;
  while((cursor=gh_call0(the_guardian)) != SCM_BOOL_F) {
    munmap(SCM_CHARS(cursor),SCM_STRING_LENGTH(cursor));
    SCM_SET_STRING_LENGTH(cursor,0);
    // whip up a fake string to satisfy scheme's reaper
    SCM_SET_CELL_WORD_1(cursor,scm_must_malloc(0,"bogus"));
  }
  return SCM_BOOL_T;
#endif
}

void
mmap_init(void)
{
  /* Place your initializations here */
#if SCM_MINOR_VERSION > 6
  the_guardian=scm_make_guardian();
#else
  the_guardian=scm_make_guardian(SCM_BOOL_T);
#endif
  scm_gc_protect_object(the_guardian);

  scm_add_hook_x(gh_eval_str("after-gc-hook"),gh_new_procedure("post-gc-mmap-reaper",post_gc_mmap_reaper,0,0,0),SCM_BOOL_T);

  gh_new_procedure("mmap-file-with-flags",guile_mmap_file_wrapper,5,0,0);
  gh_eval_str("(define-public mmap-file-with-flags mmap-file-with-flags)");

  gh_define("PROT_READ",gh_int2scm(PROT_READ));
  gh_define("PROT_WRITE",gh_int2scm(PROT_WRITE));
  gh_define("PROT_EXEC",gh_int2scm(PROT_EXEC));
  gh_eval_str("(define-public PROT_READ PROT_READ)");
  gh_eval_str("(define-public PROT_WRITE PROT_WRITE)");
  gh_eval_str("(define-public PROT_EXEC PROT_EXEC)");

#ifdef MAP_ANON
  gh_define("MAP_ANON",gh_int2scm(MAP_ANON));
  gh_eval_str("(define-public MAP_ANON MAP_ANON)");
#endif
  gh_define("MAP_FILE",gh_int2scm(MAP_FILE)); 
  gh_eval_str("(define-public MAP_FILE MAP_FILE)");
  gh_define("MAP_PRIVATE",gh_int2scm(MAP_PRIVATE));
  gh_eval_str("(define-public MAP_PRIVATE MAP_PRIVATE)");
  gh_define("MAP_SHARED",gh_int2scm(MAP_SHARED));
  gh_eval_str("(define-public MAP_SHARED MAP_SHARED)");

  return;
}
