#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/fcntl.h>

#include <poll.h>

#include <openssl/ssl.h>

/* include the guile header files */
#include <guile/gh.h>

/* This appears necessary on the mac */
char **environ=NULL;

typedef SSL *TYPE_SSL;
typedef SSL_CTX *TYPE_CTX;
typedef SSL_SESSION *TYPE_SESSION;
typedef SSL_METHOD *TYPE_METHOD;
typedef X509 *TYPE_X509;
typedef X509_NAME *TYPE_X509_NAME;
typedef EVP_PKEY *TYPE_EVP_PKEY;
typedef EVP_MD *TYPE_EVP_MD;
typedef RSA *TYPE_RSA;
typedef ASN1_INTEGER *TYPE_ASN1_INTEGER;
typedef ASN1_TIME *TYPE_ASN1_TIME;
typedef int TYPE_INT;
typedef long TYPE_LONG;
typedef unsigned long TYPE_ULONG;
typedef char *TYPE_FILENAME;
typedef char *TYPE_STR0;

static scm_t_bits ssl_method_smob_tag;
static scm_t_bits ssl_cipher_smob_tag;
static scm_t_bits ssl_ctx_smob_tag;
static scm_t_bits ssl_session_smob_tag;
static scm_t_bits ssl_smob_tag;
static scm_t_bits x509_smob_tag;
static scm_t_bits x509_store_ctx_smob_tag;
static scm_t_bits x509_name_smob_tag;
static scm_t_bits evp_pkey_smob_tag;
static scm_t_bits evp_md_smob_tag;
static scm_t_bits rsa_smob_tag;
static scm_t_bits asn1_integer_smob_tag;
static scm_t_bits asn1_time_smob_tag;

static scm_t_bits sslport_tag;

#define ASSERT_CTX(val) if(!SCM_SMOB_PREDICATE(ssl_ctx_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_X509(val) if(!SCM_SMOB_PREDICATE(x509_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_X509_STORE_CTX(val) if(!SCM_SMOB_PREDICATE(x509_store_ctx_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_X509_NAME(val) if(!SCM_SMOB_PREDICATE(x509_name_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_METHOD(val) if(!SCM_SMOB_PREDICATE(ssl_method_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_SESSION(val) if(!SCM_SMOB_PREDICATE(ssl_session_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_SSL(val) if(!SCM_SMOB_PREDICATE(ssl_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_EVP_PKEY(val) if(!SCM_SMOB_PREDICATE(evp_pkey_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_EVP_MD(val) if(!SCM_SMOB_PREDICATE(evp_md_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_RSA(val) if(!SCM_SMOB_PREDICATE(rsa_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_ASN1_INTEGER(val) if(!SCM_SMOB_PREDICATE(asn1_integer_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_ASN1_TIME(val) if(!SCM_SMOB_PREDICATE(asn1_time_smob_tag,val)) { return SCM_BOOL_F; }
#define ASSERT_INT(val) ;
#define ASSERT_LONG(val) ;
#define ASSERT_ULONG(val) ;
#define ASSERT_FILENAME(val) ;
#define ASSERT_STR0(val) ;

#define XLATE_TO_CTX(val) (SSL_CTX *)SCM_SMOB_DATA(val);
#define XLATE_TO_X509(val) (X509 *)SCM_SMOB_DATA(val);
#define XLATE_TO_X509_STORE_CTX(val) (X509_STORE_CTX *)SCM_SMOB_DATA(val);
#define XLATE_TO_X509_NAME(val) (X509_NAME *)SCM_SMOB_DATA(val);
#define XLATE_TO_METHOD(val) (SSL_METHOD *)SCM_SMOB_DATA(val);
#define XLATE_TO_SESSION(val) (SSL_SESSION *)SCM_SMOB_DATA(val);
#define XLATE_TO_SSL(val) (SSL *)SCM_SMOB_DATA(val);
#define XLATE_TO_EVP_PKEY(val) (EVP_PKEY *)SCM_SMOB_DATA(val);
#define XLATE_TO_EVP_MD(val) (EVP_MD *)SCM_SMOB_DATA(val);
#define XLATE_TO_RSA(val) (RSA *)SCM_SMOB_DATA(val);
#define XLATE_TO_ASN1_INTEGER(val) (ASN1_INTEGER *)SCM_SMOB_DATA(val);
#define XLATE_TO_ASN1_TIME(val) (ASN1_TIME *)SCM_SMOB_DATA(val);
#define XLATE_TO_INT(val) (int)gh_scm2int(val);
#define XLATE_TO_LONG(val) (int)gh_scm2long(val);
#define XLATE_TO_ULONG(val) (unsigned long)gh_scm2ulong(val);
#define XLATE_TO_FILENAME(val) ((val==SCM_BOOL_F)?NULL:(char *)SCM_CHARS(val));
#define XLATE_TO_STR0(val) ((val==SCM_BOOL_F)?NULL:(char *)SCM_CHARS(val));

#define XLATE_FROM_CTX(val,cell) if(val) { SCM_NEWSMOB(cell,ssl_ctx_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_X509(val,cell) if(val) { SCM_NEWSMOB(cell,x509_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_X509_STORE_CTX(val,cell) if(val) { SCM_NEWSMOB(cell,x509_store_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_X509_NAME(val,cell) if(val) { SCM_NEWSMOB(cell,x509_name_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_METHOD(val,cell) if(val) { SCM_NEWSMOB(cell,ssl_method_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_SESSION(val,cell) if(val) { SCM_NEWSMOB(cell,ssl_session_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_SSL(val,cell) if(val) { SCM_NEWSMOB(cell,ssl_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_EVP_PKEY(val,cell) if(val) { SCM_NEWSMOB(cell,evp_pkey_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_EVP_MD(val,cell) if(val) { SCM_NEWSMOB(cell,evp_md_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_RSA(val,cell) if(val) { SCM_NEWSMOB(cell,rsa_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_ASN1_INTEGER(val,cell) if(val) { SCM_NEWSMOB(cell,asn1_integer_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_ASN1_TIME(val,cell) if(val) { SCM_NEWSMOB(cell,asn1_time_smob_tag,(void *)val); } else { cell=SCM_BOOL_F; }
#define XLATE_FROM_INT(val,cell) cell=gh_int2scm(val);
#define XLATE_FROM_LONG(val,cell) cell=gh_long2scm(val);


#define BOGO_WRAPPER_0(name,retval_sym) SCM name ## _wrapper() { TYPE_ ## retval_sym c_retval; SCM retval; c_retval=name(); XLATE_FROM_ ## retval_sym(c_retval,retval); return retval; }
#define BOGO_WRAPPER_1(name,retval_sym,sym1) SCM name ## _wrapper(SCM arg1) { TYPE_ ## retval_sym c_retval; SCM retval; TYPE_ ## sym1 c_arg1; ASSERT_ ## sym1(arg1); c_arg1=XLATE_TO_ ## sym1(arg1); c_retval=name(c_arg1); XLATE_FROM_ ## retval_sym(c_retval,retval); return retval; }
#define BOGO_WRAPPER_2(name,retval_sym,sym1,sym2) SCM name ## _wrapper(SCM arg1,SCM arg2) { TYPE_ ## retval_sym c_retval; SCM retval; TYPE_ ## sym1 c_arg1; TYPE_ ## sym2 c_arg2; ASSERT_ ## sym1(arg1); ASSERT_ ## sym2(arg2); c_arg1=XLATE_TO_ ## sym1(arg1); c_arg2=XLATE_TO_ ## sym2(arg2); c_retval=name(c_arg1,c_arg2); XLATE_FROM_ ## retval_sym(c_retval,retval); return retval; }
#define BOGO_WRAPPER_3(name,retval_sym,sym1,sym2,sym3) SCM name ## _wrapper(SCM arg1,SCM arg2,SCM arg3) { TYPE_ ## retval_sym c_retval; SCM retval; TYPE_ ## sym1 c_arg1; TYPE_ ## sym2 c_arg2; TYPE_ ## sym3 c_arg3; ASSERT_ ## sym1(arg1); ASSERT_ ## sym2(arg2); ASSERT_ ## sym3(arg3); c_arg1=XLATE_TO_ ## sym1(arg1); c_arg2=XLATE_TO_ ## sym2(arg2); c_arg3=XLATE_TO_ ## sym3(arg3);  c_retval=name(c_arg1,c_arg2,c_arg3); XLATE_FROM_ ## retval_sym(c_retval,retval); return retval; }

size_t
ssl_smob_free(SCM val)
{
  SSL *c_val;
  if(SCM_SMOB_DATA(val)) {
    c_val=XLATE_TO_SSL(val);
    SSL_free(c_val);
  }
  return 0;
}

size_t
ssl_ctx_smob_free(SCM val)
{
  SSL_CTX *c_val;
  if(SCM_SMOB_DATA(val)) {
    c_val=XLATE_TO_CTX(val);
    SSL_CTX_free(c_val);
  }
  return 0;
}

size_t
ssl_method_smob_free(SCM val)
{
  // no need to free methods. They weren't dynamically allocated
  return 0;
}

size_t
x509_smob_free(SCM val)
{
  X509 *c_val;
  if(SCM_SMOB_DATA(val)) {
    c_val=XLATE_TO_X509(val);
    X509_free(c_val);
  }
  return 0;
}

size_t
x509_name_smob_free(SCM val)
{
  // x509 freeing is tricky
  return 0;
}

size_t
evp_pkey_smob_free(SCM val)
{
  EVP_PKEY *c_val;
  if(SCM_SMOB_DATA(val)) {
    c_val=XLATE_TO_EVP_PKEY(val);
    EVP_PKEY_free(c_val);
  }
  return 0;
}

size_t
evp_md_smob_free(SCM val)
{
  // no need to free mds. They weren't dynamically allocated
  return 0;
}

size_t
rsa_smob_free(SCM val)
{
  printf("FREEING AN RSA\n");
  return 0;
}

size_t
asn1_time_smob_free(SCM val)
{
  //ASN1_TIME *c_val;
  //if(SCM_SMOB_DATA(val)) {
  //  c_val=XLATE_TO_ASN1_TIME(val);
  //  ASN1_TIME_free(c_val);
  //}
  return 0;
}

BOGO_WRAPPER_1(SSL_accept,INT,SSL)
BOGO_WRAPPER_1(SSL_get_verify_result,LONG,SSL)
BOGO_WRAPPER_1(SSL_get_peer_certificate,X509,SSL)
BOGO_WRAPPER_3(SSL_use_certificate_file,INT,SSL,FILENAME,INT)
BOGO_WRAPPER_2(SSL_add_client_CA,INT,SSL,X509);
BOGO_WRAPPER_1(SSL_shutdown,INT,SSL)
BOGO_WRAPPER_1(SSL_connect,INT,SSL)
BOGO_WRAPPER_2(SSL_get_error,INT,SSL,INT)
BOGO_WRAPPER_2(SSL_set_fd,INT,SSL,INT)
BOGO_WRAPPER_1(SSL_get_fd,INT,SSL)

BOGO_WRAPPER_2(SSL_CTX_add_client_CA,INT,CTX,X509)
BOGO_WRAPPER_2(SSL_CTX_add_extra_chain_cert,LONG,CTX,X509)
BOGO_WRAPPER_2(SSL_CTX_add_session,INT,CTX,SESSION)
BOGO_WRAPPER_3(SSL_CTX_use_certificate_file,INT,CTX,FILENAME,INT)
BOGO_WRAPPER_1(SSL_CTX_new,CTX,METHOD)
BOGO_WRAPPER_3(SSL_CTX_load_verify_locations,INT,CTX,FILENAME,FILENAME)

BOGO_WRAPPER_2(SSL_CTX_set_cipher_list,INT,CTX,STR0)

SCM
SSL_CTX_set_session_id_context_wrapper(SCM context,SCM sid_ctx)
{
  SSL_CTX *c_context;
  const unsigned char *c_sid_ctx;
  unsigned int c_sid_ctx_len;
  int c_retval;

  ASSERT_CTX(context);

  c_context=XLATE_TO_CTX(context);
  if(sid_ctx==SCM_BOOL_F) {
    c_sid_ctx=NULL;
    c_sid_ctx_len=0;
  } else {
    c_sid_ctx=(unsigned char *)SCM_CHARS(sid_ctx);
    c_sid_ctx_len=SCM_STRING_LENGTH(sid_ctx);
  }

  c_retval=SSL_CTX_set_session_id_context(c_context,c_sid_ctx,c_sid_ctx_len);

  return gh_int2scm(c_retval);
}

BOGO_WRAPPER_2(SSL_CTX_use_PrivateKey,INT,CTX,EVP_PKEY)
BOGO_WRAPPER_3(SSL_CTX_use_PrivateKey_file,INT,CTX,FILENAME,INT)
BOGO_WRAPPER_2(SSL_CTX_use_certificate,INT,CTX,X509)

BOGO_WRAPPER_1(SSL_new,SSL,CTX)

SCM
SSL_free_wrapper(SCM ssl) {
  SSL *c_ssl;

  ASSERT_SSL(ssl);

  c_ssl=XLATE_TO_SSL(ssl);

  SSL_free(c_ssl);

  return SCM_BOOL_T;
}


SCM
SSL_write_wrapper(SCM ssl,SCM buf) {
  SSL *c_ssl;
  int c_retval;
  
  ASSERT_SSL(ssl);

  c_ssl=XLATE_TO_SSL(ssl);

  c_retval=SSL_write(c_ssl,SCM_CHARS(buf),SCM_STRING_LENGTH(buf));

  return gh_int2scm(c_retval);
}

SCM
SSL_read_wrapper(SCM ssl) {
  unsigned char read_buffer[16*1024]; // max record size for SSLv3/TLSv1
  SSL *c_ssl;
  int c_retval;
  SCM retval;

  ASSERT_SSL(ssl);
  
  c_ssl=XLATE_TO_SSL(ssl);

  c_retval=SSL_read(c_ssl,read_buffer,sizeof(read_buffer)); 

  if(c_retval>0) {
    retval=gh_str2scm((char *)read_buffer,c_retval);
  } else {
    return gh_int2scm(c_retval);
  }

  return retval;
}

BOGO_WRAPPER_0(EVP_PKEY_new,EVP_PKEY)

BOGO_WRAPPER_2(EVP_PKEY_set1_RSA,INT,EVP_PKEY,RSA)

BOGO_WRAPPER_0(EVP_sha1,EVP_MD)

static RSA *bogo_RSA_generate_key(int a,unsigned long b) {
  return RSA_generate_key(a,b,NULL,NULL);
}
BOGO_WRAPPER_2(bogo_RSA_generate_key,RSA,INT,ULONG)

static int
bogo_verify_callback(int preverify_ok,X509_STORE_CTX *x509_ctx) {
  //printf("VERIFY CALLBACK: %d %p\n",preverify_ok,x509_ctx);
  //return 0;
  // say it's OK. leave it up to guile to check the status and shut down
  return 1;
}

static int bogo_SSL_CTX_set_verify(SSL_CTX *s,int m) {
  //SSL_CTX_set_verify(s,m,NULL);
  SSL_CTX_set_verify(s,m,bogo_verify_callback);
  return 0;
}
BOGO_WRAPPER_2(bogo_SSL_CTX_set_verify,INT,CTX,INT)


BOGO_WRAPPER_0(X509_new,X509)
BOGO_WRAPPER_2(X509_set_version,INT,X509,INT)
BOGO_WRAPPER_2(X509_set_serialNumber,INT,X509,ASN1_INTEGER)
BOGO_WRAPPER_1(X509_get_issuer_name,X509_NAME,X509)
BOGO_WRAPPER_2(X509_set_issuer_name,INT,X509,X509_NAME)
BOGO_WRAPPER_1(X509_get_subject_name,X509_NAME,X509)
BOGO_WRAPPER_2(X509_set_subject_name,INT,X509,X509_NAME)
BOGO_WRAPPER_2(X509_set_pubkey,INT,X509,EVP_PKEY)
BOGO_WRAPPER_3(X509_sign,INT,X509,EVP_PKEY,EVP_MD)

BOGO_WRAPPER_1(X509_get_notBefore,ASN1_TIME,X509)
BOGO_WRAPPER_1(X509_get_notAfter,ASN1_TIME,X509)
BOGO_WRAPPER_2(X509_gmtime_adj,ASN1_TIME,ASN1_TIME,LONG)

BOGO_WRAPPER_0(X509_NAME_new,X509_NAME)

SCM
X509_NAME_add_entry_by_NID_wrapper(SCM name,SCM nid,SCM type,SCM bytes,SCM len,SCM loc,SCM set)
{
  int c_retval;
  X509_NAME *c_name;
  int c_nid;
  int c_type;
  unsigned char *c_bytes;
  int c_len;
  int c_loc;
  int c_set;

  ASSERT_X509_NAME(name);

  c_name=XLATE_TO_X509_NAME(name);
  c_nid=gh_scm2int(nid);
  c_type=gh_scm2int(type);
  c_bytes=(unsigned char *)SCM_CHARS(bytes);
  if(len == SCM_BOOL_F) {
    c_len=SCM_STRING_LENGTH(bytes);
  } else {
    c_len=gh_scm2int(len);
  }
  c_loc=gh_scm2int(loc);
  c_set=gh_scm2int(set);

  c_retval=X509_NAME_add_entry_by_NID(c_name,c_nid,c_type,c_bytes,c_len,c_loc,c_set);

  return gh_int2scm(c_retval);
}

SCM
X509_NAME_get_text_by_NID_wrapper(SCM name,SCM nid,SCM bytes,SCM len)
{
  int c_retval;
  X509_NAME *c_name;
  int c_nid;
  char *c_bytes;
  int c_len;

  ASSERT_X509_NAME(name);

  c_name=XLATE_TO_X509_NAME(name);
  c_nid=gh_scm2int(nid);
  c_bytes=(char *)SCM_CHARS(bytes);
  if(len == SCM_BOOL_F) {
    c_len=SCM_STRING_LENGTH(bytes);
  } else {
    c_len=gh_scm2int(len);
  }

  c_retval=X509_NAME_get_text_by_NID(c_name,c_nid,c_bytes,c_len);

  return gh_int2scm(c_retval);
}

BOGO_WRAPPER_0(ASN1_INTEGER_new,ASN1_INTEGER)

BOGO_WRAPPER_2(ASN1_INTEGER_set,INT,ASN1_INTEGER,INT)

SCM SSLv2_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv2_method()); return r;}
SCM SSLv2_client_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv2_client_method()); return r;}
SCM SSLv2_server_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv2_server_method()); return r;}

SCM SSLv3_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv3_method()); return r;}
SCM SSLv3_client_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv3_client_method()); return r;}
SCM SSLv3_server_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv3_server_method()); return r;}

SCM TLSv1_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,TLSv1_method()); return r;}
SCM TLSv1_client_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,TLSv1_client_method()); return r;}
SCM TLSv1_server_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,TLSv1_server_method()); return r;}

SCM SSLv23_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv23_method()); return r;}
SCM SSLv23_client_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv23_client_method()); return r;}
SCM SSLv23_server_method_wrapper() { SCM r; SCM_NEWSMOB(r,ssl_method_smob_tag,SSLv23_server_method()); return r;}

typedef struct scm_t_sslport {
  //int size_remaining;
  SCM handle;
} scm_t_sslport;

SCM
ssl_to_port_wrapper(SCM ssl)
{
  scm_t_sslport *sp;
  SCM port;
  scm_t_port *ps;

  ASSERT_SSL(ssl);

  SCM_NEWCELL(port);
  SCM_DEFER_INTS;
  ps = scm_add_to_port_table(port);
  SCM_SETPTAB_ENTRY(port,ps);
  SCM_SET_CELL_TYPE(port, (sslport_tag | scm_mode_bits("rw")));

  sp=(scm_t_sslport *)scm_must_malloc(sizeof(scm_t_sslport), "ssl_to_port_wrapper");
  sp->handle=ssl;
  //sp->size_remaining=0;
  ps->rw_random=0;
  ps->read_buf=scm_must_malloc(16*1024*1024,"ssl_to_port_wrapper");
  ps->read_pos = ps->read_end = ps->read_buf;
  ps->read_buf_size = 16*1024*1024;
  ps->write_buf=scm_must_malloc(16*1024*1024,"ssl_to_port_wrapper");
  ps->write_pos = ps->write_end = ps->write_buf;
  ps->write_buf_size = 16*1024*1024;

  SCM_SET_CELL_WORD_0(port, SCM_CELL_WORD_0(port) & ~SCM_BUF0);

  SCM_SETSTREAM(port,sp);

  SCM_SET_FILENAME(port,gh_str02scm("an-ssl-connection"));
  
  SCM_ALLOW_INTS;

  return port;
}

SCM
sslport_mark(SCM port)
{
  scm_t_sslport *sp = ((scm_t_sslport *)SCM_STREAM(port));
  scm_gc_mark(sp->handle);
  return SCM_BOOL_F;
}

static int
sslport_fill_input(SCM port)
{
  /* Guile docs say we can trust that the input buffer is empty when we're called */
  int count;
  scm_t_port *ps = SCM_PTAB_ENTRY(port);
  scm_t_sslport *sp = ((scm_t_sslport *)SCM_STREAM(port));
  SSL *c_handle;
  unsigned char *read_buf;

  c_handle=(SSL *)SCM_SMOB_DATA(sp->handle);
  read_buf = ps->read_buf;

  count=SSL_read(c_handle,ps->read_buf,(16*1024*1024));

  if(count<=0) {
    return EOF;
  } else {
    ps->read_pos=ps->read_buf;
    ps->read_end=ps->read_buf+count;
    return count;
  }
}

void
sslport_write(SCM port,const void *data,size_t size)
{
  int count;
  //scm_t_port *ps = SCM_PTAB_ENTRY(port);
  scm_t_sslport *sp = ((scm_t_sslport *)SCM_STREAM(port));
  SSL *c_handle;

  c_handle=(SSL *)SCM_SMOB_DATA(sp->handle);

  count=SSL_write(c_handle,data,size);
}

int
sslport_input_waiting(SCM port) {
  SSL *c_handle;
  scm_t_sslport *sp = ((scm_t_sslport *)SCM_STREAM(port));
  struct pollfd pfd;
  int c_retval;

  c_handle=(SSL *)SCM_SMOB_DATA(sp->handle);
  pfd.fd=SSL_get_fd(c_handle);
  pfd.events=POLLIN|POLLPRI;
  pfd.revents=0;

  c_retval=poll(&pfd,1,0);

  if(c_retval>0) {
    return 1;
  } else {
    return 0;
  }
}

#define SCM_SSLPORTP(x) (!SCM_IMP (x) && (SCM_TYP16 (x) == sslport_tag))

SCM
ssl_port_p(SCM port) {
  if(SCM_SSLPORTP(port))
    return SCM_BOOL_T;
  else
    return SCM_BOOL_F;
}

SCM
ssl_port_to_ssl(SCM port) {
  if(SCM_SSLPORTP(port)) {
    return ((scm_t_sslport *)SCM_STREAM(port))->handle;
  } else {
    return SCM_BOOL_F;
  }
}

void
openssl_init(void)
{
  SSL_load_error_strings();
  SSL_library_init();

  ssl_method_smob_tag=scm_make_smob_type("SSL_METHOD",sizeof(SSL_METHOD *));
  ssl_cipher_smob_tag=scm_make_smob_type("SSL_CIPHER",sizeof(SSL_CIPHER *));
  ssl_ctx_smob_tag=scm_make_smob_type("SSL_CTX",sizeof(SSL_CTX *));
  ssl_session_smob_tag=scm_make_smob_type("SSL_SESSION",sizeof(SSL_SESSION *));
  ssl_smob_tag=scm_make_smob_type("SSL",sizeof(SSL *));
  x509_smob_tag=scm_make_smob_type("X509",sizeof(X509 *));
  x509_store_ctx_smob_tag=scm_make_smob_type("X509_STORE_CTX",sizeof(X509_STORE_CTX *));
  x509_name_smob_tag=scm_make_smob_type("X509_NAME",sizeof(X509_NAME *));
  evp_pkey_smob_tag=scm_make_smob_type("EVP_PKEY",sizeof(EVP_PKEY *));
  evp_md_smob_tag=scm_make_smob_type("EVP_MD",sizeof(EVP_MD *));
  rsa_smob_tag=scm_make_smob_type("RSA",sizeof(RSA *));
  asn1_integer_smob_tag=scm_make_smob_type("ASN1_INTEGER",sizeof(ASN1_INTEGER *));
  asn1_time_smob_tag=scm_make_smob_type("ASN1_TIME",sizeof(ASN1_TIME *));

  scm_set_smob_free(ssl_smob_tag,ssl_smob_free);
  scm_set_smob_free(ssl_ctx_smob_tag,ssl_ctx_smob_free);
  scm_set_smob_free(ssl_method_smob_tag,ssl_method_smob_free);
  scm_set_smob_free(x509_smob_tag,x509_smob_free);
  scm_set_smob_free(x509_name_smob_tag,x509_name_smob_free);
  scm_set_smob_free(evp_pkey_smob_tag,evp_pkey_smob_free);
  scm_set_smob_free(evp_md_smob_tag,evp_md_smob_free);
  scm_set_smob_free(rsa_smob_tag,rsa_smob_free);
  scm_set_smob_free(asn1_time_smob_tag,asn1_time_smob_free);

  sslport_tag=scm_make_port_type("ssl",sslport_fill_input,sslport_write);
  scm_set_port_input_waiting(sslport_tag,sslport_input_waiting);

#define PUBLISH_CONST(c_name,scheme_name) { gh_define(scheme_name,gh_int2scm(c_name)); gh_eval_str("(define-public " scheme_name " " scheme_name ")"); }
  PUBLISH_CONST(X509_V_OK,"X509_V_OK");

  PUBLISH_CONST(SSL_VERIFY_PEER,"SSL_VERIFY_PEER");
  PUBLISH_CONST(SSL_VERIFY_CLIENT_ONCE,"SSL_VERIFY_CLIENT_ONCE");
 
  PUBLISH_CONST(SSL_ERROR_NONE,"SSL_ERROR_NONE");
  PUBLISH_CONST(SSL_ERROR_ZERO_RETURN,"SSL_ERROR_ZERO_RETURN");
  PUBLISH_CONST(SSL_ERROR_WANT_READ,"SSL_ERROR_WANT_READ");
  PUBLISH_CONST(SSL_ERROR_WANT_WRITE,"SSL_ERROR_WANT_WRITE");
  PUBLISH_CONST(SSL_ERROR_WANT_CONNECT,"SSL_ERROR_WANT_CONNECT"); 
  PUBLISH_CONST(SSL_ERROR_WANT_X509_LOOKUP,"SSL_ERROR_WANT_X509_LOOKUP");
  PUBLISH_CONST(SSL_ERROR_SYSCALL,"SSL_ERROR_SYSCALL");
  PUBLISH_CONST(SSL_ERROR_SSL,"SSL_ERROR_SSL");

  PUBLISH_CONST(SSL_FILETYPE_ASN1,"SSL_FILETYPE_ASN1");
  PUBLISH_CONST(SSL_FILETYPE_PEM,"SSL_FILETYPE_PEM");

  PUBLISH_CONST(MBSTRING_ASC,"MBSTRING_ASC");
  PUBLISH_CONST(MBSTRING_UTF8,"MBSTRING_UTF8");
  PUBLISH_CONST(V_ASN1_IA5STRING,"V_ASN1_IA5STRING");

  PUBLISH_CONST(NID_commonName,"NID_commonName");

#define PUBLISH_FUNC(c_name,scheme_name,args) { gh_new_procedure(scheme_name,c_name,args,0,0); gh_eval_str("(define-public " scheme_name " " scheme_name ")"); }

  PUBLISH_FUNC(ssl_to_port_wrapper,"SSL->port",1);
  PUBLISH_FUNC(ssl_port_p,"SSL-port?",1);
  PUBLISH_FUNC(ssl_port_to_ssl,"port->SSL",1);

  PUBLISH_FUNC(SSLv2_method_wrapper,"SSLv2_method",0);
  PUBLISH_FUNC(SSLv2_client_method_wrapper,"SSLv2_client_method",0);
  PUBLISH_FUNC(SSLv2_server_method_wrapper,"SSLv2_server_method",0);

  PUBLISH_FUNC(SSLv3_method_wrapper,"SSLv3_method",0);
  PUBLISH_FUNC(SSLv3_client_method_wrapper,"SSLv3_client_method",0);
  PUBLISH_FUNC(SSLv3_server_method_wrapper,"SSLv3_server_method",0);

  PUBLISH_FUNC(TLSv1_method_wrapper,"TLSv1_method",0);
  PUBLISH_FUNC(TLSv1_client_method_wrapper,"TLSv1_client_method",0);
  PUBLISH_FUNC(TLSv1_server_method_wrapper,"TLSv1_server_method",0);

  PUBLISH_FUNC(SSLv23_method_wrapper,"SSLv23_method",0);
  PUBLISH_FUNC(SSLv23_client_method_wrapper,"SSLv23_client_method",0);
  PUBLISH_FUNC(SSLv23_server_method_wrapper,"SSLv23_server_method",0);

  PUBLISH_FUNC(SSL_CTX_new_wrapper,"SSL_CTX_new",1);
  PUBLISH_FUNC(SSL_CTX_load_verify_locations_wrapper,"SSL_CTX_load_verify_locations",3);
  PUBLISH_FUNC(SSL_CTX_set_cipher_list_wrapper,"SSL_CTX_set_cipher_list",2);
  PUBLISH_FUNC(SSL_CTX_set_session_id_context_wrapper,"SSL_CTX_set_session_id_context",2);
  PUBLISH_FUNC(SSL_CTX_use_PrivateKey_wrapper,"SSL_CTX_use_PrivateKey",2);
  PUBLISH_FUNC(SSL_CTX_use_PrivateKey_file_wrapper,"SSL_CTX_use_PrivateKey_file",3);
  PUBLISH_FUNC(SSL_CTX_use_certificate_wrapper,"SSL_CTX_use_certificate",2);
  PUBLISH_FUNC(SSL_CTX_use_certificate_file_wrapper,"SSL_CTX_use_certificate_file",3);

  PUBLISH_FUNC(SSL_new_wrapper,"SSL_new",1);
  PUBLISH_FUNC(SSL_free_wrapper,"SSL_free",1);
  PUBLISH_FUNC(SSL_get_verify_result_wrapper,"SSL_get_verify_result",1);
  PUBLISH_FUNC(SSL_get_peer_certificate_wrapper,"SSL_get_peer_certificate",1);
  PUBLISH_FUNC(SSL_use_certificate_file_wrapper,"SSL_use_certificate_file",3);
  PUBLISH_FUNC(SSL_add_client_CA_wrapper,"SSL_add_client_CA",2);

  PUBLISH_FUNC(SSL_accept_wrapper,"SSL_accept",1);
  PUBLISH_FUNC(SSL_shutdown_wrapper,"SSL_shutdown",1);
  PUBLISH_FUNC(SSL_connect_wrapper,"SSL_connect",1);
  PUBLISH_FUNC(SSL_get_error_wrapper,"SSL_get_error",2);
  PUBLISH_FUNC(SSL_set_fd_wrapper,"SSL_set_fd",2);
  PUBLISH_FUNC(SSL_get_fd_wrapper,"SSL_get_fd",1);
  PUBLISH_FUNC(SSL_write_wrapper,"SSL_write",2);
  PUBLISH_FUNC(SSL_read_wrapper,"SSL_read",1);

  PUBLISH_FUNC(X509_new_wrapper,"X509_new",0);
  PUBLISH_FUNC(X509_set_version_wrapper,"X509_set_version",2);
  PUBLISH_FUNC(X509_set_serialNumber_wrapper,"X509_set_serialNumber",2);
  PUBLISH_FUNC(X509_get_issuer_name_wrapper,"X509_get_issuer_name",1);
  PUBLISH_FUNC(X509_set_issuer_name_wrapper,"X509_set_issuer_name",2);
  PUBLISH_FUNC(X509_get_subject_name_wrapper,"X509_get_subject_name",1);
  PUBLISH_FUNC(X509_set_subject_name_wrapper,"X509_set_subject_name",2);
  PUBLISH_FUNC(X509_set_pubkey_wrapper,"X509_set_pubkey",2);
  PUBLISH_FUNC(X509_sign_wrapper,"X509_sign",3);
  PUBLISH_FUNC(X509_get_notBefore_wrapper,"X509_get_notBefore",1);
  PUBLISH_FUNC(X509_get_notAfter_wrapper,"X509_get_notAfter",1);
  PUBLISH_FUNC(X509_gmtime_adj_wrapper,"X509_gmtime_adj",2);

  PUBLISH_FUNC(X509_NAME_new_wrapper,"X509_NAME_new",0);
  PUBLISH_FUNC(X509_NAME_add_entry_by_NID_wrapper,"X509_NAME_add_entry_by_NID",7);
  PUBLISH_FUNC(X509_NAME_get_text_by_NID_wrapper,"X509_NAME_get_text_by_NID",4);

  PUBLISH_FUNC(ASN1_INTEGER_new_wrapper,"ASN1_INTEGER_new",0);
  PUBLISH_FUNC(ASN1_INTEGER_set_wrapper,"ASN1_INTEGER_set",2);

  PUBLISH_FUNC(EVP_PKEY_new_wrapper,"EVP_PKEY_new",0);
  PUBLISH_FUNC(EVP_PKEY_set1_RSA_wrapper,"EVP_PKEY_set1_RSA",2);

  PUBLISH_FUNC(EVP_sha1_wrapper,"EVP_sha1",0);

  PUBLISH_FUNC(bogo_RSA_generate_key_wrapper,"RSA_generate_key",2);
  PUBLISH_FUNC(bogo_SSL_CTX_set_verify_wrapper,"SSL_CTX_set_verify",2);
}

