;
; declare our module and load up our library
;

(define-module (openssl))

;
; get our dynamic module
;
(load-extension "libguileopenssl" "openssl_init")

(define-public SSL-select (lambda (reads writes xs . rest)
  "select wrapper that works for Guile ports with input_waiting functions"
  ; ports that we don't bother selecting on because they are already ready.
  (define prereadres (delete #f (map 
    (lambda (v) (if (SSL-port? v) (if (char-ready? v) v #f) #f))
    reads
  )))
  (define newreads (delete #f (map
    (lambda (v) (if (SSL-port? v) (if (char-ready? v) #f (cons (SSL_get_fd (port->SSL v)) v)) #f))
    reads
  )))
  (define newwrites (delete #f (map
    (lambda (v) (if (SSL-port? v) (cons (SSL_get_fd (port->SSL v)) v) #f))
    writes
  )))
  ;(display "SSL-select!\n")
  (if (and #f (or (and (vector? writes) (eq? (vector-length writes) 0)) (null? writes))
           (or (and (vector? xs) (eq? (vector-length xs) 1)) (null? xs)))
    ; we were just selecting for read
    (list (map car newreads) '() '())
    (let ()
      ;(format #t "READS: ~a NEWREADS: ~a\n" reads newreads)
      (if (and (null? newreads) (null? newwrites))
        (begin
          ;(display "FASTPATH: ") (display rest) (newline)
          (let ()
            (define res (apply select
              (delete #f (map (lambda (v) (if (file-port? v) v #f)) reads))
              (delete #f (map (lambda (v) (if (file-port? v) v #f)) writes))
              xs
              (if (null? prereadres)
                rest      ; possibly block
                (list 0)  ; why block? we have confirmed input
              )
            ))
            ;(display "FASTPATH RES: ") (display res) (newline)
            ;(display "FASTPATH: PREREAD: ") (display prereadres) (newline)
            (list
              (append prereadres (car res))
              (cadr res)
              (caddr res)
            )
          )
        )
        (let ()
          (define res (apply select
            (append (map car newreads) (delete #f (map (lambda (v) (if (file-port? v) v #f)) reads)))
            (append (map car newwrites) (delete #f (map (lambda (v) (if (file-port? v) v #f)) writes)))
            xs
            (list 0)
          ))
          ;(format #t "INNER RES: ~a\n" res) (force-output (current-output-port))
          (list
            (append prereadres (map (lambda (v)
              (define a (assoc v newreads))
              (if a
                (cdr a)
                v
              )
            ) (car res)))
            (map (lambda (v)
              (define a (assoc v newwrites))
              (if a
                (cdr a)
                v
              )
            ) (cadr res))
            (caddr res)
          )
        )
      )
    )
  )
))
