
(display "Getting the module\n")
(use-modules (openssl))

(display "Creating an SSL context\n")
(define master-context (SSL_CTX_new (SSLv23_method)))

(display "Got the context: ") (display master-context) (newline)

(display "Setting the cipher list: ")
; >128 bits, require authentication, specifically include RC4-SHA, sort by strength
;(display (SSL_CTX_set_cipher_list master-context "HIGH:RC4-SHA:@STRENGTH"))
(display (SSL_CTX_set_cipher_list master-context "HIGH:!ADH:RC4-SHA:@STRENGTH"))
(newline)

(display "Creating a private key\n")
(define pkey (EVP_PKEY_new))
(display "got ") (display pkey) (newline)

(display "Creating an RSA key\n")
(define rsakey (RSA_generate_key 1024 #x10001))

(display "Jamming the RSA key into the private key\n")
(display "got ") (display (EVP_PKEY_set1_RSA pkey rsakey)) (newline)

(display "Using the private key\n")
(display "got ") (display (SSL_CTX_use_PrivateKey master-context pkey)) (newline)

(display "creating a cert\n")
(define cert (X509_new))
(display "got ") (display cert) (newline)

(display "creating a name\n")
(define name (X509_NAME_new))
(display "got ") (display name) (newline)

(display "add_entry_by_NID\n")
(display "got ") (display (X509_NAME_add_entry_by_NID name NID_commonName MBSTRING_ASC "localhost" -1 -1 0)) (newline)

(display "creating a serial\n")
(define serial (ASN1_INTEGER_new))
(display "got ") (display serial) (newline)

(display "setting the serial\n")
(display "got ") (display (ASN1_INTEGER_set serial (getpid))) (newline)

(display "setting the cert version\n")
(display "got ") (display (X509_set_version cert 2)) (newline)

(display "setting the cert serialNumber\n")
(display "got ") (display (X509_set_serialNumber cert serial)) (newline)

(display "setting the issuer name\n")
(display "got ") (display (X509_set_issuer_name cert name)) (newline)

(display "setting the subject name\n")
(display "got ") (display (X509_set_subject_name cert name)) (newline)

(display "gmtime adjusting the notbefore\n")
(display "got ") (display (X509_gmtime_adj (X509_get_notBefore cert) 0)) (newline)

(display "gmtime adjusting the notafter\n")
(display "got ") (display (X509_gmtime_adj (X509_get_notAfter cert) (* 60 60 24 90))) (newline)

(display "setting the public key\n")
(display "got ") (display (X509_set_pubkey cert pkey)) (newline)

(display "signing\n")
(display "got ") (display (X509_sign cert pkey (EVP_sha1))) (newline)

(display "Using it\n")
(display "got ") (display (SSL_CTX_use_certificate master-context cert)) (newline)

(display "Creating the server socket\n")
(define server-socket (socket AF_INET SOCK_STREAM 0))

(display "Got the socket: ") (display server-socket) (newline)

(display "Binding the server socket\n")
(bind server-socket AF_INET INADDR_ANY 0)


(display "Bound at: ") (display (vector-ref (getsockname server-socket) 2)) (newline)

(display "Listening\n")
(listen server-socket 1)

(display "Accepting\n")
(define connection-socket-pair (accept server-socket))
(define connection-socket (car connection-socket-pair))

(display "Accepted ") (display connection-socket) (newline)

(display "FD IS: ") (display (fileno connection-socket)) (newline)

(display "Creating an SSL\n")
(define ssl (SSL_new master-context))

(display "Got it: ") (display ssl) (newline)

(display "Setting the FD\n")

(display "we get: ") (display (SSL_set_fd ssl (fileno connection-socket))) (newline)

(display "CALLING SSL ACCEPT\n")
(display "we get: ") (display (SSL_accept ssl)) (newline)

(display "Reading\n")

(define readval (SSL_read ssl))

(display "we get: ") (write readval) (newline)

(if (eq? readval -1)
  (begin
    (display "ERROR IS: ") (display (SSL_get_error ssl readval)) (newline)
  )
)

(display "BOGO\n")

