#define _FILE_OFFSET_BITS 64
#define _XOPEN_SOURCE 600 // glibc features.h says 600 for sixth revision

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/fcntl.h>

/* include the guile header files */
#include <guile/gh.h>

/* This appears necessary on the mac */
char **environ=NULL;

/* A macro for publishing guile bindings */
#define PUBLISH(scheme_name,c_name,args) { gh_define(scheme_name,gh_new_procedure(scheme_name,c_name,args,0,0)); gh_eval_str("(define-public " scheme_name " " scheme_name ")"); }

SCM scm_grantpt(SCM filedes);
SCM scm_ptsname(SCM filedes);
SCM scm_unlockpt(SCM filedes);

#include <fcntl.h>
SCM scm_posix_openpt(SCM oflag);

static SCM ffi_wrapper(int(*c_fp)(int),char *name,SCM arg1) {
  int c_arg;
  int c_retval;

  SCM_ASSERT(scm_integer_p(arg1),arg1,SCM_ARG1,name);
   
  c_arg=scm_to_int(arg1);

  c_retval=c_fp(c_arg); 

  return scm_int2num(c_retval);
};

SCM scm_grantpt(SCM s_filedes) {
  return ffi_wrapper(grantpt,"grantpt",s_filedes);
}

SCM scm_unlockpt(SCM s_filedes) {
  return ffi_wrapper(unlockpt,"unlockpt",s_filedes);
}

SCM scm_posix_openpt(SCM s_oflag) {
  return ffi_wrapper(posix_openpt,"posix_openpt",s_oflag);
}

SCM scm_ptsname(SCM filedes) {
  int c_arg;
  char *c_retval;

  SCM_ASSERT(scm_integer_p(filedes),filedes,SCM_ARG1,"ptsname");
   
  c_arg=scm_to_int(filedes);

  c_retval=ptsname(c_arg); 

  return scm_from_locale_string(c_retval);
}

void
pty_init(void)
{
  gh_new_procedure("grantpt",scm_grantpt,1,0,0);
  gh_eval_str("(define-public grantpt grantpt)");

  gh_new_procedure("ptsname",scm_ptsname,1,0,0);
  gh_eval_str("(define-public ptsname ptsname)");

  gh_new_procedure("unlockpt",scm_unlockpt,1,0,0);
  gh_eval_str("(define-public unlockpt unlockpt)");

  gh_new_procedure("posix_openpt",scm_posix_openpt,1,0,0);
  gh_eval_str("(define-public posix_openpt posix_openpt)");

  return;
}
