(use-modules (sys unix pty))
(use-modules (ice-9 format))

(define masterfd (posix_openpt O_RDWR))

(if (and (integer? masterfd) (>= masterfd 0))
  (display "posix_openpt succeed\n")
  (display "posix_openpt fail\n")
)

(define slavename (ptsname masterfd))

(if (and (string? slavename) (> (string-length slavename) 0))
  (format #t "ptsname succeed: ~a\n" slavename)
  (format #t "ptsname fail\n")
)

(define granted (grantpt masterfd))

(if (and (integer? granted) (= granted 0))
  (display "ptgrant succeed\n")
  (display "ptgrant fail\n")
)

(define unlocked (unlockpt masterfd))

(if (and (integer? unlocked) (= unlocked 0))
  (display "unlockpt succeed\n")
  (display "unlockpt fail\n")
)

(format #t "parent pid is ~a\n" (getpid))

(flush-all-ports)
(define childpid (primitive-fork))
(cond
  ((equal? childpid 0) (let () ; the child
    (define slave (open slavename O_RDWR))
    (close masterfd)
    (format #t "slave: pid=~a ~a\n" (getpid) (port? slave))
    (force-output (current-output-port))
    (let ()
      (define slavefd (fileno slave))
      (dup2 slavefd 0)
      (dup2 slavefd 1)
      ;(dup2 slavefd 2)
      (display "slave message to tty\n\n\n")
      (force-output (current-output-port)) 
      (display "we think that worked" (current-error-port))
      (exit 0)
    )
  ))
  ((> childpid 0) (let ()
    (define masterinport (fdes->inport masterfd))
    (define masteroutport (fdes->outport masterfd))
    (display "fork succeed\n")
    (sleep 1)
    (display "ready: ") (display (char-ready? masterinport)) (newline)
    (if (char-ready? masterinport)
      (let helper ((c (read-char masterinport)))
        (format #t "C: ~a\n" c)
        (if (eof-object? c)
          (display "done with EOF!\n")
          (begin
            (display "got something")
            (display c)
            (if (char-ready? masterinport)
              (helper (read-char masterinport))
              (display "done with not ready!\n")
            )
          )
        )
      )
      (display "not ready\n")
    )
    (display "leaving")
  )) 
  ((equal? childpid -1) ; error
    (format #t "fork error! ~a\n" childpid)
    (exit -1)
  )
  (else ; the parent
    (display "fork unknown\n")
  )
)
