;
; A Branching Construction for Scheme
;

;
; This introduces new syntax into the Scheme environment
; designed to support a style of parallel computing
;

(cond-expand
  (guile

(library (proverb branch)
  (export begin-for-values global-branched* for-each-value for-each-rank call-for-values fork-begin-for-values)
  (import
    (rnrs) (proverb help) (proverb cond-expand) (proverb doc)
    (only (guile) define-macro eval-case defmacro:transformer assert-defmacro?! set-defmacro-transformer! hashq-set! gensym)
  )


(define global-branched* '()) ; list of currently branched things

(define-macro (for-each-value values)
  (define vsym (gensym))
  `(let ()
     (define ,vsym ,values)
     (call-with-current-continuation (lambda (cc)
       (set! global-branched* (cons (list #f #f cc) global-branched*))
     ))  
     ; it would be an error for global-branced* to be null here
     (let ()
       (define head (car ,vsym))
       (if (eq? (length ,vsym) 1)
         (set! global-branched* (cdr global-branched*))
         (set! ,vsym (cdr ,vsym))
       )
       head
     )
  )
)
(define-macro (for-each-rank size)
  `(if (eq? ,size 0)
    0
    (for-each-value (let helper ((c ,size) (acc '()))
      (if (eq? c 0)
        acc
        (helper (- c 1) (cons (- c 1) acc))
      )
    ))
  )
)

; a branch aware macro
(define-macro (begin-for-values . expressions)
  `(let ()
    (define results '())
    ; it would be an error for global-branched* to be not null here.
    (set! global-branched* '())
    (let ((retval (begin ,@expressions))) (let ((head (if (null? global-branched*) #f (car global-branched*))))
      (if head
        (begin
          (set! results (cons retval results))
          (call-with-current-continuation (list-ref head 2))
        )
      )
      (if (null? results)
        ; this macro always returns a list of results -- even in the singleton case
        (list retval)
        (let ((r (cons retval results)))
          (if (eq? (length r) 1)
            (car r)
            r
          )
        )
      )
    )) 
  )
)

(define call-for-values (with-wiki-docstring
  "call thunk and return a list of every value returned. See [[multi-valued-expressions | this]] for more info"
  (lambda (thunk)
    (begin-for-values (thunk))
  )
))

; a branch aware macro
(define-macro (fork-begin-for-values . expressions)
  `(let ()
    (define results '())
    ; it would be an error for global-branched* to be not null here.
    (set! global-branched* '())
    (let ((retval (begin ,@expressions))) (let ((head (if (null? global-branched*) #f (car global-branched*))))
      (if head
        (begin
          (set! results (cons retval results))
          (fork-begin (call-with-current-continuation (list-ref head 2)))
        )
      )
      (if (null? results)
        ; this macro always returns a list of results -- even in the singleton case
        (list retval)
        (let ((r (cons retval results)))
          (if (eq? (length r) 1)
            (car r)
            r
          )
        )
      )
    )) 
  )
)
)

)
(chez
  (extend-syntax (begin-for-values)
   [(begin-for-values body ...)
    (display "BEGIN FOR VALUES NOT YET AVAILABLE IN CHEZ!\n")]
  )
  (extend-syntax (fork-begin-for-values)
   [(begin-for-values body ...)
    (display "FORK BEGIN FOR VALUES NOT YET AVAILABLE IN CHEZ!\n")]
  )
)
(else
  (define-syntax begin-for-values (syntax-rules ()
    ((begin-for-values body ...)
      (display "BEGIN FOR VALUES NOT AVAILABLE IN THIS IMPLEMENTATION!\n")
    )
  ))
  (define-syntax fork-begin-for-values (syntax-rules ()
    ((fork-begin-for-values body ...)
      (display "FORK BEGIN FOR VALUES NOT AVAILABLE IN THIS IMPLEMENTATION!\n")
    )
  ))
)
)
