(define magic-write-predicate-pairs '())

(define magic-write-find-writer (lambda (val)
  (define helper (lambda (l)
    (if (null? l)
      #f
      (if ((caar l) val)
        (cdar l)
        (helper (cdr l))
      ) 
    )
  ))
  (helper magic-write-predicate-pairs)
))

;
; Pairs of the form ( <predicate> . <writer> )
;
(set! magic-write-predicate-pairs (append
  (list 
    ; things that are perfectly unreadable
    (cons null? write)
    (cons boolean? write)
    (cons symbol?  write)
    (cons unspecified? (lambda (val port) (display "#U " port)))
    (cons eof-object?  (lambda (val port) (display "#.the-eof-object " port)))
    (cons char? write)
    ; our special case for objects with associated serializers
    ;(cons passval? (lambda (val port) (passval-write val port unread)))
    ; and continue
    ;(cons string?  write)
    (cons string? (lambda (v p) (display (format "#b#~a#~a" (string-length v) v) p)))
    (cons number?  write)
    ; meta-objects
    (cons list? (lambda (val port) (display "(" port) (unread (car val) port) (for-each (lambda (v) (display " " port) (unread v port)) (cdr val)) (display ")" port)))
    (cons pair? (lambda (val port) (display "(" port) (unread (car val) port) (display " . " port) (unread (cdr val) port) (display ")" port)))
    (cons vector? (lambda (val port) (display " #(" port)
      (for-each (lambda (v) (unread v port)) (vector->list val)) 
    (display ")" port)))
    ; relatively uncommon objects
    (cons record?  (lambda (val port) (write '*record* port)))
    ;(cons module?  (lambda (val port) (write '*module* port)))
    ;(cons variable?  (lambda (val port) (write '*variable* port)))
    ;(cons macro?  (lambda (val port) (write '*macro* port)))
    ;(cons promise?  (lambda (val port) (write '*promise* port)))
    ;(cons fluid? (lambda (val port) (write '*fluid* port)))
    ;(cons environment? (lambda (val port) (write '*environment* port)))
    ;(cons closure? (lambda (val port) (display "#." port) (unread (procedure-and-environment-source val) port)))
    (cons procedure? (lambda (val port) (display "#." port) (display (car (procedure-information val)) port) (display " " port)))
    ;(cons entity? (lambda (val port) (write '*entity* port)))
    ;(cons stack? (lambda (val port) (write '*stack* port)))
    (cons (lambda (v) #t) (lambda (val port) (write (string-append "bogo-" (->string val)) port)))
  )
  magic-write-predicate-pairs
))

(define add-unreader (lambda (predicate unreader)
;  "add an unreader function for values that match <predicate>"
  (set! magic-write-predicate-pairs (append magic-write-predicate-pairs (list (cons predicate unreader))))
))

; I think the new unreader should go as far down the list as possible -- at
; the end or just before an unreader that would snarf its exemplar
(define add-unreader-with-exemplar (lambda (predicate unreader exemplar)
;  "add an unreader function for values that match <predicate>. Ensure that this is the preferred unreader for <exemplar>"
  (define new-pair (cons predicate unreader))
  (set! magic-write-predicate-pairs (let helper ((cursor magic-write-predicate-pairs) (racc '()))
    (if (null? cursor)
      (reverse (cons new-pair racc))
      ; else is there a conflict?
      (if ((caar cursor) exemplar)
        (append (reverse (cons new-pair racc)) cursor)
        ; else move on
        (helper (cdr cursor) (cons (car cursor) racc))
      )
    )
  ))
))

(define unread (lambda (val port)
;  "write bytes to port that will cause a read to return a reasonable representation of val."
  (let ()
    (define writer (magic-write-find-writer val))
    (if writer
      (writer val port)
      ; else
      (write val port)
    )
    (flush-output port)
  )
))

(define receive-value (with-docstring
  "Obtain a scheme value from <source> using the Proverb remote protocol"
  (lambda (source)
    ;(define readval (with-fluids* (list read-eval? current-descriptor-fluid) (list #t source) (lambda () (read source))))
    (define readval (with-fluids* (list current-descriptor-fluid) (list source) (lambda () (read source))))
    readval
  )
))
