; This file contains bindings for studio functionality that 
;   might be implemented very differently depending on the
;   underlying interpreter. 

; This file is a good location for implementations that would 
;   otherwise require a large number of additions to 
;   studio-dialect.scm that would not be generally useful.

(load "chicken-debug.scm")


;XXX: this only applies to strings in tire forms and should be 
;     delayed to a post-load function
(set-sharp-read-syntax! #\b 
  (lambda (port)
    (if (eq? (read-char port) #\#) 
      (let ((k (read-token (lambda (c) (not (eq? c #\#))) port )))
        (if (eof-object? k)
          k
          (begin 
            (read-char port)
            (read-string (string->number k) port)))))))

;XXX: this seems unecessary
; used for inline stringification
(set-sharp-read-syntax! #\~ (lambda (port)
  (define subordinate-form (read port))
  `(->string ,subordinate-form)
))

;XXX: this is brittle and uses chicken's non-standard read-syntax
(define current-include-filename 
  (cond 
    ((string=? (chicken-version) "2.6") ##sys#current-load-file)
    (else ##sys#current-source-filename)))


;studio provides a record predicate
;XXX: this is brittle and uses chicken's non-standard read-syntax
(define record? (lambda (r) (##sys#structure? r 'rtd)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; UTILITY:  RUN-SELECT-LOOP
;   bindings:  run-select-loop-until
;              run-select-loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; run iterations of the select loop until check returns false
(define run-select-loop-until (lambda (services body check)
  (define safe-select 
    (lambda (r w x t)
      (studio-select 
        (delete #f (map (lambda (v) (if (port-closed? v) #f v)) r))
        (delete #f (map (lambda (v) (if (port-closed? v) #f v)) w))
        (delete #f (map (lambda (v) (if (port-closed? v) #f v)) x))
        t
      )
  ))
  (define states (map car services))
  (define pres (map cadr services))
  (define posts (map caddr services))
  (define prehelper (lambda (sc prec)
    (if (null? sc)
      ; no fds
      (list '() '() '())
      ; else
      (apply (car prec) (cons (car sc) (prehelper (cdr sc) (cdr prec))))
    )
  ))
  (define posthelper (lambda (prestates sc postc)
    ; we need one select call per select-post routine -- but these
    ; inner selects run with no timeout. Since they run with no timeout
    ; they don't actually need all the prestates and we could do a better
    ; job about eliminating select-posts that are superfluous here
    (let loop ((sc sc) (postc postc) (racc '()))
      (define results (apply safe-select (append prestates '(0))))
        (if (null? sc)
          (reverse racc)
          ; else
          (loop (cdr sc) 
                (cdr postc)
                (cons (apply (car postc) (cons (car sc) results)) racc))
        )
    )
  ))
  (define thin (lambda (ball states)
    ;"remove pres and posts for loops with dead states"
    (delete #f (map (lambda (b s)
      (and s b)
    ) ball states))
  ))
  (let helper ((ball (map list states pres posts)))
    (if (null? ball)
      (begin
        #f ; we have no more states to iterate
      )
      (let ()
        (define prestates (prehelper (map car ball) (map cadr ball))) 
        (let ()
          (define results (apply safe-select (append prestates '(#f))))
          (body)
          ; sigh. one post-helper might eat the descriptor intended for another. no more than one post-helper per select call
          (let ()
            (define rargs (thin ball (posthelper prestates (map car ball) (map caddr ball))))
            (if (check)
              (helper rargs)
              #f
            )
          )
        )
      )
    )
  )
))

; run a select loop until there are no more fds
(define run-select-loop (lambda (services body)
  (run-select-loop-until services body (lambda () #t))
))

(define has-fileno?
  (lambda (p)
    (and (port? p)
         (begin 
           (##sys#check-port p 'port->fileno)
           (or (eq? 'socket (##sys#slot p 7))
               (not (zero? (##sys#peek-unsigned-integer p 0))))))))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; UTILITY: CONCURRENT PROCEDURES
;   bindings:  make-concurrent
;              make-concurrent-with-talkback
;              concurrent-apply
;              concurrent-apply-with-talkback
;              run-inferior-process
;              
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; a list of pairs of the form ( <parent-pid> . <child-pid> )
(define delinquent-children '())

(define make-concurrent-with-talkback (lambda-w-types ((sub procedure?) (talkback-token #t))
;  "accept <sub> and return a new procedure that will cause <sub> to be run in a fresh process on each invocation. The new procedure will additionally substitute a talkback port-pair as an argument to <sub> for any argument that is eq? equal to <talkback-token>. For this reason, <talkback-token> should probably be a non-immediate value. The parent caller is not provided with the result of the evaluation of <sub>. Instead, the caller receives a pair of ports (input . output) connected to the child's talkback port. This routine is most commonly used to start long-running REPL procedures in a child process. The child process will not generally have access to the parent's ports unless passed as arguments to this function. The child process will terminate upon completion of <sub>."

  ; one end of sp will be open in parent and child
  ;(define sp (socketpair AF_UNIX SOCK_STREAM 0))
  (define from-parent (studio-pipe))
  (define to-parent (studio-pipe))

  ; our ultimate return value
  (define new-lambda (lambda args
    ; let's not burden our child with our garbage.
    (gc)
    ;(run-hook after-gc-hook)
    ; or with our pending I/O
    ; flush-all-ports has complained in the past about flushing
    ; closed ports
    ;(catch #t (lambda () (flush-all-ports)) (lambda args #f))
    (let ()
      (define child-pid (process-fork))
      (cond 
        ((eq? child-pid 0) (let ()
          ;(define saved-module (current-module))
         
	  ; do our talkback-token substitution 
          (define args-copy (map (lambda (v)
            (if (eq? v talkback-token) (cons (car from-parent) (cdr to-parent)) v)
          ) args))

	  ; Throw away our ports. We ditch most ports except our talkback port and any ports
	  ; passed as arguments.
          (close (cdr from-parent)) ; -- this should be captured by the below
          (close (car to-parent)) ; -- this should be captured by the below

          (let ()
            (define protected-ports  
              (append (delete #f (map (lambda (v) (and (port? v) v)) args-copy)) 
                      (list (car from-parent) (cdr to-parent) (current-output-port) (current-error-port))))

            ;sorted list of protected fds
            (define protected-descriptors 
              (sort (append '(0 1 2) (delete #f 
                                       (map (lambda (p)
                                              (catch #t  
                                                (lambda () (fileno p))
                                                (lambda args #f)))
                                             protected-ports)))
                    <))
            (let loop ((fd 0) (pfds protected-descriptors))
              (if (< fd 1024)
                (cond 
                  ((or (null? pfds)
                       (< fd (car pfds)))
                    (if (not (port-closed? fd))
                      (file-close fd))
                    (loop (+ fd 1) pfds))

                  ((> fd (car pfds)) 
                    (loop fd (cdr pfds)))

                  ((= fd (car pfds))
                    (loop (+ fd 1) (cdr pfds))))))

            (catch #t
              (lambda ()
                (apply sub args-copy))
              (lambda args 
                (process-signal (getpid) signal/kill))))))

        ; was there an error
        ((eq? child-pid -1) (error "bogo"))
        ; else I'm the parent
        (else
          ; This finalizer stuff is a work-around for chicken not closing ports on gc.
          (let ((port-pair (cons (car to-parent) (cdr from-parent))))
            (define magic (list 'concurrent-endpoint 
                                port-pair
                                child-pid (getpid)))
            (set-finalizer! magic concurrent-cleanup)
            (close (cdr to-parent))
            (close (car from-parent))
                
            ;give each port a reference so that concurrent-cleanup will
            ;  only be called when *both* are gc'd :: does this work?
            (set-object-property! (car to-parent) 'port-guardian magic)
            (set-object-property! (cdr from-parent) 'port-guardian magic)
            ;(debugmsg "Delivering the port: ~a properties=~a\n" (car sp) (object-properties (car sp))) 
            port-pair
          )
        )
      )
    )
  ))
  (if (procedure-property sub 'args)
    (set-procedure-property! new-lambda 'args (procedure-property sub 'args))
  )
  new-lambda
))

(define make-concurrent (lambda (sub)
;  "accept <sub> and return a new procedure that will cause <sub> to be run in a fresh process on each invocation. The return value is unspecified. This routine is most commonly used to start a 'fire and forget' action."
  ; using a private non-immediate keeps us safe
  (define concurrent (make-concurrent-with-talkback sub (cons 1 1)))
  (let ()
    (define new-lambda (lambda args
      (define port (apply concurrent args))
      ; we return nothing useful
      *unspecified*
    ))
    (if (procedure-property sub 'args)
      (set-procedure-property! new-lambda 'args (procedure-property sub 'args))
    )
    new-lambda
  )
))

(define concurrent-cleanup
  (lambda (endp)
    (define from-child (car (list-ref endp 1)))
    (define to-child (cdr (list-ref endp 1)))
    
    (define child (list-ref val 2))
    (define parent (list-ref val 3))

    (if (eq? parent (getpid))
      (begin
        (debugmsg "GOT A PORT MAGIC!\n")

        (debugmsg "MAGIC CLOSING A PORT OF OURS: ~a\n" from-child) 
        (catch #t (lambda () (close from-child)) (lambda args #f))

        (debugmsg "MAGIC CLOSING A PORT OF OURS: ~a\n" to-child) 
        (catch #t (lambda () (close to-child)) (lambda args #f))
        (debugmsg "MAGIC STILL HERE:\n"))

      (begin
        (debugmsg "MAGIC CLOSING A PORT SOMEONE ELSES: ~a\n" port) 
        ; we should close the port anyway (but not shutdown)...
        (close port)))

    (let ()
      (define subhelper (lambda (c tries)
        (if (and (not (null? c)) (< tries 3))
          (let ()
            ; waitpid should never error here. We are only calling waitpid
            ; in the case that we have a set of child pids we are looking
            ; for... An error in waitpid means that the list is stale.
            (subhelper (delete #f (map (lambda (v)
              (define waitval (process-wait v #t)) ;only take the first value
              (debugmsg "WAITED FOR PID ~a: ~a\n" v waitval)
              (if (eq? (car waitval) 0) v #f)
            ) c)) (+ tries 1)))
          ; else
          (set! delinquent-children (map (lambda (v) (cons (getpid) v)) c)))))
      
      (let ((children (delete #f (map (lambda (v) (if (eq? (car v) (getpid)) (cdr v) #f)) 
                                      delinquent-children))))
        (subhelper (if child (cons child children) children) 1)))))

(define concurrent-apply (lambda (sub args)
;  "Accept a subroutine <sub> run it with <args> in a child process"
  (apply (make-concurrent sub) args)
))

(define concurrent-apply-with-talkback (lambda (sub args talkback-token)
;  "Accept a subroutine <sub> run it with <args> in a child process. substitute <talkback-token>, if in <args>, with a talkback port"
  (apply (make-concurrent-with-talkback sub talkback-token) args)
))

(define run-inferior-process (lambda-w-types ((shellfrag string?))
;  "Use the UNIX shell to run the command <shellfrag> in a child process. Return a full-duplex port connected to the standard input and standard output of the child."
  (define magic-talkback-val '(talkback))
  (let ()
    (define port-pair (concurrent-apply-with-talkback (lambda (iopair)
  	  (duplicate-fileno (port->fdes (car iopair)) 0)
	  (duplicate-fileno (port->fdes (cdr iopair)) 1)
          (execl "/bin/sh" "sh" "-c" shellfrag)
    ) (list magic-talkback-val) magic-talkback-val))
    port-pair
  )
))
