(library (proverb lexical)
  (export
    procedure-get-num-args
    procedure-get-arg-names
    consider-procedure
    arg-derive-all-from-source
  )
  (import
    (rnrs)
    (rnrs lists)
    (proverb doc) (proverb help) (proverb debug) 
    (only (proverb guilefoo) procedure-and-environment-source procedure-ref)
    (only (proverb dialect) always-false index-map)
    (guile modules)
    (only (guile)
      closure? procedure-source procedure-name
      procedure-property procedure-environment
      procedure-properties
      unspecified?
      object->string
      *unspecified*
      minor-version
      variable?
      variable-ref
      environment-module
      the-root-module
      make-list
    )
  )

;
; code for considering predicates
;

;
; this is wacky stuff 
;

;
; returns a consideration.
;
(define consider-predicate (with-docstring
  "generate a consideration of the predicate function <predicate>"
  (lambda (predicate)
    (list
      (cons 'predicate predicate)
      (cons 'predicate-source (if (closure? predicate) (procedure-source predicate) #f))
      ;(cons 'eval-module eval-module)
    )
  )
))

(define consideration-get-predicate (with-docstring
  "extract the predicate from a consideration"
  (lambda (consideration)
    (cdr (assoc 'predicate consideration))
  )
))

(define consideration-get-source (with-docstring
  "extract the source for the predicate stored in a consideration"
  (lambda (consideration)
    (cdr (assoc 'predicate-source consideration))
  )
))

;(define consideration-get-module (with-docstring
;  "extract the module for the predicate stored in a consideration"
;  (lambda (consideration)
;    (cdr (assoc 'eval-module consideration))
;  )
;))

;
; this is even wackier.
;
(define must-satisfy-predicate? (with-docstring
  "returns #t if the consideration must satisfy the predicate function <predicate>, false if we don't know"
(lambda (consideration predicate)
  (define c-predicate (consideration-get-predicate consideration))
  (define c-source (consideration-get-source consideration))

  (if (eq? c-predicate predicate)
    ; that was easy
    #t
    ; we couldn't immediately puzzle out the answer. Go look at the source.
    (if c-source
      (let ()
        (define predicate-name (procedure-name predicate))
        ; source: ( name bindings body ... )
        (if (>= (length c-source) 3) ; is there a body?
          (let ()
            (define term (list-ref c-source 2)) ; the first body term
            (define examine-term (lambda (t ancestry . limitclauses)
              (if (and (pair? t) (list? t))
                (cond
                  ; to we directly match lexically?
                  ((eq? (car t) predicate-name) #t)
                  ; do we match as a clause in an and?
                  ; (only examines the first clause at present)
                  ((eq? (car t) 'and) (let ()
                    (if (> (length t) 1)
                      (apply examine-term (cadr t) t (cdr t))
                      ; this is a broken and clause
                      #f
                    )
                  ))
                  (else #f)
                )
                #f
              )
            ))
            (if (and (string? term) (> (length c-source) 3))
              (set! term (list-ref c-source 3))
            )
            (examine-term #f term)
          )
          ; else I don't understand the form of this predicate
          #f
        )
      )
      ; else we don't have source
      #f
    )
  )
  )
))

(define predicate-curries? (with-docstring
  "returns #t if the consideration must satisfy the predicate function <predicate>, false if we don't know"
  (lambda (consideration function)
  (define consideration-predicate (cdr (assoc 'predicate consideration)))
  (define consideration-source (cdr (assoc 'predicate-source consideration)))

  (if (eq? consideration-predicate function)
    #t
    ; we couldn't immediately puzzle out the answer. Go look at the source.
    (if consideration-source
      (let ()
        (define function-name (procedure-name function))
        (if (>= (length consideration-source) 3)
          (let ()
            (define term (list-ref consideration-source 2))
            (define examine-term (lambda (t ancestry . limitclauses)
              (if (and (pair? t) (list? t))
                (cond
                  ; to we directly match lexically?
                  ((eq? (car t) function-name) #t)
                  ; do we match as a clause in an and?
                  ; (only examines the first clause at present
                  ((eq? (car t) 'and) (let ()
                    (if (> (length t) 1)
                      (if (apply examine-term (cadr t) t (cdr t))
                        ; this is a match
                        #t
                        ; we should return the stream of all matches
                        ; but (HACKJOB) we return the first
                        ;(make-dynamic-stream (list (cadr t)))
                        ; else recursively examine all the and clauses
                        (apply examine-term (cons 'and (cddr t)) t (cdr t))
                      )
                      ; this is a broken and clause
                      #f
                    )
                  ))
                  (else #f)
                )
                #f
              )
            ))
            (if (and (string? term) (> (length consideration-source) 3))
              (set! term (list-ref consideration-source 3))
            )
            (examine-term term #f)
          )
          ; else I don't understand the form of this predicate
          #f
        )
      )
      ; else we don't have source
      #f
    )
  )
  )
))

;
; ripped from verber
;
(define consider-procedure (with-docstring
  ;"Produce a <:consideration?> from <procedure: procedure?> based on <procedure> and the dynamic scope in <eval-module: module?>" 
  "Produce a <:consideration?> from <procedure: procedure?> based on <procedure>" 
  (lambda (procedure)
    (define source (if (closure? procedure) (procedure-source procedure) #f))
    (define environment-and-source (if (closure? procedure) (procedure-and-environment-source procedure) #f))
 
    (define arg-names #f)        ; a list of names for non-optional args
    (define arg-predicates #f)   ; a list of predicates for non-optional args
    (define arg-defaults #f)     ; a list of default values for non-optional args
    ;(define oarg-names #f)       ; a list of names for optional args
    ;(define oarg-predicates #f)  ; a list of predicates for optional args
    ;(define larg-name #f)
    (define larg-predicate #f)   ; a predicate for members of the larg list?        

    ; module (for context).
    ; this should use the module eval closure in guile
    ;(define module (if eval-module
    ;  eval-module
    ;  (current-module)
    ;))

    ; procedure arguments
    (define args (procedure-property procedure 'args))
    ; procedure optional arguments
    (define oargs (procedure-property procedure 'oargs))
    ; procedure remainder argument
    (define larg (procedure-property procedure 'larg))
    ; formals
    (define formals (if (closure? procedure) (list-ref (procedure-source procedure) 1) #f))

    ; We may be able to divine reasonable argument names and types
    (if args
      (begin
        (set! arg-names (map car args))
        (set! arg-predicates (map caddr args))
        (set! arg-defaults (map (lambda (v) (list-ref v 3)) args))
        (if oargs
          (set! oarg-names (map car oargs))
        )
      )
      (if (closure? procedure)
        (if source
          (let ()
            (define args (list-ref source 1))
            (if (list? args)
              (let ()
                (define all (map (lambda (v) (arg-derive-all-from-source v source procedure)) args))
                (set! arg-predicates (map car all))
	        (set! arg-defaults (map cdr all))
              )
              ; else there is a dotted term
              (let ()
                (define helper (lambda (l racc)
                  (if (null? l)
                    (reverse racc)
                    (if (pair? l)
                      (if (pair? (cdr l))
                        (helper (cdr l) (cons (car l) racc))
                        (reverse (cons (car l) racc))
                      )
                      (reverse racc)
                    )
                  )
                ))
                (let ()
                 (define all (map (lambda (v) (arg-derive-all-from-source v source procedure)) (helper args '())))
                 (set! arg-predicates (map car all))
	         (set! arg-defaults (map cdr all))
                )
              )
            )

            (if (list? arg-names)
              (set! arg-names (map object->string args))
            )
          )
          (critmsg "consider-procedure called on closure without source\n")
        )
        (let ((p (procedure-properties procedure)))
          (if p
            (let ((args (assoc 'args p)))
              (if args
                (begin
                  (set! arg-names (map car (cdr args)))
                  (set! arg-predicates (cadr (cdr args)))
                  (set! arg-defaults (caddr (cdr args)))
                )
                (let ((arity (assoc 'arity p)))
                  (if arity
                    (let ((synth-all (index-map (lambda (i v) (list (string-append "fixed-arg-" (number->string i)) #f *unspecified*)) (make-list (cadr arity)))))
                     
                      (set! arg-names (map car synth-all))
                      (set! arg-predicates (map cadr synth-all))
                      (set! arg-defaults (map caddr synth-all))
                    )
                    (critmsg "consider procedure with no useful properties")
                  )
                )
              )
            )
            (critmsg "consider-procedure called on non-closure procedure with no properties")
          )
        )
      )
    )

    ; this is our return value
    `(
      (procedure . ,procedure)
      (procedure-source . ,source)
      ;(eval-module . ,eval-module)
      (arg-names . ,(cond
          ((and arg-names (not formals))
            arg-names
          )
          ((and (not arg-names) formals)
            (map symbol->string (if (list? formals)
              formals
              (let helper ((l formals) (racc '()))
                (if (null? l)
                  (reverse racc)
                  (if (pair? l)
                    (if (pair? (cdr l))
                      (helper (cdr l) (cons (car l) racc))
                      (reverse (cons (car l) racc))
                    )
                    (reverse racc)
                  )
                )
              )
            ))
          )
          ((and (not arg-names) (not formals))
            #f
          )
          (else (map (lambda (a f)
            (if (or (not a) (unspecified? a))
              f
              a
            )
          ) arg-names formals))
      ))
      (arg-predicates . ,arg-predicates)
      (arg-defaults . ,arg-defaults)
      (larg . ,larg)
      (larg-predicate . ,larg-predicate)
      (environment-and-source . ,environment-and-source)
    )
  )
))

(define procedure-get-num-args (with-docstring
  "accept <consideration: consideration?> and return the number of arguments that the considered procedure accepts"
  (lambda (consideration)
  (if (assoc 'arg-names consideration)
    (length (cdr (assoc 'arg-names consideration)))
    #f
  )
  )
))

(define procedure-get-arg-names (with-docstring
  "Accept <consideration: consideration?> and return a list of argument names for each argument of the considered procedure" 
  (lambda (consideration)
  (if (assoc 'arg-names consideration)
    (cdr (assoc 'arg-names consideration))
    #f
  )
  )
))

(define procedure-get-arg-predicates (with-docstring
  "Accept <consideration: consideration?> and return a list of deduced predicates for each argument of the considered procedure" 
  (lambda (consideration)
  (if (assoc 'arg-predicates consideration)
    (cdr (assoc 'arg-predicates consideration))
    #f
  )
  )
))

(define procedure-get-arg-defaults (with-docstring
  "Accept <consideration: consideration?> and return a list of deduced default values for each argument of the considered procedure" 
  (lambda (consideration)
  (if (assoc 'arg-defaults consideration)
    (cdr (assoc 'arg-defaults consideration))
    #f
  )
  )
))

(define procedure-get-optional-arg-names (with-docstring
  "Accept <consideration: consideration?> and return a list of deduced default names for the optional arguments accepted by the considered procedure"
  (lambda (consideration)
  (if (assoc 'oarg-names consideration)
    (cdr (assoc 'oarg-names consideration))
    #f
  )
  )
))

(define procedure-get-optional-arg-predicates (with-docstring
  "Accept <consideration: consideration?> and return a list of deduced predicates for the optional arguments accepted by the considered procedure"
  (lambda (consideration)
  (if (assoc 'oarg-predicates consideration)
    (cdr (assoc 'oarg-predicates consideration))
    #f
  )
  )
))

;
; platform-independant code for thinking about predicates
;

(define arg-derive-all-from-source (with-docstring
  "Attempt to return a predicate and default value for the argument <arg-name: symbol?> declared in the formals of <closure: procedure?> with source <closure-source>"
  (lambda (arg-name closure-source closure)
  (define formals (if (> (length closure-source) 1)
    (cadr closure-source)
    '()
  ))
  (define improper-member (lambda (key il)
    (if (null? il)
      #f
      (if (pair? il)
        (if (eq? (car il) key)
          il
          (improper-member key (cdr il))
        )
        (eq? il key)
      )
    )
  ))
  (define get-defines (lambda (source racc)
    (define gd-helper (lambda (source racc)
      "pluck bindings out of source lines of the form (define <name> <val>)"
      (cond
        ((null? (car source)) (reverse racc))
        ((pair? (car source))
          (if (eq? (car (car source)) 'define)
            (gd-helper (cdr source) (cons (car source) racc))
            ; else
	    (reverse racc)
          )
        )
        (else
          (reverse racc)
        )
      )
    ))
    (define glr-helper (lambda (bindings)
      "pluck bindings out of source lines of the form (<name> <val)"
      (map (lambda (v) (cons 'define v)) bindings) 
    ))

    ;(display "GETTING DEFINES FROM: ") (display source) (newline)
    (if (null? source)
      (reverse racc)
      ; else
      (if (pair? source)
        (if (string? (car source))
          (get-defines (cdr source) racc)
          ; else
          (if (pair? (car source))
            (if (eq? (car (car source)) 'define)
              (gd-helper source racc)
              ; else
              (if (eq? (car (car source)) 'letrec)
                (glr-helper (cadar source))
                (reverse racc)
              )
            )
            ; else
            (reverse racc)
          )
        )
	; else
        (reverse racc)
      )
    )
  ))
  (define get-body (lambda (source racc)
    (cond
      ((null? source) (reverse racc))
      ((null? (car source)) (get-body (cdr source) racc))
      ((pair? (car source))
        (if (eq? (car (car source)) 'define)
          (get-body (cdr source) racc)
          ; else
          (get-body (cdr source) (cons (car source) racc))
        )
      )
      ((string? (car source))
        (get-body (cdr source) racc)
      )
      (else
        (get-body (cdr source) racc)
      )
    )
  ))

  (define process-arg-use (lambda (source-expression local-namespace)
;    "Scrutinize the use of an argument"
    ;;(display "PROCESS-USE-ARG SOURCE EXPRESSION: ") (display source-expression) (newline)

    (cond
      ((list? source-expression)
        (if (eq? (car source-expression) arg-name)
          (cons procedure? *unspecified*)
          (let ()
            (define member-val (member arg-name source-expression))
            (if member-val
              (let ()
                (define arg-index (- (- (length source-expression) (length (member arg-name source-expression))) 1))
                ; find the procedure which is handed our arg
                (define flexed-procedure (let ()
                  (define term (car source-expression))
                  (if (symbol? term)
                    (if (assoc term local-namespace)
                      #f
                      (if (improper-member term formals)
                        #f
                        (if (not (unspecified? (procedure-ref closure term)))
                          (procedure-ref closure term)
                          (let ()
                            (define env (procedure-environment closure))
                            (or
                              (and env (> (length env) 0) 
                                (if (string=? (minor-version) "8")
                                  (let ()
                                    ; the resolver of last resort
                                    (define lr (remove #f (map (lambda (v) (and (procedure? v) v)) env)))
                                    (let helper ((cursor lr))
                                      (if (null? cursor)
                                        #f
                                        (let ()
                                          (define res ((car cursor) term #f))
                                          (and res (variable? res) (variable-ref res))
                                        )
                                      )
                                    )
                                  )
                                  ; else
                                  (let ()
                                    ; we should probably look at the non-closure forms
                                    (define m (environment-module (remove #f (map (lambda (v) (and (procedure? v) v)) env))))
                                    ; should we also search the module-uses?
                                    (if m
                                      (catch #t (lambda () (module-ref m term)) always-false)
                                      #f
                                    )
                                  )
                                )
                              )
                              (catch #t (lambda () (module-ref the-root-module term)) always-false)
                              (begin
                                (display "UNSOUND LOOKUP: ") (display term) (newline)
                                (eval term (current-module))
                              )
                            )
                          )
                        )
                      )
                    )
                    #f
                  )
                ))    

            ;(define flexed-procedure (if (assoc (car source-expression) local-namespace) #f (if (not (unspecified? (procedure-ref closure (car source-expression)))) (procedure-ref closure (car source-expression)) (eval (car source-expression) (current-module)))))

   	    ;(display "FLEXED PROCEDURE: ") (display flexed-procedure) (newline)
   	        ;(flush-all-ports)

	        ; we don't do macros currently...
	        (if (and flexed-procedure (procedure? flexed-procedure))
                  (let ()
                    (define args (procedure-property flexed-procedure 'args))
                    (if args
                      (let ()
		        ;(display "INDEX: ") (display arg-index) (newline)
                        ;(display "ARGS: ") (display args) (newline)
                        ;(display "arg-derive-all-from-source: RETURNING: ")
                        ;(write (list-ref (list-ref args arg-index) 2))
                        ;(display " -- ")
                        ;(write (procedure-source (list-ref (list-ref args arg-index) 2)))
                        ;(newline)
                        ;(list-ref (list-ref args arg-index) 2)
                        (cons (list-ref (list-ref args arg-index) 2) (list-ref (list-ref args arg-index) 3))
                      )
                      (if (closure? flexed-procedure)
		        ; try harder.
                        (let ()
                          (define sub-formals (cadr (procedure-source flexed-procedure)))
                          (define get-fixed-args (lambda (formals)
                            (define helper (lambda (l racc)
                              (if (pair? l)
                                (helper (cdr l) (cons (car l) racc))
                                (reverse racc)
                              )
                            ))
                            (helper formals '())
                          ))
                          (if (list? sub-formals)
		            (arg-derive-all-from-source (list-ref sub-formals arg-index) (procedure-source flexed-procedure) flexed-procedure)
                            (let ()
                              (define fixed-args (get-fixed-args sub-formals))
			      (if (< arg-index (length fixed-args))
  		                (arg-derive-all-from-source (list-ref fixed-args arg-index) (procedure-source flexed-procedure) flexed-procedure)
                                ; else it is an optional arg
                                (cons #t *unspecified*)
                              )
                            )
                          )
                        )
                        ; else (we should probably examine syntax as well)
                        (cons #t *unspecified*)
                      )
                    )
	          )
                  ; else
                  (cons #t *unspecified*)
                )
              )
              ; else it wasn't used directly here
              (let ()
                (define invoked (car source-expression))
                (cond
                  ((eq? invoked 'and) (let ()
                    (define subvals (map (lambda (v) (process-arg-use v local-namespace)) (cdr source-expression)))
                    ;(display "AN AND!\n")
		    ;(display "SUBVALS: ") (display subvals) (newline)
		    (cons #t *unspecified*)
                  ))
                  (else (cons #t *unspecified*))
                )
              )
            )
          )
        )
      )
      ((symbol? source-expression) (if (eq? source-expression arg-name) (cons #t *unspecified*) (cons #t *unspecified*)))
      (else (cons #t *unspecified*))
    )
  ))

  ;(display "DERIVING ") (display arg-name) (display " from ") (display (cddr closure-source)) (newline)
  
  (let ()
    (define results (remove #f (map (lambda (v)
      ; is the argument used directly in this call?
      (define direct-use-hit (if (and (pair? (list-ref v 2)) (not (eq? (car (list-ref v 2)) 'quote)) (not (eq? (car (list-ref v 2)) 'quasiquote)) (member arg-name (cdr (list-ref v 2)))) (list-ref v 2) #f))

      ;(display "DEFINE: ") (display v) (newline)
      ;(display "HIT: ") (display direct-use-hit) (newline)

      (if direct-use-hit
        (cond
          ; we are a shadow binding and used in an if
          ((and (eq? (list-ref v 1) arg-name) (eq? (car direct-use-hit) 'if))
            ;(display "WE'RE AN IF!\n")
            (cond
              ; are we the true case?
              ((eq? (caddr direct-use-hit) arg-name) (let ()
                (define result (eval (cons lambda (list (list arg-name) (cadr direct-use-hit))) (current-module)))

                (if (or (number? (cadddr direct-use-hit)) (string? (cadddr direct-use-hit)))
                  (cons result (cadddr direct-use-hit))
                  (cons result *unspecified*)  
                )
              ))
              ; are we the false case?
              ((eq? (cadddr direct-use-hit) arg-name)
                #f
              )
              ; no idea
              (else #f)
            )
          )
          (else
            ;(display "THE FIRST CASE\n")
            (process-arg-use direct-use-hit '())
          )
        )
        #f
      )
    ) (get-defines (cddr closure-source) '()))))

    ;(display "DEFINES: ") (display (get-defines (cddr closure-source) '())) (newline)
    ;(display "bogo-RESULTS: ") (display results) (newline)

    (if (null? results)
      (let ()
        (define body-results (get-body (cddr closure-source) '()))
        (if (null? body-results)
          (cons #t *unspecified*)
          ; else
          (if (eq? (length body-results) 1)
            (if (and (pair? (car body-results)) (not (null? (car body-results))))
              (let ()
                ;(display "ABOUT TO RUN PROCESS ARG USE WHILE LOOKING FOR: ") (display arg-name) (display " ")
		;(display body-results)
		;(newline)
	        ;(cons (process-arg-use (car body-results)) *unspecified*)
	        (process-arg-use (car body-results) (map (lambda (v) (cons (cadr v) (caddr v))) (get-defines (cddr closure-source) '())))
              )
              ; else
              (cons #t *unspecified*)
            )
            ; else
            (cons #t *unspecified*)
          )
        )
      )
      ; else 
      (car results)
    )
  )
  )
))

(define arg-derive-predicate-from-source (with-docstring
  "Attempt to return a predicate for the argument <arg-name: symbol?> declared in the formals of <closure: procedure?> with source <closure-source>"
  (lambda (arg-name closure-source closure)

  (define result (arg-derive-all-from-source arg-name closure-source closure))
  
  (if result
    (car result)
    #f
  )
  )
))

)
