(library (proverb diag)
  (export
    tire-diag
    http-diag
    port-diag
    xhtml-diag
    xhtml-test-characters
    xhtml-shame
    css-syntax-sampler
  )
  (import
    (rnrs)
    (rnrs lists)
    (rnrs control)
    (rnrs arithmetic bitwise)
    (hashtable) (proverb search)
    (proverb types)
    (proverb doc) (proverb wiki) (proverb help)
    (proverb uri)
    (proverb xml) (proverb remote) (proverb mime)
    (proverb tire) (proverb filesystem)
    (proverb html) (proverb http) (proverb http-cache) (proverb javascript)
    (proverb channels)
    (proverb cond-expand) 
    (proverb debug)
    (proverb xhtml)
    (ice-9 syncase)
    (ice-9 rdelim)
    (ice-9 documentation)
    (ice-9 format)
    (proverb dialect)
    (only (proverb guilefoo)
      module-get-symbol-list
    )
    (guile modules)
    (guile errors)
    (only (guile filesystem)
      ; filesystem
      opendir readdir mkdir access?
      W_OK X_OK F_OK R_OK
      stat
      stat:type
    )
    (only (guile)
      ; module-related imports
      environment-module the-environment
      ; other imports
      with-input-from-string
      with-output-to-string
      call-with-output-string
      call-with-input-string
      define-macro eval-case defmacro:transformer *unspecified*
      port-for-each file-port?
      call-with-deferred-observers list-set!
      closure?
      procedure-documentation
      object-property procedure-environment
      struct-set! %record-type-error object-address object->string
      procedure-source
      struct-ref set-defmacro-transformer! assert-defmacro?!
      variable-ref
      add-hook! struct? procedure-property
      delete hashq-set! variable-set! hash-set!
      sort
      hash-fold 
      port-closed? fileno
      getsockname
      open-file
      close
      get-output-string
      object-properties
      ; extra debugging
      stack-ref procedure-name source-properties
      memoized? unmemoize-expr
      after-gc-hook
      gc-stats
    )
    (ice-9 pretty-print)
  )

(define tire-diag (with-docstring "an interactive tool for exploring the tire mux" (lambda ()
  (define sockname->url (lambda (sockname)
    (case (vector-ref sockname 0)
      ((2) (format #f "tire://host:~a/" (vector-ref sockname 2)))
      (else "--unknown-format--")
    )
  ))
  (define c (tire-mux-get-cookie-hash (current-tire-mux)))
  (make-wiki (let ()
    (define cookies (map (lambda (v) (cons (hashtable-ref c v #f) v)) (vector->list (hashtable-keys c))))
    (define requests (tire-mux-get-requests (current-tire-mux)))
    (define default (tire-mux-get-default-channel (current-tire-mux)))
    (define listeners (tire-mux-get-listeners (current-tire-mux)))
    (apply string-append `(
      "Your current cookie is ''``!shell.get_eval_cookie();``''\n"
      "=== Cookies ===\n"
      ,@(map (lambda (v)
      (apply string-append
        (if (eq? (car v) default)
          (format #f "* channel ~a (this channel is the default)\n" (channel-get-desc (car v)))
          (format #f "* channel ~a [[[(tire-mux-set-default-channel! (current-tire-mux) (car (remove #f (map (lambda (w) (if (eq? (object-address w) ~a) w #f)) (tire-mux-get-channels (current-tire-mux))))))|make default]]]\n" (channel-get-desc (car v)) (object-address (car v)))
        )
        (map (lambda (w)
          (define symstr (symbol->string w))
          (string-append "** cookie " symstr " [[[>" symstr "|select]]] [[[(tire-mux-unhash-cookie! (current-tire-mux) (quote " symstr "))|delete]]]\n")
        ) (cdr v))
      )
      ) (aggregate-alist-for-assoc cookies))
      "=== Listeners ===\n"
      ,@(if (null? listeners)
        '("This mux has no dedicated listeners.\n")
        (map (lambda (v)
          (html-escape-string (format #f "* listener fd ~a -- ~a\n"
              (fileno v)
              (sockname->url (getsockname v))
          ))
        ) listeners)
      )
      "=== Outstanding Requests by Source ===\n"
      ,(format #f "The source is the [[[(help channel)|channel]]] from which the 'T' form was sourced and to which the 'R' and 'I' forms should be delivered. There are ~a outstanding tire requests.\n" (length requests))
      ,@(if (null? requests)
        '()
        ; aggregated by src channel
        (apply append (map (lambda (channel-agg)
          (cons
            (format #f "* channel ~a\n" (channel-get-desc (car channel-agg)))
            (map (lambda (rest)
              (format #f "** tag ~a, form ~a\n" (car rest) (tire-request-get-form rest))
            ) (cdr channel-agg))
          )
        ) (aggregate-alist-for-assoc (map (lambda (v) (cons (tire-request-get-src v) v)) requests))))
      )
      "=== Browser Diagnostics ===\n"
      "If you are connected through Replete, try [[[!mux_to_wiki(shell,shell.tire_mux)]]]\n"
    ))
  ))
)))

(define http-diag (with-docstring "an interactive tool for exploring the HTTP server"
  (lambda ()
    (define print-vfs (lambda (path vptr)
      (make-xhtml
        (make-text (or path "/"))
        (apply with-xml-tag "<ul>" (map (lambda (v)
          (with-xml-tag "<li>"
            (make-text (car v))
            (if (procedure? (cdr v))
              (make-text "handled dynamically")
              (print-vfs (car v) (cdr v))
            )
          )
        ) vptr))
      )
    ))
    (make-xhtml
      (make-text "HTTP server diagnostics\n")
      (make-wiki (format #f "The server is running at ~a. See for yourself with [[[(http-server-get-url (current-http-server))]]]\n" (http-server-get-url (current-http-server))))
      (with-xml-tag "<br>")
      (make-wiki (format #f "Request host: ~a from [[[(http-request-get-host (current-http-request))]]]\n" (http-request-get-host (current-http-request))))
      (make-wiki (format #f "Request auth: ~a from [[[(http-request-get-auth (current-http-request))]]]\n" (http-request-get-auth (current-http-request))))
      (let helper ((req-cursor http-requests) (gets 0) (posts 0) (connects 0) (unknowns 0))
        (if (null? req-cursor)
          (make-text (format #f
            "GET requests: ~a\nPOST requests: ~a\nCONNECT requests: ~a\nunknown requests: ~a\n"
            gets
            posts
            connects
            unknowns
          ))
          (case (http-request-get-method (car req-cursor))
            ((GET) (helper (cdr req-cursor)  (+ gets 1) posts connects unknowns))
            ((POST) (helper (cdr req-cursor)  gets (+ posts 1) connects unknowns))
            ((CONNECT) (helper (cdr req-cursor) gets posts (+ connects 1) unknowns))  
            (else (helper (cdr req-cursor) gets posts connects (+ unknowns 1)))
          )
        )
      )
      (make-text (string-merge (map (lambda (v)
        (define start-time (vector-ref v 1))
        (define end-time (vector-ref v 2))
        (if (and start-time end-time)
          (let ((elapsed-time (- (+ (* 1000000 (car end-time)) (cdr end-time)) (+ (* 1000000 (car start-time)) (cdr start-time)))))
            (format #f "elapsed: ~a req: ~a\n" elapsed-time v)
          )
          (format #f "request incomplete: req ~a\n" v)
        )
      ) (reverse http-requests)) #\newline))
      (print-vfs #f (http-server-get-vfs (current-http-server)))
      (with-xml-tag "<br>")
      (make-text (format #f "There are ~a forwarded ports\n" (length http-forwarded-ports)))
      (apply with-xml-tag "<div>" (map (lambda (v)
        (make-text (format #f "~a\n" v))
      ) http-forwarded-ports))
    )
  )
))

(define port-diag (with-docstring
  "rich text dump of info on each open port"
  (lambda ()
    (define ip (current-input-port))
    (define op (current-output-port))
    (define describe-port (lambda (p)
      (if (port? p)
        (let ()
          (define pt (if (file-port? p)
            "file port" 
            "non-file-port"
          ))
          (define stats (remove #f (list
            (and (eq? p ip) "this port is the current input port")
            (and (eq? p op) "this port is the current output port")
            (and (input-port? p) (char-ready? p) (format #f "this port is ready for reading with a ~a" (object->string (peek-char p))))
            (and (port? p) (port-closed? p) "this port is closed")
            (and (file-port? p) (format #f "UNIX fd ~a" (fileno p)))
            (and (not (file-port? p)) (format #f "does this help: ~a" p))
            (string-append "Properties: " (object->string (object-properties p)))
          )))
          (if (and (input-port? p) (output-port? p))
            (make-text (apply string-append pt " opened for input and output " stats))
            (if (input-port? p)
              (make-text (format #f "~a opened for input ~a\n" pt stats))
              (if (output-port? p)
                (make-text (format #f "~a opened for output ~a: ~a\n" pt stats (catch #t
                  (lambda ()
                    (if (string=? pt "non-file-port")
                      (get-output-string p)
                      ""
                    )
                  )
                  (lambda args 
                    (format #f "bogus: ~a\n" args)
                  )
                )))
                (make-text (format #f "~a opened in an unknown way ~a\n" pt stats))
              )
            )
          )
        )
        (make-text "this isn't a port at all\n")
      )
    ))
    (define l '())
    (port-for-each (lambda (v)
      (set! l (cons v l))
    ))
    (apply make-xhtml (make-text (format #f "There are ~a ports\n" (length l))) (map describe-port l))
  )
))

(define xhtml-diag (with-docstring
  "interactive diagnostics"
  (lambda ()
  (make-wiki "
* [[[(port-diag) | port diagnostics]]]
* [[[(http-diag) | HTTP server diagnostics]]]
* [[[(tire-diag) | TIRE diagnostics]]]
")
  )
))

(define xhtml-module-grep (with-docstring
  "search the reachable source in modules for term"
  (lambda (term . modules)
    (define make-term-shamer-with-exceptions (lambda (term test exceptions message)
      (define term-helper (lambda (l)
        (if (null? l)
          #f
          (if (list? l)
            (if (eq? (car l) term)
              (if (test l)
                #t
                (any? (map term-helper l))
              )
              ; this is a bit bogus
              (if (and (eq? (car l) 'apply) (not (null? (cdr l))) (eq? (cadr l) term))
                (if (test (cdr l))
                  #t
                  (any? (map term-helper l))
                )
                (if (eq? (car l) 'quasiquote)
                  #f
                  (any? (map term-helper l))
                )
              )
            )
            #f
          )
        )
      ))
      (define procedure-term-audit (lambda (proc-name proc-source)
        (cond
          ((null? proc-source) (make-text (format #f "proc ~a has an empty body\n" proc-name)))
          ((not (eq? (car proc-source) 'lambda))
            (make-text (format #f "proc ~a starts with something other than lambda\n" proc-name))
          )
          ((term-helper (cddr proc-source)) (make-wiki (format #f "proc [[[(xhtml-examine ~a)|~a]]] has ~a\n" proc-name proc-name message)))
          (else #f)
        )
      ))
      (lambda ()
        (define s (get-shame-symbols))
        (let ()
          (define r (remove #f (map (lambda (v)
            (define val (variable-ref (cdr v)))
            (if (closure? val)
              (if (member (car v) exceptions)
                #f
                (procedure-term-audit (car v) (procedure-source val))
              )
              #f
            )
          ) (sort s (lambda (v1 v2) (string>=? (symbol->string (car v1)) (symbol->string (car v2))))))))
          (make-xhtml
            (make-text (format #f "There are ~a procedures with ~a:\n" (length r) message))
            (apply make-xhtml (map (lambda (v)
              v
              ;(make-text (string-append v "\n"))
            ) r))
          )
        )
      )
    ))
    (define make-term-shamer (lambda (term test message)
      (make-term-shamer-with-exceptions term test '() message)
    ))
    (define make-simple-term-shamer (lambda (term)
      (make-term-shamer-with-exceptions term always-true '() (symbol->string term))
    ))
    (define get-shame-symbols (lambda ()
      (apply append (map (lambda (v)
        (module-get-symbol-list v)
      ) (if (null? modules) (list (current-module)) (map (lambda (v)
        (if (module? v) v (resolve-interface v))
      ) modules))))
    ))
    ((make-simple-term-shamer term))
  )
))

(define xhtml-shame (with-docstring
  "an interactive routine to help the authors of Proverb identify areas for improvement in the system"
  (lambda modules
    (define get-shame-symbols (lambda ()
      (apply append (map (lambda (v)
        (module-get-symbol-list v)
      ) (if (null? modules) (delete #f (map (lambda (v) (if (equal? (module-name v) '(guile)) #f v)) (cons (current-module) (module-uses (current-module))))))))
    ))
    (define undocumented-functions (lambda ()
    (define m (sort (get-shame-symbols) (lambda (v1 v2) (string<=? (symbol->string (car v1)) (symbol->string (car v2))))))
      (remove #f (map (lambda (v)
        (define val (variable-ref (cdr v)))
        (if (procedure? val)
          (if (procedure-documentation val)
            #f
            (if (object-documentation val)
              #f
              (car v)
            )
          )
        #f
        )
      ) m))
    ))
    (define untested-functions (lambda ()
      (define m (get-shame-symbols))
      (remove #f (map (lambda (v)
        (define val (variable-ref (cdr v)))
        (if (and (procedure? val) (closure? val))
          (if (object-property val 'unit-test)
            #f
            (car v)
          )
          #f
        )
      ) m))
    ))
    (define all-functions (lambda ()
    (define m (get-shame-symbols))
      (remove #f (map (lambda (v)
        (define val (variable-ref (cdr v)))
        (if (procedure? val)
          (car v)
          #f
        )
      ) m))
    ))

    (define single-char-shame (lambda ()
      (define procedure-singlechar-audit (lambda (proc-name proc-source)
        (cond
          ((null? proc-source) (format #f "proc ~a has an empty body" proc-name))
          ((not (eq? (car proc-source) 'lambda))
            (format #f "proc ~a starts with something other than lambda" proc-name)
          )
          ((and (list? (cadr proc-source)) (not (null? (cadr proc-source))))
            (if (not (all? (map (lambda (v) (> (string-length (symbol->string v)) 1)) (cadr proc-source))))
              (format #f "proc ~a has some single-character formals" proc-name)
              #f
            )
          )
          (else #f)
        )
      ))
      (define s (get-shame-symbols))
      (let ()
        (define r (remove #f (map (lambda (v)
          (define val (variable-ref (cdr v)))
          (if (closure? val)
            (procedure-singlechar-audit (car v) (procedure-source val))
            #f
          )
        ) (sort s (lambda (v1 v2) (string>=? (symbol->string (car v1)) (symbol->string (car v2))))))))
        (make-xhtml
          (make-text (format #f "There are ~a procedures with single-character formals\n" (length r)))
          (apply make-xhtml (map (lambda (v)
            (make-text (string-append v "\n"))
          ) r))
        )
      )
    ))

    (define make-term-shamer-with-exceptions (lambda (term test exceptions message)
      (define term-helper (lambda (l)
        (if (null? l)
          #f
          (if (list? l)
            (if (eq? (car l) term)
              (if (test l)
                #t
                (any? (map term-helper l))
              )
              ; this is a bit bogus
              (if (and (eq? (car l) 'apply) (not (null? (cdr l))) (eq? (cadr l) term))
                (if (test (cdr l))
                  #t
                  (any? (map term-helper l))
                )
                (if (eq? (car l) 'quasiquote)
                  #f
                  (any? (map term-helper l))
                )
              )
            )
            #f
          )
        )
      ))
      (define procedure-term-audit (lambda (proc-name proc-source)
        (cond
          ((null? proc-source) (make-text (format #f "proc ~a has an empty body\n" proc-name)))
          ((not (eq? (car proc-source) 'lambda))
            (make-text (format #f "proc ~a starts with something other than lambda\n" proc-name))
          )
          ((term-helper (cddr proc-source)) (make-wiki (format #f "proc [[[(xhtml-examine ~a)|~a]]] has ~a\n" proc-name proc-name message)))
          (else #f)
        )
      ))
      (lambda ()
        (define s (get-shame-symbols))
        (let ()
          (define r (remove #f (map (lambda (v)
            (define val (variable-ref (cdr v)))
            (if (closure? val)
              (if (member (car v) exceptions)
                #f
                (procedure-term-audit (car v) (procedure-source val))
              )
              #f
            )
          ) (sort s (lambda (v1 v2) (string>=? (symbol->string (car v1)) (symbol->string (car v2))))))))
          (make-xhtml
            (make-text (format #f "There are ~a procedures with ~a: ~a:\n" (length r) term message))
            (apply make-xhtml (map (lambda (v)
              v
              ;(make-text (string-append v "\n"))
            ) r))
          )
        )
      )
    ))
    (define make-term-shamer (lambda (term test message)
      (make-term-shamer-with-exceptions term test '() message)
    ))
    (define make-simple-term-shamer (lambda (term)
      (make-term-shamer-with-exceptions term always-true '() "This is the interactive Proverb style guide and we don't like it!")
    ))

    (let ()
      (define needless-begin-shame (make-term-shamer 'begin (lambda (l) (eq? (length l) 2)) "single term begin"))
      (define bindless-let-shame (make-term-shamer 'let (lambda (l) (and (> (length l) 3) (null? (cadr l)) (or (not (list? (caddr l))) (not (eq? (car (caddr l)) 'define))))) "let without bindings"))
      (define elseless-if-shame (make-term-shamer 'if (lambda (l)
        (and
          (eq? (length l) 3)
          (let ()
            (define truecase (list-ref l 2))
            (and 
              (list? truecase)
              (not (null? truecase))
              (symbol? (car truecase))
              ; let them skate if its a conditional debug message
              (not (eq? (car truecase) 'debugmsg))
              (not (eq? (car truecase) 'critmsg))
              (let ()
                (define symstr (symbol->string (car truecase)))
                (and
                  (> (string-length symstr) 0)
                  ; let them skate if their if is for a side effect
                  (not (eq? (string-ref symstr (- (string-length symstr) 1)) #\!))
                )
              )
            )
          )
        )
      ) "elseless if (without debugmsg or obvious side effect)"))
      (define elseless-or-shame (make-term-shamer 'or (lambda (l) (eq? (length l) 2)) "elseless or"))
      (define elseless-and-shame (make-term-shamer 'and (lambda (l) (eq? (length l) 2)) "elseless and"))
      (define make-raw-shame (make-term-shamer-with-exceptions 'make-raw (lambda (l) #t) '(with-xml-tag make-raw-text) "make-raw"))
      (define static-lambda-shame (make-term-shamer 'lambda (lambda (l)
        (if (and (list? l) (> (length l) 2))
          (if (or (eq? (cadr l) '()) (symbol? (cadr l)))
            (if (or
                  (equal? (cddr l) '(#t))
                  (equal? (cddr l) '(#f))
                  (equal? (cddr l) '(()))
                  (and (eq? (length l) 3) (number? (caddr l)))
                  (and (eq? (length l) 3) (string? (caddr l)))
                )
              #t
              (begin
                ;(display l emergency-port) (newline emergency-port)
                #f
              )
            )
            #f
          )
          #f
        )
      ) "statically computable thunk"))
      (define needless-curry-thunk-shame (make-term-shamer 'lambda (lambda (l)
        (if (and (list? l) (eq? (length l) 3))
          (if (and (list? (caddr l))
                   ; it's at least a list
                   (>= (length (caddr l)) 1)
                   ; it could be a call
                   (symbol? (car (caddr l)))
                   ; it exactly passes our args
                   (equal? (cdr (caddr l)) (cadr l)))
            #t
            #f
          )
          #f
        )
      ) "needless curry thunk"))
      (define simple-shamers (map (lambda (v)
        (cons v (if (pair? v)
          (make-term-shamer (car v) always-true (cdr v))
          (make-simple-term-shamer v)
        ))
      ) '(
          (select . "select does not work well on non-fd ports and is difficult for mortals to use well")
          call-with-current-continuation studio-select write-char read read-char with-output-to-string
          with-input-from-string getenv setenv format sleazy-tag sleazy-root-mux sleazy-continuation sleazy-escape
          read-hash-extend procedure-property
          (fileno . "fileno behaves likes its namesake in the C library and works well only on some Scheme ports. Its presence is evidence that you are probably about to get in trouble.")
         )))
      (define a (all-functions))
      (define u (undocumented-functions))
      (define t (untested-functions))
      (make-xhtml
        (make-wiki (apply string-append
          (format #f "There are ~a undocumented functions: \n\n" (length u))
          (append
            (map (lambda (v)
              (format #f "* [[[(xhtml-examine ~a)|~a]]]\n" v v)
            ) u)
            (list (format #f "There are ~a out of ~a functions without a test case: \n" (length t) (length a)))
            (map (lambda (v)
              (format #f "* ~a\n" v) 
            ) (sort t (lambda (v1 v2) (string<=? (symbol->string v1) (symbol->string v2)))))
            (list "\n\n")
          )
        ))
        (with-xml-tag "<br>")
        (single-char-shame)
        (with-xml-tag "<br>")
        (needless-begin-shame)
        (with-xml-tag "<br>")
        (elseless-if-shame)
        (with-xml-tag "<br>")
        (elseless-and-shame)
        (with-xml-tag "<br>")
        (elseless-or-shame)
        (with-xml-tag "<br>")
        (bindless-let-shame)
        (with-xml-tag "<br>")
        (static-lambda-shame)
        (with-xml-tag "<br>")
        (needless-curry-thunk-shame)
        (with-xml-tag "<br>")
        (make-raw-shame)
        (with-xml-tag "<br>")
        (apply with-xml-tag "<span>" (map (lambda (v)
          (with-xml-tag "<span>"
            ((cdr v))
            (with-xml-tag "<br>")
          )
        ) simple-shamers))
      )
    )
  )
))

(define xhtml-test-characters (lambda ()
  (define grid-helper (lambda (inner)
    (apply with-xml-tag 'table (let row-helper ((row-cursor 0) (racc '()))
      (if (equal? row-cursor 16)
        (reverse racc)
        (row-helper
          (+ row-cursor 1)
          (cons
            (apply with-xml-tag 'tr (with-xml-tag 'td (make-text (format #f "~a: " row-cursor))) (let col-helper ((col-cursor 0) (col-racc '()))
              (if (equal? col-cursor 16)
                (reverse col-racc)
                (col-helper
                  (+ col-cursor 1)
                  (cons
                    (with-xml-tag 'td (inner (+ (* 16 row-cursor) col-cursor)))
                    col-racc
                  )
                )
              )
            ))
            racc
          )
        )
      )
    ))
  ))
  (with-xml-tag 'div
    (make-text "Character integer values\n")
    (grid-helper (lambda (v) v))
    (make-text "Character printable representations\n")
    (grid-helper (lambda (v) (integer->char v)))
    (make-text "Strings with embedded characters\n")
    (grid-helper (lambda (v) (string (integer->char v))))
    (make-text "xml-body-escaped-string version\n")
    (grid-helper (lambda (v) (xml-body-escape-string (string (integer->char v)))))
    (make-text "text node version with embedded characters\n")
    (grid-helper (lambda (v) (make-text (string (integer->char v)))))
  )
))

(define css-syntax-sampler (with-wiki-docstring
  "(css-syntax-sampler) -- show a list of styled Scheme expressions"
  (lambda ()
    (make-xhtml
      (with-xml-tag "<h4>" (make-text "A list"))
      (list 1 2 3)
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A vector"))
      (vector 1 2 3)
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A pair"))
      (cons 1 2)
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A boolean"))
      #t 
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A number"))
      128000 
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A character"))
      #\a 
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A string"))
      "a" 
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "A symbol"))
      'a 
      (with-xml-tag "<br>")
      (with-xml-tag "<h4>" (make-text "An unspecified"))
       *unspecified*
      (with-xml-tag "<br>")
    )
  )
))


)
