;
; Set up some supplemental runtime documentation
;

(for-each (lambda (v) (set-procedure-property! v 'has-side-effect 'no)) (list
  cdr cddr car cons pair? port? string? null? modulo object-address eq? > >= < <=
  length vector-length make-vector + null? vector not assoc logand eof-object?
  integer->char string-length uniform-vector-length string-ref vector-ref
  reverse procedure-source procedure-name procedure-documentation procedure-property
  defined? closure? object->string delete current-module list? list struct?
  string<? symbol->string string-split
))

(for-each (lambda (v) (set-procedure-property! v 'has-side-effect #t)) (list
  display flush-all-ports newline write-char read-char socketpair open-file force-output open-input-file
  open-output-file
))

(set-procedure-property! display 'args (list
  (list "object" "the object to display" #t *unspecified*)
))
(set-procedure-property! display 'oargs (list
  (list "port" "display object through port ..." port? *unspecified*)
))
(set-procedure-property! display 'rtype unspecified?)

;
; Standard Guile funcs
;

(define supplement-docs (lambda (v d)
  (if (not (or (procedure-documentation v) (procedure-property v 'documentation) (object-property v 'documentation)))
    (set-object-property! v 'documentation d)
  )
))

(supplement-docs * "* -- (* args ...) -- standard Scheme -- multiplies its arguments")
(supplement-docs + "+ -- (+ args ...) -- standard Scheme -- adds its arguments")
(supplement-docs - "- -- (+ args ...) -- standard Scheme -- subtracts its arguments")
(supplement-docs / "/ -- (+ args ...) -- standard Scheme -- divides its arguments")
(supplement-docs < "< -- (< args ...) -- standard Scheme -- is each argument less than the following argument?")
(supplement-docs <= "<= -- (<= args ...) -- standard Scheme -- is each argument less than or equal to the following argument?")
(supplement-docs = "= -- (= args ...) -- standard Scheme -- is each argument numerically equivalent to the following argument?")
(supplement-docs > "> -- (> args ...) -- standard Scheme -- is each argument greater than the following argument?")
(supplement-docs >= ">= -- (>= args ...) -- standard Scheme -- is each argument greater than or equal to the following argument?")
(supplement-docs acos "acos -- (acos arg) -- standard Scheme -- return arccosine of arg")
(supplement-docs angle "angle -- (angle arg) -- standard Scheme -- return the angle portion of the polar coordinate arg")
(supplement-docs magnitude "magnitude -- (magnitude arg) -- standard Scheme -- return the magnitude portion of the polar coordinate arg")
(supplement-docs apply "apply -- (apply proc (args ...)) -- standard Scheme -- apply the arguments args ... to the Scheme procedure proc. For example (apply + l) would sum the numbers in the list l.")
(supplement-docs car "car -- (car arg) -- standard Scheme -- return the first element of a pair")
(supplement-docs caar "caar -- (caar arg) -- standard Scheme -- return (car (car arg))")
(supplement-docs cadar "cadar -- (cadar arg) -- standard Scheme -- return (car (cdr (car arg)))")
(supplement-docs caddar "caddar -- (caddar arg) -- standard Scheme -- return (car (cdr (cdr (car arg))))")
(supplement-docs cadddr "cadddr -- (cadddr arg) -- standard Scheme -- return (car (cdr (cdr (cdr arg))))")
(supplement-docs caddr "caddr -- (caddr arg) -- standard Scheme -- return (car (cdr (cdr arg)))")
(supplement-docs cadr "cadr -- (cadr arg) -- standard Scheme -- return (car (cdr arg))")
(supplement-docs cdr "cdr -- (cdr arg) -- standard Scheme -- return the second element of a pair")
(supplement-docs cdar "cdar -- (cdar arg) -- standard Scheme -- return (cdr (car arg))")
(supplement-docs cddr "cddr -- (cddr arg) -- standard Scheme -- return (cdr (cdr arg))")
(supplement-docs cddar "cddar -- (cddar arg) -- standard Scheme -- return (cdr (cdr (car arg)))")
(supplement-docs cdddr "cdddr -- (cdddr arg) -- standard Scheme -- return (cdr (cdr (cdr arg)))")
(supplement-docs cdddar "cdddar -- (cdddar arg) -- standard Scheme -- return (cdr (cdr (cdr (car arg))))")
(supplement-docs cddddr "cddddr -- (cddddr arg) -- standard Scheme -- return (cdr (cdr (cdr (cdr arg))))")
(supplement-docs call-with-current-continuation "call-with-current-continuation -- (call-with-current-continuation arg) -- standard Scheme -- see a Scheme language reference")
(supplement-docs call-with-values "call-with-values -- ... -- standard Scheme -- see a Scheme language reference")
(supplement-docs call/cc "call/cc -- (call/cc arg) -- semi-standard Scheme -- shorthand for call-with-current-continuation")
(supplement-docs map "map -- standard Scheme")
(supplement-docs max "max -- standard Scheme")
(supplement-docs min "min -- standard Scheme")
(supplement-docs modulo "modulo -- standard Scheme")
(supplement-docs quotient "quotient -- standard Scheme")
(supplement-docs remainder "remainder -- standard Scheme")
(supplement-docs denominator "denominator -- standard Scheme")
(supplement-docs numerator "numerator -- standard Scheme")
(supplement-docs exact->inexact "exact->inexact -- standard Scheme")
(supplement-docs expt "expt -- standard Scheme")
(supplement-docs for-each "for-each -- standard Scheme")
(supplement-docs gcd "gcd -- standard Scheme")
(supplement-docs asin "asin -- (asin arg) -- standard Scheme -- return arcsine of arg")
(supplement-docs atan "atan -- (atan arg) -- standard Scheme -- return the arctangent of arg")
(supplement-docs tan "tan -- (tan arg) -- standard Scheme -- return tangent of arg")
(supplement-docs cos "cos -- (cos arg) -- standard Scheme -- return cosine of arg")
(supplement-docs sin "sin -- (sin arg) -- standard Scheme -- return sine of arg")
(supplement-docs lcm "lcm -- standard Scheme")
(supplement-docs zero? "zero? -- standard Scheme")
(supplement-docs negative? "negative? -- standard Scheme")
(supplement-docs positive? "positive? -- standard Scheme")
(supplement-docs display "display -- standard Scheme")
(supplement-docs write "write -- standard Scheme")
(supplement-docs real-part "real-part -- standard Scheme")
(supplement-docs imag-part "imag-part -- standard Scheme")
(supplement-docs vector-ref "vector-ref -- standard Scheme -- (vector-ref vec ind) -- return the indth element of vector vec")
(supplement-docs vector-set! "vector-set! -- standard Scheme -- (vector-set! vec ind obj) -- set the indth element of vector vec to the Scheme value obj")
(supplement-docs vector-length "vector-length -- standard Scheme -- (vector-length vec) -- return the length of the vector vec")

(use-modules (rnrs arithmetic bitwise))
(supplement-docs bitwise-arithmetic-shift "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-bit-count "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-bit-field "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-bit-set? "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-copy-bit "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-copy-bit-field "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-first-bit-set "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-if "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-length "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-reverse-bit-field "from R6RS (r6rs arithmetic bitwise)")
(supplement-docs bitwise-rotate-bit-field "from R6RS (r6rs arithmetic bitwise)")

(set-procedure-property! $cos 'args (list
  (list "operand" "a real number in radians" real? *unspecified*)
))
(set-procedure-property! $cos 'rtype real?)

(set-procedure-property! cos 'args (list
  (list "operand" "a real or imaginary number in radians" number? *unspecified*)
))
(set-procedure-property! cos 'rtype number?)

(set-procedure-property! > 'largs (list
  (list "operand" "a real number" real? *unspecified*)
))
(set-procedure-property! > 'rtype boolean?)

(set-procedure-property! < 'largs (list
  (list "operand" "a real number" real? *unspecified*)
))
(set-procedure-property! < 'rtype boolean?)

(set-procedure-property! >= 'largs (list
  (list "operand" "a real number" real? *unspecified*)
))
(set-procedure-property! >= 'rtype boolean?)

(set-procedure-property! <= 'largs (list
  (list "operand" "a real number" real? *unspecified*)
))
(set-procedure-property! <= 'rtype boolean?)


(set-procedure-property! / 'rtype number?)
(set-procedure-property! * 'rtype number?)
(set-procedure-property! + 'rtype number?)
(set-procedure-property! - 'rtype number?)


(set-procedure-property! vector-ref 'args (list
  (list "source vector" "" vector? *unspecified*)
  (list "index" "an integer" integer? *unspecified*)
))
(set-procedure-property! vector-ref 'has-side-effect 'no)

(set-procedure-property! vector-length 'args (list
  (list "v" "" vector? *unspecified*)
))
(set-procedure-property! vector-length 'has-side-effect 'no)


(set-procedure-property! load 'documentation "load a Scheme file")
(set-procedure-property! load 'args (list
  (list "Scheme file" "" (lambda (v) (access? v R_OK)) *unspecified*)
))

(set-procedure-property! chdir 'documentation "change directory")
(set-procedure-property! chdir 'args (list
  (list "Directory name" "" file-is-directory? *unspecified*)
))
(set-procedure-property! chdir 'has-side-effect #t)

(set-procedure-property! open 'args (list
  (list "File" "" (lambda (v) (access? v R_OK)) *unspecified*)
  ;(list "Mode" "" number? *unspecified*)
  (list "Mode" "" (lambda (v) (member v
                    (list 
                      O_APPEND O_CREAT O_EXCL O_NDELAY O_NOCTTY
		      O_NONBLOCK O_RDONLY O_RDWR O_TRUNC O_WRONLY
                    )
                  ))
  *unspecified*)
))
(set-procedure-property! open 'has-side-effect #t)

(set-procedure-property! opendir 'args (list
  (list "Dir path" "" (lambda (v) (access? v R_OK)) *unspecified*)
))
(set-procedure-property! opendir 'has-side-effect #t)

(set-procedure-property! open-file 'args (list
  (list "Filename" "" (lambda (v) (or (access? v R_OK) (string? v))) *unspecified*)
  (list "Filename" "" (lambda (v) (or (access? v R_OK) (string? v))) *unspecified*)
  (list "Mode" "" string? *unspecified*)
))
(set-procedure-property! open-file 'has-side-effect #t)

;
; List calls
;
(set-procedure-property! list-ref 'args (list
  (list "list" "" list? *unspecified*)
  (list "index" "" integer? *unspecified*)
))
(set-procedure-property! list-ref 'has-side-effect 'no)

;
; String manipulation calls
;
(set-procedure-property! string-length 'args (list
  (list "string" "" string? *unspecified*)
))
(set-procedure-property! string-length 'has-side-effect 'no)

;
; Port calls
;
(set-procedure-property! close 'args (list
  (list "port" "" port? *unspecified*)
))
(set-procedure-property! close 'has-side-effect #t)

;
; Networking calls
;
(set-procedure-property! socket 'args (list
  (list "family" "" (lambda (v) (member v (list AF_UNIX AF_INET AF_INET6))) *unspecified*)
  (list "style" "" (lambda (v) (member v (list SOCK_STREAM SOCK_DGRAM SOCK_RAW))) *unspecified*)
  (list "proto" "" (lambda (v) (integer? v)) *unspecified*)
))
(set-procedure-property! socket 'has-side-effect #t)

(set-procedure-property! connect 'args (list
  (list "socket" "" (lambda (v) (port? v)) *unspecified*)
  (list "family" "" (lambda (v) (member v (list AF_UNIX AF_INET AF_INET6))) *unspecified*)
  (list "address" "" #t *unspecified*)
))
(set-procedure-property! connect 'has-side-effect #t)


;
; system calls
;
(set-procedure-property! primitive-fork 'has-side-effect #t)
(set-procedure-property! primitive-fork 'rtype integer?)

(set-procedure-property! make-regexp 'args (list
  (list "exp" "" string? *unspecified*)
))


(set-procedure-property! procedure-documentation 'args (list
  (list "p" "" procedure? *unspecified*)
))
(set-procedure-property! procedure-documentation 'has-side-effect #f)

(set-procedure-property! read-char 'oargs (list
  (list "p" "" input-port? *unspecified*)
))
(set-procedure-property! read-char 'has-side-effect #t)
(set-procedure-property! read-char 'rtype (lambda (v) (or (char? v) (eof-object? v))))

(set-procedure-property! write-char 'args (list
  (list "v" "" char? *unspecified*)
))
(set-procedure-property! write-char 'oargs (list
  (list "p" "" output-port? *unspecified*)
))
(set-procedure-property! write-char 'has-side-effect #t)
