(library (proverb games)
  (export
    make-chessboard
    default-chessboard-bindings
  )
  (import
    (rnrs)
    (ice-9 format)
    (proverb doc) (proverb wiki)
    (proverb xml)
    (proverb filesystem)
    (proverb html) (proverb javascript)
    (proverb xhtml)
    (proverb debug)
    (only (proverb bootstrap) support-dir)
  )

(define default-chessboard-bindings '(
  (a 1 WR) (b 1 WNG) (c 1 WB) (d 1 WQ) (e 1 WK) (f 1 WB) (g 1 WNG) (h 1 WR)
  (a 8 BR) (b 8 BNG) (c 8 BB) (d 8 BQ) (e 8 BK) (f 8 BB) (g 8 BNG) (h 8 BR)
  (a 2 WP) (b 2 WP) (c 2 WP) (d 2 WP) (e 2 WP) (f 2 WP) (g 2 WP) (h 2 WP)
  (a 7 BP) (b 7 BP) (c 7 BP) (d 7 BP) (e 7 BP) (f 7 BP) (g 7 BP) (h 7 BP)
))

(define make-chessboard (with-docstring
  "generate an XHTML chessboard with pieces at the locations in <bindings>. A sample binding is '(a 1 WR) -- placing the white rook in A 1. The entire set of default chessboard bindings is available in default-chessboard-bindings."
  (lambda (bindings)
    (define sym->icon (lambda (s)
      (case s
        ((BB) "/artwork/chess/bb.gif") ; black bishop
        ((BK) "/artwork/chess/bk.gif") ; black king
        ((BNB) "/artwork/chess/bnb.gif") ; black knight facing left
        ((BNG) "/artwork/chess/bng.gif") ; black knight facing right
        ((BP) "/artwork/chess/bp.gif") ; black pawn
        ((BQ) "/artwork/chess/bq.gif") ; black queen
        ((BR) "/artwork/chess/br.gif") ; black rook
        ((WB) "/artwork/chess/wb.gif") ; white bishop
        ((WK) "/artwork/chess/wk.gif") ; white king
        ((WNB) "/artwork/chess/wnb.gif") ; white knight facing left
        ((WNG) "/artwork/chess/wng.gif") ; white knight facing right
        ((WP) "/artwork/chess/wp.gif") ; white pawn
        ((WQ) "/artwork/chess/wq.gif") ; white queen
        ((WR) "/artwork/chess/wr.gif") ; white rook
        (else "/bogo")
      )
    ))
    (define lookup (lambda (row col)
      (let helper ((c bindings))
        (if (null? c)
          (make-text "")
          (if (and (eq? col (caar c)) (eq? row (cadar c)))
            (make-image (filename->string (string-append support-dir "/" (sym->icon (caddar c)))) 'width 32)
            (helper (cdr c))
          )
        )
      )
    ))
    (define make-styled-row (lambda (row-num even-style odd-style)
      (with-xml-tag 'tr
        (with-xml-tag `(td class ,even-style width 32 height 32) (lookup row-num 'a))
        (with-xml-tag `(td class ,odd-style width 32 height 32) (lookup row-num 'b))
        (with-xml-tag `(td class ,even-style width 32 height 32) (lookup row-num 'c))
        (with-xml-tag `(td class ,odd-style width 32 height 32) (lookup row-num 'd))
        (with-xml-tag `(td class ,even-style width 32 height 32) (lookup row-num 'e))
        (with-xml-tag `(td class ,odd-style width 32 height 32) (lookup row-num 'f))
        (with-xml-tag `(td class ,even-style width 32 height 32) (lookup row-num 'g))
        (with-xml-tag `(td class ,odd-style width 32 height 32) (lookup row-num 'h))
      )
    ))
    (define make-row (lambda (row-num)
      (if (even? row-num)
        (make-styled-row row-num 'lightSquare 'darkSquare)
        (make-styled-row row-num 'darkSquare 'lightSquare)
      )
    ))
    (define chessboard-id (make-dom-id))

    (make-xhtml
      (with-xml-tag "<style>" (make-raw-text (format #f "
        #~a .lightSquare { 
          background-color: #ffffff;
        }
        #~a .darkSquare { 
          background-color: #446644;
        }
      " chessboard-id chessboard-id)))
      (with-xml-tag `(table border 0 cellspacing 0 id ,chessboard-id)
        (make-row 8)
        (make-row 7)
        (make-row 6)
        (make-row 5)
        (make-row 4)
        (make-row 3)
        (make-row 2)
        (make-row 1)
        (with-xml-tag "<tr>"
          (with-xml-tag "<td align='center'>" (make-text "a"))
          (with-xml-tag "<td align='center'>" (make-text "b"))
          (with-xml-tag "<td align='center'>" (make-text "c"))
          (with-xml-tag "<td align='center'>" (make-text "d"))
          (with-xml-tag "<td align='center'>" (make-text "e"))
          (with-xml-tag "<td align='center'>" (make-text "f"))
          (with-xml-tag "<td align='center'>" (make-text "g"))
          (with-xml-tag "<td align='center'>" (make-text "h"))
        )
      )
    )
  )
))

)
