(library (proverb getopt)
  (export
    getopt 
    make-boolean-option*
    make-boolean-variable-option
    make-integer-variable-option
    make-string-variable-option
  )
  (import
    (rnrs)
    (proverb debug)
    (ice-9 format)
    (ice-9 syncase)
    (only (guile) cond-expand string-index procedure-properties list-head)
  )

(define make-boolean-option* (lambda (name true-text set-true set-false)
  (list
    (list (string-append "--" name) set-true true-text)
    (list (string-append "--no" name) set-false (string-append "do not " true-text))
  )
))

(define-syntax make-boolean-variable-option (syntax-rules ()
  ((foo name true-text var)
    (make-boolean-option* name true-text
      (lambda () (set! var #t))
      (lambda () (set! var #f))
    )
  )
  ((foo name true-text var setvar)
    (make-boolean-option* name true-text
      (lambda () (critmsg "SETTING\n") (set! var #t) (set! setvar #t))
      (lambda () (critmsg "UNSETTING\n") (set! var #f) (set! setvar #t))
    )
  )
))

(define-syntax make-integer-variable-option (syntax-rules ()
  ((foo name text var)
    (list (string-append "--" name) (lambda (v) (set! var (string->number v))) text)
  )
))

(define-syntax make-string-variable-option (syntax-rules ()
  ((foo name text var)
    (list (string-append "--" name) (lambda (v) (set! var v)) text)
  )
))



(define getopt (lambda (options arguments)

  (let cmdline-helper ((cursor arguments))
    (cond
      ((null? cursor)
        'done
      )
      ((assoc (car cursor) options) => (lambda (a)
(cond-expand
(guile
        (define n (cadr (assoc 'arity (procedure-properties (cadr a)))))
        (define m (cadddr (assoc 'arity (procedure-properties (cadr a)))))
)
(else
        (define n 0)
        (define m #f)
        (format #t "WISH YOU HAD ARITY!\n")
)
)
        (if m
          (let ((next (apply (cadr a) (cdr cursor))))
            (cmdline-helper next)
          )
          (begin
            (apply (cadr a) (cdr (list-head cursor (+ n 1))))
            (cmdline-helper (list-tail cursor (+ n 1)))
          )
        )
      ))
      ; split option names with an '='
      ((and (string? (car cursor)) (string-index (car cursor) #\=)) => (lambda (s)
        (critmsg "FOO!\n")
        (cmdline-helper (cons (substring (car cursor) 0 s) (cons (substring (car cursor) (+ s 1)) (cdr cursor))))
      ))
      (else 'done)
    )
  )

))

)
