(define-module (rnrs))
(use-modules (guile))

; from the index of R5
(module-re-export! (current-module) '(

  ; ' and , are excluded
  ;! * + , - / < <= = => > >= ?
  ;* + , - / < <= = => > >= ?
  ;* + - / < <= = => > >= ?
  ;* + - / < <= = > >= ?
  * + - / < <= = > >=

  abs acos and angle append apply asin assoc assq assv atan

  begin boolean?

  caar cadr cadar caddr caddar cadddr call-with-current-continuation call-with-input-file
  call-with-output-file call-with-values call/cc
  car case catch cdar cddr cdddr cdddar cddddr cdr ceiling char->integer ;; cdar may not actually be defined
  char-alphabetic? char-ci<=? char-ci<? char-ci=?
  char-ci>=? char-ci>? char-downcase char-lower-case?
  char-numeric? char-ready? char-upcase char-upper-case?
  char-whitespace? char<=? char<? char=? char>=? char>?
  char? close-input-port close-output-port
  complex? cond cons cos current-input-port current-output-port

  ;define define-syntax delay denominator display do
  define delay denominator display do
  ;dynamic-wind else eof-object? eq? equal? eqv? eval
  dynamic-wind eof-object? eq? equal? eqv? eval
  even? exact->inexact exact? exp expt

  ;#f floor for-each force
  floor for-each force

  gcd

  if imag-part inexact->exact inexact?
  input-port? integer->char integer? 
  interaction-environment lambda
  ;lcm length let let* let-syntax letrec
  lcm length let let* letrec
  ;letrec-syntax list list->string
  list list->string
  list->vector list-ref list-tail
  list? load log

  magnitude make-polar make-rectangular make-string make-vector
  map max member memq memv min modulo

  ;negative? newline nil not null-environment
  ;negative? newline not null-environment
  negative? newline not
  null? number->string number? numerator

  odd? open-input-file open-output-file or output-port?

  pair? peek-char port? positive? procedure?

  quasiquote quote quotient

  rational? rationalize read read-char real-part real? remainder reverse
  round

  ;scheme-report-environment set! set-car! set-cdr! setcar sin sqrt
  ;set! set-car! set-cdr! setcar sin sqrt
  set! set-car! set-cdr! sin sqrt
  string string->list string->number string->symbol string-append
  string-ci<= string-ci<? string-ci=? string-ci>=? string-ci>?
  string-copy string-fill! string-length string-ref string-set!
  string<=? string<? string=? string>=? string>? string? substring
  ;symbol->string symbol? syntax-rules
  symbol->string symbol? 

  ;tan transcript-off transcript-on
  ;tan transcript-on
  tan
  truncate

  ;unquote unquote-splicing
  ;unquote-splicing

  values vector vector-set! vector? vector-ref vector->list vector-length

  with-input-from-file with-output-to-file write write-char

  zero?
))

(define-macro (library name export import . body)
  `(eval
    '(begin
      (set-object-property!
        (current-module)
        'library-filename
        (cond
          ((current-load-port) => (lambda (p) (port-filename p)))
          (else #f)
        )
      )
      (set-module-uses! (current-module) '())
      (,define library ,library)
      ,@(if (null? (cdr import))
        '()
        (delete #f (map (lambda (v)
          (if (list? v)
            (if (null? v)
              #f
              (if (eq? (car v) 'only)
                `(,begin ,@(map (lambda (w) 
                   (let ((m (resolve-interface (cadr v))))
                     (if m
                       (begin
                         (let ((u (module-variable m w)))
                           (if u
                             `(,module-add! (,current-module) (,quote ,w) (,module-variable ,m (,quote ,w)))
                             `(error ,(format #f "error creating library ~a: no symbol ~a in interface ~a" name w (cadr v)))
                           )
                         )
                       )
                       `(error ,(format #f "couldn't resolve interface: ~a" (cadr v)))
                     )
                   )
                 ) (cddr v)))
                `(,process-use-modules (,quote ((,v))))
              )
            )
            #f
          )
        ) (cdr import)))
      )
      ,@body
      ,@(if (null? (cdr export))
        '()
        `((let ((mmm (,current-module))) ,@(map (lambda (v)
          `(if (,module-locally-bound? mmm (,quote ,v))
            (,module-export! mmm (,quote (,v)))
            (,module-re-export! mmm (,quote (,v)))
           )
        ) (cdr export))))
      )
    )
    (process-define-module (list (quote ,name)))
  )
)


; our R6isms
;(module-export! (current-module) '(library))
(export-syntax library)

(define-public vector-map (lambda (proc . vectors)
  (if (null? vectors)
    (error "vector map must be passed at least one vector\n")
    (if (apply equal? (map vector-length vectors))
      (list->vector (apply map proc (map vector->list vectors)))
      (error "vectors must be the same length")
    )
  )
))

(define-module (rnrs lists))
(use-modules (guile))
(define-public remove delete)

(define-module (rnrs files))
  (use-modules (guile))
  (if (defined? 'delete-file)
    (re-export delete-file)
    (define-public delete-file (lambda (filename)
      (system (string-append "rm '" pathname "'"))
    ))
  )
  (if (defined? 'file-exists?)
    (re-export file-exists?)
    (define-public file-exists? (lambda (filename)
      (or
        (access? filename R_OK)
        (access? filename W_OK)
        (access? filename X_OK)
        (catch #t
          (lambda ()
            (and (stat filename) #t)
          )
          (lambda args
            #f
          )
        )
      )
    ))
  )

(define-module (rnrs control))
(use-modules (guile))
(define-macro (unless test . expressions) `(if (not ,test) (begin ,@expressions) *unspecified*))
(define-public unless unless)

(define-macro (when test . expressions) `(if ,test (begin ,@expressions) *unspecified*))
(define-public when when)

(define-module (rnrs ports))
(use-modules (guile))
(define-public eof-object (lambda () the-eof-object))
(re-export eof-object? eof-object?)
(re-export current-input-port)
(re-export current-output-port)
(re-export current-error-port)

(define-module (rnrs io simple))
(use-modules (guile))
(use-modules (rnrs ports))
(re-export eof-object)
(re-export eof-object?)
(re-export current-input-port)
(re-export current-output-port)
(re-export current-error-port)

(define-module (rnrs programs))
(use-modules (guile))
(re-export command-line)
(re-export exit)
