;
; R6 compatible hashtables
;
(library (hashtable)
  (export
    make-eq-hashtable
    make-eqv-hashtable
    make-hashtable
    hashtable?
    hashtable-size
    hashtable-ref
    hashtable-set!
    hashtable-delete
    hashtable-contains?
    hashtable-update!
    hashtable-hash-function
    hashtable-mutable?
    hashtable-keys
    equal-hash
    string-hash
    string-ci-hash
    symbol-hash
  )
  (import (rnrs) (proverb doc) (proverb debug) (guile))

(define guile-make-hash-table make-hash-table)
(define guile-hash-for-each hash-for-each)
(define guile-hash-fold hash-fold)

(define make-eq-hashtable (with-docstring "R6RS compatibility"
  (lambda args
    (if (null? args)
      (vector 'r6-hashtable eq? hashq hashq-ref hashq-set! hashq-remove! (guile-make-hash-table))
      (vector 'r6-hashtable eq? hashq hashq-ref hashq-set! hashq-remove! (guile-make-hash-table (car args)))
    )
  )
))

(define make-eqv-hashtable (with-docstring "R6RS compatibility" 
  (lambda args
    (if (null? args)
      (vector 'r6-hashtable eqv? hashv hashv-ref hashv-set! hashv-remove! (guile-make-hash-table))
      (vector 'r6-hashtable eqv? hashv hashv-ref hashv-set! hashv-remove! (guile-make-hash-table (car args)))
    )
  )
))

(define make-hashtable (with-docstring "R6RS compatibility" 
  (lambda (hash-function equiv . args)
    (if (null? args)
      (vector 'r6-hashtable equiv hash-function #f #f #f (guile-make-hash-table))
      (vector 'r6-hashtable equiv hash-function #f #f #f (guile-make-hash-table (car args)))
    ) 
  )   
))      

(define hashtable? (with-docstring "R6RS compatibility"
  (lambda (v)
    (and (vector? v) (eqv? (vector-length v) 7) (eq? (vector-ref v 0) 'r6-hashtable))
  )
))

(define hashtable-size (with-docstring "R6RS compatibility"
  (lambda (v)
    #f
  )
))

(define hashtable-ref (with-docstring "R6RS compatibility"
  (lambda (hashtable key default)
    ((vector-ref hashtable 3) (vector-ref hashtable 6) key default)
  )
))

(define hashtable-set! (with-docstring "R6RS compatibility"
  (lambda (hashtable key obj)
    ((vector-ref hashtable 4) (vector-ref hashtable 6) key obj)
  )
))

(define hashtable-delete (with-docstring "R6RS compatibility"
  (lambda (hashtable key)
    ((vector-ref hashtable 5) (vector-ref hashtable 6) key)
  )
))

(define hashtable-contains? (with-docstring "R6RS compatibility"
  (lambda (hashtable key)
    (critmsg "hashtable-contains\n")
    #f
  )
))

(define hashtable-update! (with-docstring "R6RS compatibility"
  (lambda (hashtable key proc default )
    (critmsg "hashtable-update!")
    #f
  )
))

(define hashtable-hash-function (with-docstring "R6RS compatibility"
  (lambda (hashtable)
    (critmsg "hashtable-hash-function!")
    #f
  )
))

(define hashtable-mutable? (with-docstring "R6RS compatibility"
  (lambda (hashtable)
    (critmsg "hashtable-mutable?")
    #f
  )
))

(define hashtable-keys (with-docstring "R6RS compatibility"
  (lambda (hashtable)
    (list->vector (guile-hash-fold (lambda (k v a)
      (cons k a)
    ) '() (vector-ref hashtable 6)))
  )
))

(define equal-hash (with-docstring "R6RS compatibility"
  (lambda (obj)
    (critmsg "equal-hash")
    #f
  )
))

(define string-hash (with-docstring "R6RS compatibility"
  (lambda (obj)
    (critmsg "string-hash")
    #f
  )
))

(define string-ci-hash (with-docstring "R6RS compatibility"
  (lambda (obj)
    (critmsg "string-ci-hash")
    #f
  )
))

(define symbol-hash (with-docstring "R6RS compatibility"
  (lambda (obj)
    (critmsg "symbol-hash")
    #f
  )
))

)
