;
; Routines for inspecting large strings
;
;
(library (proverb hexdump)
  (export
    hexdump-string hexdump-strings hexdump-anything hexalign hexdump-inner*
    hexdump-live-string
    linedump-string
  )
  (import
    (rnrs) (proverb xml)
    (proverb doc) (proverb passval) (proverb dialect) (proverb html)
    (proverb remote)
    (proverb itree)
    (proverb xhtml)
    (proverb debug)
    (only (guile)
      make-regexp the-environment object-address list-set! eval-case
      run-hook regexp-exec substring/shared
      make-list
      list-head list-tail
      string-join
    )
    (guile modules)
    (guile errors)
    (ice-9 regex)
    (ice-9 format)
  )

(define hexdump-char (lambda (c)
  (case c
    ((
      #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9
      #\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
      #\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z
      #\A #\B #\C #\D #\E #\F #\G #\H #\I #\J #\K #\L #\M
      #\N #\O #\P #\Q #\R #\S #\T #\U #\V #\W #\X #\Y #\Z
      #\! #\" #\# #\$ #\% #\& #\' #\( #\) #\* #\+ #\, #\-
      #\. #\/ #\: #\; #\< #\= #\> #\? #\@ #\[ #\\ #\] #\^
      #\_ #\` #\{ #\| #\} #\~ #\space
     )
      c
    )
    (else #\.)
  )  
))

(define ffixed (lambda (n w)
  (define ns (number->string n 16))
  (string-append (make-string (- w (string-length ns)) #\0) ns)
))

(define hexdump-line-formatter (lambda (index width total-size total-width obj)
  (define l (string->list obj))
  (with-xml-tag "<span>"
    (with-xml-tag "<span style='color: grey'>" (make-text (string-append (ffixed index total-size) ": ")))
    (make-text (hexline (map char->integer l)))
    (make-text "  ")
    (make-text (string-append (list->string (map hexdump-char l)) (if (< width total-width) "" "")))
  )
))

(define hexdump-inner* (lambda (objsub objlen start len width)
  (define sln (string-length (number->string (+ 1 objlen) 16)))
  (define helper (lambda (index hlen racc)
    (define w (if (< hlen width) hlen width))
    (if (or (>= index objlen) (<= w 0))
      (reverse racc)
      (helper (+ index w) (- hlen w) (cons (hexdump-line-formatter index w sln width (objsub index (+ index w))) racc))
    )
  ))
  (define breaker (lambda (l rest)
    (let breaker-helper ((cursor l) (racc '()))
      (if (null? cursor)
        (if (null? rest)
          (reverse racc)
          (append (reverse racc) rest)
        )
        (breaker-helper (cdr cursor) (cons (make-raw "<br/>") (cons (car cursor) racc)))
      )
    ) 
  ))
  (apply with-xml-tag "<div style='font-family:courier,fixed;'>"
    (if (>= start objlen)
      (list (format #f "start index ~a exceeds length ~a" start objlen))
      (if (> (+ start len) objlen)
        ; truncate
        (breaker (helper start (- objlen start) '()) (list (with-xml-tag "<span style='color: red;'>" (make-text "END"))))
        (breaker (helper start len '()) '())
      )
    )
  )
))

(define hexdump-inner-wrap (lambda (objlen objsub start len width func)
  (catch #t (lambda ()
    (func objlen objsub start len width)
  ) (lambda args
    (with-xml-tag "<div>"
	(apply format #f (string-append "HEXDUMP ERROR: " (list-ref args 2)) (list-ref args 3))
    )
  ))
))

(define hexdump-string (lambda (str)
  (hexdump-anything (lambda (start finish) (substring str start finish))
                    (string-length str)
                    make-regexp
                    (lambda (re i)
                      (regexp-exec re str i))
                    hexdump-inner*
  )
))

(define hexdump-live-string (lambda (str)
  (dribble (hexdump-anything (lambda (start finish) (substring str start finish))
                    (string-length str)
                    make-regexp
                    (lambda (re i)
                      (regexp-exec re str i))
                    hexdump-inner*
  ))
  (lambda args
    #f
  )
))

(define hexdump-strings (lambda strings
  (hexdump-anything (lambda (start finish) "FOO")
                    (apply min (map string-length strings))
                    #f
                    #f
                   (lambda (objsub objlen start len width)
                     (apply sidebyside width start (map (lambda (v)
                       (substring v start (+ start len))
                     ) strings))
                   )
  )
))

(define linedump-string (lambda (s)
  (define lines (studio-string-split s #\newline))
  (hexdump-anything
    (lambda (start finish) "FOO")
    (length lines)
    #f
    #f 
    (lambda (objsub objlen start len width)
      (critmsg "LEN: ~a\n" len)
      (with-xml-tag '(span style "font-family: fixed;")
        (make-text (string-join (list-head (list-tail lines start) len) "\n"))
      )
    )
  )
))

(define hexdump-anything (lambda (objsub objlen objcomp objexec supplied-dumper)
  (define current-index 0)
  (define search-re #f)
  (define search-itree #f)
  (define dumper (lambda (start len width)
    (set! current-index start)
    (list 
      (hexdump-inner-wrap objsub objlen start len width (or supplied-dumper hexdump-inner*))
      start
    )
  ))
  (define bogo-search (lambda (i)
    (define search-result (objexec search-re i))
    (if search-result
      (begin
        (set! current-index (car (vector-ref search-result 1)))
        (itree-insert-interval search-itree i current-index current-index)
      )
      (begin
        (itree-insert-interval search-itree i #f #f)
      )
    )
    current-index
  ))
  (define update-search (lambda (term)
    (set! search-re (objcomp term))
    (set! search-itree (make-itree))
    (bogo-search current-index)
  ))
  (define clear-search (lambda ()
    (set! search-re #f)
    (set! search-itree #f)
    current-index
  ))
  (define next-match (lambda (j)
    (define i (+ j 1))
    (if search-itree
      (let ()
        (define n (itree-next search-itree i))
        (if n
          (if (eq? n #t)
            ; the tree uncertain about the next match
            ; try to fill it in starting from the last match
            (let ()
              (define p (itree-prev search-itree i))
              (if p
                (if (eq? p #t)
                  ; this is bogus. This returns the earliest match, not
                  ; necessarily the preceeding match
                  (bogo-search 0)
                  (bogo-search (+ p 1))
                )
                ; there was a node with value f
                #f
              )
            )
            (set! current-index n)
          )
          ; there was a node with value #f
          #f
        )
      )
      ; else there is no search term
    )
    current-index  
  ))
  (define prev-match (lambda (i)
    (if search-itree
      (let ()
        (define n (itree-prev search-itree i))
        (if n
          (if (eq? n #t)
            (bogo-search 0)
            (set! current-index n)
          )
          ; there was no previous match
          #f
        )
        current-index
      )
      ; else there is no search term. Go to the beginning?
      0
    )
  ))

  ; flex some bindings to make these variables appear used
  prev-match next-match
  clear-search update-search
  dumper

  (widgetscript-apply "
    function(sl,hassearch,chunk_size) {

      var current=false;
      var current_index=0;
      var width=16;
      var div=widget.div;
      var current_search='';
      var match_index=0;
      var lock=0;
      var font_size=12;
      var dumpdiv;

      dumpdiv=document.createElement('div');
      dumpdiv.style.fontSize=font_size+'px';

      var painter=function(v) {
        try {
          try {
            if(current) {
              dumpdiv.removeChild(current);
            }
          } catch (e) { 123; };
          dumpdiv.appendChild(v);
          dheight=dumpdiv.clientHeight;
          current=v;
        } catch (e) { 
          123;
        }
        lock--;
      };

      div.addEventListener('DOMMouseScroll',function (e) {
        lock++;
        var increment;
        if(portable_shift_key_p(e)) {
          increment=1;
        } else {
          increment=chunk_size;
        }
        e.stopPropagation();
        e.preventDefault();
        if(e.detail) {
          var delta;
          delta = -e.detail/3;
          if(delta > 0) {
            current_index-=increment;
          } else if(delta < 0) {
            current_index+=increment;
          }
          if(lock==1) {
            redo();
          }
        }
      },false);

      var wrapfunc=function(scheme_name) {
        return function() {
          var s='(' + scheme_name + ' ';
          var x,v;
          for(x=0 ; x < arguments.length ; x++) {
            v=arguments[x];
            if(typeof(v) == 'string') {
              s+=''+scheme_stringify(v)+' ';
            } else {
              s+=''+v+' ';
            }
          }
          s+=')';
          widget.shell.inner_evaluator.eval_to_js('] ' + widget.wrapper(s),widget.shell,function(v) {
            painter(v[0]);
            current_index=v[1];
          });
        }
      };

      var dumper=wrapfunc('dumper');

      var next,prev,bumpup,bumpdown;
      var bigger,smaller;
      var diag,reload;
      var nextmatch,prevmatch;
      var width_box,chunk_box;
      var search_box;

      var dwidth=false;
      var dheight=false;
    
      function redo() {
        if(current_index>=sl) {
          current_index-=chunk_size;
        }
        if(current_index < 0) {
          current_index=0;
        }
        dumper(current_index,chunk_size,width);
      };

      function make_button(name,handler) {
        var button=document.createElement('input');
        button.type='button';
        button.value=name;
        button.onclick=function(e) {
          handler(e); redo();
        };
        return button;
      }

      function make_stringbox(handler,v) {
        var box=document.createElement('input');
        box.type='text';
        box.value=v;
        box.onchange=function(e) {
          handler(box.value);
          redo();
        }
        return box; 
      }

      function make_intbox(handler,v) {
        var box=document.createElement('input');
        box.type='text';
        box.value=v;
        box.width=12;
        box.onchange=function(e) {
          handler(parseInt(box.value));
          redo();
        }
        return box; 
      }

      width_box=make_intbox(function(v) { width=v; },width);
      chunk_box=make_intbox(function(v) { chunk_size=v; },chunk_size);

      next=make_button('next',function(e) { current_index+=chunk_size; });
      prev=make_button('prev',function(e) { current_index-=chunk_size; });

      bigger=make_button('bigger',function(e) {
        font_size++;
        dumpdiv.style.fontSize=font_size+'px';
        dheight=dumpdiv.clientHeight;
      });
      smaller=make_button('smaller',function(e) {
        font_size--;
        dumpdiv.style.fontSize=font_size+'px';
        dheight=dumpdiv.clientHeight;
      });
      {
        var resizeo_lock=false;
        var resizeo_timeout=400;
        var resizeo = function(preheight,mytries) {
          var resizeo_finalize = function() {
            dwidth=dumpdiv.clientWidth;
            dheight=dumpdiv.clientHeight;
            resizeo_lock=false;
            /*alert('resizeo done -- out of tries');*/
          }
          resizeo_lock=true;
          if(mytries > 0 && font_size > 0 && dumpdiv.clientHeight > dheight) {
            /*alert('resizeo -- height increased with work to do and ' + mytries + ' tries');*/
            font_size--;
            dumpdiv.style.fontSize=font_size+'px';
            if(dumpdiv.clientHeight == preheight) {
              /* the font size reduction may not
               * reduce the actual size in a single
               * step.
               */
              if(mytries-- > 0) {
                setTimeout(function() { resizeo(preheight,mytries); },resizeo_timeout);
              } else {
                resizeo_finalize();
              }
            } else {
              setTimeout(function() { resizeo(preheight,3); },resizeo_timeout);
            }
          } else {
            resizeo_finalize();
          }
        }
        window.addEventListener('resize',function(e) {
          if(!dheight) {
            dheight=dumpdiv.clientHeight;
          }
          if(dheight) {
            if(!resizeo_lock) {
              resizeo(dumpdiv.clientHeight,3);
            } else {
              /*alert('resizing locked');*/
            }
          }
        },false);
      }
      diag=make_button('diagnostics',function(e) {
        alert('diagnostics: width: ' + dumpdiv.clientWidth +
              '  scroll width: ' + dumpdiv.scrollWidth +
              '  height:       ' + dumpdiv.clientHeight +
              '  scroll height: ' + dumpdiv.scrollHeight +
              '  font size: ' + font_size
        );
      });

      if(hassearch) {
        function update_search(term) {
          var c;
          if(term == '') {
            c = widget.wrapper('(clear-search)');
          } else {
            c = widget.wrapper('(update-search ' + scheme_stringify(term) + ')');
          }
          widget.shell.inner_evaluator.eval_to_js('] ' + c,widget.shell,function(nci) {
            if(nci) current_index=nci;
            redo();
            return false;
          });
        }

        search_box=make_stringbox(function(v) {
          current_search=v;
          match_index=0;
          update_search(v);
        },current_search);

        nextmatch=make_button('next match',function(e) { 
          widget.shell.inner_evaluator.eval_to_js('] ' + widget.wrapper('(next-match ' + current_index + ')'),widget.shell,function(nci) {
            if(nci)
              current_index=nci;
            redo();
            return false;
          });
        });
        prevmatch=make_button('prev match',function(e) {
          widget.shell.inner_evaluator.eval_to_js('] ' + widget.wrapper('(prev-match ' + current_index + ')'),widget.shell,function(nci) {
          if(nci)
              current_index=nci;
            redo();
            return false;
          });
        });
      }

      bumpup=make_button('+1',function(e) { current_index+=1; });
      bumpdown=make_button('-1',function(e) { current_index-=1; });
     
      {
        var buttons = [chunk_box,width_box,bumpdown,bumpup,prev,next,bigger,smaller,diag];
        for ( var i in buttons ) {
          div.appendChild(buttons[i]);
        }
      }

      div.appendChild(document.createElement('br'));

      if(hassearch) {
        div.appendChild(search_box);
        div.appendChild(prevmatch);
        div.appendChild(nextmatch);
         
        div.appendChild(document.createElement('br'));
      }

      div.appendChild(dumpdiv);

      redo();

    }
  " (list objlen (if objexec #t #f) (min objlen 512)))
))

(define hexalign (with-docstring
  "accept a string M, a list of indicies I, and constants BEFORE and AFTER. return a presentation of the substrings from IsubN - BEFORE to IsubN + AFTER."
  (lambda (m i before after)
    (with-xml-tag "<div style='font-family:courier,fixed;'>"
      ;(make-raw (apply string-append `(
      ;  "<table><tr><td>INDEX</td><td>BYTES</td></tr>"
      ;  ,@(map (lambda (v)
      ;    (string-append "<tr><td>" (number->string v 16) "</td>" (apply string-append (map (lambda (v) (format #f "<td>~a</td>" (hexdump-char v))) (string->list (substring m (- v before) (+ v after))))) "</tr>")
      ;  ) i)
      ;  "</table>"
      ;)))
      (apply string-append `(
        "INDEX,BYTES\n"
        ,@(map (lambda (v)
          (string-append (number->string v 16) ":" (apply string (map hexdump-char (string->list (substring m (- v before) (+ v after))))) "\n")
        ) i)
      ))
    )
  )
))

(define hexit (lambda (i)
  (string-append (if (< i 16)
    (string-append "0" (number->string i 16))
    (number->string i 16)
  ) " ")
))

(define hexit-nospace (lambda (i)
  (if (< i 16)
    (string-append "0" (number->string i 16))
    (number->string i 16)
  )
))

(define hexline (lambda (l)
  (let helper ((cursor l) (racc '()))
    (if (null? cursor)
      (apply string-append (reverse racc))
      (helper (cdr cursor) (cons ((if (null? (cdr cursor))
        hexit-nospace
        hexit
      ) (car cursor)) racc))
    )
  )
))

(define sidebyside (lambda (w actual-start . strings)
  (define sls (map string-length strings))
  
  (let ()
    (define spanify (lambda (l mode)
      (with-xml-tag (if mode "<span style='color: green'>" "<span style='color: red'>")
        (make-text (hexline l))
      )
    ))
    (define maxl (apply min sls))
    (define maxlsl (string-length (number->string (+ actual-start (apply min sls) 16))))
    (apply with-xml-tag "<div style='font-family: courier,fixed'>"
      (let helper ((i 0) (racc '()))
        (define subs (map (lambda (v)
          (map char->integer (string->list (substring/shared v i (+ i w))))
        ) strings))
        (if (>= (+ i w) maxl)
          (apply append (reverse racc))
          (helper (+ i w) (cons 
            (let subhelper ((cursors subs) (raccs (make-list (length subs) '())) (spanraccs (make-list (length subs) '())) (mode #t))
              (if (member '() cursors)
                (list (apply with-xml-tag "<span>"
                  (with-xml-tag "<span style='color: grey'>" (make-text (string-append (ffixed (+ actual-start i) maxlsl) ": ")))
                  (append
                    (apply append (map (lambda (r s)
                      (list
                        (apply with-xml-tag "<span>" (reverse (cons (spanify (reverse r) mode) s)))
                        (make-text " ")
                      )
                    ) raccs spanraccs))
                    (list (make-raw "<br/>"))
                  )
                ))
                (if (eq? mode (apply eq? (map car cursors)))
                  (subhelper (map cdr cursors)
                             (map (lambda (c r) (cons (car c) r)) cursors raccs)
                             spanraccs
                             mode
                  )
                  (subhelper (map cdr cursors)
                             (map (lambda (c) (list (car c))) cursors)
                             (map (lambda (r s)  
                               (cons (spanify r mode) s)
                             ) raccs spanraccs)
                             (not mode)
                  )
                )
              )
            )
          racc))
        )
      )
    )
  )
))

)
