;
; Support routines for HTML on ports
;

(library (proverb html)
  (export html-escape-char html-escape-string html-quote-escape-string)
  (import (rnrs) (proverb doc) (only (guile) cond-expand))

(define html-escape-char (with-docstring
  "accept a character and return the string containing an HTML-safe form"
  (lambda (char)
    (case char
      ((#\<) "&lt;")
      ((#\>) "&gt;")
      ((#\&) "&amp;")
      ((#\") "&quot;")
      ((#\') "&apos;")
      ; this may not be quite the right thing
      ((#\nul) "")
      (else (string char))
    )
  )
))

(define html-escape-string (with-docstring
  "accept a string and return the string containing an HTML-safe form"
  (lambda (str)
    (cond-expand 
      (chicken 
        ;chicken is sensitive to apply calls with parameters >120 
        ;some options exist to remedy this however. see chicken README
        (string-translate*
  	  str
          '(("<"  . "&lt;")
            (">"  . "&gt;")
            ("&"  . "&amp;")
            ("\"" . "&quot;")
            ("'"  . "&apos;")))
      )
      (else
        (apply string-append (map html-escape-char (string->list str)))
      )
    )
  )
))

(define html-quote-escape-string (with-docstring
  "accept a Scheme string <str> and escape any double quotes to the HTML quot entity"
  (lambda (str)
    (let helper ((cursor (string->list str)) (racc '()))
      (if (null? cursor)
        (apply string (reverse racc))
        (cond
          ((eq? (car cursor) #\")
            (helper (cdr cursor) (append '(#\; #\t #\o #\u #\q #\&) racc))
          )
          ((eq? (car cursor) #\')
            (helper (cdr cursor) (append '(#\; #\9 #\3 #\# #\&) racc))
          )
          (else
            (helper (cdr cursor) (cons (car cursor) racc))
          )
        )
      )
    )
  )
))

)
