(library (proverb http-upload)
  (export
    http-build-uploader
    http-build-system-uploader
    http-uploads
  )
  (import
    (rnrs)
    (proverb doc)
    (proverb patterns)
    (proverb http)
    (proverb debug)
  )

(define sleazy-global-uploads '())

(define http-uploads (lambda ()
  sleazy-global-uploads
))

(define extract-boundary (lambda (ct)
  (define t "multipart/form-data; boundary=")
  (if (> (string-length ct) (string-length t))
    (if (string=? t (substring ct 0 (string-length t)))
      (substring ct (string-length t))
      #f
    )
    #f
  )
))

(define extract-entity (lambda (l)
  (let helper ((cursor l))
    (cond
      ((bogo-string-match "\r\n\r\n" cursor) => (lambda (match)
        (cons '() (list->string match))
      ))
      (else (if (null? cursor)
        #f
        (helper (cdr cursor))
      ))
    )
  )
))

(define split-multipart (lambda (multipart boundary)
  (define l (string->list multipart))
  (define lb (string->list (string-append "--" boundary)))

  (let helper ((lcursor l) (bcursor lb) (mcount 0) (restarts '()) (body '()) (bodies '()))
    (if (null? lcursor)
      (begin
        (if (null? bcursor)
          (critmsg "TERMINATED INPUT ON AN BOUNDARY\n")
          (critmsg "TERMINATED INPUT NOT ON A BOUNDARY\n")
        )
        bodies
      )
      (if (null? bcursor)
        (begin
          (critmsg "COMPLETED A MATCH ~a\n" mcount)
          (cond
            ((bogo-string-match "\r\n" lcursor) => (lambda (match)
              (helper (cddr lcursor) lb 0 '() '() (cons (extract-entity (reverse body)) bodies))
            ))
            ((bogo-string-match "--\r\n" lcursor) => (lambda (match)
              (if (null? match)
                (reverse (cons (extract-entity (reverse body)) bodies))
                (begin
                  (critmsg "MULTIPART WITH TRAILING CRAP!\n")
                  #f
                )
              )
            ))
            (else (critmsg "ERROR!\n"))
          )
        )
        (if (eq? (car lcursor) (car bcursor))
          (begin
            (if (eq? bcursor lb)
              (helper (cdr lcursor) (cdr bcursor) (+ mcount 1) (cons (cons (cdr lcursor) (cons (car lcursor) body)) restarts) body bodies)
              (helper (cdr lcursor) (cdr bcursor) (+ mcount 1) restarts body bodies)
            )
          )
          (if (> mcount 0)
            (begin
              (if (null? restarts)
                (helper (cdr lcursor) lb 0 '() (cons (car lcursor) body) bodies)
                (helper (caar restarts) lb 0 (cdr restarts) (cdar restarts) bodies) 
              )
            )
            (helper (cdr lcursor) lb 0 restarts (cons (car lcursor) body) bodies)
          )
        )
      )
    )
  )
))

(define http-build-uploader (with-docstring
  "construct an multipart upload gizmo that passes uploaded elements to the function <val>"
  (lambda (val)
    (lambda (continuation inport outport path args)
      ; args contains the body of the post
      (define r (current-http-request))
      (let ((headers (vector-ref r 8)))
        (let ((ct (assoc 'content-type headers)))
          (if ct
            (let ((boundary (extract-boundary (cdr ct))))
              (if (not boundary)
                (begin
                  (critmsg "http upload multipart -- no boundary!\n")
                  (display "HTTP/1.1 200 OK\n" outport)
                  (display "\n" outport)
                  (continuation)
                )
                (let ((m (split-multipart args boundary)))
                  (val m)
                  (display "HTTP/1.1 200 OK\n" outport)
                  (display "\n" outport)
                  (continuation)
                )
              )
            )
            (begin
              (display "HTTP/1.1 200 OK\n" outport)
              (display "\n" outport)
              (continuation)
            )
          )
        )
      )
    )
  )
))

(define http-build-system-uploader (lambda ()
  (http-build-uploader (lambda (val)
    (set! sleazy-global-uploads (cons val sleazy-global-uploads))
  ))
))

)
