;
; I/O forwarding 
;

;
; What's in a forwarding context?
;
; a from descriptor, a to descriptor, a buffer, a size, and whether or
; not the to is local
;

(library (proverb ioforward)
  (export)
  (import
    (rnrs)
    (proverb tire) (proverb channels) (proverb doc) (proverb remote)
    (proverb dialect)
    (proverb debug)
    (only (guile) close fileno getpid)
    (ice-9 rdelim)
  )

;(define make-forwarding-context (with-docstring
;  "one of a series of routines for managing IO forwarding between ports. Accepts a from port or fd, a to port or fd, and a boolean that specified whether the 'to' endpoint is local."
;  (lambda (from to to-local?)
;    (define from-fd (if (integer? from) from (fileno from)))
;    (define to-fd (if (integer? to) to (fileno to)))
;    (list from to (make-string 4096 (integer->char 0)) 0 0 to-local?)
;  )
;))

;(define forward-from (with-docstring
;  "return the from value from a forwarding context"
;   (lambda (c) (list-ref c 0))
;))
;(define forward-to (with-docstring
;  "return the to value from a forwarding context"
;  (lambda (c) (list-ref c 1))))
;(define forward-buffer (with-docstring
;  "return the buffer from a forwarding context"
;  (lambda (c) (list-ref c 2))))
;(define forward-count (with-docstring
;  "return the count (length of pending IO stored in the buffer) from a forwarding context"
;  (lambda (c) (list-ref c 3))))
;(define forward-start (with-docstring
;  "return the start index of pending IO stored in the buffer of a forwarding context" (lambda (c) (list-ref c 4))))
;(define forward-to-local (with-docstring
;  "return the value of the to-local? flag in the forwarding context"
;  (lambda (c) (list-ref c 5))))
;
;(define forward-write (with-docstring
;  "do the write IO necessary to drain the buffer of the forwarding context c"
;  (lambda (c)
;  (define l (forward-count c))
;  (let ()
;    (define retval (if (> l 0)
;      (let () 
;        (define count (catch #t (lambda () (write-string/partial (forward-buffer c) (forward-to c) 0 l)) (lambda args 0)))
;        ;(if (not (eq? count 1))
;        ;)
;        count
;      )
;      0
;    ))
;    (list-set! c 4 0)
;    (list-set! c 3 0)
;    retval
;  )
;)))

;(define forward-read (with-docstring
;  "do the read IO necessary to drain the read ports of the forwarding context c"
;  (lambda (c)
;  ; catch the 'connection reset by peer case'
;  (define count (catch #t
;    (lambda ()
;      (read-string!/partial (forward-buffer c) (forward-from c))
;    )  
;    (lambda args
;      0
;    )
;  ))
;  (list-set! c 3 (or count 0))
;  count
;)))
;(define SHUT_RD 0)
;(define SHUT_WR 1)
;(define SHUT_RDWR 2)
;(define forward-close (with-docstring
;  "close the endpoints associated with the forwarding context"
;  (lambda (c)
;  (catch #t (lambda () (shutdown (forward-from c) SHUT_RD)) (lambda args #f))
;  (catch #t (lambda () (shutdown (forward-to c) SHUT_WR)) (lambda args #f))
;  )
;))

(define port->cookie (with-docstring
  "turn the Scheme port <port> into a cookie in the root tire mux. This call is made inside a TIRE evaluation environment. See the procedure source for up-to-date details on the supported cookie commands."
  (lambda (port)
  (define mux (current-tire-mux))
  (define forward-reading? #f)
  (define reading-queue '())
  (define forward-writing? #f)
  (define writing-queue '())
  (define channel-locked? #f)
  (define run-if-unlocked (lambda (thunk)
    (if channel-locked?
      #f
      (begin
        (set! channel-locked? #t)
        (catch #t thunk (lambda args (debugmsg "channel error while locked\n")))
        (set! channel-locked? #f)
      )
    )
  ))
  (define dispatch (lambda (form)
    (case (car form)
      ((t) (let ()
        (define tag (cadr form))
        (define cookie (caddr form))
        (define command (cadddr form))
        (define respond (lambda (v)
          (tire-mux-enqueue mux channel `(r ,tag ,v))
        ))
        (case command
          ((close) (begin
            (close port)
            (set! forward-reading? #f)
            (set! forward-writing? #f)
            (respond 'ok)
          ))
          ((ready?) (if (char-ready? port)
              (respond #t)
              (respond #f)
          ))
          ((read-char) 
            (if (and port (char-ready? port))
              (respond (read-char port))
              (begin
                (set! forward-reading? #t)
                (set! reading-queue (append reading-queue (list
                  (lambda ()
                    (define val (read-char port))
                    (if (eof-object? val)
                      (begin
                        (set! port #f)
                        (tire-mux-enqueue mux channel `(r ,tag EOF))
                      )
                      (respond val)
                    )
                  )
                )))
              ) 
          ))
          ((read-stream) (let ()
            (define helper (lambda ()
              (define val (read-char port))
              (if port
                (if (eof-object? val)
                  (begin
                    (set! port #f)
                    (tire-mux-enqueue mux channel `(r ,tag done))
                  )
                  (begin
                    (tire-mux-enqueue mux channel `(i ,tag stream ,(string val)))
                    (set! reading-queue (append reading-queue (list helper)))
                  )
                )
                (tire-mux-enqueue mux channel `(r ,tag done))
              )
            ))
            (set! forward-reading? #t)
            (set! reading-queue (append reading-queue (list helper)))
          ))
          ((read-stream-buf) (let ()
            (define buflen (list-ref form 4))
            (let ()
              (define buf (make-string buflen))
            (define helper (lambda ()
              (define val (studio-read-string!/partial buf port))
              (if val
                (begin
                  (tire-mux-enqueue mux channel `(i ,tag stream ,(substring buf 0 val)))
                  (set! reading-queue (append reading-queue (list helper)))
                )
                (tire-mux-enqueue mux channel `(r ,tag done))
              )
            ))
            (set! forward-reading? #t)
            (set! reading-queue (append reading-queue (list helper)))
            )
          ))
          (else (tire-mux-enqueue mux channel `(e ,tag (no-such-command ,command))))
        )
      ))
      ((e)
        (debugmsg "io-forward channel -- got e message ~a\n" form)
      )
      (else (tire-mux-enqueue mux channel `(e ,(list-ref form 1) bogus-form)))
    )
  ))
  (define channel (make-channel
    ; dispatch
    (lambda (mux form continuation)
      (dispatch form)
      ; your return value is irrelevant
      (and continuation (continuation))
    )
    ; has-queue
    ; forms are dispatched immediately... no queue
    #f
    ; select-pre
    (lambda (mux)
      (list
        (if (and (not channel-locked?) port forward-reading?) (list port) '())
        (if (and (not channel-locked?) port forward-writing?) (list port) '())
        '()
      )
    )
    ; select-post
    (lambda (mux read-fds write-fds x-fds)
      (debugmsg "ioforward select-post\n")
      (run-if-unlocked (lambda ()
        (if (and forward-writing? (not (null? writing-queue)))
          (begin
            ((car writing-queue))
            (set! writing-queue (cdr writing-queue))
            (and (null? writing-queue) (set! forward-writing? #f))
          )
        )
        (if (and forward-reading? (not (null? reading-queue)))
          (begin
            (debugmsg "ioforward select-post servicing read queue ~a\n" (car reading-queue))
            ((car reading-queue))
            (debugmsg "ran the handler\n")
            (set! reading-queue (cdr reading-queue))
            (and (null? reading-queue) (set! forward-reading? #f))
          )
        )
      ))
      #t
    )
    ; open
    #f
    ; close
    #f
    ; desc
    "port->cookie channel"
  ))

  (tire-mux-add-channel! mux channel)
  (let ()
    (define cookie (string->symbol (string-append (number->string (getpid)) "-" (number->string (fileno port)))))
    (tire-mux-hash-cookie! mux cookie channel)
    ; this presumes that sleazy continuation is available!
    (sleazy-tire-responder 'i 'route cookie)
    cookie
  )
  )
))

)
