(library (proverb javascript)
  (export scheme->javascript scheme->javascript* js-quote-escape-string)
  (import (rnrs) (ice-9 syncase) (proverb dialect) (proverb doc))

(define-syntax scheme->javascript (syntax-rules ()
  ((scheme->javascript form) (scheme->javascript* 'form))
))

(define scheme->javascript* (lambda (form)
  (cond
    ((eq? form #f)
      "false"
    )
    ((eq? form #t)
      "true"
    )
    ((integer? form)
      (number->string form)
    )
    ((number? form)
      (number->string (exact->inexact form))
    )
    ((null? form)
      "[]"
    )
    ((string? form)
      (string-append "\"" (js-quote-escape-string form) "\"")
    )
    ((list? form)
      (if (eq? (car form) 'quote)
        (apply string-append "[" (let helper ((cursor (cadr form)) (racc '()))
          (if (null? cursor)
            (reverse (cons "]" racc))
            (if (null? (cdr cursor))
              (helper (cdr cursor) (cons (scheme->javascript* (car cursor)) racc))
              (helper (cdr cursor) (cons "," (cons (scheme->javascript* (car cursor)) racc)))
            )
          )
        ))
        (case (car form)
          ((lambda)
            "function (x,y,z) {
               alert(10);
             }"
          )
          ((begin)
            (string-append
              "{"
              (apply string-append (map (lambda (v)
                (string-append (scheme->javascript* v) "; ")
              ) (cdr form)))
              "}"
            )
          )
          (else
        (apply string-append (object->string (car form)) "(" (let helper ((cursor (cdr form)) (racc '()))
          (if (null? cursor)
            (reverse (cons ")" racc))
            (if (null? (cdr cursor))
              (helper (cdr cursor) (cons (scheme->javascript* (car cursor)) racc))
              (helper (cdr cursor) (cons "," (cons (scheme->javascript* (car cursor)) racc)))
            )
          )
        ))
          )
        )
      )
    )
    (else "foo")
  )
))

(define js-quote-escape-string (with-docstring
  "accept a Scheme string <str> and escape it for use inside a javascript quoted string"
  (lambda (str)
    (let helper ((cursor (string->list str)) (racc '()))
      (if (null? cursor)
        (apply string (reverse racc))
        (cond
          ((eq? (car cursor) #\")
            (helper (cdr cursor) (append '(#\" #\\) racc))
          )
          ((eq? (car cursor) #\')
            (helper (cdr cursor) (append '(#\' #\\) racc))
          )
          ((eq? (car cursor) #\\)
            (helper (cdr cursor) (append '(#\\ #\\) racc))
          )
          ((eq? (car cursor) #\newline)
            (helper (cdr cursor) (append '(#\n #\\) racc))
          )
          (else
            (helper (cdr cursor) (cons (car cursor) racc))
          )
        )
      )
    )
  )
))

)
