;
; This library is only required for environments that
; are initializing a Javascript Shell. These are
; typically environments that provide an HTTP server
;
; The routine 'proverb-jsshell-site' provides
; a welcome form for the Javascript Shell to
; run. You should ensure that this form is
; legal in your RPC environment.
;
; see support/jsshell for more information on
; the Javascript shell distributed as part of
; Proverb.
;

(library (proverb jsshell)
  (export
    proverb-jsshell-site
    proverb-jsshell-jsprefs
  )
  (import
    (rnrs)
    (proverb remote) (proverb help) (proverb xhtml) (proverb xml)
    (proverb javascript)
    (proverb html)
    (only (guile) gensym the-environment object-address list-set!)
  )

(define proverb-jsshell-jsprefs (lambda (filename)
  (if filename
    (let ((f (open-input-file filename)))
      (if f
        (let ()
          (define h (let helper ((v (read f)) (racc '()))
            (if (eof-object? v)
              (reverse racc)
             (helper (read f) (cons v racc))
            )   
          ))
          (apply string-append (map (lambda (v)
            (string-append "jsprefs.loadpref(\"" (js-quote-escape-string (car v)) "\",\"" (js-quote-escape-string (cdr v)) "\");")
          ) h))
        )
        ""
      )
    )
    ""
  )
))


  (define proverb-jsshell-site (lambda args
(apply string-append "
document.title='Proverb';
var jsshell_current_evaluator=the_hybrid_evaluator;
hybrid_default_evaluator=the_scm_rpc_evaluator;
jsprefs.setpref('welcome_form','](welcome)');
jsprefs.setpref('welcome_back_form','](welcome-back)');
var rpc_url='../rpc';

// each of these wikis use the class internal anchor because the get-article
// should be approximately side-effect free

make_scheme_wiki_ns = function(name,getsym,putsym) {
  var putfunc;
  if(putsym) {
    putfunc=function(name,body,shell,cont) {
      shell.inner_evaluator.eval_to_js('](' + putsym + ' (quote ' + name + ') ' + scheme_stringify(body) + ')',shell,function(e) {
        cont(e);
      });
    };
  } else {
    putfunc=false;
  }
  make_wiki_ns(name,
    false,
    function(name,shell,cont) {
      shell.inner_evaluator.eval_to_js('](' + getsym + ' (quote ' + name + '))',shell,function(e) {
        cont(e);
      });
    },
    putfunc,
    false,
    false,
    false
  );
};

make_proverb_wiki_ns = function(namespace) {
  var oldval=jspref_wikis[namespace];
  if(!oldval) { /* don't whack existing name spaces */
    var putfunc;
    var getfunc;
    var sfunc;
    var ifunc;
    var home;
    if(oldval && oldval.put_article) {
      putfunc=oldval.put_article;
    } else {
      putfunc=function(name,body,shell,cont) {
        name=name.replace(/^ */,'').replace(/ *$/,'');
        shell.inner_evaluator.eval_to_js('](wiki-put \"' + namespace + '\" (quote ' + name + ') ' + scheme_stringify(body) + ')',shell,function(e) {
          cont(e);
        });
      };
    }
    if(oldval && oldval.get_article) {
      getfunc=oldval.get_article;
    } else {
      getfunc=function(name,shell,cont) {
        name=name.replace(/^ */,'').replace(/ *$/,'');
        shell.inner_evaluator.eval_to_js('](wiki-get \"' + namespace + '\" (quote ' + name + '))',shell,function(e) {
          cont(e);
        });
      };
    }
    if(oldval && oldval.search) {
      s=oldval.search;
    } else {
      sfunc=false;
    }
    if(oldval && oldval.index) {
      ifunc=oldval.ifunc;
    } else {
      ifunc=false;
    }
    if(oldval && oldval.home) {
      home=oldval.home;
    } else {
      home='default';
    }

    make_wiki_ns(namespace,
      false,
      getfunc,
      putfunc,
      sfunc,
      ifunc,
      home
    );

  }
};

" (apply string-append (map (lambda (v)
  (string-append "make_proverb_wiki_ns('" v "');")
) (map car wiki-namespaces))) "

jsprefs.setpref('default_wiki','Proverb');

"
args)
  ))

(define bogo-matches '())

(define jsshell-search-help-articles (lambda (query-string)
  (if dribble
    (begin
      (let ()
        (define key (gensym "yield-key-"))
        (define resumer (lambda (v) (set! bogo-matches v) (bogo-resume key v) "ok"))

        (dribble 
          (widgetscript-apply "function(querystring) {
            function schemeify_list(l) {
              var x,r;
              r='(';
              for(x=0;x<l.length;x++) {
                r += l[x] + ' ';
              }
              r+=')';

              return r;  
            };
            jspref_wikis['Jsshell'].search(querystring,function(results) {
              widget.shell.inner_evaluator.eval_to_string(widget.wrapper('(resumer (quote ' + schemeify_list(results)  + '))'),widget.shell,function(v) { alert(v); });
            });
          }" (list query-string))
        )
        (tire-eval-yield key)
      )
    )
    '()
  )
))

(define jsshell-help-articles-index (lambda ()
  (if dribble
    (begin
      (let ()
        (define key (gensym "yield-key-"))
        (define resumer (lambda (v) (set! bogo-matches v) (bogo-resume key v) "ok"))

        (dribble
          (widgetscript-apply "function() {
            function schemeify_list(l) {
              var x,r;
              r='(';
              for(x=0;x<l.length;x++) {
                r += l[x] + ' ';
              }
              r+=')';

              return r;  
            };
            jspref_wikis['Jsshell'].index(function(results) {
              widget.shell.inner_evaluator.eval_to_string(widget.wrapper('(resumer (quote ' + schemeify_list(results)  + '))'),widget.shell,function(v) { alert(v); });
            });
          }" '())
        )
        (tire-eval-yield key)
      )
    )
    '()
  )
))

(define jsshell-wiki-ns `(
  (search . ,jsshell-search-help-articles)
  (index . ,jsshell-help-articles-index)
))

(register-wiki-namespace "Jsshell" jsshell-wiki-ns)



)
