(library (proverb main)
  (export studio-main)
  (import
    (rnrs)
    (rnrs programs)
    (rnrs io simple)
    (rnrs lists)
    (proverb http) (proverb http-rpc) (proverb http-cache) (proverb http-upload)
    (proverb dialect) (proverb tire) (proverb mime)
    (proverb javascript) (proverb jsshell)
    (proverb remote)
    (proverb debug) (proverb filesystem)
    (proverb doc)
    (proverb md5)
    (proverb persist)
    (proverb getopt)
    (proverb help)
    (proverb cond-expand)
    (only (proverb guilefoo) delete-regexp run-select-loop-until)
    (only (proverb bootstrap) support-dir)
    (ice-9 format)
    (guile modules)
    (guile errors)
    (only (guile filesystem) access? R_OK)
    (only (guile)
      ; I/O
      force-output
      port-closed?
      dup2 fdopen dup close
      lstat stat:size stat:mtime
      ; system/networking
      getpid gethostname
      system
      getenv
      getpwuid
      getuid
      uname
      ; string functions
      string-join
      string-index
      ; guile internals
      eval-case defmacro:transformer current-error-port procedure-properties with-input-from-string call-with-new-thread assert-defmacro?! set-defmacro-transformer! fluid-set! gensym
    )
  )

; "recursively build a tree of filenames, mime types"
(define find-useful-files (lambda (dirname)
  (define filenames (delete-regexp "~$" (delete-regexp "^[.]" (dir->filelist dirname))))
  (map (lambda (f)
    (define fqf (string-append dirname "/" f))
    (if (directory? fqf)
    (cons f (find-useful-files fqf))
    (cons f (filename->mime-type f))
  )) filenames)
))

(define export-protected-directory (lambda (real-name protector)
  (http-build-protected-directory-responder real-name (find-useful-files real-name) protector)
))

(define export-directory (lambda (real-name)
  (export-protected-directory real-name #f)
))

(cond-expand
  (guile #f)
  (else (define support-dir "/Users/dsridge/projects/proverb/support/"))
)

(define studio-static-vfs `(
  ( "/" ,(http-build-symlink "jsshell/jsshell.html"))
  ( "sexp.xsl" ,(string-append support-dir "sexp.xsl") "text/xml")
  ;( "jsshell" ,(export-directory (string-append support-dir "jsshell")))
))



; HTTP digest authentication
(define make-sleazy-digest-auth (lambda (realm users)
  (lambda (username cred)
    (and
      (string? username)
      (let ((r (assoc username users)))
        (and
          r
          (md5 (string-append username ":" realm ":" (cadr r)))
        )
      )
    )
  )
))

(define promulgate-report-fake (lambda (arch memsize cores loadavg)
(string-append "<?xml version=\"1.0\" encoding=\"UTF-8\"?>
  <promulgate>
    <host>
      <hostname>" (gethostname) "</hostname>
      <architecture>" arch "</architecture>
      <loadavg>
        <timepoint stamp='now'><load val='" (number->string (if (inexact? loadavg) loadavg (exact->inexact loadavg))) "' /></timepoint>
      </loadavg>
      <cpusocket id='0'>
        "
        (let helper ((count 0) (left cores) (acc ""))
          (if (equal? left 0)
            acc
            (helper (+ count 1) (- left 1) (string-append acc (format #f "<core id='~d' />" count)))
          )
        )
        "<memory>
          <dimm id='0' size='" (number->string memsize) "' />  
        </memory>
      </cpusocket>
      <disk size='128GB' />
    </host>
  </promulgate>
")
))

(define promulgate-report-darwin (lambda ()
  (promulgate-report-fake (vector-ref (uname) 4) (* 4 1024 1024 1024) 2 .5)
))

(define promulgate-report-proc (lambda ()
  (define m (filename->string "/proc/meminfo"))
  (define c (filename->string "/proc/cpuinfo"))
  (define l (filename->string "/proc/loadavg"))
  (let ()
    (define ml (studio-string-split m #\newline))
    (define cl (studio-string-split c #\newline))
    (define ll (studio-string-split l #\space))
    (define parseint (lambda (v)
      (let preamble-helper ((cursor (string->list v)))
        (if (null? cursor)
          0 
          (if (member (car cursor) '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
            (let digit-helper ((cursor cursor) (acc 0))
              (if (null? cursor)
                acc
                (if (member (car cursor) '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))
                  (digit-helper (cdr cursor) (+ (* 10 acc) (string->number (string (car cursor)))))
                  (if (eq? (car cursor) #\space)
                    (let suffix-helper ((cursor (cdr cursor)))
                      (if (null? cursor)
                        acc
                        (if (eq? (car cursor) #\k)
                          (* 1024 acc)
                          acc
                        )
                      )
                    )
                    acc
                  )
                )
              )
            )
            (if (eq? (car cursor) #\space)
              (preamble-helper (cdr cursor))
              0
            )
          )
        )
      )
    ))
    (let ()
      ; no space padding in /proc/meminfo
      (define m-kvl (remove #f (map (lambda (v)
        (define tmp (studio-string-split v #\:))
        (if (or (null? tmp) (null? (cdr tmp)))
          #f
          (cons (car tmp) (parseint (cadr tmp)))
        )
      ) ml)))
      ; space padding in /proc/cpuinfo
      (define c-kvl (remove #f (map (lambda (v)
        (define tmp (studio-string-split v #\:))
        (if (or (null? tmp) (null? (cdr tmp)))
          #f
          (cons (apply string (remove #\ht (remove #\space (string->list (car tmp))))) (cadr tmp))
        )
      ) cl)))
      (let ()
        (define memsize (assoc "MemTotal" m-kvl))
        (define cores (assoc "processor" (aggregate-alist-for-assoc c-kvl)))
        (let ()
          (define load (if (and ll (not (null? ll)))
            (let ()
              (define f (string->number (car ll)))
              (if f
                f
                0
              )
            )
            0 
          ))
          (promulgate-report-fake
            (vector-ref (uname) 4)
            (if memsize
              (cdr memsize)
              0
            )
            (if cores (length (cdr cores)) 1)
            load
          )
        )
      )
    )
  )
))

;
; This report reveals details of your
; system.
; 
(define promulgate-report (lambda ()
  (if (and (access? "/proc/cpuinfo" R_OK)
           (access? "/proc/loadavg" R_OK)
      )
    (promulgate-report-proc)
    (if (access? "/System" R_OK)
      (promulgate-report-darwin)
      (promulgate-report-fake (vector-ref (uname) 4) 2 1 .5)
    )
  )
))

(define studio-main (with-docstring
  "(studio-main (args ...)) -- principal entry point for Proverb"
  (lambda (args)
  (define command (if (null? args) "studio-web" (car args)))
  (define rest (if (null? args) '() (cdr args)))
  (define hostname #f)            ; server hostname
  (define debug #f)               ; enable more debugging output
  (define verbose #f)             ; switch on some verbosity flags
  (define web #t)                 ; run an HTTP server
  (define webthread #f)           ; run the HTTP server in a new thread
  (define http-state-val #f)      ; the HTTP server object
  (define http-port #f)           ; HTTP server port
  (define http-xss #f)            ; HTTP allow cross-site scripting on RPC target
  (define do-promulgate #t)       ; allow participation in promulgate
  (define http-keepalive #t)      ; HTTP use connection keepalive
  (define http-keepalive-set #f)
  (define http-socket #f)         ; use a pre-bound listener on fd X as HTTP server socket
  (define http-local #f)         ; tell the web server to listen only on localhost
  (define publish-redir #f)       ; publish a redirect in the filesystem to the web server
  (define start-browser #f)       ; run a companion program to start a web browser 
  (define start-tire-listener #f) ; run a TIRE-over-TCP server
  (define default-channel #f)     ; the default TIRE channel
  (define add '())
  (define vfs-add '())
  (define tire-port #f)           ; TCP port number for TIRE-over-TCP server
  (define tire-fds #f)            ; pre-connected file descriptors to use for TIRE
  (define thunks '())
  (define threadable #f)          ; enable the macro form 'threadable let'
  (define cookie-alias #f)        ; use X as an alias for 'local'
  (define tirestdio #f)           ; use stdio as an inherited tire connection
  (define back #f)                ; we should phone back to another tire server
  (define pure #f)                ; our default eval environment should be rnrs
  (define welcome #f)             ; we should run X on first connect
  (define welcome-back #f)        ; we should run X on subsequent connect
  (define stylesheet #f)          ; the default stylesheet to use
  (define sleazy-realm "Proverb") ; the main realm protecting the web app
  (define sleazy-loadbalance-realm "Promulgate") ; the secondary realm for the loadbalancer
  (define default-user (let ()
    (define e (getpwuid (getuid)))
    (if e 
      (vector-ref e 0)
      (or
        (getenv "USER")
        "user"
      )
    )
  ))                              ; Let in the user who ran Proverb
  (define sleazy-default-pass "bogo") ; with a lame default password
  (define sleazy-users '())       ; list of users
  (define history (let ()         ; filename containing command-line history or #f
    (define home (getenv "HOME"))
    (if home
      (if (access? (string-append home "/.proverb/history") R_OK)
        (string-append home "/.proverb/history")
        #f
      )
      #f
    )
  ))
  (define profile (let ()
    (define home (getenv "HOME"))
    (if home
      (if (access? (string-append home "/.proverb/profile.scm") R_OK)
        (string-append home "/.proverb/profile.scm")
        #f
      )
      #f
    )
  ))
  (define jsprefs (let ()
    (define home (getenv "HOME"))
    (if home
      (if (access? (string-append home "/.proverb/jsprefs") R_OK)
        (string-append home "/.proverb/jsprefs")
        #f
      )
      #f
    )
  ))

  (define display-help (lambda ()
    (display "proverb-web.sh:\n")
(cond-expand
(guile
    (for-each (lambda (v)
      (define n (cadr (assoc 'arity (procedure-properties (cadr v)))))
      (if (eq? n 0)
        (begin (display (car v)))
        (begin (display (car v)) (display " ") (display n) (display " arguments"))
      )
      (display "  ") (display (list-ref v 2))
      (newline)
    ) opts)
)
(else #f)
)
    (exit 0)
  ))
  (define display-wiki-help (lambda ()
    (display "== Proverb command-line option summary ==\n")
(cond-expand
(guile
    (for-each (lambda (v)
      (define n (cadr (assoc 'arity (procedure-properties (cadr v)))))
      (if (eq? n 0)
        (begin (display "* ") (display (car v)))
        (begin (display "* ") (display (car v)) (display " ") (display n) (display " arguments"))
      )
      (display "\n** ") (display (list-ref v 2))
      (newline)
    ) opts)
)
(else #f)
)
    (exit 0)
  ))

  (define dologin #t)
  (define login-method (lambda (username cred) #t))
  (define promulgate-login-method (lambda (username cred) #t))
  (define interactive-module (resolve-module '(proverb interactive)))

  (define opts `(
    ;
    ; HTTP,Web options
    ; 
    ,@(make-boolean-variable-option "web" "run an HTTP server" web)
    ,@(make-boolean-variable-option "redir" "create a file containing a redirection to the HTTP server" publish-redir)
    ,@(make-boolean-variable-option "http-xss" "allow cross-site HTTP requests" http-xss)
    ,@(make-boolean-variable-option "promulgate" "allow participation in the promulgate load balancer" do-promulgate)
    ("--login" ,(lambda ()
      (set! dologin #t)
    ) "require a login for web. credentials are printed to proverb stdout")
    ("--pass" ,(lambda (s)
      (set! sleazy-default-pass s)
    ) "set the password for the default user")
    ,(make-string-variable-option "realm" "set the HTTP authentication realm to ARG" sleazy-realm)
    ("--nologin" ,(lambda () (set! dologin #f)) "don't require a login for web")
    ("--secure" ,(lambda () (eval '(use-modules (openssl)) (current-module))) "use SSL for web")
    ,(make-integer-variable-option "port" "use specified port for HTTP" http-port)
    ,(make-string-variable-option "hostname" "use the specified name as the hostname" hostname)
    ,(make-integer-variable-option "http-socket" "use specified filedescriptor number for HTTP" http-socket)
    ,(make-integer-variable-option "http-local" "tell the http server to bind only to localhost" http-local)
    ,@(make-boolean-variable-option "keepalive" "use keepalive for HTTP connections" http-keepalive http-keepalive-set)
    ,@(make-boolean-variable-option "webthread" "allow the HTTP server to run in a thread" webthread)
    ,(make-string-variable-option "start-browser" "use the expression to start a browser with the URL" start-browser)
    ("--vfs" ,(lambda (name value mime)
      (cond
        ((string? value)
          (set! vfs-add (cons (list name (http-build-file-responder value mime)) vfs-add))
        )
        ((procedure? value)
          (set! vfs-add (cons (list name value) vfs-add))
        )
        (else #f)
      )
    ) "--vfs NAME VALUE MIME -- extend the HTTP server's namespace")
    ("--vfseval" ,(lambda (name value)
      (set! thunks (cons
        (lambda ()
          (define handler (eval (with-input-from-string value read) interactive-module))
          (set! vfs-add (cons (list name handler) vfs-add))
        )
        thunks
      ))
    ) "--vfseval NAME VALUE -- extend the HTTP server's namespace with NAME handled by the handler returned by evaluating VALUE")
    ("--vfsdir" ,(lambda (name value)
      (cond
        ((string? value)
          (let ()
            (define l (studio-string-split name #\/))
            (set! vfs-add (cons (let ()
              (list name (http-build-directory-responder value (find-useful-files value)))
            ) vfs-add))
          )
        )
        (else #f)
      )
    ) "--vfsdir NAME VALUE -- extend the HTTP server's namespace with the directory NAME handled by the files in VALUE")
    ("--webroot" ,(lambda (name)
      (set! studio-static-vfs (cons
        (list "/"  (http-build-symlink name))
        (cdr studio-static-vfs)
      ))
    ) "create a redir from '/' to NAME in the HTTP server")
    ("--jsprefs" ,(lambda (filename)
      (set! jsprefs filename)
    ) "load javascript shell preferences from filename")
    ("--nojsprefs" ,(lambda ()
      (set! jsprefs #f)
    ) "do not load javascript shell preferences from default location or any previous command line specification")
    ;
    ; TIRE server options
    ; 
    ,@(make-boolean-variable-option "tire" "run a tire server" start-tire-listener)
    ,(make-integer-variable-option "tireport" "use specified port for TIRE" tire-port)
    ("--tirefds" ,(lambda (in out) (set! tire-fds (cons (string->number in) (string->number out)))) "use the specified fds as the default TIRE channel")
    ("--tirefd" ,(lambda (fd) (set! tire-fds (cons (string->number fd) (string->number fd)))) "use the specified full-duplex fd as the default TIRE channel")
    ("--tirestdio" ,(lambda () (set! tirestdio #t) (set! web #f) (set! start-tire-listener #f) (set! tire-fds (cons 0 1))) "use stdio as the default TIRE channel. process will end when stdin closes")
    ,(make-string-variable-option "default" "set the default tire channel to URL" default-channel)
    ("--add" ,(lambda (p) (set! add (cons p add))) "add a TIRE cookie to the mux")
    ,(make-string-variable-option "back" "connect back to the tire cookie 'c' with our cookie" back)
    ("--alias" ,(lambda (name)
      (set! cookie-alias name)
    ) "alias the cookie 'local' to NAME")
    ;
    ; Wiki and help related options
    ;
    ("--fswiki" ,(lambda (namespace path)
      (set! thunks (cons
        (lambda ()
          (register-wiki-namespace namespace (make-filesystem-wiki-ns path))
        )
        thunks
      ))
    ) "--fswiki NAMESPACE PATH -- add a new filesystem wiki rooted at PATH to Proverb")
    ;
    ; General Proverb options
    ;
    ,@(make-boolean-variable-option "debug" "enable some debugging output" debug)
    ,@(make-boolean-variable-option "verbose" "enable some verbose output" verbose)
    ,@(make-boolean-variable-option "threadable" "allow certain threadable code to execute in new threads" sleazy-threadable)
    ("--load" ,(lambda (name)
      (set! thunks (cons (lambda () (eval `(load ,name) interactive-module)) thunks))
    ) "load a Scheme file. processed together with loads and evals in the order they appear on the command line" )
    ("--eval" ,(lambda (formstr)
      (define form (with-input-from-string formstr read))
      (set! thunks (cons (lambda () (eval form interactive-module)) thunks))
    ) "evaluate a Scheme form. processed together with loads and evals in the order they appear on the command line" )
    ("--persist" ,(lambda (name)
      (set! thunks (cons (lambda () (if (access? name R_OK) (eval `(load ,name) interactive-module)) (persist-open name)) thunks))
    ) "set the persist file to name")
    ("--stylesheet" ,(lambda (name)
      ;(set! vfs-add (cons (list "jsshell-site.css" (http-build-file-responder name "text/css")) vfs-add))
      (set! stylesheet name)
    ) "use NAME as the default CSS stylesheet")
    ,(make-string-variable-option "welcome" "use FORM as the welcome form" welcome)
    ,(make-string-variable-option "welcome-back" "use FORM as the welcome form" welcome-back)
    ,(make-string-variable-option "history" "load interactive command line history from FILENAME" history)
    ("--nohistory" ,(lambda ()
      (set! history #f)
    ) "do not load interactive command line history from default location or any previous command line specification")
    ,(make-string-variable-option "profile" "load profile from FILENAME" profile)
    ("--noprofile" ,(lambda ()
      (set! profile #f)
    ) "do not load a profile file")
    ("--trace" ,(lambda (name)
      (eval `(trace! ,(string->symbol name)) interactive-module)
    ) "trace the Scheme function NAME")
    ("--trace-args" ,(lambda (name)
      (eval `(trace-args! ,(string->symbol name)) interactive-module)
    ) "trace the Scheme function NAME. print the arguments.")
    ("--trace-ret" ,(lambda (name)
      (eval `(trace-ret! ,(string->symbol name)) interactive-module)
    ) "trace the Scheme function NAME. print the return value")
    ("--trace-through" ,(lambda (name)
      (eval `(trace-through! ,(string->symbol name)) interactive-module)
    ) "trace the Scheme function NAME. also trace the returned function")
    ("--pure" ,(lambda ()
      (set! thunks (cons
        (lambda ()
          (define m (make-module))
          (set-module-name! m `(proverb session ,(gensym "session-")))
          (module-use! interactive-module (resolve-interface '(rnrs)))
          (set! interactive-module m)
        )
        thunks
      ))
    ) "make the default interactive Scheme evaluation environment 'pure' -- with fewer Guileisms")
    ;("--safer" ,(lambda ()
    ;  (set! pure '(sandbox))
    ;) "make the default Scheme evaluation slightly safer")
    ("--new-session-module" ,(lambda ()
      (set! thunks (cons
        (lambda ()
          (define m (make-module))
          (set-module-name! m `(proverb session ,(gensym "session-")))
          (set! interactive-module m)
        )
        thunks
      ))
    ) "create a completely blank interactive environment")
    ("--session-import" ,(lambda (modname)
      (set! thunks (cons
        (lambda ()
          (define r (with-input-from-string modname read))
          (if r
            (let ((m (resolve-interface r)))
              (if m
                (module-use! interactive-module m)
              )
            )
          )
        )
        thunks
      ))
    ) "add the library MODNAME to the interactive environment")
    ("--deps" ,(lambda () (display "Dependancies flexed. Exiting.\n") (exit 0)) "go through the motions of starting up and exit")
    ("--help" ,(lambda () (display-help)) "display this message")
    ("--wiki-help" ,(lambda () (display-wiki-help)) "generate a wiki article with a summary of these options")
    ("--" ,(lambda args (set! command-line (lambda () args)) '()) "stop argument processing and assign the remaining arguments to the command-line procedure")
  ))

  (define setup-vfs (lambda (tire-state)
    (http-insert-handlers-from-template! '() `(
      ,@studio-static-vfs
      ("jsshell" ,(export-protected-directory (string-append support-dir "jsshell") (lambda (v)
        (http-write-protect v sleazy-realm login-method)
      )))
      ("jsshell/jsshell.manifest" ,(http-protect (lambda (continuation inport outport path args)
          (define fuf (find-useful-files (string-append support-dir "jsshell")))
          (define prepreamble (string-append support-dir "jsshell"))
          (define helper (lambda (v preamble)
            (apply append (map (lambda (w)
              (define postamble (if preamble
                (string-append preamble "/" (car w))
                (car w)
              ))
              (if (list? (cdr w))
                (helper (cdr w) postamble)
                (list (cons postamble (cdr w)))
              )
            ) v))
          ))
          (let* ((files-n-mime (helper fuf #f))
                 (sum (md5 (apply string-append (map (lambda (v)
                        (define fname (string-append prepreamble "/" (car v)))
                        (define mime (cdr v))
                        (if (access? fname R_OK)
                          (let ((s (catch #t (lambda () (lstat fname)) (lambda args #f))))
                            (if s
                              (string-append (car v) ":" mime ":" (number->string (stat:size s)) ":" (number->string (stat:mtime s)))
                              ""
                            )
                          )
                          ""
                        )
                      ) files-n-mime))))
                )
        
            (http-string-responder (apply string-append
"CACHE MANIFEST
# random version: " sum "
"
(append (map (lambda (v)
  (string-append (car v) "\n")
) files-n-mime) '(
"NETWORK:
*
/
"
            ))) "text/cache-manifest" outport)
      )
          (continuation)
      ) sleazy-realm login-method))
      ("login.html" ,(http-protect (lambda (continuation inport outport path args)
        (http-string-responder "Welcome to Proverb. You have authenticated successfully\n" "text/plain" outport)
	(continuation)
      ) sleazy-realm login-method))
      ("authslash" ,(http-protect
        (http-build-symlink "jsshell/jsshell.html")
        sleazy-realm login-method
      ))
      ("jsshell-site.js" ,(http-protect (lambda (continuation inport outport path args)
       (http-string-responder (string-append
         (proverb-jsshell-site)
          (if default-channel
            (let ()
              (define p (tire-parse-url default-channel))
              (if (and p (>= (length p) 4) (list-ref p 3))
                (format #f "var default_eval_cookie=\"~a\";" (list-ref p 3))
                ""
              )
            )
            ""
          )
          (if history
            (catch #t (lambda () (let ()
              (define f (open-input-file history))
              (if f
                (let ()
                  (define h (let helper ((v (read f)) (racc '()))
                    (if (eof-object? v)
                      (reverse racc)
                      (helper (read f) (cons v racc))
                    )
                  ))
                  (string-append "jsshell_readline.default_history=["
                    (string-join (map (lambda (v)
                      (string-append "jsshell_readline.make_history_element(\"" (js-quote-escape-string (car v)) "\",false,false)")
                    ) h) ",")
                  "];")
                )
                ""
              )
            )) (lambda args (critmsg "ERROR LOADING HISTORY")))
            ""
          )
          (if jsprefs
            (proverb-jsshell-jsprefs jsprefs)
            ""
          )
          (if stylesheet
            (format #f "jsprefs.loadpref('theme','~a');\n" (js-quote-escape-string stylesheet))
            ""
          )
          (if welcome
            (format #f "jsprefs.loadpref('welcome_form','~a');\n" (js-quote-escape-string welcome))
            ""
          )
          (if welcome-back
            (format #f "jsprefs.loadpref('welcome_back_form','~a');\n" (js-quote-escape-string welcome-back))
            (if welcome
              (format #f "jsprefs.loadpref('welcome_back_form','~a');\n" (js-quote-escape-string welcome))
              ""
            )
          )
       ) "text/javascript" outport)
        (continuation)
      ) sleazy-realm login-method))
      ("jsshell-site.css" ,(lambda (continuation inport outport path args)
        (http-string-responder "" "text/css" outport)
	(continuation)
      ))
      ,@(reverse vfs-add)
      ("forward" ,(http-protect (lambda (continuation inport outport path args)
        (if (not bogo-forward)
          (let ()
            (define s (socket AF_INET SOCK_STREAM 0))
            (connect s AF_INET (inet-aton "127.0.0.1") 5901)
            (set! bogo-forward s)
          )
        )
        (critmsg "forward ~a\n" args)
        (let ()
          (define mode (cdr (safe-assoc "mode" args '(("mode" . #f)))))
          (cond
            ((equal? mode "read") (let ()
              (define bogo-buffer (make-string 10000))
              (let ()
                (define count (read-string!/partial bogo-buffer bogo-forward))
                (critmsg "count: ~a\n" count)
                (if count
                  (http-string-responder (substring bogo-buffer 0 count) "text/octet-stream" outport)
                )
              )
            ))
            ((equal? mode "write") (let ()
              (define bogo-buffer (make-string 10000))
              (let ()
                (define count (read-string!/partial bogo-buffer inport))
                (critmsg "count: ~a\n" count)
              )
            ))
            (else (critmsg "UNKNOWN MODE!\n"))
          )
        )
        (continuation)
      ) sleazy-realm login-method))
      ("rpc" ,(http-protect (if http-xss
        (lambda (continuation inport outport path args)
          (define origin (http-request-get-origin (current-http-request)))
          (http-tire-rpc tire-state
                         (if webthread #f http-state-val)
                         continuation inport outport path args
                         (if origin 
                           `(,(string-append "Access-Control-Allow-Origin: " origin)
                             "Access-Control-Allow-Credentials: true")
                           '()
                         )
          )
        )
        (lambda (continuation inport outport path args)
          (http-tire-rpc tire-state
                         (if webthread #f http-state-val)
                         continuation inport outport path args '()
          )
        )
      ) sleazy-realm login-method))
      ("cache" ,(http-protect (lambda (continuation inport outport path args)
        (http-tire-cache tire-state (if webthread #f http-state-val) continuation inport outport path args)
      ) sleazy-realm login-method))
      ("upload" ,(http-protect (http-build-system-uploader) sleazy-realm login-method))
      ,@(if do-promulgate
        `(
          ("promulgate-xsscheck" ,(lambda (continuation inport outport path args)
            (define origin (http-request-get-origin (current-http-request)))
            (http-string-responder-with-headers
              "<?xml version=\"1.0\" encoding=\"UTF-8\"?><ok />"
              "text/xml"
              (if origin 
                `(,(string-append "Access-Control-Allow-Origin: " origin)
                  ; these will be credentials along for the ride. The point
                  ; of this piece of namespace is for the Javascript in 
                  ; the client to diagnose a failure to load
                  ; the main promulgate entry.
                  "Access-Control-Allow-Credentials: true")
                '()
              )
              outport
            )
	    (continuation)
          ))
          ("promulgate" ,(http-protect (lambda (continuation inport outport path args)
            (define origin (http-request-get-origin (current-http-request)))
            (http-string-responder-with-headers
              (promulgate-report)
              "text/xml"
              (if origin
                `(,(string-append "Access-Control-Allow-Origin: " origin)
                  "Access-Control-Allow-Credentials: true")
                '()
              )
              outport
            )
	    (continuation)
          ) sleazy-loadbalance-realm promulgate-login-method))
        )
        '()
      )
    ))
  ))

  (cond-expand
    (guile
      (set! vfs-add (cons
        (list "guile/manual/index.html" (http-build-redir "http://www.gnu.org/software/guile/manual/html_node/index.html"))
        vfs-add
      ))
    )
    (else #t)
  )

  (getopt opts rest) 

  (if verbose (set! proverb-verbose-debugging? #t))

  (if tire-fds
    (let ()
      (define inp (if (eq? (car tire-fds) 0)
        (let ()
          (define newfd (dup 0))
          (close 0)
          (fdopen newfd "r")
        )
        (fdopen (car tire-fds) "r")
      ))
      (define outp (if (eq? (cdr tire-fds) 1)
        (let ()
          (define newfd (dup 1))
          ; here we move our existing stdout to stderr
          ; so that we never generate crap on the tire stdout
          (dup2 2 1)
          (fdopen newfd "w")
        )
        (fdopen (cdr tire-fds) "w")
      ))
      (set-car! tire-fds inp)
      (set-cdr! tire-fds outp)
    )
  )

  (if profile
    (eval `(load ,profile) interactive-module)
  )

  (for-each (lambda (v)
    (v)
  ) (reverse thunks))

  (if web
    (begin
      (if (null? sleazy-users)
        (set! sleazy-users `(
          (,default-user ,sleazy-default-pass)
        ))
      )
      (if dologin
        (begin
          (set! login-method (make-sleazy-digest-auth sleazy-realm sleazy-users))
          (set! promulgate-login-method (make-sleazy-digest-auth sleazy-loadbalance-realm sleazy-users))
        )
      )
    )
  )

  (let* ((tire-state (tire-mux-setup '()))
         (server-state (if web
           (http-setup-server
             (setup-vfs tire-state)
             (apply append
               (if http-port `((port . ,http-port)) '())
               (if hostname `((hostname . ,hostname)) '())
               (if http-keepalive-set `((keepalive . ,keepalive)) '())
               (if http-socket `((socketfd . ,http-socket)) '())
               (if http-local `((local . ,#t)) '())
               '()
             )
           )
           #f
         ))
         (cpair (cookie-create-proverb-channel tire-state (if pure (resolve-interface pure) interactive-module)))
        )

    (if (and web server-state)
      (begin
        (if (not start-browser)
          (begin
            (if dologin
              (let ((ua (http-server-get-url-alist server-state)))
                (display "The Proverb web server is running at:\n")
                (let ()
                  (define method (assoc 'method ua))
                  (define hostname (assoc 'hostname ua))
                  (define port (assoc 'port ua))
                  (if (or (not method) (not hostname) (not port))
                    #f
                    (begin
                      (display (string-append (cdr method) "://"
                                              default-user ":" sleazy-default-pass "@"
                                              (cdr hostname) ":"
                                              (number->string (cdr port))
                                              ; some browsers get nervous if you provide auth in the URL to
                                              ; an entry point that does not require it.
                                              "/authslash"
                      ))
                      (newline)
                    )
                  )
                )
                (format #t "You may log in with username=~a, password=~a\n" default-user sleazy-default-pass)
              )
              (begin
                (display "The Proverb web server is running at:\n")
                (display (http-server-get-url server-state))
                (newline)
              )
            )
            (force-output)
          )
        )
        (set! http-state-val server-state)
        (if publish-redir
          (let ()
            (define redir-filename (http-write-redir server-state publish-redir))
            (display "WEB REDIR AT: file://") (display redir-filename) (newline)
          )
        )
        (if start-browser
          (system (string-append start-browser " " (http-server-get-url server-state)))
        )
      )
    )

    ; Our tire mux needs buffering. A user may try to add a channel
    ; to the mux from another thread during a select
    (tire-mux-setup-buffering! tire-state)

    (tire-mux-add-channel! tire-state (cdr cpair))
    ; we call ourselves local and a local cookie
    (tire-mux-hash-cookie! tire-state 'local (cdr cpair))
    (tire-mux-hash-cookie! tire-state (generate-local-cookie) (cdr cpair))
    (if cookie-alias (format #t "ALIAS! ~a\n" cookie-alias))
    (if cookie-alias
      (tire-mux-hash-cookie! tire-state (string->symbol cookie-alias) (cdr cpair))
    )
    (if cookie-alias (format #t "ALIAS DONE\n"))

    (if start-tire-listener
      (let ()
        (define l (tire-mux-add-tcp-listener! tire-state tire-port))
        (format #t "TIRE URL: tire://~a:~a/~a\n" (or hostname (gethostname)) l (generate-local-cookie))
      )
    )

    (if tire-fds
      (tire-mux-add-channel! tire-state (make-tire-fport-channel (car tire-fds) (cdr tire-fds)))
    )

    (if (not (null? add))
      (for-each (lambda (v)
        (define p (tire-parse-url v))
        ; ensure that we got a cookie
        (if (and p (list-ref p 3))
          (let ()
            (define c ((car p)))
            (tire-mux-add-channel! tire-state c)
            (tire-mux-hash-cookie! tire-state (list-ref p 3) c)
          )
        )
      ) (reverse add))
    )

    (if default-channel
      (let ()
        (define c (make-tire-url-channel default-channel))
        (if c
          (begin
            (tire-mux-add-channel! tire-state c)
            (tire-mux-set-default-channel! tire-state c)
          )
          (error "couldn't resolve")
        )
      )
    )

    (if back
      (tire-slave-connect-back tire-state back (lambda args
        #t
      ))
    )

    (let ()
      (define terminator (if tirestdio
        (lambda ()
          (not (port-closed? (car tire-fds)))
        )
        (lambda ()
          #t
        )
      ))
      (if web
        (if webthread
          (begin
            (call-with-new-thread (lambda ()
              (run-select-loop-until (list
                (list server-state http-server-select-pre http-server-select-post)
              ) (if debug (lambda () (format #t "%~a%" (getpid))) (lambda () #t)) terminator)
            ))
            (run-select-loop-until (list
              (list tire-state tire-mux-select-pre tire-mux-select-post)
            ) (if debug (lambda () (format #t ".~a." (getpid))) (lambda () #t)) terminator)
          )
          (run-select-loop-until (list
            (list tire-state tire-mux-select-pre tire-mux-select-post)
            (list server-state http-server-select-pre http-server-select-post)
          ) (if debug (lambda () (format #t ".~a." (getpid))) (lambda () #t)) terminator)
        )
        (run-select-loop-until (list
          (list tire-state tire-mux-select-pre tire-mux-select-post)
        ) (if debug (lambda () (format #t ".~a." (getpid))) (lambda () #t)) terminator)
      )
    )
  )
  )
))

)
