;
; passvals are objects with associated serializers
;

(library (proverb passval)
  (export make-passval passval? passval-write
          make-global* global? global-write define-global)
  (import
    (proverb doc)
    (proverb debug)
    (ice-9 syncase)
    (guile)
  )

(define passval? (with-docstring
  "is <val> a passval?"
  (lambda (val)
    (and
      (object-property val 'serializer)
      #t
    )
  )
))

(define passval-get-val (with-docstring
  "retrieve the Scheme value protected by the passval <passval>"
  (lambda (passval) passval)
))

(define passval-get-writer (with-docstring
  "retrieve the serializer for the passval <passval>"
  (lambda (passval)
    (cdr (assoc 'serializer (object-properties passval)))
  )
))

(define make-passval (with-docstring
  "create a new passval to protect the Scheme value <val>. The serializer is <writer>."
  (lambda (val writer)
    (set-object-property! val 'serializer writer)
    val
  )
))

(define passval-write (with-docstring
  "(passval-write <pval> <port> <writer>) -- serialize the value <pval> to the port <port> using <writer> as the subordinate writer for interior values"
  (lambda (pval port writer)
    (define serval (let ((v (assoc 'now-serializing (object-properties pval))))
      (and v (cdr v))
    ))
    (if serval
      (writer pval port)
      (let ()
        (define pwrite (passval-get-writer pval))
        (set-object-property! pval 'now-serializing #t)
        (catch #t (lambda () 
          (display "#.(make-passval " port)
          (pwrite (passval-get-val pval) port writer)
          (writer (procedure-source pwrite) port)
          (display ")" port)
        ) (lambda args (critmsg "passval-write failure")))
        (set-object-property! pval 'now-serializing #f)
      )
    )
  )
))

(define global-reader (lambda (char ip)
  (let ()
    (define gname (read ip))
    (let ()
      (define gval (hashq-ref global-hash gname))
      gval
    )
  )
))

; install the binary reader
(call-with-input-string "#groot" (lambda (ip)
  (catch #t (lambda ()
    (read ip)
  ) (lambda args
    (read-hash-extend #\g global-reader)
  ))
  (close-input-port ip)
))

(define global? (with-docstring
  "is <val> a global?"
  (lambda (val)
    (and
      (object-property val 'global)
      #t
    )
  )
))

(define global-hash (make-weak-value-hash-table))

(define make-global* (with-docstring
  "create a new global to protect the Scheme value <val>. emitted by define-global"
  (lambda (name val)
    (set-object-property! val 'global name)
    (hashq-set! global-hash name val)
    val
  )
))

(define-syntax define-global (syntax-rules ()
  ((define-global name value) (begin
    (define name (make-global* (quote name) value))
  ))
))

(define global-get-name (with-docstring
  "return the name of the global value <val>."
  (lambda (val)
    (object-property val 'global)
  )
))

(define global-write (with-docstring
  "(global-write <gval> <port> <writer>) -- serialize the value <gval> to the port <port> using <writer> as the subordinate writer for interior values"
  (lambda (gval port writer)
    (display (string-append "#g" (symbol->string (global-get-name gval)) " ") port)
  )
))

)
