; Macros to allow procedures to be defined with type predicates

(cond-expand
  (guile
   ;
   ; This macro is intense.
   ;
   (define-module (proverb types))
(re-export set-procedure-property!)
(define-public procedure-types-list-set! list-set!)
(define-macro (lambda-w-types typedformals . body)
  (define (dotdepth? p)
    (define (depth obj)
      (if (pair? obj) (1+ (depth (car obj))) 0))
    (<= (depth (cdr p)) (depth (car p)))
  )
  (define (split-at-dot dotlist . accum)
    (cond
      ; end of list without dot
      ((null? dotlist) (cons (reverse accum) '()))
      ; reached dotted end 
      ((dotdepth? dotlist) 
       (cons (reverse (cons (car dotlist) accum)) 
	     (cdr dotlist)))
      ; normal case: recurse
      (else (apply split-at-dot (cdr dotlist) (cons (car dotlist) accum)))
  ) )
  (define (make-spec-prop spec)
    (list 
     (symbol->string (car spec)) "" 
     (if (> (length spec) 1)
	 (list-ref spec 1)
	 #t)
     (if (> (length spec) 2)
	 (list-ref spec 2)
	 (cond-expand 
           (guile *unspecified*)
           (chicken (void))))))
  (define (make-opt-arg optspec)
    (if (pair? optspec)
	(make-spec-prop optspec)
	(list optspec)))
  (define (make-opt-name optspec)
    (if (pair? optspec)
	(car optspec) optspec))
  (define (cndr-smack! l v)
    (if (eq? l '())
	#f
	(if (pair? l)
	    (if (pair? (cdr l))
		(cndr-smack! (cdr l) v)
		(if (eq? (cdr l) '())
		    (set-cdr! l v)
		    #f))
	    #f))
  )
    
  (let* ((formalslist (split-at-dot typedformals))
	 (props (car formalslist))
	 (formals (map car props))
	 (doit
	  (if (not (null? (cdr formalslist)))
	      (cndr-smack! formals (make-opt-name (cdr formalslist)))))
	 (argproplist 
	  (map make-spec-prop props))
	 (optargproplist 
	  (list (make-opt-arg (cdr formalslist))))
        )
  (append
    `(let ((bogobogofunc (lambda ,formals ,@body))
           (argstobe ',argproplist)
	  ))

    (let ()
      (define adjust (lambda (prop index)
        `(procedure-types-list-set! (list-ref argstobe ,index) 2 ,(list-ref (list-ref argproplist index) 2))
      ))
      (define helper (lambda (nl c)
        (if (null? nl)
          '()
          (cons (adjust (car nl) c) (helper (cdr nl) (+ c 1)))
        )
      ))
      (helper props 0)
    )

   `((set-procedure-property! bogobogofunc 'args argstobe)
     (set-procedure-property! bogobogofunc 'largs ',optargproplist)
     bogobogofunc)
  )
))
(define-public lambda-w-types lambda-w-types)
)
(chez
  (extend-syntax (lambda-w-types)
    [(lambda-w-types ([name predicate] ...) body-start body-rest ...)
      (andmap symbol? '(name ...))
      (lambda (name ...) body-start body-rest ...)
    ]
    [(lambda-w-types formals body-start body-rest ...)
      (lambda args
        (display "YOU WISH LAMBDA-W-TYPES WITH IMPROPER PROTOTYPE!\n")
      )
    ]
  )
)
(else
  (display "DEFINING PORTABLE LAMBDA-W-TYPES!")
  (library (proverb types)
    (export lambda-w-types)
    (import (rnrs))
    (define-syntax lambda-w-types (syntax-rules ()
      ((lambda-w-types ((name predicate) ...) body-start body-rest ...)
        (lambda (name ...) body-start body-rest ...)
      )
      ((lambda-w-types formals body-start body-rest ...)
        (lambda args
          (display "YOU WISH LAMBDA-W-TYPES WITH IMPROPER PROTOTYPE!\n")
        )
    )   
    ))
  )
)
)


