
(make-wiki-article 'pseudoparse "

Proverb provides a daffy set of routines for pseudo-evaluating Scheme source fragments.

These routines are provided principally to allow partial evaluation in support of 
readline-style tab completion but they may be useful for other purposes. At present, this
code is unused in Proverb and left over from an earlier version of the system.

")

(define pseudoparse (with-docstring
  "pseudoparse the form <sexpr> in the context of the function <func> and module <module>"
  (lambda (sexpr func module)
    (master-pseudoparse sexpr func module 'eval)
  )
))

(define pseudoparse-quote (with-docstring
  "pseudoparse the quoted form <sexpr> in the context of func and module. func and module presumably have no effect"
  (lambda (sexpr func module)
    (master-pseudoparse sexpr func module 'quote)
  )
))

(define pseudoparse-quasiquote (with-docstring
  "pseudoparse the quasiquoted form <sexpr> in the context of func and module"
  (lambda (sexpr func module)
    (master-pseudoparse sexpr func module 'quasiquote)
  )
))

(define object->pseudoval (with-docstring
  "convert a Scheme object into a pseudoval. pseudovals are used in pseudoparsing"
  (lambda (o)
    (cond
      ((number? o) (cons 'NUMBER o))
      ((boolean? o) (cons 'BOOLEAN o))
      ((null? o) (cons 'NULL o))
      ((char? o) (cons 'CHAR o))
      ((string? o) (cons 'STRING o))
      ((procedure? o) (cons 'PROCEDURE o))
      ((symbol? o) (cons 'SYMBOL (cons o o)))
      ((list? o) (cons 'LIST (map object->pseudoval o)))
      ((vector? o) (cons 'VECTOR (vector-map object->pseudoval o)))
      ((pair? o) (cons 'PAIR (cons (object->pseudoval (car o)) (object->pseudoval (cdr o)))))
      (else 
       (display "TRYING TO CONVERT UNKNOWN type object to a pseudoval: ") (display o) (newline)
       (cons 'UNKNOWN o))
    )
  )
))

(define pseudoval->object (with-docstring
  "convert a pseudoval into a regular Scheme object. pseudovals are generated by pseudoparsing"
  (lambda (pv)
  (cond
    ((symbol? pv) pv)
    ((pair? pv)
      (cond
        ((eq? (car pv) 'NUMBER) (cdr pv))
	((eq? (car pv) 'BOOLEAN) (cdr pv))
	((eq? (car pv) 'NULL) (cdr pv))
	((eq? (car pv) 'STRING) (cdr pv))
	((eq? (car pv) 'CHAR) (cdr pv))
	((eq? (car pv) 'PROCEDURE) (cdr pv))
	((eq? (car pv) 'SYMBOL) (if (eq? (cadr pv) (cddr pv))
          ; a literal symbol
          (cadr pv)
          (pseudoval->object (cddr pv))
        ))
        ((eq? (car pv) 'LIST)
          (map pseudoval->object (cdr pv))
        )
        ((eq? (car pv) 'VECTOR)
          (vector-map pseudoval->object (cdr pv))
        )
        ((eq? (car pv) 'PAIR)
          (cons (pseudoval->object (cadr pv)) (pseudoval->object (cddr pv)))
        )
        ((eq? (car pv) 'UNKNOWN)
          (cdr pv)
        )
        (else *unspecified*)
      )
    )
    (else *unspecified*)
  )
  )
))

(define master-pseudoparse (with-docstring
  "a low-level pseudoparse function wrapped by other pseudoparse functions"
 (lambda (sexpr func module mode)
  (cond
    ; easy cases for those objects that evaluate to themselves
    ((number? sexpr) (cons 'NUMBER sexpr))
    ((boolean? sexpr) (cons 'BOOLEAN sexpr))
    ((null? sexpr) (cons 'NULL sexpr))
    ((string? sexpr) (cons 'STRING sexpr)) 
    ;;;;
    ; tougher cases
    ;;;;
    ; a symbol
    ((symbol? sexpr) 
     (if (eq? mode 'eval)
      (let ()
        (define result (procedure-ref func sexpr))
	(display sexpr)
	(newline)
        (if (unspecified? result)
          (let ()
	    (define modresult (module-ref module sexpr))
	    (display "PROCEDURE REF GAVE NO JOY. TRY MODULE REF.\n")
	    (display "MODULE IS: ") (display module) (newline)
	    (if (unspecified? modresult)
		(begin
		  (display "MODULE REF GAVE NO JOY. TRY EVAL.\n")
		  (if (eval `(defined? (quote ,sexpr)) module)
		      (let ()
			(define pv (object->pseudoval (eval sexpr module)))
			(display "TRYING TO EVAL IT: ") (display pv) (newline)
			(cons 'SYMBOL (cons sexpr pv))
	        ) )   )
		(begin 
		  (display "SYMBOL ") (display sexpr) (display " IS NOT DEFINED IN MODULE ")
		  (display (module-name module)) (display "!\n")
		  (cons 'SYMBOL (cons sexpr *unspecified*))  ; this symbol isn't defined 
             )  )
	    (cons 'SYMBOL (cons sexpr (object->pseudoval modresult)))
          ) 
	  (cons 'SYMBOL (cons sexpr (object->pseudoval result)))
      ) )
      ; else we're in a literal mode
      (cons 'SYMBOL (cons sexpr sexpr))
    ))
    ((list? sexpr)
      (newline)
      (cond
        ((and (eq? (car sexpr) 'list) (eq? mode 'eval))
          (cons 'LIST (map (lambda (v) (master-pseudoparse v func module 'eval)) (cdr sexpr)))
        )
        ((and (eq? (car sexpr) 'quote) (eq? mode 'eval))
          (master-pseudoparse (cadr sexpr) func module 'quote)
        )
        ((and (eq? (car sexpr) 'quasiquote) (eq? mode 'eval))
          (master-pseudoparse (cadr sexpr) func module 'quasiquote)
        )
        ((and (eq? (car sexpr) 'unquote) (eq? mode 'quasiquote))
          (master-pseudoparse (cadr sexpr) func module 'eval)
        )
        ((or (eq? mode 'quote) (eq? mode 'quasiquote))
          (cons 'LIST (map (lambda (v) (master-pseudoparse v func module mode)) sexpr))
        )
        (else
          ; this must be a function call
	 (cons 'UNKNOWN *unspecified*)
        )
      )
    )
    ((pair? sexpr)
      (cond
        ((or (eq? mode 'quote) (eq? mode 'quasiquote))
          (cons 'PAIR 
		(cons (master-pseudoparse (car sexpr) func module mode)
		      (master-pseudoparse (cdr sexpr) func module mode)))
        )
        (else
	 (cons 'UNKNOWN *unspecified*)
        )
      )
    )
    ((vector? sexpr)
      (cons 'VECTOR (vector-map (lambda (v) (master-pseudoparse v func module 'eval)) sexpr))
    )
    (else
      ; this is a language term we don't understand
      (cons 'UNKNOWN *unspecified*)
    )  
  )
  )
))

; This is just a first cut and only works on some cases
;  Want the treatment of args to do all the required look ups but also to 
;   do the right thing with composite objects.  This appears to be
;   a master-pseudoparse mode with some properties of 'eval (eg. lookup symbols)
;     and some of 'quote (eg. handle lists)
(define (pseudoparse-eval sexpr func module)
  (if (not (and (list? sexpr) (>= (length sexpr) 2)))
      (throw 'wrong-type-arg "pseudoparse-eval" 
	     "Expected proc of value in position 1 ~A" (list sexpr) #f))
  (let* ((proc (pseudoval->object (master-pseudoparse (car sexpr) func module 'eval)))
	 (args 
	  (map (lambda (arg) (pseudoval->object (master-pseudoparse arg func module 'eval)))
	       (cdr sexpr)))
        )
    (display args) (newline)
    (eval `(apply ,proc (quote ,args)) module)
) )
