;;; A portable queue implementation
;;; Owen Barnett

(library (proverb queue)
  (export studio-queue-add)
  (import (rnrs) (proverb doc) (only (guile) error))

(define studio-queue-add
  (with-wiki-docstring "
Add an element ''el'' to queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue el)
    ((vector-ref queue 1) 'add el))))

(define studio-queue-push
  (with-wiki-docstring "
Push an element ''el'' onto front of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue el) 
    ((vector-ref queue 1) 'push el))))

(define studio-queue-add-list
  (with-wiki-docstring "
Add list ''alist'' to queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue el) 
    ((vector-ref queue 1) 'add-list el))))

(define studio-queue-push-list
  (with-wiki-docstring "
Push list ''alist'' onto front of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue el) 
    ((vector-ref queue 1) 'push-list el))))

(define studio-queue-peek
  (with-wiki-docstring "
Peek at the first element of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'peek))))

(define studio-queue-peek-all
  (with-wiki-docstring "
Peek at all elements of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'peek-all))))

(define studio-queue-get
  (with-wiki-docstring "
Dequeue and return the first element of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'get))))

(define studio-queue-get-all
  (with-wiki-docstring "
Dequeue and return all elements of queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'get-all))))

(define studio-queue-size
  (with-wiki-docstring "
Returns the number of elements in queue ''queue''.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'size))))

(define studio-queue-empty?
  (with-wiki-docstring "
True if ''queue'' has no elements.

See also 
* [[make-queue]].
"
  (lambda (queue)
    ((vector-ref queue 1) 'empty?))))

(define studio-queue?
  (with-wiki-docstring "
True if ''queue''is a queue object. 

See also 
* [[make-queue]].
"
  (lambda (queue)
    (and (vector? queue) 
         (eq?  (vector-ref queue 0) 'queue)))))

(define make-studio-queue
  (with-wiki-docstring "
Makes a queue object portable between studio implementations.
"
  (lambda ()
    (let ((head '())
          (tail '()))
      (vector 'queue
        (lambda (msg . args)
          (case msg
            ((add)
              (set! tail (cons (car args) tail)))
            ((push)
              (set! head (cons (car args) head)))
            ((add-list)
              (set! tail (append (reverse (car args)) tail)))
            ((push-list)
              (set! tail (append (car args) head)))
            ((get) 
              (and (null? head)
                (begin 
                  (and (null? tail) (error 'empty-queue))
                  (set! head (reverse tail))
                  (set! tail '())))
              (let ((rv (car head)))
                (set! head (cdr head))
                rv))
            ((get-all)
              (let ((rv (append head (reverse tail))))
                (set! head '())
                (set! tail '())
                rv))
            ((peek)
             (and (null? head)
                (begin 
                  (and (null? tail) (error 'empty-queue))
                  (set! head (reverse tail))
                  (set! tail '())))
             (car head))
            ((peek-all)
             (append head (reverse tail)))
            ((size)
             (+ (length head) (length tail)))
            ((empty?)
             ;(debugmsg "Q:  ~a ~a\n" head tail)
             (and (null? head) (null? tail)))
            (else (error 'unknown-queue-command)))))))))


;(define-test make-studio-queue studio-queue? studio-queue-empty? studio-queue-size studio-queue-get-all studio-queue-get studio-queue-peek-all studio-queue-peek  (lambda ()
;  (define q (make-studio-queue))
;  (define sentinel (cons 1 2))
;  (define test-emptiness (lambda (queue)
;    (studio-queue-empty? queue)
;    (equal? (studio-queue-peek-all queue) '())
;    (equal? (studio-queue-get-all queue) '())
;    (equal? (studio-queue-size queue) 0)
;    (eq? sentinel (catch #t (lambda () (studio-queue-peek queue)) (lambda args sentinel)))
;    (eq? sentinel (catch #t (lambda () (studio-queue-get queue)) (lambda args sentinel)))
;  ))
;  (and
;    (studio-queue? q)
;    (not (studio-queue? (vector 1 2)))
;    (not (studio-queue? #t))
;    (test-emptiness q)
;    (let ()
;      (studio-queue-add q 1)
;      (and
;        (equal? (studio-queue-size q) 1)
;        (eq? (studio-queue-peek q) 1)
;        (equal? (studio-queue-peek-all q) '(1))
;        (equal? (studio-queue-get-all q) '(1))
;        (test-emptiness q)
;        (begin
;          (studio-queue-add q 1)
;          (studio-queue-add q 2)
;          (and
;            (equal? (studio-queue-size q) 2)
;            (eq? (studio-queue-peek q) 1)
;            (equal? (studio-queue-peek-all q) '(1 2))
;            (begin
;              (studio-queue-push q 3)
;              (and
;                (eq? (studio-queue-size q) 3)
;                (eq? (studio-queue-peek q) 3)
;                (equal? (studio-queue-peek-all q) '(3 1 2))
;                (equal? (studio-queue-get q) '3)
;                (equal? (studio-queue-get q) '1)
;                (equal? (studio-queue-get q) '2)
;                (test-emptiness q)
;              )
;            )
;          )
;        )
;      )
;    )
;  )
;))

)
