;
; Code for searching Guile objects.
;
(library (proverb search)
  (export studio-search)
  (import
    (rnrs)
    (proverb doc) (proverb help) (proverb regexp) (proverb debug)
    (guile)
    (only (proverb guilefoo) module-get-symbol-list)
  )

(define studio-search (with-docstring
  "return a list of objects that match the regular expression string query-string"
  (lambda (query-string)

  (define re (make-regexp (regexp-escape-string query-string)))
  (define all-modules (cons (current-module) (module-uses (current-module))))
  (define safe-variable-ref (lambda (variable name)
    (catch #t (lambda ()
      (variable-ref variable)
    ) (lambda args (critmsg "variable ref fails for variable ~a\n" name) *unspecified*))
  ))

  ; every match is actually a list of matching provenance ...
  ; in this way we can recall why we matched in the first place.
  (let ()
    (define module-name-matches (delete #f (map (lambda (v)
      (define n (module-name v))
      (if (regexp-exec re (object->string n))
        (cons 'module v)
        #f
      )
    ) all-modules)))

    (define help-article-matches (map (lambda (v) (cons 'article v)) (search-help-articles query-string)))

    (define binding-matches (delete #f (apply append (map (lambda (m)
      (map (lambda (v)
        (define name-match (regexp-exec re (object->string (car v))))
        (if name-match
          (list 'binding v m)
          (let ()
            (define value (safe-variable-ref (cdr v) (car v)))
            (cond
              ((and (string? value) (regexp-exec re value))
                (list 'binding v m)
              )
              ((and (procedure? value) (procedure-documentation value) (regexp-exec re (procedure-documentation value)))
                (list 'binding v m)
              )
              ((let ((d (object-property value 'documentation))) (and d (regexp-exec re d)))
                (list 'binding v m)
              )
              (else #f)
            )
          )
        )
      ) (module-get-symbol-list m))
    ) all-modules))))

    (append
      module-name-matches
      help-article-matches
      binding-matches
    )
  )
)))

)
