;
; module shame
;

; Example: (find-undefined-in-module (resolve-module '(proverb http)))

(define-public xhtml-find-undefined (with-docstring
  "interactively find undefined references in Scheme libraries"
  (lambda ()
    (make-wiki
      (apply string-append
        "=== Find undefined references in Scheme libraries ===\n"
        "* [[[(xhtml-find-undefined-in-all) | Check all libraries]]]\n" 
        (map (lambda (v)
          (format #f "* [[[(xhtml-find-undefined-in-module (resolve-module '~a)) | ~a]]]\n" v v)
        ) (map module-name (map resolve-module (map module-name (module-uses (current-module))))))
      )
    )
  )
))

(define colorize-tree (lambda (t)
  (if (null? (cdr t))
    (with-xml-tag "<span style='font-size: 300%;'>"
      (car t)
    )
    (if (list? (car t))
      (with-xml-tag "<span>" (map (lambda (v)
        (if (eq? v (cadr t))
          (with-xml-tag "<span style='font-size: 110%;'>"
            (colorize-tree (cdr t))
          )
          (with-xml-tag "<span style=''>" v)
        )
      ) (car t)))
    )
  )
))

(define-public xhtml-find-undefined-in-all (with-docstring
  "find undefined references in every loaded Scheme library"
  (lambda ()
    (define m (module-uses (current-module)))
    (apply with-xml-tag "<span>" (map (lambda (v)
      (xhtml-find-undefined-in-module v)
    ) (map resolve-module (map module-name m))))
  )
))

(define-public xhtml-find-undefined-in-module (with-docstring
  "find undefined references in the module module"
  (lambda (module)
    (define w (find-undefined-in-module module))
    (if (null? w)
      (make-text (format #f "Module ~a has a clean bill of health\n" (module-name module)))
      (apply with-xml-tag "<span>" 
        (make-text (format #f "Module ~a has some trouble...\n" (module-name module)))
        (map (lambda (y)
          (apply with-xml-tag "<span>"
            (make-wiki (string-append "* [[Scheme:" (symbol->string (car y)) "|" (symbol->string (car y)) "]]\n"))
            (map (lambda (x) 
              (if (pair? x)
                (case (car x)
                  ((notfound) (with-xml-tag "<span>" (make-wiki (format #f "** couldn't find symbol ~a\n" (car (cdr x)))) (colorize-tree (cdr (cdr x))) (make-text "\n")))
                  ((notused) (with-xml-tag "<span>" (make-wiki (format #f "** didn't use symbol ~a\n" (car (cdr x)))) (colorize-tree (cdr (cdr x))) (make-text "\n")))
                  (else (make-wiki (format #f "** bogus problem with symbol ~a\n" x)))
                )
                (make-wiki (format #f "** some other problem with symbol ~a\n" y))
              )
            ) (cdr y))
          )
        ) w)
      )
    )
  )
))

(define rewrite-lame-define (lambda (form)
  ;(dribble "Rewriting a lame define!\n")
  (cons 'lambda (cons (cdr (cadr form)) (cddr form)))
))

(define make-binding (lambda (name)
  (cons name (vector #f '()))
  ;      |           |   |
  ;      |           |   + constraints
  ;      |           |
  ;      |           + referenced?
  ;      |
  ;      + name
))

(define binding-referenced? (lambda (binding)
  (vector-ref (cdr binding) 0)
))

(define reference-binding! (lambda (binding)
  (vector-set! (cdr binding) 0 #t)
))

;
; Returns a list of unbound symbols
; 
(define-public shame-eval (with-docstring
  "accept a source expression s, a module m, and a list of lexical bindings bindings, deferred lexical bindings deferred-bindings, and
produce a report about variables that can't be found"
  (lambda (source module bindings deferred-bindings)
  (define fakefunc (lambda ()
    'fakefunc 
  ))
  (define promote-sources (lambda (l)
    (map (lambda (v) (if (and (pair? v) (eq? (car v) 'notfound)) (cons 'notfound (cons (car (cdr v)) (cons source (cdr (cdr v))))) v)) l)
    (map (lambda (v) (if (and (pair? v) (eq? (car v) 'notused)) (cons 'notused (cons (car (cdr v)) (cons source (cdr (cdr v))))) v)) l)
  ))
  (define mec (module-eval-closure module))
  (define lookup (lambda (sym)
    (define b (assoc sym bindings))
    (if b
      (begin
        (reference-binding! b)
        fakefunc
      )
      (let ((v (mec sym #f)))
        (and v (or (variable-ref v) #t))
      )
    )
  ))
  (let shame-helper ((racc '()))
    ;(format #t "SHAME HELPER: ~a\n" source)
    (cond
      ((null? source)
        '() ; succeeds but is binding neutral
      )
      ((list? source)
        (cond
          ((list? (car source))
            ; then this is certainly a function
            (apply append (map (lambda (v) (promote-sources (shame-eval v module bindings deferred-bindings))) source))
          )
          ; else this may be a function or syntax
          ((symbol? (car source))
            (let ((w (lookup (car source))))
              (cond
                ((not w)
                  (list (cons 'notfound (cons (car source) (list source))))
                )
                ((closure? w)
                  (apply append (map (lambda (v) (promote-sources (shame-eval v module bindings deferred-bindings))) (cdr source)))
                )
                ((procedure? w)
                  (apply append (map (lambda (v) (promote-sources (shame-eval v module bindings deferred-bindings))) (cdr source)))
                )
                ((macro? w) (case (car source)
                  ((lambda) (let ()
                    (define lambda-helper (lambda (body bogo-bindings)
                      (let d-helper ((cursor body) (private-bindings '()) (d-racc '()))
                        (if (null? cursor)
                          (list (cons 'bad-lambda source))
                          (let ((t (car cursor)))
                            (if (and (pair? t) (eq? (car t) 'define))
                              (if (pair? (cdr t))
                                (begin
                                  (if (pair? (cadr t))
                                    (begin
                                      ;(dribble (make-text (format #f "evil! failed to recurse! ~a\n" t)))
                                      ;(dribble (make-text (format #f "BOGO evil! failed to recurse! ~a\n" (list 'define (car (cadr t)) (rewrite-lame-define t)))))
                                      (d-helper (cdr cursor) (cons (make-binding (car (cadr t))) private-bindings) (cons (list 'define (car (cadr t)) (rewrite-lame-define t)) d-racc))
                                    )
                                    (d-helper (cdr cursor) (cons (make-binding (cadr t)) private-bindings) (cons t d-racc))
                                  )
                                )
                                (list (cons 'bad-define source))
                              )
                              (if (and (string? t) (not (null? (cdr cursor))))
                                ; this is an old-fashioned doc string
                                (d-helper (cdr cursor) private-bindings d-racc)
                                (let ((res (apply append (append
                                  ; shame-eval now that we have accumulated all the define bindings
                                  (map (lambda (t)
                                    (promote-sources (shame-eval (caddr t) module bogo-bindings private-bindings))
                                  ) d-racc)
                                  ; and the body terms
                                  (let ((these-bindings (append private-bindings bogo-bindings)))
                                    (map (lambda (x)
                                      (promote-sources (shame-eval x module these-bindings '()))
                                    ) cursor)
                                  )
                                ))))
                                (append
                                (delete #f (map (lambda (v)
                                 (if (not (binding-referenced? v))
                                   (cons 'notused (cons (car v) (list (car v))))
                                   #f
                                 )
                                ) private-bindings))
                                res
                                )
                                )
                              )
                            )
                          )
                        )
                      )
                    ))
                    (let ((nb (let helper ((cursor (cadr source)) (acc (append deferred-bindings bindings)))
                                (if (null? cursor)
                                  acc
                                  (if (symbol? cursor)
                                    ; this is the varargs portion
                                    (cons (make-binding cursor) acc)
                                    ; this is the fixed portion
                                    (helper (cdr cursor) (cons (make-binding (car cursor)) acc))
                                  )
                                )
                              )))
                      (lambda-helper (cddr source) nb)
                    )
                  ))
                  ((let) (let ()
                    (define let-body-helper (lambda (body bogo-bindings)
                      (let d-helper ((cursor body) (private-bindings '()) (d-racc '()))
                        (if (null? cursor)
                          (cons
                            (cons 'let-without-body source)
                            (reverse d-racc)
                          )
                          (if (and (pair? (car cursor)) (eq? (caar cursor) 'define))
                            (let ((t (cadr (car cursor))))
                              (if (pair? t)
                                ; this is a (define (foo) ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding (car t)) private-bindings)
                                  (cons 
                                    (rewrite-lame-define (car cursor))
                                    d-racc
                                  )
                                )
                                ; this is a (define foo ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding t) private-bindings)
                                  (cons 
                                    (caddr (car cursor))
                                    d-racc
                                  )
                                )
                              )
                            )
                            ; else this and the rest of these terms are real let body
                            (apply append (append
                              (reverse (map (lambda (v) (promote-sources (shame-eval v module bogo-bindings (append private-bindings deferred-bindings)))) d-racc))
                              (map (lambda (x)
                                (promote-sources (shame-eval x module (append private-bindings bogo-bindings) deferred-bindings))
                              ) cursor)
                            ))
                          )
                        )
                      )
                    ))
                    (if (symbol? (cadr source))
                      ; named let
                      (let ()
                        (define these-bindings (map make-binding (map car (caddr source))))
                        (append
                          (apply append (map (lambda (v)
                            (promote-sources (shame-eval (cadr v) module bindings (append these-bindings deferred-bindings)))
                          ) (caddr source)))
                          (let-body-helper (cdddr source) (cons (make-binding (cadr source)) (append these-bindings bindings)))
                        )
                      )
                      ; regular let
                      (let ()
                        (define these-bindings (map make-binding (map car (cadr source))))
                        (append
                          (apply append (map (lambda (v)
                            (promote-sources (shame-eval (cadr v) module bindings (append these-bindings deferred-bindings)))
                          ) (cadr source))) 
                          (let-body-helper (cddr source) (append these-bindings bindings))
                        )
                      )
                    )
                  ))
                  ; it would sure be nice if we could detect needless let*
                  ((let*) (let ()
                    (define let*-body-helper (lambda (body bogo-bindings)
                      (let d-helper ((cursor body) (private-bindings '()) (d-racc '()))
                        (if (null? cursor)
                          (cons
                            (cons 'let-without-body source)
                            (reverse d-racc)
                          )
                          (if (and (pair? (car cursor)) (eq? (caar cursor) 'define))
                            (let ((t (cadr (car cursor))))
                              (if (pair? t)
                                ; this is a (define (foo) ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding (car t)) private-bindings)
                                  (cons 
                                    (rewrite-lame-define (car cursor))
                                    d-racc
                                  )
                                )
                                ; this is a (define foo ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding t) private-bindings)
                                  (cons 
                                    (caddr (car cursor))
                                    d-racc
                                  )
                                )
                              )
                            )
                            ; else this and the rest of these terms are real let body
                            (apply append (append
                              (reverse (map (lambda (v) (promote-sources (shame-eval v module bogo-bindings (append private-bindings deferred-bindings)))) d-racc))
                              (map (lambda (x)
                                (promote-sources (shame-eval x module (append private-bindings bogo-bindings) deferred-bindings))
                              ) cursor)
                            ))
                          )
                        )
                      )
                    ))
                    ; there is no named let*
                    (let let*-binding-helper ((cursor (cadr source)) (these-bindings '()) (racc '()))
                      (if (null? cursor)
                        (append racc (let*-body-helper (cddr source) (append these-bindings bindings)))
                        (let ()
                          (define this-binding (make-binding (car (car cursor))))
                          (let*-binding-helper (cdr cursor)
                                               (cons this-binding these-bindings)
                                               (append (promote-sources (shame-eval (cadr (car cursor))
                                                                                    module
                                                                                    (append these-bindings bindings)
                                                                                    (cons this-binding deferred-bindings)
                                               )) racc)
                          )
                        )
                      ) 
                    )
                  ))
                  ((letrec) (let ()
                    (define letrec-body-helper (lambda (body bogo-bindings)
                      (let d-helper ((cursor body) (private-bindings '()) (d-racc '()))
                        (if (null? cursor)
                          (cons
                            (cons 'let-without-body source)
                            (reverse d-racc)
                          )
                          (if (and (pair? (car cursor)) (eq? (caar cursor) 'define))
                            (let ((t (cadr (car cursor))))
                              (if (pair? t)
                                ; this is a (define (foo) ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding (car t)) private-bindings)
                                  (cons 
                                    (rewrite-lame-define (car cursor))
                                    d-racc
                                  )
                                )
                                ; this is a (define foo ...) form
                                (d-helper
                                  (cdr cursor)
                                  (cons (make-binding t) private-bindings)
                                  (cons 
                                    (caddr (car cursor))
                                    d-racc
                                  )
                                )
                              )
                            )
                            ; else this and the rest of these terms are real let body
                            (apply append (append
                              (reverse (map (lambda (v) (promote-sources (shame-eval v module bogo-bindings (append private-bindings deferred-bindings)))) d-racc))
                              (map (lambda (x)
                                (promote-sources (shame-eval x module (append private-bindings bogo-bindings) deferred-bindings))
                              ) cursor)
                            ))
                          )
                        )
                      )
                    ))
                    ; there is no named letrec
                    (let ()
                      (define these-bindings (map make-binding (map car (cadr source))))
                      (append
                        (apply append (map (lambda (v)
                          (promote-sources (shame-eval (cadr v) module bindings (append these-bindings deferred-bindings)))
                        ) (cadr source))) 
                        (letrec-body-helper (cddr source) (append these-bindings bindings))
                      )
                    )
                  ))
                  ((define)
                    (if (pair? (cadr source))
                      (promote-sources (shame-eval (rewrite-lame-define source) module bindings (cons (make-binding (car (cadr source))) deferred-bindings)))    
                      (promote-sources (shame-eval (caddr source)  module bindings (cons (make-binding (cadr source)) deferred-bindings)))
                    )
                  )
                  ((delay)
                    (promote-sources (shame-eval (cadr source) module bindings deferred-bindings))
                  )
                  ; do is a library function in r6 but primitive in guile 1.8
                  ((do)
                    (if (pair? (cdr source))
                      (let* (  (bogo-bindings (cadr source))
                               (binding-names (map car bogo-bindings))  )
                        (apply append (map (lambda (x)
                          (append
                            (promote-sources (shame-eval (cadr x) module bindings (append (map make-binding binding-names) deferred-bindings)))
                            (if (null? (cddr x))
                              '()
                              (promote-sources (shame-eval (caddr x) module (append (map make-binding binding-names) bindings) deferred-bindings))
                            )
                            (if (null? (caddr source))
                              (list (cons 'do-without-test))
                              (let ()
                                (define test-clause (car (caddr source)))
                                (append
                                  (promote-sources (shame-eval test-clause module (append (map make-binding binding-names) bindings) deferred-bindings))
                                  (if (null? (cdr (caddr source)))
                                    '()
                                    (promote-sources (shame-eval (cadr (caddr source)) module (append (map make-binding binding-names) bindings) deferred-bindings))
                                  )
                                  (if (null? (cdddr source))
                                    ; guile seems to tolerate do without body.
                                    '()
                                    (apply append (map (lambda (y)
                                      (promote-sources (shame-eval y module (append (map make-binding binding-names) bindings) deferred-bindings))
                                    ) (cdddr source)))
                                  )
                                )
                              )
                            )
                          )
                        ) bogo-bindings))
                      )
                      (list (cons 'do-without-bindings source))
                    )
                  )
                  ((set!)
                    (promote-sources (shame-eval (caddr source) module bindings deferred-bindings))
                  )
                  ((and begin or)
                    (apply append (map (lambda (x)
                      (promote-sources (shame-eval x module bindings deferred-bindings))
                    ) (cdr source)))
                  )
                  ((if)
                    (if (null? (cdr source))
                      (list (cons 'if-without-test source))
                      (if (and (list? (cadr source)) (pair? (cadr source)))
                        (begin
                        (if (and (eq? (car (cadr source)) 'defined?) (not (null? (cdr (cadr source)))) (list? (cadr (cadr source))) (pair? (cadr (cadr source))) (eq? (car (cadr (cadr source))) 'quote) (not (null? (cdr (cadr (cadr source))))) (symbol? (cadr (cadr (cadr source)))))
                          ; this if has introduced a new binding.
                          (append
                            (promote-sources (shame-eval (cadr source) module bindings deferred-bindings))
                            (if (null? (cdr (cdr source)))
                              (list (cons 'if-without-true source))
                              (append
                                (promote-sources (shame-eval (caddr source) module (cons (make-binding (cadr (cadr (cadr source)))) bindings) deferred-bindings))
                                (if (null? (cdr (cdr (cdr source))))
                                  '()
                                  (promote-sources (shame-eval (cadddr source) module bindings deferred-bindings))
                                )
                              )
                            )
                          )
                          ; else
                          (apply append (map (lambda (x) (promote-sources (shame-eval x module bindings deferred-bindings))) (cdr source)))
                        )
                        )
                        ; else
                        (apply append (map (lambda (x) (promote-sources (shame-eval x module bindings deferred-bindings))) (cdr source)))
                      )
                    )
                  )
                  ((quote)
                    '()
                  )
                  ((quasiquote)
                    (let helper ((cursor (cadr source)) (acc '()))
                      (if (null? cursor)
                        (reverse acc)
                        (helper (cdr cursor) (if (pair? (car cursor))
                          (case (caar cursor)
                            ((unquote)
                              (append (promote-sources (shame-eval (cadar cursor) module bindings deferred-bindings)) acc)
                            )
                            ((unquote-splicing) (append
                              (apply append (map (lambda (v)
                                (promote-sources (shame-eval v module bindings deferred-bindings))
                              ) (cdar cursor)))
                              acc
                            ))
                            (else
                              ;(dribble (format #f "unquote bogo ~a" (car cursor)))
                              acc
                            )
                          )
                          acc
                        ))
                      )
                    )
                  )
                  ((case)
                    (apply append (promote-sources (shame-eval (cadr source) module bindings deferred-bindings)) (map (lambda (x)
                      (if (list? x)
                        (if (null? x)
                          (begin (dribble "BAD CASE!") '())
                          (apply append (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) (cdr x)))
                        )
                        (begin
                          (dribble "BAD CASE!")
                          '()
                        )
                      )
                    ) (cddr source)))
                  )
                  ((cond)
                    (apply append (map (lambda (x)
                      (if (list? x)
                        (if (null? x)
                          (begin
                            (dribble "BAD COND!")
                            '()
                          )
                          (if (eq? (car x) 'else)
                            (apply append (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) (cdr x)))
                            (if (and (> (length x) 2) (eq? (cadr x) '=>))
                              (apply append (promote-sources (shame-eval (car x) module bindings deferred-bindings)) (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) (cddr x)))
                              (apply append (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) x))
                            )
                          )
                        )
                        (begin
                          (dribble "BAD COND!")
                          '()
                        )
                      )
                    ) (cdr source)))
                  )
                  ((cond-expand)
                    (let ((gcase (assoc 'guile (cdr source))))
                      (if gcase
                        (apply append (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) (cdr gcase)))
                        (let ((ecase (assoc 'else (cdr source))))
                          (if ecase
                            (apply append (map (lambda (y) (promote-sources (shame-eval y module bindings deferred-bindings))) (cdr ecase)))
                            (begin (dribble "BAD COND EXPAND!\n") '())
                          )
                        )
                      )
                    )
                  )
                  ((the-environment)
                    '()
                  )
                  (else (if (primitive-macro? w)
                    (begin (dribble (list "UNEXPANDED PRIMITIVE MACRO: " (car source))) '())
                    (if (defmacro? w)
                      (let ((e (macroexpand source)))
                        (if (equal? e source)
                          (begin
                            (dribble (list "EXPANSION WAS THE SAME..."))
                            '()
                          )
                          (promote-sources (shame-eval e module bindings deferred-bindings))
                        )
                      )
                      (let ()
                        (define t (macro-transformer w))
                        (if t 
                          (let ((e (t source (the-environment))))
                            (if (equal? e source)
                              (dribble "MACRO-TRANSFORMER EXPANSION WAS THE SAME")
                            )
                            (promote-sources (shame-eval (t source (the-environment)) module bindings deferred-bindings)) ; the-environment is not right
                          )
                          (begin (dribble (list "UNEXPANDEDABLE NON-DEFMACRO " (car source))) '())
                        )
                      )
                    )
                  ))
                ))
                (else (begin (dribble (list "FOUND BAD TARGET OF CALL!" (car source) w)) '()))
              )
            )
          )
          (else
            ; so we're not a symbol. That happens when an unhygenic
            ; macro has an odd expansion with literal non-immediates
            ; Who writes macros like this? Oh, wait, I do.
            (if (macro? (car source))
              (case (macro-name (car source))
                ((begin) (apply append (map (lambda (v) (promote-sources (shame-eval v module bindings deferred-bindings))) (cdr source))))
                (else (list (cons 'unhandled-macro (car source))))
              )
              (list (cons 'bad-function-preamble (car source)))
            )
          )
        )
      )
      ((vector? source)
        (apply append (vector->list (vector-map (lambda (v)
          (promote-sources (shame-eval v module bindings deferred-bindings))
        ) source)))
      )
      ((symbol? source)
        (let ((v (lookup source)))
          (if (not v)
            (list (cons 'notfound (cons source (list source))))
            '()
          )
        )
      )
      ; else it succeeds but is binding neutral
      (else '())
    )
  )
  )
)) 

(define-public find-undefined-in-procedure (lambda (procedure procedure-name module)
  (define s (procedure-and-environment-source procedure))
  (define rm (environment-module (procedure-environment procedure)))
  (if s
    (or
      (if rm
        (shame-eval s rm '() '())
        (begin
          ; must be a very low-level Guile procedure
          '()
        )
      )
      (begin
        (cons 'fail procedure-name)
        '()
      )
    )
    '()
  )
))

;
; Returns a list of the form ( symbol . ( problem ... ) ) with no entries
; for symbols with no problems
;
(define-public find-undefined-in-module (lambda (module)
  (define l (module-get-symbol-list module))   
  (remove #f (map (lambda (v)
    (define val (catch #t
      (lambda () (variable-ref (cdr v)))
      (lambda args (dribble (format #f "ref fails on ~a\n" (car v))) #f)
    ))
    (if (and (procedure? val) (closure? val))
      (let ((w (find-undefined-in-procedure val (car v) module)))
        (if (not (list? w))
          (dribble (format #f "BROKE ~a\n" (car v)))
        )
        (if (null? w)
          #f
          (cons (car v) w)
        )
      )
      #f
    )
  ) l))
))
