;
; cond-expand some bindings that are present in some dialects but which
; are widely implementable
;
(library (proverb dialect)
  (export
    once
    studio-make-hash-table
    file-size file-execute-access? any? studio-string-split studio-pipe
    exclusive directory? current-load-dir append-load-path! dir->filelist studio-make-weak-value-hash-table
    ;with-maybe-mutex studio-select accept-io-socket studio-read-string!/partial write-string/full studio-flush-port studio-read-string!/full threadable-let sleazy-threadable lambda-w-types value? all? vector-head vector-tail string-head string-tail vector-for-each string-prefix-match unique uniq alist? duplicate-alist
    with-maybe-mutex studio-select accept-io-socket studio-read-string!/partial write-string/full studio-flush-port studio-read-string!/full threadable-let sleazy-threadable value? all? vector-head vector-tail string-head string-tail vector-for-each string-prefix-match unique uniq alist? duplicate-alist
    index-map vector-index-map
    aggregate-alist-for-assoc aggregate-alist-for-assq 
    always-true always-false
    identify-scheme-type
    safe-assoc
    object->string
    ; used by threadable let macro
    sleazy-threads-mutex
    sleazy-threads-left
    sleazy-threads-incr
    ; queer omissions from R6RS
    list-head
  )
  (import
    (guile)
    (ice-9 rw)
    (proverb doc)
    (proverb debug)
    (proverb types)
  )

(load "scheme-util.scm")

(cond-expand
  (guile
    ;place all imports necessary to support studio here
    (define once (with-docstring "accept a thunk, run it, and complain if it returns more than once"
      (lambda (inner-func)
        (lambda args
          (define count 0)
          (let ()
            (define retval (apply inner-func args))
            (set! count (+ count 1))
            (if (eq? count 1)
              retval 
              (begin
                (critmsg "RETURNED MULTIPLY!\n")
               (error "multiple return")
              )   
            )
          )
        )
      )
    ))

    ;];        
    ;[ files ] ;[;
    ; chicken and guile just don't agree about filesystem stuff
    (define dir->filelist (with-docstring
      "accept a directory name and return a list of filenames inside it"
      (lambda (dirname)
        (catch #t (lambda ()
          (define d (opendir dirname))
          (let helper ((name (readdir d)) (racc '()))
            (if (eof-object? name)
              (begin
                (closedir d)
                (reverse racc)
              )
              (helper (readdir d) (cons name racc))
            )
          )
        ) (lambda args '()))
      )
    ))
    ; chicken provides this (and not stat)
    (define file-size (with-docstring
      "return the size (in bytes) of the file named by <filename>"
      (lambda (filename)
        (stat:size (stat filename))
      )
    ))
    ; chicken provides this (and not access?)
    (define file-execute-access? (with-docstring
      "return true if the file named by <filename> can be executed, false otherwise"
      (lambda (filename)
        (access? filename X_OK)
      )
    ))
    ; chicken provides directory? -- but not stat
    (define directory? (with-docstring
      "return true if the file named by <filename> is a directory"
      (lambda (filename)
        (catch #t (lambda ()
          (define v (stat (string-append filename)))
          (eq? (stat:type v) 'directory)
        ) always-false)
      )
    ))
    ;];
    ;[ I/O ] ;[;
    
    ;standardize the semantics of select across schemes (and avoid name collisions)
    (begin
      ; try-use-modules from guile-util is not available yet
      (catch #t (lambda () (process-use-modules '(((openssl))))) (lambda args #f))
      (if (defined? 'SSL-select)
        (begin
          (define studio-select SSL-select)
          (define ssl-available #t)
        )
        (begin
          (define studio-select select)
          (define ssl-available #f)
        )
      )
    )

    (if ssl-available
      (define bogus-write-string/partial (lambda (str port from to)
        (if (file-port? port)
          (write-string/partial str port from to)
          (begin
            (let ((s (substring str from to)))
              (display str port)
              (- to from)
            )
          )
        )
      ))
      (define bogus-write-string/partial write-string/partial)
    )
    
    ; (ice-9 rw) provides write-string/partial but not full. 
    (define write-string/full (with-docstring
      "write the contents of the string <str> to port <port> from index <start> inclusive to index <stop> exclusive."
      (lambda (str port start stop)
      (define j (or stop (string-length str)))
      (let helper ((i (or start 0)))
        (if (eq? i j)
          (- j (or start 0))
          (let ()
            (define c (bogus-write-string/partial str port i j))
            (if (eq? c 0)
              #f
              (if c
                (helper (+ i c))
                #f
              )
            )
          )
        )
      )
    )))

    (define studio-read-string!/full (with-docstring
      "read the contents of the port <port> into the string <buffer> and return the number of bytes read"
      (lambda (buffer port)
        (let helper ((i 0) (l (string-length buffer)))
          (define res (studio-read-string!/partial buffer port i))
          (if res
            (if (eq? res l)
              res
              (begin
                ;(critmsg "partial read in studio-read-string!/full ~a of ~a" res l)
                (helper (+ i res) (- l res))
              )
            )
            #f
          )
        )
      )
    ))

    (define studio-read-string!/partial (with-docstring
      "wrapper around read-string!/partial. Guile's implementation only works on fports"
      (lambda (buf port . rest)
        (if (file-port? port)
          (apply read-string!/partial buf port rest)
          (let ()
            (define starting-index (if (pair? rest) (car rest) 0))
            (define post-index (if (and (pair? rest) (pair? (cdr rest))) (cadr rest) (string-length buf)))
            (let helper ((i starting-index))
              (if (eq? i post-index)
                (- i starting-index)
                (if (char-ready? port)
                  (let ()
                    (define c (read-char port))
                    (if (eof-object? c)
                      (- i starting-index)
                      (begin
                        (string-set! buf i c)
                        (helper (+ i 1))
                      )
                    )   
                  )
                  (- i starting-index)
                )
              )
            )
          )
        )
      )
    ))

    (define accept-io-socket (with-docstring
      "accepts on a socket descriptor <sockfd> and returns a pair (inport . outport)"
      (lambda (sockfd)
        (let ((new (accept sockfd)))
          (cons (car new) (car new)))
    )))

    (define studio-flush-port (with-docstring
      "Proverb function for flushing Scheme ports"
      (lambda (port)
        ; guile soft ports won't flush if guile doesn't think there
        ; is any buffered I/O
        (define op (object-property port 'extra-flush)) 
        ; only guile can actually push unseen characters to the soft port
        ; implementation. Then we can force.
        (catch #t (lambda ()
          (force-output port)
          (and op (op))
        ) (lambda args (critmsg "error in studio-flush-port")))
        *unspecified*
      )
    ))
    ;];
    ;[ load system ] ;[;
    (define current-load-filename (with-docstring "return the name of the filename currently being loaded" (lambda () (port-filename (current-load-port)))))
    (define current-load-dir (with-docstring "return the directory name of the file currently being loaded" (lambda ()
      (dirname (current-load-filename))
    )))
    (define append-load-path! (with-docstring "add a new directory to the load path search by load-from-path" (lambda (new-dir)
      (set! %load-path (append %load-path (list new-dir))))))
    ;];
    ;[ misc ] ;[;
    ; chicken requires define-for-syntax when defining syntax-rules
    (define define-for-syntax define)
    ; chicken provides this as a builtin 'extended'
    (define any? (with-docstring "does list <l> have any non-false members?" (lambda (list)
      (if (null? list)
          #f
          (or (car list) (any? (cdr list)))
      )
    )))
    ;];
    ;[ studio functions ] ;[;
    ;XXX only accepting characters
    (define studio-string-split string-split)

    ;XXX only using size
    (define studio-make-hash-table (with-docstring
      "portable hash table construction put here to span cross-dialect differences. accepts <hash-func> used to hash keys for look up, <eq-func> used to compare two keys for equality, and a minimum number of buckets <size>"
      (lambda (hash-func eq-func . size)
        (if (null? size)
          (make-hash-table)
          (make-hash-table (car size))
        )
      )
    ))

    (define studio-make-weak-value-hash-table (with-docstring
     "portable weak-value hash table construction put here to span cross-dialect differences. accepts <hash-func> used to hash keys for look up, <eq-func> used to compare two keys for equality, and a minimum number of buckets <size>. The key/value pair will be removed from the hash table when value becomes otherwise unavailable"
      (lambda (hash-func eq-func . size)
        (if (null? size)
          (make-weak-value-hash-table)
          (make-weak-value-hash-table (car size))
        )
      )
    ))
    (define studio-pipe pipe)
    ;];
    (re-export list-head list-head)
  )
  (chicken
    ;place all imports necessary to support studio here
    (use syntax-case)
    (use lolevel)
    (use numbers)
    (use posix)
    (use proplist)
    (use records)
    (use regex)
    (use srfi-13)
    (use srfi-18)
    (use tcp)
    ;[ exceptions ] ;[;
    (define catch
      (lambda (e thunk handler)
        (condition-case (thunk)
          (e () (handler ((condition-property-accessor 'exn 'message) e) 
                         ((condition-property-accessor 'exn 'arguments) e))))))
    (define lazy-catch
      (lambda (e thunk handler)
        (condition-case (thunk)
          (e () (begin 
                  (handler ((condition-property-accessor 'exn 'message) e)
                           ((condition-property-accessor 'exn 'arguments) e))
                  (signal e))))))
    ;];
    ;[ files ] ;[; 
    (define dirname pathname-directory) 

    (define dir->filelist  
      (lambda (dirname)
        (directory dirname #t)))

    (define fileno port->fileno)

    (define R_OK 'R_OK)
    (define W_OK 'W_OK)
    (define X_OK 'X_OK)
    (define access?
      (lambda (filename kind)
        (case kind
          ((R_OK)
            (file-read-access? filename))
          ((W_OK)
            (file-write-access? filename))
          ((X_OK)
            (file-execute-access? filename))
          (else (error (format "Unknown kind ~a" kind))))))
    
    (define readlink read-symbolic-link)
    ;]; 
    ;[ fluids ] ;[; 
    ;chicken uses parameters
    (define make-fluid (lambda ()  (make-parameter #f)))
    (define fluid-set! (lambda (f v) (f v)))
    (define fluid-ref (lambda (v) (v)))
    
    (define-macro (with-fluids* fluids vals thunk)
      `(parameterize ,(map (lambda (x y) (list x y)) fluids vals) (,thunk))
    )
    ;]; 
    ;[ hash-tables ] ;[;
    ;r6rs-like make-hashtable
    (define studio-make-hash-table 
      (lambda (hash-func eq-func . size)
        ;hash-function equiv [initial-size]
        (apply make-hash-table eq-func hash-func size)))

    (define studio-make-weak-value-hash-table 
      (lambda (hash-func eq-func . size)
        (let ((ht (apply studio-make-hash-table 
                    hash-func eq-func size)))
          (setprop! ht 'weak-values #t)
          ht)))
                    
    ;studio hash-* functions operate on weak hashes too
    (define hash-set! 
      (lambda (ht key value)
        ;(debugmsg "HASH: ~a\n" (hash-table-keys ht))
        (cond 
          ((getprop 'weak-values ht)
            (hash-table-set! ht key (make-weak-locative value)))
          (else
            (hash-table-set! ht key value)))))
     
    ;studio hashes return false if a key has no value
    (define hash-ref 
      (lambda (ht key)
        (cond
          ((getprop 'weak-values ht)
            ;(display "Returning weak value!\n")
            (if (hash-table-exists? ht key)
              (locative->object (hash-table-ref ht key))
              #f))
          (else 
            (if (hash-table-exists? ht key)
              (hash-table-ref ht key)
              #f)))))

    (define hash-remove! hash-table-delete!)
    
    (define hash-fold 
      (lambda (p init ht)
        (hash-table-fold ht p init)))
    ;]; 
    ;[ hooks ] ;[; 
    (define make-hook
      (lambda (num-args)
        (let ((hook-list '()))
          `(hook .
            ,(lambda (cmd . args)
               (case cmd  
                 ((run)
                     (for-each (lambda (hook) 
                               (apply hook args))
                               hook-list))
                 ((getlist)
                   hook-list)
                 ((reset)
                   (set! hook-list '()))
                 ((add)
                   (set! hook-list (cons (car args) hook-list)))))))))
        
    (define hook?
      (lambda (hook)
        (and (list? hook ) (eq? (car hook) 'hook))
    ))
    
    (define run-hook
      (lambda (hook . args)
        (apply (cdr hook) 'run args)
    ))
    
    (define hook->list
      (lambda (hook)
        ((cdr hook) 'getlist)
    ))
    
    (define reset-hook!
      (lambda (hook)
        ((cdr hook) 'reset)
    ))
    
    (define add-hook!
      (lambda (hook thunk)
        ((cdr hook) 'add thunk)
    ))
    ;]; 
    ;[ I/O ] ;[;
    ; chicken provides flush-output to force output on a port
    (define force-output flush-output)
    (define studio-flush-port flush-output)

    (define close-port 
      (lambda (p)
        ;(debugmsg "CLOSE-PORT ~a:~a\n" p (or (port-closed? p) (fileno p)))
        (if (input-port? p)
          (close-input-port p)
          (begin
            (flush-output p)
            (close-output-port p)))
        ;This shouldn't be necessary (but is).
        ;  It may happen because proverb expects to be able to shutdown 
        ;  both sides of a socket even though it only calls 'shutdown' on 
        ;  one half-duplex port.  I've only seen this loop execute once 
        ;  and only on tcp ports.  If it ever executed more than once, 
        ;  then it is something else.
        (let loop () 
          (if (and (has-fileno? p) (not (port-closed? p)))
            (begin 
              (debugmsg "HAD TO FILE-CLOSE PORT: ~a\n" p)
              (file-close (fileno p))
              (loop))))))
    
    (define close
      (lambda (p-or-fd)
        ;(debugmsg "CLOSE\n")
        (if (port? p-or-fd)
          (close-port p-or-fd)
          (file-close p-or-fd))))

    (define shutdown 
      (lambda (port how)
        ;(debugmsg "SHUTDOWN ~a ~a\n" port how)
        (close-port port)))
    
    ; chicken does full writes anyway
    (define write-string/full
      (lambda (str port start end)
        (write-string (substring str (or start 0)) end port)
        (- end (or start 0))))

    ; this ok?
    (define write-string/partial write-string/full)

    (define read-string!/partial 
      (lambda (s . p-start-stop)
        (cond
          ((null? p-start-stop)
            (read-string! #f s))
          ((= (length p-start-stop) 1)
            (read-string! #f s (car p-start-stop)))
          ((= (length p-start-stop) 2)
            (read-string! #f s (car p-start-stop) (cadr p-start-stop)))
          ((= (length p-start-stop) 3)
            (read-string! (caddr p-start-stop) s (car p-start-stop) (cadr p-start-stop)))
          (else
            (error 'wrong-number-of-args)))))

    ; accepts on a sockfd and returns a pair: (inport . outport)
    ; for chicken, sockfd must have a 'tcp-listener property
    ; set to a 'tcp-listener' returned by tcp-listen
    (define accept-io-socket
      (lambda (sockfd)
        (let ((listener (getprop sockfd 'tcp-listener)))
          (if (and listener (tcp-listener? listener))
            (let-values (((in out) (tcp-accept listener)))
              (cons in out))
            (error "socket fd has no tcp-listener property!\n")))))

    ;chicken provides create-pipe that returns two values
    (define studio-pipe
      (lambda ()
        (let-values (((p1 p2) (create-pipe)))
          (cons 
            (open-input-file* p1) 
            (open-output-file* p2)))))

    (define with-error-to-string
      (lambda (thunk)
        (define error-string (open-output-string))
        (define stderr (current-error-port))
        (let ()
          (current-error-port error-string)
          (thunk)
          (current-error-port stderr)
          (get-output-string error-string))))

    (define with-error-to-port with-error-output-to-port)

    ;XXX not a complete implementation: no r/w ports
    (define make-soft-port
      (lambda (v mode)
        (case (string->symbol mode)
          ((r) (make-input-port 
                 (vector-ref v 3) 
                 (lambda args #t) 
                 (or (vector-ref v 4)
                     (lambda args #t))))
          ((w) (make-output-port 
                 (vector-ref v 0) 
                 (or (vector-ref v 4)
                     (lambda args #t))))
          (else (error (format "Unsupported mode: ~a" mode))))))
           
    (define port-closed? 
      (lambda (p) 
        ;(debugmsg "PORT-CLOSED? ~a\n" p)
        ;is there a better way than this?
        (and (or (port? p) (integer? p))
          (condition-case 
            (let ()
              (file-control 
                (cond 
                  ((integer? p) p )
                  ((port? p) (port->fileno p))
                  (else (error 'invalid-type)))
                fcntl/getfl)
              #f
            )
            (e () #t)))))

    (define port->fileno* 
      (lambda (p)
        (cond 
          ((integer? p) p)
          ((and (port? p) (has-fileno? p))
            (port->fileno p))
          (else #f))))

    ;XXX: chicken's file-select only accepts fds and doesn't select on exceptions
    ; maybe studio select should not accept port values.  what does it mean to select
    ; on a port that isn't backed by a fileno?
    ; 
    (define studio-select
      (lambda (rfds wfds xfds . timeout)
        ;make a mapping of all scheme ports w/ fds
        (let ((input-alist
               (delete #f (map (lambda (p)
                                 (if (has-fileno? p)
                                   (cons (port->fileno p) p)
                                   #f))
                               rfds)))
              (output-alist
               (delete #f (map (lambda (p)
                                 (if (has-fileno? p)
                                   (cons (port->fileno p) p)
                                   #f))
                               wfds)))
              (ready-input-ports
                ;shortcut scheme-ports that may have buffered input
                ;  otherwise file-select might block
                (let loop ((inrfds rfds) (racc '()))
                  (if (null? inrfds)
                    racc
                    (if (and (port? (car inrfds))
                             (char-ready? (car inrfds)))
                      (loop (cdr inrfds)
                            (cons (car inrfds) racc))                      
                      (loop (cdr inrfds) racc))))))

          (if (null? ready-input-ports) 
            (let* ((irfds (if rfds (delete #f (map port->fileno* rfds))))
                   (iwfds (if wfds (delete #f (map port->fileno* wfds)))))
              (let-values (((rset wset) (apply file-select irfds iwfds timeout)))
                ;(debugmsg "SELECT RETURNED: ~a ~a\n" rset wset)
                (list 
                  ;gross
                  (fold (lambda (fd rest)
                          (let ((p (assoc fd input-alist)))
                            (if p 
                              (append 
                                (let loop ((alist (cdr (member p input-alist)))
                                           (racc (list (cdr p))))
                                  (let ((p (assoc fd alist)))
                                    (if p 
                                       (loop (cdr (member p alist))
                                            (cons (cdr p) racc))
                                      racc)))
                                rest)
                              (cons fd rest))))
                        '()
			(or rset '()))
                  (fold (lambda (fd rest)
                          (let ((p (assoc fd output-alist)))
                            (if p 
                              (append 
                                (let loop ((alist (cdr (member p output-alist)))
                                           (racc (list (cdr p))))
                                  (let ((p (assoc fd alist)))
                                    (if p 
                                       (loop (cdr (member p alist))
                                            (cons (cdr p) racc))
                                      racc)))
                                rest)
                              (cons fd rest))))
                        '()
			(or wset '()))
                  '() )))
	    ;short circuit file-select 
	    ;XXX: this could be dangerous if any wfds block, so we only return rfds
            (begin 
              (debugmsg "SHORTCUTTING! ~a\n" ready-input-ports)
              (list ready-input-ports '() '()))))))
	    
    ;];
    ;[ lists ] ;[; 
    (define list-head take)

    (define list-set! 
      (lambda (l i v)
        (let loop ((ll l) (count 0))
          (if (null? l)
            (error "out of range")
            (if (= count i)
              (set-car! ll v)
              (loop (cdr ll) (+ count 1)))))))

    ;]; 
    ;[ load system ] ;[; 
    (define current-load-filename 
      (lambda () 
        (cond 
          ((string=? (chicken-version) "2.6") 
            (eval (with-input-from-string "##sys#current-load-file" read)))
          (else
            (eval (with-input-from-string "##sys#current-source-filename" read))))))

    (define current-load-dir (lambda ()
      (define clf (current-load-filename))
      (or (and clf (pathname-directory clf)) (current-directory) ".")))

    (define append-load-path! (lambda (p)
      ;(set! %load-path (append %load-path (list (eval (with-input-from-string "##sys#current-load-path" read)))))
      (debugmsg "CHICKEN: append-load-path! not sorted out yet\n")))

    (define-syntax include*
      (lambda (x)
        (define read-file
          (lambda (fn k)
            (let ((file-to-include (string-append (pathname-directory current-include-filename) "/" fn)))
              (let ((p (open-input-file file-to-include)))
                (set! current-include-filename file-to-include)
                ;(display current-include-filename )(newline)
                (let f ((x (read p)))
                  (if (eof-object? x)
                      (begin (close-input-port p) '())
                      (cons (datum->syntax-object k (macroexpand x))
                            (f (read p)))))))))
        (syntax-case x ()
           ((k filename)
            (let ((fn (syntax-object->datum (syntax filename))))
              (with-syntax (((exp ...) (read-file (eval fn) (syntax k))))
                (syntax (begin exp ...))))))))
    
    (define-syntax define-alias
      (syntax-rules ()
        [(_ x y)
         (define-syntax x
           (identifier-syntax y))]))

    ;(define-alias load include*)
    
    (define-syntax define-module (syntax-rules ()
      ((define-module name) #f)
    ))

    ;]; 
    ;[ misc ] ;[;
    (define sync void) ;is this like disk sync?

    (define the-environment interaction-environment)

    (define-macro (1+ v) `(+ 1 ,v))

    (define *unspecified* (void))

    (define defined? global-bound?)

    (define unspecified? 
      (lambda (val)
        (eq? val (void))))

    (define make-guardian
      (lambda ()
        (let ((guarded-objects (make-queue)))
          (lambda args
            (if (null? args)
              (queue-get guarded-objects)
              (for-each 
                (lambda (obj)
                  (set-finalizer! 
                    obj 
                    (lambda (obj) 
                      (queue-add guarded-objects obj))))
                args))))))

    (define read-hash-extend
      (lambda (c handler)
        (set-sharp-read-syntax! c (lambda (port) (handler c port)))))
    
    (define getpwnam
      (lambda (user)
        (list->vector (user-information user))))

    (define-macro (tilde-wrap function)
      (define s (gensym))
      `(let ()
        (define ,s ,function)
        (lambda (arg)
          (if (and (string? arg) (> (string-length arg) 1) (eq? (string-ref arg 0) #\~))
            (if (eq? (string-ref arg 1) #\/)
              (,s (string-append (getenv "HOME") (substring arg 1)))
              (let ()
                (define i (string-index arg #\/))
                (if i
                  (let ()
                    (define pw (getpwnam (substring arg 1 i)))
                    (if pw
                      (,s (string-append (vector-ref pw 5) (substring arg i)))
                      (,s arg)
                    )
                  )
                  (,s arg)
                )
              )
            )
            (,s arg)
          )
        )
      )
    )

    ;; Make unicode lambda work. All Hail Guy Steele.
    (define-macro (λ . body) `(lambda ,@body))
    ;];
    ;[ net ] ;[; 
    (define gethostname get-host-name)
    (define getsockname
      (lambda (sock)
        (let-values (((local-addr remote-addr) (tcp-addresses sock))
                     ((local-port remote-port) (tcp-port-numbers sock)))
        (vector AF_INET local-addr local-port))))
    (define getpeername
      (lambda (sock)
        (let-values (((local-addr remote-addr) (tcp-addresses sock))
                     ((local-port remote-port) (tcp-port-numbers sock)))
        (vector AF_INET remote-addr remote-port))))
    ;];
    ;[ object properties ] ;[; 
    (define object->string ->string)

    ;XXX: should this be refactored?
    (define procedure-source  
      (lambda (obj) 
        (if (extended-procedure? obj)
          (let ((src (getprop obj 'procedure-source)))
            (if src src #f))
          #f)))

    ; this is a weak value hash, but the studio weak value hash is implemented
    ;  with properties, so this can't use it. 
    (define global-object-property-hash (studio-make-hash-table string-hash string=?))

    (define set-object-property! 
      (lambda (obj key val)
        (if (not (symbol? key)) (error 'invalid-property-name))
        (let ((props-hash (hash-ref global-object-property-hash
                               (number->string (object-address obj)))))
          (if props-hash
            (hash-set! props-hash key val)
            (let ()
              (set-finalizer! 
                obj  
                (lambda (o)
                  (hash-remove!  
                    global-object-property-hash
                    (number->string (object-address o)))))
              (let ((new-props-hash (studio-make-hash-table hash eq?)))
                (hash-set! new-props-hash key val)
                (hash-set! 
                  global-object-property-hash
                  (number->string (object-address obj))
                  new-props-hash)))))))
                
    (define object-properties 
      (lambda (obj)
        (let ((props-hash (hash-ref global-object-property-hash
                               (number->string (object-address obj)))))
          (if props-hash
            (hash-table->alist props-hash)
            '()))))
       
    (define object-property 
      (lambda (obj key)
        (let ((props-hash (hash-ref global-object-property-hash
                               (number->string (object-address obj)))))
          (if props-hash
            (hash-ref props-hash key)
            #f))))
    
    (define object-address 
      (lambda (obj)
        (let ((p (object->pointer obj)))
          (if p 
            (pointer->address p)
            #f))))
    
    (define procedure-property
      (lambda (obj prop)
        (if (eq? prop 'arity)
          (list
            (- (length (procedure-information obj)) 1)
            0
            (not (null? (cdr (last-pair (procedure-information obj))))))
          (object-property obj prop))))
    
    (define set-procedure-property! set-object-property!)
        
    (define procedure-properties 
      (lambda (obj)
        (append
          (if (procedure-information obj)
            `((arity ,(- (length (procedure-information obj)) 1)
                     0
                     ,(not (null? (cdr (last-pair (procedure-information obj)))))))
             '())
          (object-properties obj))))
    
    ;]; 
    ;[ processes ] ;[; 
    (define getpid current-process-id)
    (define setpgid set-process-group-id!)
    (define getpgid process-group-id)

    ;]; 
    ;[ regex ] ;[;
    (define make-regexp regexp)
    (define regexp-exec string-match)
    (define regexp-substitute/global
      (lambda (port regexp target . items)
        (string-substitute 
          regexp 
          (apply string-append 
                 (map 
                   (lambda (item) 
                     (cond ((number? item) (format "\\~a" item))
                             ((string? item) item)
                           (else "")))
                   items))
          target)))
    (define delete-regexp (lambda (re l)
      ;"delete elements from <l> that match the regular expression <re>"
      (define fail (cons 1 1))
      (if (null? l)
        l
        (if (string? re)
          (delete-regexp (regexp re) l)
          (let helper ((cursor l) (racc '()))
            (if (null? cursor)
              (reverse racc)
              (if (string-match re (car cursor))
                (helper (cdr cursor) racc)
                (helper (cdr cursor) (cons (car cursor) racc))
              )
            )
          )
        )
      )
    ))
    ;];
    ;[ strings ] ;[; 
    (define studio-string-split 
      (lambda (str . delstr-and-flag)
        (if (not (null? delstr-and-flag))
          (if (char? (car delstr-and-flag))
            (apply string-split str 
                                     (string (car delstr-and-flag))
                                     (cdr delstr-and-flag))
            (apply string-split str delstr-and-flag)))))

    ;]; 
    ;[ threads ] ;[; 
    ;chicken has mutex-lock! and mutex-unlock!
    (define lock-mutex mutex-lock!)
    (define unlock-mutex mutex-unlock!)
    (define join-thread thread-join!)
    (define call-with-new-thread make-thread)
    (define-macro (with-mutex mutex . body)
      `(dynamic-wind
        (lambda () (lock-mutex ,mutex))
        (lambda () ,@body)
        (lambda () (unlock-mutex ,mutex))))
    ;]; 
  )
  (chez
    (display "LOADING CHEZ-SPECIFIC DIALECT CODE!\n")
    (define-syntax define-module (syntax-rules ()
      ((define-module name) #f)
    ))
    (display "DONE!\n")
  )
  (else
    (display "LOADING IMPLEMENTATION-SPECIFIC DIALECT CODE!\n")
  )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Place experimental dialect code below this line
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;[ threadable forms]
;[;
(cond-expand
  (guile
    (use-modules (ice-9 syncase))
    (use-modules (ice-9 threads))
    (define-syntax with-maybe-mutex (syntax-rules ()
      ((with-maybe-mutex mutex body ...)
        ;(if mutex
        ;  (dynamic-wind
        ;     (lambda () (lock-mutex mutex))
        ;     (lambda () (begin body ...))
        ;     (lambda () (unlock-mutex mutex)))
          (begin body ...)
        ;)
      )
    ))
    (define sleazy-threadable #f)
    (define sleazy-threads-left 8)
    (define sleazy-threads-mutex (make-mutex))
    (define sleazy-threads-incr (with-docstring
      "private routine to atomically increment the number of pre-spawned threads available"
      (lambda ()
        (lock-mutex sleazy-threads-mutex)
        (set! sleazy-threads-left (+ sleazy-threads-left 1))
        (unlock-mutex sleazy-threads-mutex)
      )
    ))
    (define sleazy-threads-decr (with-docstring
      "private routine to atomically decrement the number of pre-spawned thread available"
      (lambda ()
        (lock-mutex sleazy-threads-mutex)
        (set! sleazy-threads-left (- sleazy-threads-left 1))
        (unlock-mutex sleazy-threads-mutex)
      )
    ))
    (define-syntax threadable-let (syntax-rules ()
      ((threadable-let bindings body ...)
        (if sleazy-threadable
          (let bindings
            (begin
              (lock-mutex sleazy-threads-mutex)
              (if (> sleazy-threads-left 0)
                (begin
                  (set! sleazy-threads-left (- sleazy-threads-left 1))
                  (critmsg "threadable-let: ~a threads left\n" sleazy-threads-left)
                  (unlock-mutex sleazy-threads-mutex) 
                  (let ()
                    (define thread (call-with-new-thread
                      (lambda ()
                        (define retval ((lambda () body ...)))
                        (sleazy-threads-incr)
                        retval
                      )
                    ))
                    (delay (join-thread thread))
                  )
                )
                (begin
                  (unlock-mutex sleazy-threads-mutex) 
                  ;(let bindings body ...)
                  (let () body ...)
                )
              )
            )
          )
          (let bindings body ...)
        ) 
      )
    ))
    (define exclusive (with-docstring
      "accept a Scheme function <func> and return a new function that wraps <func> and which will complain if entered recursively"
      (lambda (func)
        (define in? #f)
        (lambda args
          (if in?
            (begin
              (critmsg "AHH! ENTERED AN EXCLUSIVE FORM MULTIPLY ~a\n" func)
              (error 'reentered-exclusive-form)
            )
            (begin
              (set! in? #t)
              (let ()
                (define retval (apply func args))
                (set! in? #f)
                retval
              )
            )
          )   
        )
      )
    ))
  )
  (chicken
    (define-syntax with-maybe-mutex (syntax-rules ()
      ((with-maybe-mutex mutex body ...)
        ;(if mutex
        ;  (dynamic-wind
        ;     (lambda () (lock-mutex mutex))
        ;     (lambda () (begin body ...))
        ;     (lambda () (unlock-mutex mutex)))
          (begin body ...)
        ;)
      )
    ))
    (define sleazy-threadable #f)
    (define sleazy-threads-left 8)
    (define sleazy-threads-mutex (make-mutex))
    (define sleazy-threads-incr (lambda ()
      (lock-mutex sleazy-threads-mutex)
      (set! sleazy-threads-left (+ sleazy-threads-left 1))
      (unlock-mutex sleazy-threads-mutex)
    ))
    (define sleazy-threads-decr (lambda ()
      (lock-mutex sleazy-threads-mutex)
      (set! sleazy-threads-left (- sleazy-threads-left 1))
      (unlock-mutex sleazy-threads-mutex)
    ))
    (define-syntax threadable-let (syntax-rules ()
      ((threadable-let bindings body ...)
        (if sleazy-threadable
          (let bindings
            (begin
              (lock-mutex sleazy-threads-mutex)
              (if (> sleazy-threads-left 0)
                (begin
                  (set! sleazy-threads-left (- sleazy-threads-left 1))
                  (unlock-mutex sleazy-threads-mutex) 
                  ;(format emergency-port "LEFT: ~a\n" sleazy-threads-left)
                  (let ()
                    (define thread (call-with-new-thread (lambda () (define retval ((lambda () body ...))) (sleazy-threads-incr) retval)))
                    (delay (join-thread thread))
                  )
                )
                (begin
                  (unlock-mutex sleazy-threads-mutex) 
                  (let bindings body ...)
                )
              )
            )
          )
          (let bindings body ...)
        ) 
      )
    ))
    (define exclusive (lambda (func)
      (define in? #f)
      (lambda args
        (if in?
          (critmsg "AHH! ENTERED AN EXCLUSIVE FORM MULTIPLY\n")
          (begin
            (set! in? #t)
            (let ()
              (define retval (apply func args))
              (set! in? #f)
              retval
            )
          )
        ) 
      )
    ))
  )
  (chez
    (display "EXPANDING!\n")
    (extend-syntax (threadable-let)
     [(threadable-let bindings body ...)
       (let bindings body ...)]
    )
    (display "DONE EXPANDING!\n")
    (display "CHEZ LOADING EXPERIMENTAL FORMS!\n")
  )
  (else
    (display "ADDING THREADING CODE FOR NON-SPECIFIC DIALECTS!\n")
  )
)
;];

;[ chicken modules ] -- WARNING: USE at own risk
;[; 
(cond-expand
  (chicken 
    (define current-module interaction-environment)

    (define-syntax define-public (syntax-rules ()
      ((define-public name value)
        (define name value)
      )))
  )
  (else
    #t
  )
)
;]; 

)
