(define svgsample (with-docstring
  "create a sample SVG tag that creates a red circle with a black border. center X and center Y are independent random variables between 0 and 99. radius is an random value between 0 and 19. must be called in a context like that created by svgwrap in order to be visible."
  (lambda ()
    (with-xml-tag `(circle xmlns "http://www.w3.org/2000/svg"
      cx ,(random 100)
      cy ,(random 100)
      r ,(random 20)
      stroke "black"
      stroke-width 2
      fill "red")
    )
  )
))

(define svgsamplerect (with-docstring
  "create a sample SVG tag"
  (lambda (ulx uly lrx lry)
    (with-xml-tag `(circle xmlns "http://www.w3.org/2000/svg"
      cx ,(+ ulx (/ (- lrx ulx) 2))
      cy ,(+ uly (/ (- lry uly) 2))
      r
      ,(if (> (- lrx ulx) (- lry uly))
        (/ (- lry uly) 2)
        (/ (- lrx ulx) 2)
      )
      stroke "black"
      stroke-width 2 
      fill "red")
    )
  )
))

(define svg2 (with-docstring
  "a function like svgsample. creates a green circle"
  (lambda ()
    (with-xml-tag `(circle xmlns "http://www.w3.org/2000/svg"
      cx ,(random 100)
      cy ,(random 100)
      r  ,(random 20)
      stroke "black"
      stroke-width 2
      fill "green")
    )
  )
))

(define object_update_tracking "
")

(define svgdemo (with-docstring
  "a complete interactive demo of SVG in a Proverb widget"
  (lambda ()
    (define scale 3)
    (define width 100)
    (define height 100)

  (widgetscript-apply "
  function(width,height,scale) {

     var is_ipad=function() {
       if(navigator && navigator.platform && navigator.platform == 'iPad') {
         return true;
       } else {
         return false;
       }
     }

     function make_button(label) {
       var b = document.createElement('input');
       b.type='button';
       b.value=label;
       return b;
     }; 

     var current_width;
     current_width=width*scale;
     var current_height;
     current_height=height*scale;

     var d=document.createElement('div');
     d.style.backgroundColor='yellow';
     d.width=current_width+'px';
     widget.div.appendChild(d);
     var s=document.createElementNS('http://www.w3.org/2000/svg','svg'); /* rewrite with (pixels (* width bytesize)) */
     s.setAttribute('width',current_width+'px');
     s.setAttribute('height',(height*scale)+'px');
     s.setAttribute('viewbox','0 0 100 100'); /* rewrite with width, height */
     var g=document.createElementNS('http://www.w3.org/2000/svg','g');
     s.appendChild(g);

     var subd=document.createElement('div');
     subd.style.backgroundColor='pink';
     subd.style.position='relative';
     subd.style.width=current_width+'px';
     subd.style.height=current_height+'px';
     s.style.position='absolute';
     s.style.left='0px';
     s.style.top='0px';
     var stretchy=document.createElement('div');
     stretchy.style.position='absolute';
     if(is_ipad()) {
       stretchy.style.width='30px';
       stretchy.style.height='30px';
     } else {
       stretchy.style.width='10px';
       stretchy.style.height='10px';
     }
     stretchy.style.left=( (width*scale) - 10) + 'px';
     stretchy.style.top=( (width*scale) - 10) + 'px';
     stretchy.style.backgroundColor='green';
     var bogo_scale_matrix=false;
     var mk_stretchy_stardown=function(element,movename,upname,getxy) {
       return function(e) {
         e.preventDefault();
         e.stopPropagation();
         var base_width=current_width;
         var base_height=current_height;
         var e_xy;
         var start_x;
         var start_y;
         e_xy=getxy(e);
         start_x=e_xy[0];
         start_y=e_xy[1];
         var common=function(e) {
           var e_xy=getxy(e);
           e.preventDefault();
           e.stopPropagation();
           current_width=base_width + (e_xy[0] - start_x);
           current_height=base_height + (e_xy[1] - start_y);
           if(current_width < 50) {
             current_width=50;
           }
           if(current_height < 50) {
             current_height=50;
           }
           var m=s.createSVGMatrix();
           bogo_scale_matrix=m.scale(current_width/(width*scale),current_height/(height*scale));
           make_bogo_matrix();
           s.setAttribute('width',current_width+'px');
           s.setAttribute('height',current_height+'px');
         subd.style.width=current_width+'px';
         subd.style.height=current_height+'px';
         element.style.left=(current_width - 10) + 'px';
         element.style.top=(current_height - 10) + 'px';
         var t = s.createSVGTransform();
         t.setMatrix(bogo_scale_matrix);
         if(g.transform.baseVal.numberOfItems == 0) {
           g.transform.baseVal.appendItem(t);
         } else {
           g.transform.baseVal.replaceItem(t,0);
         }
       }
       var move_handler=function(e) {
         common(e);
         var m=s.createSVGMatrix();
       };
       var up_handler=function(e) {
         var m=s.createSVGMatrix();
         window.removeEventListener(upname,up_handler,true);
         window.removeEventListener(movename,move_handler,true);
         common(e);
       };
       window.addEventListener(upname,up_handler,true);
       window.addEventListener(movename,move_handler,true);
       }
     };
     stretchy.addEventListener('mousedown',mk_stretchy_stardown(stretchy,'mousemove','mouseup',function(e) { return [e.clientX,e.clientY]}),false);
     var touchfoo=mk_stretchy_stardown(stretchy,'touchmove','touchend',function(e) { var t0=e.changedTouches.item(0); return [t0.clientX,t0.clientY]});
     stretchy.addEventListener('touchstart',function(e) { touchfoo(e); },false);


        
     var ns='http://www.w3.org/2000/svg';
     var ulx,uly,lrx,lry;
     ulx=0; uly=0; lrx=100; lry=100;

     var back=document.createElementNS('http://www.w3.org/2000/svg','rect');
     back.setAttribute('width',(width*scale)+'px');
     back.setAttribute('height',(height*scale)+'px');
     back.setAttribute('style','fill:white;');
     g.appendChild(back);
     var pointer1=document.createElementNS('http://www.w3.org/2000/svg','rect');
     pointer1.setAttribute('x',10); pointer1.setAttribute('y',10); pointer1.setAttribute('style','fill:red;');
     pointer1.setAttribute('width',10); pointer1.setAttribute('height',10);
     g.appendChild(pointer1);
     var pointer2=document.createElementNS('http://www.w3.org/2000/svg','rect');
     pointer2.setAttribute('x',20); pointer2.setAttribute('y',20); pointer2.setAttribute('style','fill:green;');
     pointer2.setAttribute('width',5); pointer2.setAttribute('height',5);
     g.appendChild(pointer2);
     var left_button=make_button('left');
     d.appendChild(left_button);
     var right_button=make_button('right');
     d.appendChild(right_button);
     var foreground_button=make_button('foreground');
     d.appendChild(foreground_button);
     var new_button=make_button('new');
     d.appendChild(new_button);
     var delete_button=make_button('delete');
     d.appendChild(delete_button);
     var rpc_button=make_button('rpc');
     d.appendChild(rpc_button);
     var info_button=make_button('info');
     d.appendChild(info_button);
     d.appendChild(document.createElement('hr'));

     var zoomin_button=make_button('zoom in');
     d.appendChild(zoomin_button);
     var zoomout_button=make_button('zoom out');
     d.appendChild(zoomout_button);
 
     function make_color_button(color) {
       var b;
       b=document.createElement('div');
       b.style.width='20px';
       b.style.height='20px';
       if(is_ipad()) {
         b.className='touchable';
         b.addEventListener('touchstart',function(e) {
           current_color=color;
         },false);
       } else {
       }
       b.style.background=color;
       b.addEventListener('click',function(e) {
         current_color=color;
       },false);
       return b;
     };
     d.appendChild(make_color_button('black'));
     d.appendChild(make_color_button('red'));
     d.appendChild(make_color_button('orange'));
     d.appendChild(make_color_button('yellow'));
     d.appendChild(make_color_button('green'));
     d.appendChild(make_color_button('indigo'));
     d.appendChild(make_color_button('blue'));
     d.appendChild(make_color_button('violet'));

     var make_bogo_buttons = function(values,def,setter) {
       var buttons;
       function make_value_button(value) {
         var a;
         a=document.createElement('span');
         if(value == def) {
           a.style.backgroundColor='red';
         } else {
           a.style.backgroundColor='blue';
         }
         a.style.paddingLeft='1em';
         a.style.paddingRight='1em';
         var b;
         b=document.createElement('a');
         b.style.height='20px';
         if(is_ipad()) {
           b.className='touchable';
         }
         b.appendChild(document.createTextNode('' + value));
         b.addEventListener('click',function(e) {
           setter(value);
           var j;
           for(j in values) { 
             if(values[j] == value) {
               buttons[j].style.backgroundColor='red';
             } else {
               buttons[j].style.backgroundColor='blue';
             }
           }
         },false);
         b.addEventListener('touchstart',function(e) {
           setter(value);
           var j;
           for(j in values) { 
             if(values[j] == value) {
               buttons[j].style.backgroundColor='red';
             } else {
               buttons[j].style.backgroundColor='blue';
             }
           }
         },false);
         a.appendChild(b);
         return a;
       }
       var i;
       buttons=new Array;
       for(i in values) { 
         buttons[i]=make_value_button(values[i]);
         d.appendChild(buttons[i]);
       }
     }

     d.appendChild(document.createTextNode('opacity: '));

     var current_opacity=75;
     make_bogo_buttons([100,75,50,25,0],current_opacity,function(value) { current_opacity=value; });

     d.appendChild(document.createElement('hr'));

     d.appendChild(document.createTextNode('mode: '));

     var current_mode='line';
     make_bogo_buttons(['line','polygon','rect','spline','rpc'],current_mode,function(value) { current_mode=value; });

     d.appendChild(document.createElement('hr'));

     var eval_text=document.createElement('input');
     eval_text.type='text';
     d.appendChild(eval_text);
     d.appendChild(document.createElement('hr'));

     //var fake_input=document.createElement('input');
     //fake_input.style.width='0px';
     //fake_input.style.height='0px';
     //d.appendChild(fake_input);
     //s.addEventListener('mouseover',function(e) {
     //  fake_input.focus(); 
     //},false);
     //s.addEventListener('mouseout',function(e) {
     //  fake_input.blur();
     //},false);
     //fake_input.addEventListener('keypress',function(e) {
     //  subshell.string_evaluator('] (svgsample)',false);
     //},false);

     d.appendChild(subd);
     subd.appendChild(s);
     subd.appendChild(stretchy);

     var selected_object=false;

     var subshell=shell.subshell(function(v) {
       if(v) {
         make_trackable(v);
         g.appendChild(v);
       }
     });

     eval_text.addEventListener('change',function(e) {
       subshell.string_evaluator(eval_text.value,false);
       eval_text.value='';
     },false);

     left_button.addEventListener('click',function(e) {
       var span=lrx-ulx;
       lrx=lrx-(span*.1);  
       ulx=ulx-(span*.1);  
       s.setAttribute('viewBox',ulx + ' ' + uly + ' ' + lrx + ' ' + lry);
     },false);

     right_button.addEventListener('click',function(e) {
       var span=lrx-ulx;
       lrx=lrx+(span*.1);  
       ulx=ulx+(span*.1);  
       s.setAttribute('viewBox',ulx + ' ' + uly + ' ' + lrx + ' ' + lry);
     },false);

     zoomin_button.addEventListener('click',function(e) {
       var span=lrx-ulx;
       var newspan=span*.9
       lrx=lrx-((span-newspan)/2);
       ulx=ulx+((span-newspan)/2);

       lry=lry-((span-newspan)/2);
       uly=uly+((span-newspan)/2);

       s.setAttribute('viewBox',ulx + ' ' + uly + ' ' + lrx + ' ' + lry);
     },false);

     zoomout_button.addEventListener('click',function(e) {
       var span=lrx-ulx;
       var newspan=span*.9
       lrx=lrx+((span-newspan)/2);
       ulx=ulx-((span-newspan)/2);

       lry=lry+((span-newspan)/2);
       uly=uly-((span-newspan)/2);

       s.setAttribute('viewBox',ulx + ' ' + uly + ' ' + lrx + ' ' + lry);
     },false);

     foreground_button.addEventListener('click',function(e) {
       if(selected_object) {
         selected_object.parentNode.appendChild(selected_object);
       }
     },false);

     rpc_button.addEventListener('click',function(e) {
       subshell.string_evaluator('] (svgsample)',false);
     },false);

     new_button.addEventListener('click',function(e) {
       var n=document.createElementNS(ns,'rect');
       n.setAttribute('y',''+40);
       n.setAttribute('x',''+40);
       n.setAttribute('height',''+10);
       n.setAttribute('width',''+10);
       n.setAttribute('style','fill: blue;');
       g.appendChild(n);
       make_trackable(n);
     },false);

     delete_button.addEventListener('click',function(e) {
       if(selected_object) {
         selected_object.parentNode.removeChild(selected_object); 
       }
     },false);

     info_button.addEventListener('click',function(e) {
       if(selected_object) {
         var m = s.createSVGMatrix();
         var m2;
         cm =offset_matrix;
         alert('pre cm: ' + cm.a + ' ' + cm.b + ' ' + cm.c + ' ' + cm.d + ' ' + cm.e + ' ' + cm.f );
         alert('pre m: ' + m.a + ' ' + m.b + ' ' + m.c + ' ' + m.d + ' ' + m.e + ' ' + m.f );
         m2=m.rotate(20.0);
         m.translate(5,5);
         alert('post m: ' + m2.a + ' ' + m2.b + ' ' + m2.c + ' ' + m2.d + ' ' + m2.e + ' ' + m2.f );
         var t = s.createSVGTransform();
         t.setMatrix(m2);
         alert('still here: ' + t);

         var t = selected_object.transform.baseVal.appendItem(t);
         var t = selected_object.transform.animVal.appendItem(t);

         alert('still still here'); 
             
       } else {
         alert('no selected object');
       }
     },false);

     var pointer_tracking=false;
     var pointer_drawing=false;
     var start_point;
     var start_transform;

     var bogo_matrix=false;
     var bogo_inverse_matrix=false;
     var make_bogo_matrix=function() {
       bogo_matrix=s.getScreenCTM();
       if(bogo_scale_matrix) {
         bogo_matrix=bogo_matrix.multiply(bogo_scale_matrix);
       }
       bogo_inverse_matrix=bogo_matrix.inverse();
     };
     var get_bogo_inverse_matrix=function() {
       if(bogo_inverse_matrix) {
         return bogo_inverse_matrix;
       } else {
         make_bogo_matrix();
         return bogo_inverse_matrix;
       }
     };

     function start_tracking(o,x,y) {
       make_bogo_matrix();
       selected_object=o;
       var p = s.createSVGPoint();
       p.x = x;
       p.y = y;
       p = p.matrixTransform(get_bogo_inverse_matrix());

       start_point=p;

       if( selected_object.transform.baseVal.numberOfItems == 0) {
         var nm=s.createSVGMatrix();
         var t=s.createSVGTransform();
         t.setMatrix(nm);
         selected_object.transform.baseVal.appendItem(t);
         start_transform=t;
       } else {
         start_transform=selected_object.transform.baseVal.consolidate();
       }

       pointer_tracking=true;
     }

     function object_update_tracking(o,sp,st,p) {
       var nm = s.createSVGMatrix();
       var t = s.createSVGTransform();

       var t = s.createSVGTransform();
       t.setMatrix(st.matrix.translate(p.x-sp.x,p.y-sp.y));
       o.transform.baseVal.replaceItem(t,0);

       if(o.onmove) {
         if(o.onmoving) {
           //alert('already moving: ' + o.onmoving);
           1;
         } else {
           o.onmoving=true;
           o.onmove(sp,st,p);
           o.onmoving=false;
         }
       }
     }

     function update_tracking(x,y) {
       var p = s.createSVGPoint();
       p.x = x;
       p.y = y;
       p = p.matrixTransform(get_bogo_inverse_matrix());
       object_update_tracking(selected_object,start_point,start_transform,p);
     }

     var current_line;
     var current_color='black';
     var current_points;
     var current_segs;
     function continue_drawing(p) {
       var l;
       current_points.unshift(p)
       l=document.createElementNS(ns,'line');
       current_segs.unshift(l);
       current_line=l;
       current_line.setAttribute('x1',p.x);
       current_line.setAttribute('y1',p.y);
       current_line.setAttribute('x2',p.x);
       current_line.setAttribute('y2',p.y);
       current_line.setAttribute('stroke',current_color);
       current_line.setAttribute('stroke-width',1);
       current_line.setAttribute('stroke-opacity',current_opacity/100);
       g.appendChild(current_line);
     }

     function update_drawing(x,y) {
       var p = s.createSVGPoint();
       p.x = x;
       p.y = y;
       p = p.matrixTransform(get_bogo_inverse_matrix());

       current_line.setAttribute('x2',p.x);
       current_line.setAttribute('y2',p.y);
       continue_drawing(p);
     }

     function start_drawing(x,y) {
       make_bogo_matrix();
       var p = s.createSVGPoint();
       p.x = x;
       p.y = y;
       p = p.matrixTransform(get_bogo_inverse_matrix());

       current_points=Array();
       current_segs=Array();
       pointer_drawing=true;
       continue_drawing(p);
     }

     function make_absorbing_redrawer(handler) {
       return function() {
         var x;
         var r=handler(current_points);
         while(x=current_segs.pop()) {
           g.removeChild(x);
         } 
         return r;
       };
     }

     var polygon_redrawer=make_absorbing_redrawer(function(points) {
       var x,p;
       p=document.createElementNS(ns,'polyline');
       p.style.stroke=current_color;
       p.setAttribute('opacity',current_opacity/100);
       p.setAttribute('fill',current_color);
       while(x=points.pop()) {
         p.points.appendItem(x);        
       }
       return p;
     });

     var spline_redrawer=make_absorbing_redrawer(function(points) {
       var x,p,s,start,last;
       p=document.createElementNS(ns,'path');
       start=x=points.pop();
       s=p.createSVGPathSegMovetoAbs(x.x,x.y);
       p.pathSegList.appendItem(s);
       p.style.stroke=current_color;
       p.setAttribute('opacity',current_opacity/100);
       while(x=points.pop()) {
         last=x;
         //s=p.createSVGPathSegLinetoAbs(x.x,x.y);
         s=p.createSVGPathSegCurvetoQuadraticSmoothAbs(x.x,x.y);
         p.pathSegList.appendItem(s);
       }
       //p.pathSegList.appendItem(p.createSVGPathSegMovetoAbs(last.x,last.y));
       p.pathSegList.appendItem(p.createSVGPathSegMovetoAbs(start.x,start.y));
       //p.pathSegList.appendItem(p.createSVGPathSegClosePath());
       return p;
     });

     var make_absorbing_bb_redrawer=function(handler) {
       return make_absorbing_redrawer(function(points) {
           var ul,ur,ll,lr;
           var x;
           var p;
           ul=s.createSVGPoint();
           lr=s.createSVGPoint();
           x=points.pop();
           ul.x=lr.x=x.x;
           ul.y=lr.y=x.y;
           while(x=points.pop()) {
             if(x.x < ul.x) {
               ul.x=x.x;
             } else if(x.x>lr.x) {
               lr.x=x.x;
             }
             if(x.y < ul.y) {
               ul.y=x.y; 
             } else if(x.y>lr.y) {
               lr.y=x.y;
             }
           }
           return handler(ul,lr);
       });
     };

     var rect_redrawer=make_absorbing_bb_redrawer(function(ul,lr) {
           var ur,ll;
           var x;
           var p;
           ur=s.createSVGPoint();
           ll=s.createSVGPoint();
           ur.y=ul.y; ur.x=lr.x;
           ll.x=ul.x; ll.y=lr.y;

           p=document.createElementNS(ns,'polyline');
           p.style.stroke=current_color;
           p.setAttribute('opacity',current_opacity/100);
           p.setAttribute('fill',current_color);
           p.points.appendItem(ul);
           p.points.appendItem(ur);
           p.points.appendItem(lr);
           p.points.appendItem(ll);
           return p;
     });

     var rpc_redrawer=make_absorbing_bb_redrawer(function(ul,lr) {
       subshell.string_evaluator('] (svgsamplerect ' + ul.x + ' ' + ul.y + ' ' + lr.x + ' ' + lr.y + ' )',false);
     });

     function stop_drawing() {
         var p;
         pointer_drawing=false;
         if(current_mode == 'polygon') {
           p=polygon_redrawer();
           make_trackable(p);
           g.appendChild(p); 
           selected_object=p;
         } else if(current_mode == 'rect') {
           p=rect_redrawer();
           make_trackable(p);
           g.appendChild(p);
           selected_object=p;
         } else if(current_mode == 'rpc') {
           rpc_redrawer();
         } else if(current_mode == 'spline') {
           p=spline_redrawer();
           make_trackable(p);
           g.appendChild(p);
           selected_object=p;
         }
     }

     function make_trackable(o) {
       if(o.onmove) {
         alert('making trackable an onmove object');
       }
       if(is_ipad()) {
         o.addEventListener('touchstart',function(e) {
           var t0=e.changedTouches.item(0);
           start_tracking(o,t0.clientX,t0.clientY);
         },false);
       } else {
         o.addEventListener('click',function(e) {
           e.stopPropagation();
           return false;
         },true);
         o.addEventListener('mousedown',function(e) {
           start_tracking(o,e.clientX,e.clientY);
         },false);
       }
     }

     make_trackable(pointer1);
     make_trackable(pointer2);

     if(is_ipad()) {
       s.addEventListener('touchend',function(e) {
         var t0=e.changedTouches.item(0);
         if(pointer_tracking) {
           update_tracking(t0.clientX,t0.clientY);
           pointer_tracking=false;
         }
         if(pointer_drawing) {
           update_drawing(t0.clientX,t0.clientY);
           stop_drawing();
         }
         // leave it selected
         //selected_object=false;
         e.stopPropagation();
         e.preventDefault();
         return false;
       },true);
       s.addEventListener('touchmove',function(e) {
         var t0=e.changedTouches.item(0);
         if(pointer_tracking) {
           update_tracking(t0.clientX,t0.clientY);
         }
         if(pointer_drawing) {
           update_drawing(t0.clientX,t0.clientY);
         }
         e.stopPropagation();
         e.preventDefault();
         return false;
       },true);
       s.addEventListener('touchstart',function(e) {
         var t0=e.changedTouches.item(0);
         if(!pointer_tracking)
           start_drawing(t0.clientX,t0.clientY);
         e.stopPropagation();
         e.preventDefault();
       },false);
     } else {
       s.addEventListener('mouseup',function(e) {
         if(pointer_tracking) {
           update_tracking(e.clientX,e.clientY);
           pointer_tracking=false;
         }
         if(pointer_drawing) {
           update_drawing(e.clientX,e.clientY);
           stop_drawing();
         }
         // leave it selected
         //selected_object=false;
         e.stopPropagation();
         e.preventDefault();
         return false;
       },true);

       s.addEventListener('mousemove',function(e) {
         if(pointer_tracking) {
           update_tracking(e.clientX,e.clientY);
         }
         if(pointer_drawing) {
           update_drawing(e.clientX,e.clientY);
         }
         e.stopPropagation();
         e.preventDefault();
         return false;
       },true);

       s.addEventListener('mousedown',function(e) {
         if(!pointer_tracking)
           start_drawing(e.clientX,e.clientY);
         e.stopPropagation();
         e.preventDefault();
       },false);
     }

     subd.addEventListener('keypress',function(e) {
       d.appendChild(document.createTextNode('kp'));
     },false);
     subd.addEventListener('keydown',function(e) {
       d.appendChild(document.createTextNode('kd'));
     },false);
     subd.addEventListener('keyup',function(e) {
       d.appendChild(document.createTextNode('ku'));
     },false);

    }
    "
    (list width height scale))
  )
))

(define svgboxchart (with-docstring
  "creates an SVG box chart for the numbers in <lst>. Must be wrapped with something like svgwrap"
  (lambda (lst)
  (define numify (lambda (v)
    (if (exact? v)
      (exact->inexact v)
      v
    )
  ))
  (define m (apply max lst))
  (with-xml-tag `(g xmlns "http://www.w3.org/2000/svg" transform ,(format #f "scale(~f ~f)" (numify (/ 100 (length lst))) (numify (/ 50 m))))
    (with-xml-tag `(path fill "#555" d 
      ,(let helper ((cursor lst) (count 0) (racc '()))
        (if (null? cursor)
          (string-append "M 0 0 " (apply string-append (reverse racc)) "V 0 0")
          (let ()
            (define x count)
            (define y (car cursor))
            (helper (cdr cursor) (+ count 1) (cons (format #f "L ~f ~f ~f ~f " x y (+ x 1) y) racc))
          )
        )
      )
    ))
  )
  )
))

(define svglinechart (with-docstring
  "creates an SVG line chart for the numbers in <lst>. Must be wrapped with something like svgwrap"
  (lambda (lst)
  (define numify (lambda (v)
    (if (exact? v)
      (exact->inexact v)
      v
    )
  ))
  (define m (apply max lst))
  (with-xml-tag `(g xmlns "http://www.w3.org/2000/svg" transform ,(format #f "scale(~f ~f)" (numify (/ 100 (length lst))) (numify (/ 50 m))))
    (with-xml-tag `(path fill none stroke "#922" d 
      ,(let helper ((cursor lst) (count 0) (racc '()))
        (if (null? cursor)
          (string-append "M 0 0 " (apply string-append (reverse racc)) "V 0 0")
          (let ()
            (define x count)
            (define y (car cursor))
            (helper (cdr cursor) (+ count 1) (cons (format #f "L ~f ~f ~f ~f " x y (+ x 1) y) racc))
          )
        )
      )
    ))
  )
  )
))

(define svgdotchart (with-docstring
  "creates an SVG line chart for the numbers in <lst>. Must be wrapped with something like svgwrap"
  (lambda (maxval lst)
    (apply with-xml-tag '(g xmlns "http://www.w3.org/2000/svg")
      (map (lambda (v)
        (with-xml-tag `(circle r 1 cx ,(* 100 (/ v maxval)) cy 50 stroke "green" stroke-width 1 fill "yellow"))
      ) lst)
    )
  )
))

(define svgwrap (with-docstring
  "wrap value <val> with the necessary XML context to support SVG rendering"
  (lambda (val)
    (with-xml-tag "<svg xmlns='http://www.w3.org/2000/svg' width='400' height='400' viewBox='0 0 100 100'>" (with-xml-tag "<g>" val))
  )
))

(define svgmodule (with-docstring
  "accept a Guile module or module name and return an SVG of the inter-object relationships."
  (lambda (module-or-name)
  (define s (module-get-symbol-list (if (module? module-or-name) module-or-name (resolve-interface module-or-name))))
  (define n (if (module? module-or-name) (module-name module-or-name) module-or-name))
  (define termcounter (lambda (v)
    (if (list? v)
      (apply + (map termcounter v))
      1
    )
  ))
  (define termsizer (lambda (v)
    (define w (variable-ref (cdr v)))
    (cond
      ((list? w) (termcounter w))
      ((string? w) (string-length w))
      ((closure? w) (termcounter (procedure-source w)))
      (else 1)
    )
  ))
  (define colorize (lambda (v) (cond
    ((procedure? v) "green")
    ((list? v) "orange")
    ((string? v) "pink")
    ((vector? v) "blue")
    ((pair? v) "red")
    (else "yellow")
  )))
  (define colorize-arc (lambda (conditional?)
    (if conditional?
      "blue"
      "black"
    )
  ))
  (define arcify (lambda ()
    (define names (map car s))     
    (define arcify-source (lambda (name source conditional?)
      (cond
        ((null? source) '())
        ((list? source)
          (delete '() (cond
            ((eq? (car source) 'if)
              (append (arcify-source name (car source) conditional?) (apply append (delete '() (map (lambda (v) (arcify-source name v #t)) (cdr source)))))
            )
            ((eq? (car source) 'or)
              (append (arcify-source name (car source) conditional?) (apply append (delete '() (map (lambda (v) (arcify-source name v #t)) (cdr source)))))
            )
            ((eq? (car source) 'and)
              (append (arcify-source name (car source) conditional?) (apply append (delete '() (map (lambda (v) (arcify-source name v #t)) (cdr source)))))
            )
            ((eq? (car source) 'case)
              (append (arcify-source name (cadr source) conditional?) (apply append (delete '() (map (lambda (v) (arcify-source name v #t)) (cdr source)))))
            )
            (else (apply append (delete '() (map (lambda (v) (arcify-source name v conditional?)) source))))
          ))
        )
        ((symbol? source) (if (member source names) (list (list name source (colorize-arc conditional?))) '()))
        (else '())
      )
    ))
    (apply append (delete #f (map (lambda (v)
      (define val (variable-ref (cdr v)))
      (define name (car v))
      (cond
        ((closure? val) (arcify-source name (procedure-source val) #f))
        (else #f)
      )
    ) s)))
  ))
  (let ()
    (define locs (map (lambda (v)
      (cons (random 100) (random 100))
    ) s))
  (with-xml-tag "<div>"
    (make-text (format #f "Static control flow graph for module ~a with ~a terms\n" n (length s)))
    (make-text "Node legend:\n")
    (apply with-xml-tag 'table (map (lambda (exemplar name scale)
      (with-xml-tag "<tr>"
        (with-xml-tag `(td width 5 height 5 bgcolor ,(colorize exemplar)))
        (with-xml-tag 'td (make-text name))
        (with-xml-tag 'td (make-text (format #f "area proportional to ~a" scale)))
      )
    ) (list (list 1 2 3) (cons 1 2) (lambda () #t) "foo")
      (list "list"       "pair"     "function"     "string")
      (list "length"     "1"        "number of terms" "length")
    ))
    (with-xml-tag "<br>")
  (svgwrap (apply with-xml-tag "<g>"
    (append
      (delete #f (let ()
        (define bound-locs (map (lambda (n l) (cons (car n) l)) s locs))
        (map (lambda (v)
          (define l1 (assoc (car v) bound-locs))
          (define l2 (assoc (cadr v) bound-locs))
          (define color (caddr v))
          (with-xml-tag `(line xmlns "http://www.w3.org/2000/svg"
            x1 ,(car (cdr l1))
            y1 ,(cdr (cdr l1))
            x2 ,(car (cdr l2))
            y2 ,(cdr (cdr l2))
            stroke ,color
            stroke-width .1
          ))
        ) (arcify))
      ))
      (map (lambda (v l)
        (define color (colorize (variable-ref (cdr v))))
        (with-xml-tag `(circle xmlns "http://www.w3.org/2000/svg"
          cx ,(car l)
          cy ,(cdr l)
          r  ,(/ (sqrt (termsizer v)) 5)
          stroke "black"
          stroke-width .5
          fill ,color
          onclick ,(format #f "alert('~a');" (js-quote-escape-string (symbol->string (car v))))
        ))
      ) s locs)  
    )
  ))
  )
  )
  )
))
