;
; Full of guileisms
;
(define-module (proverb testcases))

(use-modules (rnrs))
(use-modules (rnrs files))
(use-modules (rnrs lists))
(use-modules (rnrs arithmetic bitwise))
(use-modules (proverb test))
(use-modules (proverb uri))
(use-modules (proverb xml))
(use-modules (proverb dialect))
(use-modules (proverb wiki))
(use-modules (proverb tire))
(use-modules (proverb http))
(use-modules (proverb html))
(use-modules (proverb mime))
(use-modules (proverb remote))
(use-modules (proverb http-cache))
(use-modules (proverb xhtml))
(use-modules (proverb filesystem))
(use-modules (proverb lexical))
(use-modules (proverb doc))
(use-modules (proverb concurrent))
(use-modules (proverb interactive-completion))
(use-modules (proverb guilefoo))
(use-modules (proverb debug))
(use-modules (proverb md5))
(use-modules (proverb dataconv))
(use-modules (proverb games))
(use-modules (proverb diag))
(use-modules (proverb bootstrap))
(use-modules (guile))
(use-modules (ice-9 documentation))

(load "shame.scm")

(define-public studio-test (with-docstring
  "accept a Scheme value <v> and run an associated test routine. Returns non-false if there is a test and the test succeeds. Returns false otherwise. Many tests write messages to the current-output-port."
  (lambda (value)
  (define p (object-property value 'unit-test))
  (if p  
    ((cadr p))
    (begin
      (display "No test case available\n")
      #f
    )
  )
  )
))

(define-public studio-test-any-except (with-docstring
  "return a generator that, when called, will test a random testable procedure except for the procedures in <exceptions> and return a pair of the form (result . generator) where result is the result of the test and generator is a new generator of the same type"
  (lambda (exceptions)
    (define tested-functions (lambda ()
      (define m (apply append (map module-get-symbol-list (cons (current-module) (module-uses (current-module))))))
      (delete #f (map (lambda (v)
        (define val (variable-ref (cdr v)))
        (if (procedure? val)
          (let ((t (object-property val 'unit-test)))
            (if (and t (eq? 'automatic (car t)))
              (if (member val exceptions)
                #f
                (cons (car v) val)
              )
              #f
            )
          )
          #f
        )
      ) m))
    ))
    (let ()
      (define v (list->vector (tested-functions)))
      (define generator (lambda ()
        (define e (vector-ref v (random (vector-length v))))
        (let ()
          (define t (object-property (cdr e) 'unit-test))
          (format #t "testing ~a\n" (car e))
          (cons ((cadr t)) generator)
        )
      ))
      (if (equal? (vector-length v) 0)
        #f
        generator
      )
    )
  )
))

(define-public studio-test-until-fault (lambda ()
  (define g (studio-test-any-except '()))
  (let helper ((tester g))
    (define r (tester))
    (if (car r)
      (helper (cdr r))
      r
    )
  )
))

(define-public studio-test-all-except (with-docstring
  "test every Scheme value bound at the top level of Proverb with an associated test case thunk with the exception of the functions in <exceptions>"
  (lambda (exceptions)
  (define tested-functions (lambda ()
    (define m (apply append (map module-get-symbol-list (cons (current-module) (module-uses (current-module))))))
    (delete #f (map (lambda (v)
      (define val (variable-ref (cdr v)))
      (if (procedure? val)
        (let ((t (object-property val 'unit-test)))
          (if (and t (eq? 'automatic (car t)))
            (if (member val exceptions)
              #f
              (cons (car v) val)
            )
            #f
          )
        )
        #f
      )
    ) m))
  ))
  (for-each (lambda (v)
    (display "testing ") (display (car v)) (newline)
    (or
      (studio-test (cdr v))
      (format (current-error-port) "~a failed\n" (car v))
    )
  ) (tested-functions))
  (display "Done\n")
  )
))

(define-public studio-test-all (with-docstring
  "test every Scheme value bound at the top level of Proverb with an associated test case thunk."
  (lambda ()
    (studio-test-all-except '())
  )
))

(cond-expand

(guile

(define-test concurrent-apply-with-talkback (lambda ()
  ; test concurrent-apply
  (define t (concurrent-apply-with-talkback (lambda (p)
    (let helper ()
      (define r (read (car p)))
      (if (eof-object? r)
        ; exit
        #f
        (if (equal? r '(exit))
          #f
          (begin
            (write r (cdr p))
            (studio-flush-port (cdr p))
            (helper)
          )
        )
      )
    )
  ) '(bogo) 'bogo))

  (display "testing concurrent-apply-with-talkback\n")
  (if (char-ready? (car t))
    (begin
      (display "child is ready prematurely\n")
      #f 
    )
    (begin
      (display '(bogo) (cdr t))
      (studio-flush-port (cdr t))
      (let ()
        (define r (read (car t)))
        (if (eof-object? r)
          (begin (display "child died without echoing\n") #f)
          (if (equal? r '(bogo))
            (begin
              (display "child echoes OK\n")
              (display '(exit) (cdr t)) (studio-flush-port (cdr t))
              (let ()
                (define r2 (read (car t)))
                (if (eof-object? r2)
                  (begin
                    (display "child dies OK!\n")
                    (gc)
                    #t
                  )
                  (begin
                    (display "child didn't die\n")
                    #f 
                  )
                ) 
              ) 
              #t
            )
            (begin (display "child out-of-sync\n") #f)
          )
        )
      )
    )
  )
))

(define-test fork-add (lambda ()
  (if (current-tire-mux)
    (let ()
      (define t (fork-add))
      (if t
        (let ()
          (define r (remote-eval t '(list 1 2 3)))
          (if (equal? r '(1 2 3))
            (begin
              (display "remote-eval succeeded!\n")
              (fork-delete! t)
              (display "test succeeded!\n")
              #t
            )
            (begin
              (display "remote-eval failed!\n")
              #f
            )
          )
        )
        (begin
          (display "fork-add failed!\n")
          #f
        )
      )
    )
    (begin
      (display "fork-add can only be tested in a TIRE evaluation environment\n")
      #f
    )
  )
))

(define-test fork-delay* (lambda ()
  (let ()
    (define res (fork-delay* (lambda () (let () (define a (fork-delay* (lambda () (list 1 2 3)))) (define b (fork-delay* (lambda () (list 4 5 6)))) 123))))
    (if (eq? (force res) 123)
      (begin
        (display "succeeded simple!\n")
        (if (eq? (force res) 123)
          (begin
            (display "succeeded reforce!\n")
            #t
          )
          (begin
            (display "failed reforce!\n")
            #t
          )
        )
      )
      (begin
        (display "failed simple!\n" (current-error-port))
        #f
      )
    )
  )
  (display "now forcing subordinate...\n")
  (let ()
    (define res (fork-delay* (lambda () (let () (define a (fork-delay* (lambda () (list 1 2 3)))) (define b (fork-delay* (lambda () (list 4 5 6)))) (force a) 123))))
    (if (eq? (force res) 123)
      (begin
        (display "succeeded simple!\n")
        (if (eq? (force res) 123)
          (begin
            (display "succeeded reforce!\n")
            #t
          )
          (begin
            (display "failed reforce!\n")
            #t
          )
        )
      )
      (begin
        (display "failed simple!\n" (current-error-port))
        #f
      )
    )
  )
  (display "now whacking...\n")
  (let ()
    (define res (fork-delay* (lambda () (let () (define a (fork-add-helper* (current-tire-mux) sleazy-tire-responder (lambda () (list 1 2 3)) #f)) (define b (fork-delay* (lambda () (list 4 5 6)))) (let () (define result (tire-eval `(t ,(gensym "bogo-force") ,a force))) result) 123))))
    (if (eq? (force res) 123)
      (begin
        (display "succeeded simple!\n")
        (if (eq? (force res) 123)
          (begin
            (display "succeeded reforce!\n")
            #t
          )
          (begin
            (display "failed reforce!\n")
            #t
          )
        )
      )
      (begin
        (display "failed simple!\n" (current-error-port))
        #f
      )
    )
  )
))

(define-macro-test fork-begin (lambda ()
  (if (equal? (fork-begin (list 1 2 3)) '( 1 2 3))
    (begin
      (display "simple fork begin succeeds!\n")
      (if (equal? (fork-begin (fork-begin (list 4 5 6))) '(4 5 6))
        (begin
          (display "nested fork-begin succeeds!\n")
          (let ()
            (define res (fork-begin (let () (define a (fork-begin (list 1 2 3))) (define b (fork-begin (list 4 5 6))) (list a b))))
            (if (equal? res '( ( 1 2 3 ) (4 5 6) ))
              (begin
                (display "reducing fork-begin succeeds!\n")
                #t
              )
              (begin
                (format #t "reducing fork-begin fails with: ~a\n" res)
                #f
              )
            )
          )
        )
        (begin
          (display "nested fork-begin failed!\n")
          #f
        )
      )
    )
    (begin
      (display "simple fork begin test failed!\n")
      #f
    )
  )
))

(define-test fork-delete! (lambda ()
  ; can we add and delete a single child
  (define singleton-test (lambda ()
    (let ((c (fork-add)))
      (format #t "child created ~a\ndeleting child...\n" c)
      (let ((v (fork-delete! c)))
        (format #t "child deleted with ~a\n" v)
        v
      )
    )
  ))
  ; can we add and delete a grandchild and still speak to the child
  (define double-test (lambda ()
    (let ((c (fork-add)))
      (let ((gc (remote-eval c '(fork-add))))
        (format #t "child ~a and grandchild created ~a\n" c gc)
        (let ((v (fork-delete! gc)))
          (format #t "grandchild deleted with ~a\n" v)
          (if (equal? (remote-eval c '456) 456)
            (begin
              (format #t "child is still alive!\ndeleting child...\n")
              (if (fork-delete! c)
                (begin
                  (format #t "child deleted!\n")
                  #t
                )
                (begin
                  (format #t "child delete failed!\n")
                  #f
                )
              )
            )
            (begin (format #t "child is broken!\n") #f)
          )
        )
      )
    )
  ))
  (display "running singleton add/delete! pair...\n")
  (if (singleton-test)
    (begin
      (display "running singleton add/delete! pair again ...\n")
      (if (singleton-test)
        (begin 
          (display "singleton succeeds again. running double test\n")
          (if (double-test)
            (begin
              (display "double test succeeds. running double test again\n")
              (if (double-test)
                (begin (format #t "double test succeeds again\n" #t))
                (begin (format #t "double test failes the second time\n" #f))
              )
              #t
            ) 
            (begin
              (display "double test fails\n")
              #f
            ) 
          )
        )
        (begin
          (display "singleton test fails second time\n")
          #f
        )
      )
    )
    (begin
      (display "singleton test fails\n")
      #f
    )
  )
))

(define-test remote-eval (lambda ()
  (define foo (fork-add))
  (display "created test target\n")
  (let helper ((c 400))
    (if (> c 0)
      (begin
        (if (eq? (modulo c 50) 0)
          (format #t "testing... ~a left\n" c)
        )
        (if (not (equal? (remote-eval foo `(+ 1 2 3 ,c)) (+ 1 2 3 c)))
          (format #t "incorrect result at rep ~a\n" c)
        )
        (helper (- c 1))
      )
      (begin
        (fork-delete! foo)
        #t
      )
    )
  )
))

(define-test file-size (lambda ()
  (define l (random (* 1024 1024)))
  (let ()
    (define s (make-string l (integer->char (random 256))))
    (define f (open-output-file "/tmp/file-size-test"))
    (let ()
      (display s f)
      (close f)
      (let ((r (file-size "/tmp/file-size-test")))
        (delete-file "/tmp/file-size-test")
        (eq? r l)
      )
    )
  )
))

(define-test file-execute-access? (lambda ()
  (and
    (file-execute-access? "/bin/sh")
    (not (file-execute-access? "/etc/passwd"))
  )
))

(define-test value? (lambda ()
  (and
    (value? 123)
    (value? '123)
    (value? #t)
  )
))

(define-test all? (lambda ()
  (and
    (all? '( #t ))
    (all? '( 123 #t))
    (not (all? '( #f)))
    (not (all? '( #t #f)))
    (not (all? '( #f #t)))
  )
))

(define-test vector-head (lambda ()
  (and
    (equal? (vector-head #() 0) #())
    (equal? (vector-head #( 1 ) 0) #())
    (equal? (vector-head #( 1)  1) #( 1))
    (equal? (vector-head #( 1 2) 1) #( 1))
    (equal? (vector-head #( 1 2) 2) #( 1 2))
  )
))

(define-test vector-tail (lambda ()
  (and
    (equal? (vector-tail #() 0) #())
    (equal? (vector-tail #( 1 ) 0) #( 1))
    (equal? (vector-tail #( 1)  1) #())
    (equal? (vector-tail #( 1 2) 0) #( 1 2))
    (equal? (vector-tail #( 1 2) 1) #( 2))
  )
))

(define-test string-head (lambda ()
  (and
    (string=? (string-head "" 0) "")
    (string=? (string-head "a" 0) "")
    (string=? (string-head "a" 1) "a")
    (string=? (string-head "ab" 1) "a")
    (string=? (string-head "ab" 2) "ab")
  )
))

(define-test string-tail (lambda ()
  (and
    (string=? (string-tail "" 0) "")
    (string=? (string-tail "a" 0) "a")
    (string=? (string-tail "a" 1) "")
    (string=? (string-tail "ab" 0) "ab")
    (string=? (string-tail "ab" 1) "b")
  )
))

(define-test vector-map (lambda ()
  (and
    (equal? (vector-map (lambda (v) #t) #()) #())
    (equal? (vector-map (lambda (v) (+ v 1)) #(1 2 3)) #(2 3 4))
  )
))

(define-test vector-for-each (lambda ()
  ; test that each element is visited and in the right order
  (define vtest (lambda (l)
    (equal? (reverse l) (let () (define lprime '()) (vector-for-each (lambda (v) (set! lprime (cons v lprime))) (list->vector l)) lprime))
  ))
  (and
    (vtest '())
    (vtest '(1))
    (vtest '(1 2))
  )
))

(define-test string-prefix-match (lambda ()
  (and
    (string-prefix-match "" "")
    (string-prefix-match "f" "f")
    (not (string-prefix-match "f" ""))
    (string-prefix-match "f" "fo")
  )
))

(define-test with-docstring (lambda ()
  (string=? (object-documentation (with-docstring "foofoo" (lambda (x y) (+ 1 2 3 x y)))) "foofoo")
))

(define-test uniq (lambda ()
  (and
    (equal? (uniq '(1)) '(1))
    (equal? (uniq '(1 1)) '(1))
    (equal? (uniq '(1 2)) '(1 2))
    (equal? (uniq '(1 1 2)) '(1 2))
    (equal? (uniq '(1 2 2)) '(1 2))
    (equal? (uniq '(1 2 2 3)) '(1 2 3))
    (equal? (uniq '(1 2 2 3 2)) '(1 2 3 2))
    (not (equal? (uniq '("foo" "foo")) '("foo")))
  )
))

(define-test unique (lambda ()
  (and
    (equal? (unique '(1)) '(1))
    (equal? (unique '(1 1)) '(1))
    (equal? (unique '(1 2)) '(1 2))
    (equal? (unique '(1 1 2)) '(1 2))
    (equal? (unique '(1 2 2)) '(1 2))
    (equal? (unique '(1 2 2 3)) '(1 2 3))
    (equal? (unique '(1 2 2 3 2)) '(1 2 3 2))
    (equal? (unique '("foo" "foo")) '("foo"))
  )
))

(define-test any? (lambda ()
  (and
    (any? (list #t))
    (not (any? (list #f)))
    (any? (list #t #f))
    (any? (list #f #t))
    (not (any? (list #f #f)))
  )
))

(define-test string-merge (lambda ()
  (and
    (string=? (string-merge '("a" "b") #\/) "a/b")
    (string=? (string-merge '("" "") #\/) "/")
    (string=? (string-merge '("a" "") #\/) "a/")
    (string=? (string-merge '("" "b") #\/) "/b")
  )
))

(define-test studio-string-split (lambda ()
  (and
    (equal? (studio-string-split "" #\/) '(""))
    (equal? (studio-string-split "/" #\/) '("" ""))
    (equal? (studio-string-split "//" #\/) '("" "" ""))
    (equal? (studio-string-split "a/" #\/) '("a" ""))
    (equal? (studio-string-split "/b" #\/) '("" "b"))
    (equal? (studio-string-split "a/b" #\/) '("a" "b"))
    (equal? (studio-string-split "a//b" #\/) '("a" "" "b"))
    (equal? (studio-string-split "abc" #\/) '("abc"))
  )
))

(define-test studio-pipe (lambda ()
  (define p (studio-pipe))
  (and
    (input-port? (car p))
    (output-port? (cdr p))
    (not (char-ready? (car p)))
    (begin
      (display "f" (cdr p))
      (studio-flush-port (cdr p))
      (and
        (char-ready? (car p))
        (eq? (read-char (car p)) #\f)
        (not (char-ready? (car p)))
        (begin
          (close (cdr p))
          (and
            (char-ready? (car p))
            (eof-object? (read-char (car p)))
          )
        )
      )
    )
  )
))

(define-test html-escape-string (lambda ()
  (and
    (string=? "" (html-escape-string ""))
    (string=? "F" (html-escape-string "F"))
    (string=? "&lt;" (html-escape-string "<"))
    (string=? "&gt;" (html-escape-string ">"))
    (string=? "&amp;" (html-escape-string "&"))
    (string=? "FOO&lt;BAR&gt;BAZ&amp;BARTOO;" (html-escape-string "FOO<BAR>BAZ&BARTOO;"))
  )
))

(define-test alist? (lambda ()
  (and
    (alist? '())
    (alist? '((foo . bar)))
    (not (alist? '((foo . bar) 1)))
    (not (alist? '((foo . bar) . (bar . baz))))
  )
))

(define-test duplicate-alist (lambda ()
  (define a '((a . 1) (b . 2)))
  (and 
    (not (any? (map eq? a (duplicate-alist a))))
    (all? (map equal? a (duplicate-alist a)))
  )
))

(define-test exclusive (lambda ()
  (define a (exclusive (lambda (reenter?)
    (if reenter?
      (begin
        (display "reentering...\n")
        (a #f)
      )
      'done
    )
  )))
  (define sentinel (cons 'a 'b))
  (and
    (eq? (a #f) 'done)
    (eq? (catch #t (lambda () (a #t)) (lambda args sentinel)) sentinel)
  )
))

(define-test directory? (lambda ()
  (and
    (directory? "/")
    (directory? "/bin")
    (not (directory? "/bin/ls"))
  )
))

(define-test which (lambda ()
  (and
    (which "ls")
    (file-execute-access? (which "ls"))
    (eq? (system (which "ls")) 0)
  )
))

(define-test predicate? (lambda ()
  (and
    (not (predicate? 123))
    (not (predicate? (lambda () #t)))
    (predicate? (lambda (x) #t))
    (not (predicate? (lambda (x y) #t)))
  )
))

(define-test find-largest-root (lambda ()
  (and
    (string=? (find-largest-root "" "") "")
    (string=? (find-largest-root "a" "") "")
    (string=? (find-largest-root "" "a") "")
    (string=? (find-largest-root "a" "a") "a")
    (string=? (find-largest-root "ab" "a") "a")
    (string=? (find-largest-root "a" "ab") "a")
    (string=? (find-largest-root "a" "b") "")
  )
))

(define-test read-uint8 (lambda ()
  (and
    (equal? (call-with-input-string (string (integer->char 0)) read-uint8) 0)
    (equal? (call-with-input-string (string (integer->char 255)) read-uint8) 255)
  )
))

(define-test read-uint16-be (lambda ()
  (and
    (equal? (call-with-input-string (string (integer->char 1) (integer->char 0)) read-uint16-be) 256)
    (equal? (call-with-input-string (string (integer->char 0) (integer->char 1)) read-uint16-be) 1)
    (equal? (call-with-input-string (string (integer->char 1) (integer->char 0) #\M) read-uint16-be) 256)
    (equal? (call-with-input-string (string (integer->char 0) (integer->char 1) #\M) read-uint16-be) 1)
  )
))

(define-test read-uint16-le (lambda ()
  (and
    (equal? (call-with-input-string (string (integer->char 1) (integer->char 0)) read-uint16-le) 1)
    (equal? (call-with-input-string (string (integer->char 0) (integer->char 1)) read-uint16-le) 256)
    (equal? (call-with-input-string (string (integer->char 1) (integer->char 0) #\M) read-uint16-le) 1)
    (equal? (call-with-input-string (string (integer->char 0) (integer->char 1) #\M) read-uint16-le) 256)
  )
))

(define-test string->uint8 (lambda ()
  (and
    (equal? (string->uint8 (string (integer->char 0))) 0)
    (equal? (string->uint8 (string (integer->char 255))) 255)
  )
))

(define-test string->uint16-be (lambda ()
  (and
    (equal? (string->uint16-be (string (integer->char 1) (integer->char 0))) 256)
    (equal? (string->uint16-be (string (integer->char 0) (integer->char 1))) 1)
    (equal? (string->uint16-be (string (integer->char 1) (integer->char 0) #\M)) 256)
    (equal? (string->uint16-be (string (integer->char 0) (integer->char 1) #\M)) 1)
  )
))

(define-test string->uint16-le (lambda ()
  (and
    (equal? (string->uint16-le (string (integer->char 1) (integer->char 0))) 1)
    (equal? (string->uint16-le (string (integer->char 0) (integer->char 1))) 256)
    (equal? (string->uint16-le (string (integer->char 1) (integer->char 0) #\M)) 1)
    (equal? (string->uint16-le (string (integer->char 0) (integer->char 1) #\M)) 256)
  )
))

(define-test ref->uint8 (lambda ()
  (and
    (equal? (ref->uint8 (string (integer->char 0)) 0) 0)
    (equal? (ref->uint8 (string (integer->char 255)) 0) 255)
    (equal? (ref->uint8 (string #\M (integer->char 0)) 1) 0)
    (equal? (ref->uint8 (string #\M (integer->char 255)) 1) 255)
  )
))

(define-test uint8s->uint16 (lambda ()
  (and
    (equal? (uint8s->uint16 0 1) 256)
    (equal? (uint8s->uint16 1 0) 1)
    (equal? (uint8s->uint16 2 3) #x0302)
  )
))

(define-test uint8s->uint32 (lambda ()
  (and
    (equal? (uint8s->uint32 0 0 0 1) #x01000000)
    (equal? (uint8s->uint32 1 0 0 0) #x00000001)
    (equal? (uint8s->uint32 1 2 3 4) #x04030201)
  )
))

(define-test index-map (lambda ()
  (and
    (equal? (index-map cons '()) '())
    (equal? (index-map cons '(a)) '((0 . a)))
    (equal? (index-map cons '(a b)) '((0 . a) (1 . b)))
  )
))

(define-test vector-index-map (lambda ()
  (and
    (equal? (vector-index-map cons #()) #())
    (equal? (vector-index-map cons #('a)) #((0 . 'a)))
    (equal? (vector-index-map cons #('a 'b)) #((0 . 'a) (1 . 'b)))
    (equal? (vector-index-map list #('a 'b) #('c 'd)) #((0 'a 'c) (1 'b 'd)))
  )
))

(define-test get-module-list (lambda ()
  (define l (get-module-list)) 
  (and
    (not (null? l))
    (all? (map module? l))
    (member (current-module) (get-module-list))
    #t
  )
))

(define-test aggregate-alist-for-assoc (lambda ()
  (and
    (equal? (aggregate-alist-for-assoc '()) '())
    (equal? (aggregate-alist-for-assoc '((a . 1))) '((a . (1))))
  )
))

(define-test char->url-hex (lambda ()
  (define vtest (lambda (s)
    (string=? s (url-decode-string (apply string-append (map char->url-hex (string->list s)))))
  ))
  (and
    (vtest "")
    (vtest "a")
    (vtest " ")
    (vtest "a ") 
    (vtest " a") 
    (vtest "a b") 
  )
))

(define-test tab-complete-string (lambda ()
  (and
    (string=? (tab-complete-string "(tab-complete-strin" (current-module)) "(tab-complete-string")
    (string=? (tab-complete-string "(tab-complete-string " (current-module)) "(tab-complete-string str")
  )
))

(define-test procedure-get-num-args (lambda ()
  (define p-0-test (lambda () (+ 1 2 3)))
  (define p-3-test (lambda (a b c) (+ 1 2 3)))
  (define p-var-test (lambda (a b c . rest) (+ 1 2 3)))
  (let ()
    (define c-0-test (consider-procedure p-0-test))
    (define c-3-test (consider-procedure p-3-test))
    (define c-var-test (consider-procedure p-var-test))
    (and
      (eq? (procedure-get-num-args c-0-test) 0)
      (eq? (procedure-get-num-args c-3-test) 3)
      (eq? (procedure-get-num-args c-var-test) 3)
    )
  )
))

(define-test procedure-get-arg-names (lambda ()
  (define p-0-test (lambda () (+ 1 2 3)))
  (define p-3-test (lambda (a b c) (+ 1 2 3)))
  (define p-var-test (lambda (a b c . rest) (+ 1 2 3)))
  (let ()
    (define c-0-test (consider-procedure p-0-test))
    (define c-3-test (consider-procedure p-3-test))
    (define c-var-test (consider-procedure p-var-test))
    (and
      (equal? (procedure-get-arg-names c-0-test) '())
      (equal? (procedure-get-arg-names c-3-test) '("a" "b" "c"))
      (equal? (procedure-get-arg-names c-var-test) '("a" "b" "c"))
    )
  )
))

(define-test url-decode-string (lambda ()
  (define str (apply string (map (lambda (v) (integer->char (random 256))) (make-list 1001))))
  (and 
    (string=? str (url-decode-string (url-encode-string str)))
    (equal? (url-decode-string "") "")
    (equal? (url-decode-string "a") "a")
    (equal? (url-decode-string "%20") " ")
    (equal? (url-decode-string "a%20") "a ")
    (equal? (url-decode-string "%20a") " a")
    (equal? (url-decode-string "b%20a") "b a")
  )
))

(define-test url-encode-string (lambda ()
  (define str (apply string (map (lambda (v) (integer->char (random 256))) (make-list 1001))))
  (string=? str (url-decode-string (url-encode-string str)))
))

(define-test diagnose-eof (lambda ()
  (define tester (lambda (str)
    (diagnose-eof str (catch #t (lambda () (with-input-from-string str read)) (lambda args args)))
  ))
  (and
    (equal? (tester "(foo") '(dangling-call foo "" () #f))
    (equal? (tester "(foo bar") '(dangling-call foo "" (bar) #f))
    (equal? (tester "(foo bar (baz") '(dangling-call baz "(foo bar " () #f))
    ; why do we convert the head of the string to a symbol?
    (equal? (tester "(foo bar (baz \"bartoo") '(open-string bartoo 15))
    (equal? (tester "(foo bar (baz \"bar too") '(open-string bar 15))
    ; and why the eof object?
    (equal? (tester "(foo bar (baz \"") `(open-string ,the-eof-object 15))
  )
))

(define-test always-true (lambda ()
  (eq? (always-true) #t)
))

(define-test always-false (lambda ()
  (eq? (always-false) #f)
))

(define-test path-expand path-decompose canonicalize-path build-path (lambda ()
  (define h (getenv "HOME"))
  (and
    ; path-decompose
    (equal? (path-decompose "/") '("/"))
    (equal? (path-decompose "//") '("/"))
    (equal? (path-decompose "/foo") '("/" "foo"))
    (equal? (path-decompose "//foo") '("/" "foo"))
    (equal? (path-decompose "/foo/") '("/" "foo"))
    (equal? (path-decompose "/foo//") '("/" "foo"))
    (equal? (path-decompose "/foo/bar") '("/" "foo" "bar"))
    (equal? (path-decompose "foo/bar") '("foo" "bar"))
    (equal? (path-decompose "foo//bar") '("foo" "bar"))
    (equal? (path-decompose "foo/") '("foo"))
    (equal? (path-decompose "foo") '("foo"))
    ; path-expand
    (equal? (path-expand (path-decompose "~")) (path-decompose h))
    (equal? (path-expand (path-decompose "~/")) (path-decompose h))
    (equal? (path-expand (path-decompose "~/foo")) (append (path-decompose h) '("foo")))
    (equal? (path-expand (path-decompose "/~/foo")) '("/" "~" "foo"))
    ; build-path
    (equal? (build-path '("/")) "/")
    (equal? (build-path '("/" "etc" "passwd")) "/etc/passwd")
    (equal? (build-path '("etc" "passwd")) "etc/passwd")
    (equal? (canonicalize-path "/") '("/"))
    ; we want to know something that is a symlink
    (if (and (file-exists? "/dev/stdin") (file-exists? "/dev/fd") (file-is-directory? "/dev/fd"))
      (or 
        ; this is actually a bad test -- but we fail this anyway
        (equal? (canonicalize-path "/dev/stdin") '("/" "proc" "self" "fd" "0"))
        (equal? (canonicalize-path "/dev/stdin") '("/" "dev" "fd" "0"))
      )
      (begin
        (display "NEED A SYMLINK FOR THIS TEST TO SUCCEED" (current-error-port))
        #f
      )
    )
  )
))

(define-test procedure-ref (lambda ()
  (define z 2)
  (define a 1)
  (define sample-1 (let ()
    (define a 5)
    (define b 6)
    (let ()
      (define b 7)
      (define c 8)
      (let ()
        (define c 9)
        (define d 10)
        (lambda (e)
          (define f 11)
          (+ e f)
        )
      )
    )
  ))
  (and
    (equal? (procedure-ref sample-1 'd) 10)
    (equal? (procedure-ref sample-1 'c) 9)
    (equal? (procedure-ref sample-1 'b) 7)
    (equal? (procedure-ref sample-1 'a) 5)
    (equal? (procedure-ref sample-1 'z) 2)
  )
))

(define-test tire-parse-url (lambda ()
  (define a (tire-parse-url "tire://localhost:1000"))
  (define b (tire-parse-url "tire://localhost:1000/foo"))
  (define c (tire-parse-url "tire://foo.bar.com:1000"))
  (and
    (and (list? a) (eq? (length a) 4) (equal? (list-ref a 1) "localhost")
                                      (equal? (list-ref a 2) 1000)
                                      (equal? (list-ref a 3) #f))
    (and (list? b) (eq? (length b) 4) (equal? (list-ref b 1) "localhost")
                                      (equal? (list-ref b 2) 1000)
                                      (equal? (list-ref b 3) 'foo))
    (and (list? c) (eq? (length c) 4) (equal? (list-ref c 1) "foo.bar.com")
                                      (equal? (list-ref c 2) 1000)
                                      (equal? (list-ref c 3) #f))
  )
))

(define-test audio-guess-mime-type (lambda ()
  (and
    ; leading bytes from a real mp3
    (equal? (audio-guess-mime-type (string #\377 #\373 #\240 #\d)) "audio/mpeg")
    ; leading bytes from a real MIDI file
    (equal? (audio-guess-mime-type (string #\M #\T #\h #\d)) "audio/midi")
    ; You call that a test case?
  )
))

(define-test find-undefined-in-procedure shame-eval (lambda ()
  (define bogus-predicate-undefined (lambda (v)
    (and (list? v) (let ()
      (define a (aggregate-alist-for-assq v))
      (and a
           (let ()
             (define v (assoc 'notfound a))
             (and v (equal? (length (cdr v)) 1) (pair? (car (cdr v))) (eq? (car (car (cdr v))) 'bogus-undefined-symbol))
           )
      )
    ))
  ))
  (define clear-predicate-undefined (lambda (v)
    (and (list? v) (not (assoc 'notfound v)))
  ))
  (define procon (lambda (p1 p2)
    (define r1 (find-undefined-in-procedure p1 #f #f))
    (define r2 (find-undefined-in-procedure p2 #f #f))
    (let ()
      (define ex (clear-predicate-undefined r1))
      (define counterex (bogus-predicate-undefined r2))
      (if (not (and ex counterex))
        (format #t "~a on example ~a:\n ~a\nand ~a on counterexample ~a:\n ~a\n"
                   (if ex "succeeded" "failed") (if ex "" (format #f "(with ~a)" r1)) (procedure-source p1)
                   (if counterex "succeeded" "failed") (if counterex "" (format #f "(with ~a)" r2)) (procedure-source p2))
      )
      (and ex counterex)
    )
  ))
  (define procons (lambda (s1 s2)
    (procon (eval s1 (current-module)) (eval s2 (current-module)))
  ))
  (and
    (bogus-predicate-undefined (shame-eval 'bogus-undefined-symbol (current-module) '() '()))
    (bogus-predicate-undefined (shame-eval '(begin bogus-undefined-symbol) (current-module) '() '()))
    (bogus-predicate-undefined (shame-eval '(+ bogus-undefined-symbol 1) (current-module) '() '()))
    (procons `(lambda () 1) 
             `(lambda () bogus-undefined-symbol))
    (procons `(lambda () (+ 1))
             `(lambda () (+ bogus-undefined-symbol)))
    (procons `(lambda () (+ 1 2))
            `(lambda () (+ 1 bogus-undefined-symbol)))
    (procons `(lambda () (+ 1))
            `(lambda () (bogus-undefined-symbol 1)))
    (procons `(lambda () (define foo 10) foo)
            `(lambda () (define foo 10) bogus-undefined-symbol))
    (procons `(lambda () (define foo (lambda () foo)) foo)
            `(lambda () (define foo (lambda () bogus-undefined-symbol)) foo))
    (procons `(lambda () (define foo 10) (define bar (lambda () foo)) bar)
            `(lambda () (define foo 10) (define bar (lambda () bogus-undefined-symbol)) bar))
    (procons `(lambda () (define bar (lambda () foo)) (define foo 10) bar)
            `(lambda () (define bar (lambda () bogus-undefined-symbol)) (define foo 10) bar))
    (procons `(lambda (bar) (define foo 10) foo)
            `(lambda (bar) (define foo 10) bogus-undefined-symbol))
    (procons `(lambda (bar) (define foo bar) bar)
            `(lambda (bar) (define foo bogus-undefined-symbol) bar))
    (procons `(lambda () (let ((foo #t)) foo))
            `(lambda () (let ((foo bogus-undefined-symbol)) foo)))
    (procons `(lambda () (let helper ((foo #t)) (cons foo helper)))
            `(lambda () (let helper ((foo bogus-undefined-symbol)) (cons foo helper))))
    (procons `(lambda () (let ((foo 10)) foo))
            `(lambda () (let ((foo 10)) bogus-undefined-symbol)))
    ; foo
    (procons `(lambda (bar) (let () (define foo bar) foo))
            `(lambda (bar) (let () (define foo bogus-undefined-symbol) foo)))
    (procons `(lambda () (let ((foo 10) (bar 20)) foo))
            `(lambda () (let ((foo 10) (bar 20)) bogus-undefined-symbol)))
    (procons `(lambda () (let ((foo 10) (bar 20)) bar))
            `(lambda () (let ((foo 10) (bar 20)) bogus-undefined-symbol)))
    (procons `(lambda () (let () (define foo 10) foo))
            `(lambda () (let () (define foo 10) bogus-undefined-symbol)))
    (procons `(lambda () (let () (define foo 10) (define bar 20) foo))
            `(lambda () (let () (define foo 10) (define bar 20) bogus-undefined-symbol)))
    (procons `(lambda () (let () (define foo 10) (define bar 20) bar))
            `(lambda () (let () (define foo 10) (define bar 20) bogus-undefined-symbol)))
    (procons `(lambda () (let () (define foo 10) (define bar (lambda () foo)) bar))
             `(lambda () (let () (define foo 10) (define bar (lambda () bogus-undefined-symbol)) bar)))
    (procons `(lambda () (let () (define bar (lambda () foo)) (define foo 10) bar))
             `(lambda () (let () (define bar (lambda () bogus-undefined-symbol)) (define foo 10) bar)))
    (procons `(lambda () (let* ((a 10) (b a)) b))
             `(lambda () (let* ((a bogus-undefined-symbol) (b a)) b)))
    (procons `(lambda () (let* ((a 10) (b a)) (cons a b)))
             `(lambda () (let* ((a 10) (b bogus-undefined-symbol)) (cons a b))))
    (procons `(lambda () (let* ((a 10) (b a)) b))
             `(lambda () (let* ((a 10) (b a)) bogus-undefined-symbol)))
    (procons `(lambda (foo) (lambda (bar) bar))
            `(lambda (foo) (lambda (bar) bogus-undefined-symbol)))
    (procons `(lambda (foo) (lambda (bar) foo))
            `(lambda (foo) (lambda (bar) bogus-undefined-symbol)))
    (procons `(lambda (foo foo2) (lambda (bar bar2) bar))
            `(lambda (foo foo2) (lambda (bar bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo foo2) (lambda (bar bar2) foo))
            `(lambda (foo foo2) (lambda (bar bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo foo2) (lambda (bar bar2) bar2))
            `(lambda (foo foo2) (lambda (bar bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo foo2) (lambda (bar bar2) foo2))
            `(lambda (foo foo2) (lambda (bar bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo . foo2) (lambda (bar . bar2) bar))
            `(lambda (foo . foo2) (lambda (bar . bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo . foo2) (lambda (bar . bar2) foo))
            `(lambda (foo . foo2) (lambda (bar . bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo . foo2) (lambda (bar . bar2) bar2))
            `(lambda (foo . foo2) (lambda (bar . bar2) bogus-undefined-symbol)))
    (procons `(lambda (foo . foo2) (lambda (bar . bar2) foo2))
            `(lambda (foo . foo2) (lambda (bar . bar2) bogus-undefined-symbol)))
    (procons `(lambda () (define (loop) 123) loop)
            `(lambda () (define (loop) 123) bogus-undefined-symbol))
    (procons `(lambda () (define (loop foo bar) (+ 123 foo)) loop)
            `(lambda () (define (loop foo bar) (+ 123 bogus-undefined-symbol)) loop))
    (procons `(lambda () (define (loop foo bar) (+ 123 bar)) loop)
            `(lambda () (define (loop foo bar) (+ 123 bogus-undefined-symbol)) loop))
    (procons `(lambda () (let () (define (loop) 123) loop))
            `(lambda () (let () (define (loop) 123) bogus-undefined-symbol)))
    (procons `(lambda () (let () (define (loop foo bar) (+ 123 foo)) loop))
            `(lambda () (let () (define (loop foo bar) (+ 123 bogus-undefined-symbol)) loop)))
    (procons `(lambda () (let () (define (loop foo bar) (+ bar foo)) loop))
            `(lambda () (let () (define (loop foo bar) (+ 123 bogus-undefined-symbol)) loop)))
    (procons `(lambda () (let foo () (if #f (foo))))
            `(lambda () (let foo () (if #f (bogus-undefined-symbol)))))
    (procons `(lambda () (let foo () (if #f (foo))))
            `(lambda () (let foo () (if #f (bogus-undefined-symbol)))))
    (procons `(lambda () (let foo () (if #f 1 (foo))))
            `(lambda () (let foo () (if #f 1 (bogus-undefined-symbol)))))
    (procons `(lambda () (let foo () (if #f 1 (foo))))
            `(lambda () (let foo () (if #f 1 (bogus-undefined-symbol)))))
    (procons `(lambda () (let foo () (if #f (foo) 1)))
            `(lambda () (let foo () (if #f (bogus-undefined-symbol) 1))))
    (procons `(lambda () (let foo () (if #f (foo) 1)))
            `(lambda () (let foo () (if #f (bogus-undefined-symbol) 1))))
    (procons `(lambda () (do ((i 0 (1+ i))) ((>= i 10)) (+ 1 2 3)))
            `(lambda () (do ((i 0 (1+ i))) ((>= bogus-undefined-symbol)) (+ 1 2 3))))
    (procons `(lambda () (do ((i 0 (1+ i))) ((>= i 10)) (+ i 2 3)))
            `(lambda () (do ((i 0 (1+ bogus-undefined-symbol))) ((>= i 10)) (+ 1 2 3))))
    (procons `(lambda () (do ((i 0 (1+ i))) ((>= i 10)) (+ i 2 3)))
            `(lambda () (do ((i 0 (1+ i))) ((>= bogus-undefined-symbol 10)) (+ 1 2 3))))
    (procons `(lambda (foo) (begin foo))
            `(lambda (foo) (begin bogus-undefined-symbol)))
    (procons `(lambda (foo) (begin 1 foo))
            `(lambda (foo) (begin 1 bogus-undefined-symbol)))
    (procons `(lambda (foo) (begin foo 1))
            `(lambda (foo) (begin bogus-undefined-symbol 1)))
    (procons `(lambda (foo) `(,foo))
            `(lambda (foo) `(,bogus-undefined-symbol)))
    (procons `(lambda (foo) `(1 ,foo))
            `(lambda (foo) `(1 ,bogus-undefined-symbol)))
    (procons `(lambda (foo) `(1 ,foo 2))
            `(lambda (foo) `(1 ,bogus-undefined-symbol 2)))
    (procons `(lambda (foo) `(1 ,@foo 2))
            `(lambda (foo) `(1 ,@bogus-undefined-symbol 2)))
    (procons `(lambda (foo bar) `(1 ,@foo ,bar))
            `(lambda (foo bar) `(1 ,@bogus-undefined-symbol ,bar 2)))
    (procons `(lambda (foo bar) `(1 ,@foo ,bar))
            `(lambda (foo bar) `(1 ,@foo ,bogus-undefined-symbol 2)))
    (procons `(lambda (foo bar) `(1 ,foo ,@bar))
            `(lambda (foo bar) `(1 ,bogus-undefined-symbol ,@bar 2)))
    (procons `(lambda (foo bar) `(1 ,foo ,@bar))
            `(lambda (foo bar) `(1 ,foo ,@bogus-undefined-symbol 2)))
    (procons `(lambda (foo) (cond ((foo) #t) (else #f)))
            `(lambda (foo) (cond ((bogus-undefined-symbol) #t) (else #f))))
    (procons `(lambda (foo) (cond (#f foo) (else #f)))
            `(lambda (foo) (cond (#f bogus-undefined-symbol) (else #f))))
    (procons `(lambda (foo) (cond (#f #t) (else foo)))
            `(lambda (foo) (cond (#f #t) (else bogus-undefined-symbol))))
    (procons `(lambda (foo) (case foo ((bar) #t) (else #f)))
            `(lambda (foo) (case bogus-undefined-symbol ((bar) #t) (else #f))))
    (procons `(lambda (foo bar) (case foo ((baz) bar) (else #f)))
            `(lambda (foo bar) (case foo ((baz) bogus-undefined-symbol) (else #f))))
    (procons `(lambda (foo bar) (case foo ((baz) #t) (else bar)))
            `(lambda (foo bar) (case foo ((baz) #t) (else bogus-undefined-symbol))))
    ; these are tricky and esoteric cases
    (procons `(lambda () (if (defined? 'foo) foo))
            `(lambda () (if (defined? 'foo) bogus-undefined-symbol)))
    (procons `(lambda () (if (defined? 'bogus-undefined-symbol) bogus-undefined-symbol #t))
            `(lambda () (if (defined? 'bogus-undefined-symbol) #t bogus-undefined-symbol)))
    ;(procons `(lambda () (and (defined? 'foo) foo))
    ;        `(lambda () (and (defined? 'foo) bogus-undefined-symbol)))
  )
))

; verbatim from R6
(define-test bitwise-arithmetic-shift (lambda ()
  (and (equal? (bitwise-arithmetic-shift -6 -1) -3)
       (equal? (bitwise-arithmetic-shift -5 -1) -3)
       (equal? (bitwise-arithmetic-shift -4 -1) -2)
       (equal? (bitwise-arithmetic-shift -3 -1) -2)
       (equal? (bitwise-arithmetic-shift -2 -1) -1)
       (equal? (bitwise-arithmetic-shift -1 -1) -1)
  )
))

; these lame vectors generated manually by dsridge
(define-test md5 (lambda ()
  (and (equal? (md5 "") "d41d8cd98f00b204e9800998ecf8427e")
       (equal? (md5 "foo") "acbd18db4cc2f85cedef654fccc4a4d8")
       (equal? (md5 (make-string 512 (integer->char 0))) "bf619eac0cdf3f68d496ea9344137e8b")
       (equal? (md5 (make-string 4096 (integer->char 0))) "620f0b67a91f7f74151bc5be745b7110")
  )
))

(define-interactive-test critmsg (lambda ()
  (begin
    (critmsg "bogo\n")
    (make-xhtml
      (make-text "You should have just seen a critical message that says 'bogo'\n")
    )
  )
))

(define-interactive-test xhtml-show-stack (lambda ()
  (begin
    (dribble (make-text "This test case will create an error. Navigate it. It should be obvious that the current error is that bogobogofunc is not defined."))
    (list (list (list (if (+ (bogobogofunc 2 2 2) "foo") #t #f) 1 2 3)))
  )
))

(define-interactive-test welcome (lambda ()
  (welcome)
))

(define-interactive-test xhtml-more* (lambda ()
  (xhtml-more* "/etc/passwd")
))

(define-interactive-test tire-diag (lambda ()
  (tire-diag)
))

(define-interactive-test clear (lambda ()
  (clear)
))

(define-interactive-test retrieve-file (lambda ()
  (retrieve-file "/etc/passwd")
))

(define-interactive-test make-chessboard (lambda ()
  (make-chessboard default-chessboard-bindings)
))

(define-interactive-test xhtml-shame (lambda ()
  (xhtml-shame)
))

(define-interactive-test xhtml-ls* (lambda ()
  (xhtml-ls* "/etc")
))

(define-interactive-test xhtml-help* (lambda ()
  (xhtml-help* "help")
))

(define-interactive-test xrun (lambda ()
  (xrun (bogo))
))

(define-interactive-test colorspark (lambda ()
  ; the documentation flexes the interface
  (make-wiki "{{Proverb:colorspark}}")
))

(define-interactive-test svgdemo (lambda ()
  (make-xhtml
    (make-text "Try typing (svg2) into the interpreter area of the SVG widget")
    (svgdemo)
  )
))

(define-interactive-test svgsample (lambda ()
  (make-xhtml
    (make-text "You should see an SVG thingy")
    (svgwrap (svgsample))
  )
))

(define-interactive-test svgboxchart (lambda ()
  (define points (random 1000))
  (define max (random 1000))
  (define min (random 1000))
  (make-xhtml
    (make-text (format #f "Barchart test: points: ~a max: ~a min: ~a\n" points max min))
    (svgwrap (svgboxchart (map (lambda (v) (+ min (random (- max min)))) (make-list points))))
  )
))

(define-interactive-test make-wiki (lambda ()
  (make-xhtml
    (make-text "The wiki syntax sampler should make some sense...")
    (make-wiki "{{Jsshell:wiki-syntax-sampler}}")
  )
))

(define-interactive-test xmlize xmlize-value css-syntax-sampler (lambda ()
  (make-xhtml
    (make-text "The CSS sampler should demonstrate the xmlization")
    (css-syntax-sampler)
  )
))

(define-interactive-test svg2 (lambda ()
  (make-xhtml
    (make-text "You should see an SVG thingy")
    (svgwrap
      (svg2)
    )
  )
))

(define-interactive-test svgwrap (lambda ()
  (make-xhtml
    (make-text "You should see an SVG thingy")
    (svgwrap
      (svg2)
    )
  )
))

(define-interactive-test make-raw (lambda ()
  (make-xhtml
    (make-text "The following should be an HTML header")
    (make-raw "<h3>FOO</h3>")
  )
))

(define-interactive-test make-caching-xml-tag (lambda ()
  (make-xhtml
    (with-xml-tag (make-caching-xml-tag 'img 'src (make-http-cache-fragment (filename->string (string-append support-dir "/artwork/ida.png")) "image/png") 'width 64))
    (with-xml-tag (make-xml-tag 'br))
    (make-text "You should see the IDA logo just above this")
    (with-xml-tag (make-xml-tag 'br))
    (with-xml-tag (make-caching-xml-tag 'img 'src (make-ephemeral-http-cache-fragment (filename->string (string-append support-dir "/artwork/crumple.png")) "image/png") 'width 64))
    (with-xml-tag (make-xml-tag 'br))
    (make-text "You should see a crumpled piece of paper just above this")
  )
)) 

(define-interactive-test with-caching-xml-tag (lambda ()
  (make-xhtml
    (with-caching-xml-tag `(img src ,(make-http-cache-fragment (filename->string (string-append support-dir "/artwork/ida.png")) "image/png") width 64))
    (with-xml-tag 'br)
    (make-text "You should see the IDA logo just above this")
    (with-xml-tag 'br)
    (with-caching-xml-tag `(img src ,(make-ephemeral-http-cache-fragment (filename->string (string-append support-dir "/artwork/crumple.png")) "image/png") width 64))
    (with-xml-tag (make-xml-tag 'br))
    (make-text "You should see a crumpled piece of paper just above this")
  )
))

)
(else
  (display "TEST CASES NOT SET UP!\n")
)

)
