(library (proverb vt100)
  (export vt100-test measure-the-screen)
  (import
    (proverb doc) (guile)
  )

(define vt100-terminal-lines #f)
(define vt100-terminal-columns #f)

; guile provides #\esc for escape
; but SISC won't even read it
(define esc-char #\033)

(define get-num-lines (lambda ()
  (define env-lines (getenv "LINES"))
  (define prop-lines (assoc 'li terminal-properties))
  (if vt100-terminal-lines
    vt100-terminal-lines
    (if env-lines
      (with-input-from-string env-lines read)
      (if prop-lines
        (cadr prop-lines)
        24
      )
    )
  )
))

;
; termcap helper routines
;

(define termcap-string->string (with-docstring
  "take a termcap string and optional arguments for parametized strings and generate the result"
  (lambda (s . args)
  (define sl (string->list s))
  (define helper (lambda (l a stack state racc)
    (cond
      ((eq? state 'normal)
        (if (null? l)
          (reverse racc)
          (cond 
            ((eq? (car l) #\\)
              (helper (cdr l) a stack 'backslash racc)
            )
            ((eq? (car l) #\%)
              (helper (cdr l) a stack 'percent racc)
            )
            (else (helper (cdr l) a stack state (cons (car l) racc)))
          )
        )
      ) 
      ((eq? state 'backslash)
        (if (null? l)
          (reverse racc)
          (cond 
            ((eq? (car l) #\E)
              (helper (cdr l) a stack 'normal (cons esc-char racc))
            )
            (else (helper (cdr l) a stack 'normal racc))
          )
        )
      ) 
      ((eq? state 'percent)
        (if (null? l)
          (reverse racc)
          (cond
            ((eq? (car l) #\%)
              ; a literal percent
              (helper (cdr l) a stack 'normal (cons #\% racc))
            )
            ((eq? (car l) #\d)
              ; grab the first item off the stack
              (helper (cdr l) a (cdr stack) 'normal (append (reverse (string->list (object->string (car stack)))) racc))
            )
            ((eq? (car l) #\p) (let ()
              ; push the specified item onto the stack
              ;(define number (- (char->integer (cadr l)) (char->integer 0)))
              ;(helper (cddr l) a (cons (list-ref a number) stack) 'normal racc)
              (helper (cddr l) a stack 'normal racc)
            ))
            ((eq? (car l) #\i)
              ; add one to the first two items on the stack
              (helper (cdr l) a (cons (+ 1 (car stack)) (cons (+ 1 (cadr stack)) (cddr stack))) 'normal racc)
            )
            (else (helper (cdr l) a stack 'normal racc))
          )
        )
      )
      (else (if (null? l) (reverse racc) (helper (cdr l) a stack state (cons (car l) racc))))
    )   
  ))
  (list->string (helper sl args args 'normal '()))
  )
))

;
; %d -- a number as a string
; %i -- add one to the first two items on the stack
; %p[0-9] -- push argument [0-9] onto the stack


;
; terminal properties taken from xterm-xfree86
;
(define terminal-properties '(
  ; these three lines are stolen from another entry
  (AB "\\E[4%p1%dm")          ; set background color (???)
  (AF "\\E[3%p1%dm")          ; set foreground color (???)
  (op "\\E[37;40m")           ; unset the attributes?

  (is "\\E7\\E[r\\E[m\\E[?7h\\E[?1;3;4;6l\\E[4l\\E8\\E>") ; initialization string 2 at login
  (rs "\\E7\\E[r\\E[m\\E[?7h\\E[?1;3;4;6l\\E[4l\\E8\\E>") ; reset string
  (AL "\\E[%dL")                       ; insert %1 lines
  (DL "\\E[%dM")		       ; delete %1 lines
  (DC "\\E[%dP")                       ; delete %c characters
  (DO "\\E[%dB")                       ; cursor down %1 lines
  (UP "\\E[%dA")                       ; cursor up %1 lines
  (LE "\\E[%dD")                       ; cursor left %1 characters
  (RI "\\E[%dC")                       ; cursor right %1 characters
  (al "\\E[L")			       ; insert one line
  (am)				       ; automatic margins -- automatic line wrap
  (bl "^G")			       ; audio bell
  (bs)                                 ; performs a backspace
  (cd "\\E[J")                         ; clear to end of screen
  (ce "\\E[K")                         ; clear to end of line
  (cl "\\E[H\\E[2J")                   ; clear screen and cursor home
  (cm "\\E[%i%d;%dH")                  ; cursor move to row %1 and column %2 on screen
  (co 80)                              ; number of columns
  (cs "\\E[%i%d;%dr")                  ; scroll region from line %1 to %2
  (ct "\\E[3g")			       ; clear tabs
  (dc "\\E[P")                         ; delete one character
  (dl "\\E[M")                         ; delete one line
  (ho "\\E[H")                         ; cursor home
  (im "\\E[4h")                        ; begin insert mode
  (ei "\\E[4l")                        ; end insert mode
  (mi)                                 ; cursor movement works in insert mode
  (ks "\\E[?1h\\E=")                   ; turn keypad on
  (ke "\\E[?1l\\E>")                   ; turn keypad off
  (@7 "\\E[4~")			       ; end key
  (kh "\\E[1~")                        ; cursor home key
  (k1 "\\EOP")                         ; function key 1
  (k2 "\\EOQ")                         ; function key 2
  (k3 "\\EOR")                         ; function key 3
  (k4 "\\EOS")                         ; function key 4
  (k5 "\\E[15~")                       ; function key 5
  (k6 "\\E[17~")                       ; function key 6
  (k7 "\\E[18~")                       ; function key 7
  (k8 "\\E[19~")                       ; function key 8
  (k9 "\\E[20~")                       ; function key 9
  (k-semicolon "\\E[21~")              ; function key 10
  (F1 "\\E[23~")                       ; function key 11
  (F2 "\\E[24~")                       ; function key 12
  (F3 "\\E[25~")                       ; function key 13
  (F4 "\\E[26~")		       ; function key 14
  (F5 "\\E[28~")		       ; function key 15
  (F6 "\\E[29~")	               ; function key 16
  (F7 "\\E[31~")                       ; function key 17
  (F8 "\\E[32~")                       ; function key 18
  (F9 "\\E[33~")                       ; function key 19
  (FA "\\E[34~")                       ; function key 20
  (kn 20)
  (@0 "\\E[1~")			       ; find key
  (kI "\\E[2~")                        ; insert character/insert mode key
  (kD "\\E[3~")                        ; key for delete char under cursor
  (*6 "\\E[4~")                        ; select key
  (kP "\\E[5~")	                       ; previous page key
  (kN "\\E[6~")                        ; next page key
  (km)                                 ; terminal has a meta key
  (kb "^H")	                       ; backspace key
  (ku "\\EOA")                         ; cursor up key
  (kd "\\EOB")                         ; cursor down key
  (kr "\\EOC")                         ; cursor right key
  (kl "\\EOD")                         ; cursor left key
  (li 24)                              ; number of lines
  (md "\\E[1m")                        ; start bold mode
  (me "\\E[m")                         ; end all modes
  (mr "\\E[7m")                        ; start reverse mode
  (ms)                       ; cursor movement works in standout/underline mode
  (nd "\\E[C")                         ; cursor right one character
  (pt)                                 ; 
  (eA "\\E)0")			       ; enable alternate character set
  (as "^N")                            ; start alternate character set
  (ae "^O")                            ; end alternate character set
  (ml "\\El")                          ; 
  (mu "\\Em")
  (sc "\\E7")                          ; save cursor position
  (rc "\\E8")                          ; restore saved cursor position
  (sf "\\n")                           ; normal scroll one line
  (so "\\E[7m")                        ; start standout mode
  (se "\\E[m")                         ; end standout mode
  (sr "\\EM")                          ; reverse scroll
  (ti "\\E7\\E[?47h")                  ; begin program that uses cursor motion
  (te "\\E[2J\\E[?47l\\E8")            ; end program that uses cursor motion
  (up "\\E[A")                         ; cursor up one line
  (us "\\E[4m")                        ; start underlining
  (ue "\\E[m")                         ; end underlining
  (u9 "\\E[c")                         ; terminal enquire string
  (xn)                                 ; newline/wraparound glitch
))

(define get-num-columns (lambda ()
  (define env-columns (getenv "COLUMNS"))
  (define prop-columns (assoc 'co terminal-properties))
  (if vt100-terminal-columns
    vt100-terminal-columns
    (if env-columns
      (with-input-from-string env-columns read)
      (if prop-columns
        (cadr prop-columns)
        80
      )
    )
  )
))

(define save-cursor-position (let ()
  (define magic-string (termcap-string->string (cadr (assoc 'sc terminal-properties))))
  (lambda (vt100-out-port)
    (display magic-string vt100-out-port)
  )
))

(define restore-cursor-position (let ()
  (define magic-string (termcap-string->string (cadr (assoc 'rc terminal-properties))))
  (lambda (vt100-out-port)
    (display magic-string vt100-out-port)
  )
))
 
(define clear-the-screen (let ()
  (define magic-string (termcap-string->string (cadr (assoc 'cl terminal-properties))))
  (lambda (vt100-out-port)
    (display magic-string vt100-out-port)
  )
))

(define colors '(
  (red 1)
  (green 2)
  (orange 3)
  (blue 4)
  (purple 5)
  (cyan 6)
  (gray 7)
  (grey 7)
  ; unknown
  (black 8)
  (white 9)
))

(define foreground-colors '(
  (Black 30) 
  (Red 31) 
  (Green 32) 
  (Yellow 33) 
  (Blue 34) 
  (Magenta 35) 
  (Cyan 36) 
  (White 37) 
))
(define background-colors '(
  (Black 40) 
  (Red 41) 
  (Green 42) 
  (Yellow 43) 
  (Blue 44) 
  (Magenta 45) 
  (Cyan 46) 
  (White 47) 
))


(define make-string-with-colors (lambda (string fg-color bg-color)
  (if (symbol? fg-color)
    (set! fg-color (cadr (assoc fg-color colors)))
  )
  (if (symbol? bg-color)
    (set! bg-color (cadr (assoc bg-color colors)))
  )

  (if vt100-uses-colors 
    (string-append
      (termcap-string->string (cadr (assoc 'AF terminal-properties)) fg-color)
      (termcap-string->string (cadr (assoc 'AB terminal-properties)) bg-color)
      string
      ;(termcap-string->string (cadr (assoc 'op terminal-properties)))
      (termcap-string->string "\\E[0m")
    )
    ; else
    string
  )
))

(define goto-xy (lambda (vt100-out-port x y)
  (display (termcap-string->string (cadr (assoc 'cm terminal-properties)) y x) vt100-out-port)
))

(define scroll-region (lambda (vt100-out-port from to)
  (display (termcap-string->string (cadr (assoc 'cs terminal-properties)) from to) vt100-out-port)
))

;
; measure the screen
;
(define measure-the-screen (lambda (vt100-in-port vt100-out-port)
  (define absorb-a-number (lambda (l)
    (let helper ((cursor l) (acc #f))
      (if (null? cursor)
        (cons cursor acc)
        (case (car cursor)
          ((#\0) (helper (cdr cursor) (+ (* 10 acc) 0)))
          ((#\1) (helper (cdr cursor) (+ (* 10 acc) 1)))
          ((#\2) (helper (cdr cursor) (+ (* 10 acc) 2)))
          ((#\3) (helper (cdr cursor) (+ (* 10 acc) 3)))
          ((#\4) (helper (cdr cursor) (+ (* 10 acc) 4)))
          ((#\5) (helper (cdr cursor) (+ (* 10 acc) 5)))
          ((#\6) (helper (cdr cursor) (+ (* 10 acc) 6)))
          ((#\7) (helper (cdr cursor) (+ (* 10 acc) 7)))
          ((#\8) (helper (cdr cursor) (+ (* 10 acc) 8)))
          ((#\9) (helper (cdr cursor) (+ (* 10 acc) 9)))
          (else (cons cursor acc))
        )
      )
    )
  ))
  (define absorb-helper (lambda (racc)
    (define char (read-char vt100-in-port))
    (if (eq? char #\R)
      (reverse racc)
      (absorb-helper (cons char racc))
    )
  ))
  ; turn off scrolling areas
  (display (termcap-string->string "\033[r") vt100-out-port)
  ; send the cursor to a position beyond the edge of the screen real estate
  (goto-xy vt100-out-port 998 998)
  ; send the enquire string
  (display (termcap-string->string "\\E[6n") vt100-out-port)
  ; get the report string
  (let ()
    (define absorb-response (absorb-helper '()))
    (let ()
      (define res1 (absorb-a-number (cddr absorb-response)))
      res1
      (cdr res1)
      (set! vt100-terminal-lines (cdr res1))
      (let ()
        (define res2 (absorb-a-number (cdar res1)))
        (set! vt100-terminal-columns (cdr res2))
       #t
      )
    )
  )
))

(define vt100-enable-colors (lambda ()
  (set! vt100-uses-colors #t)
))
(define vt100-disable-colors (lambda ()
  (set! vt100-uses-colors #f)
))
(define vt100-set-terminal-geometry (lambda (columns lines)
  (set! vt100-terminal-lines lines)
  (set! vt100-terminal-columns columns)
))

(define vt100-uses-colors #t)

(define vt100-test (lambda (vt100-out-port)
  (clear-the-screen vt100-out-port)
  (display (termcap-string->string "\\E7\\E[?47h") vt100-out-port)   ; begin program that uses cursor motion
  (goto-xy vt100-out-port 1 1)
  (display "FOO at 1,1" vt100-out-port)
  (goto-xy vt100-out-port 2 2)
  (display "BAR AT 2,2" vt100-out-port)
  (goto-xy vt100-out-port 3 3)
  (display (make-string-with-colors "COLORED BAZ AT 3,3" 'blue 'green) vt100-out-port)
))

)
