(library (proverb xhtml)
  (export
    wikify-source
    widgetscript-apply widgetscript-apply-with-attrs
    xhtml-environment-pickler
    xhtml-widget-pickler
    xhtml-widget-stores make-widgetscript
    xhtml-widget-formals xhtml-widget-eval
    make-dom-id
    filter-widget decbase-filter hexbase-filter pretty-print-filter
    context-menu-filter-widget
    context-menu-selection-widget
    ; svg functions
    svgdemo svg2 svgwrap
    svgboxchart svglinechart svgdotchart
    svgsample svgsamplerect
    ; general 
    proverb-jsshell-background-widget
    welcome welcome-back
    xhtml-more* clear
    retrieve-file retrieve-file-with-mimetype
    xhtml-ls*
    xrun xrun*
    make-image make-script make-video
    make-video-from-url
    xhtml-show-stack
    xhtml-examine
    push-cookie
    xhtml-read-indirect
    xhtml-popen
    wikipoint
    make-preformatted-text
    ; help functions
    xhtml-search-browser
    xhtml-help-browser xhtml-help* help
    xhtml-help-wrapper* xhtml-help-index
    xhtml-save-history xhtml-auto-save-history
    xhtml-save-prefs
    ; foo
    xhtml-make-color-dictionary
    colortable
    colorspark
    make-bookmarklet
    make-fullscreen
  )
  (import
    (rnrs)
    (rnrs lists)
    (rnrs control)
    (rnrs arithmetic bitwise)
    (hashtable) (proverb search)
    (proverb types)
    (proverb doc) (proverb wiki) (proverb help)
    (proverb uri)
    (proverb xml) (proverb remote) (proverb mime)
    (proverb tire) (proverb filesystem)
    (proverb html) (proverb http) (proverb http-cache) (proverb javascript)
    (proverb channels)
    (proverb cond-expand) 
    (only (proverb bootstrap) support-dir)
    (proverb debug)
    (ice-9 syncase)
    (ice-9 rdelim)
    (ice-9 documentation)
    (ice-9 format)
    (proverb dialect)
    (only (proverb guilefoo)
      module-get-symbol-list
    )
    (guile modules)
    (guile errors)
    (only (guile filesystem)
      ; filesystem
      opendir readdir mkdir access?
      W_OK X_OK F_OK R_OK
      stat
      stat:type
    )
    (only (guile)
      ; module-related imports
      environment-module the-environment
      ; other imports
      with-input-from-string
      with-output-to-string
      call-with-output-string
      call-with-input-string
      define-macro eval-case defmacro:transformer *unspecified*
      call-with-deferred-observers list-set!
      local-eval usleep
      gensym closure?
      beautify-user-module! run-hook gettimeofday procedure-documentation
      object-property procedure-environment
      struct-set! %record-type-error object-address object->string
      procedure-source
      struct-ref set-defmacro-transformer! assert-defmacro?!
      variable-ref
      add-hook! struct? procedure-property
      delete hashq-set! variable-set! hash-set!
      sort
      hash-fold fluid-set!
      random list-head fluid-ref
      port-closed? fileno
      getenv open-file
      close
      get-output-string
      object-properties
      ; extra debugging
      stack-ref procedure-name source-properties
      memoized? unmemoize-expr
      call-with-new-thread
      force-output
      after-gc-hook
      gc-stats
    )
    (ice-9 pretty-print)
  )

(define make-dom-id (with-docstring
  "Create a string suitable for use as a DOM id in a browser. This string should be unlikely to collide with an existing DOM id."
  (lambda ()
    (define t (gettimeofday))
    (let ()
      (define s (string-append (number->string (car t)) (number->string (cdr t))))
      (string-append "dom-" s (number->string (object-address s)))
    )
  )
))

(define xhtml-help* (with-docstring
  "this is a private function wrapped by the XHTML help browser and the xhtml-help macro"
  (lambda (query-string)
    (if query-string
      (let ()
        (define results (studio-search query-string))
        (let ()
          (define aalist (aggregate-alist-for-assoc results))
          (let ()
            (define module-name-matches (assoc 'module aalist))
            (define help-article-matches (assoc 'article aalist)) 
	    (define binding-matches (assoc 'binding aalist))
          
            (make-wiki (string-append
              "== Matching modules ==\n"          
              (if module-name-matches
                (if (null? module-name-matches)
                  "none\n"
                  (apply string-append (map (lambda (v)
                    (define n (module-name v))
                   (string-append "* [[Scheme:" (object->string n) "|" (object->string n) "]]\n")
                  ) (cdr module-name-matches)))
                )
                "none\n"
              )
              "== Help article matches ==\n"
              (if help-article-matches
                (if (null? help-article-matches) 
                  "none\n"
                  (apply string-append (map (lambda (v)
                    (define article-name (string-append (car v) ":" (symbol->string (cdr v))))
                    (string-append "* [[" article-name " | " article-name "]]\n")
                  ) (cdr help-article-matches)))
                )
                "none\n"
              )
              "== Binding matches ==\n"
              (if (and binding-matches (not (null? binding-matches)))  
                (let ()
                  (define matches-by-module (aggregate-alist-for-assoc (map (lambda (v) (cons (cadr v) v)) (cdr binding-matches))))
                  (apply string-append (map (lambda (v)
                    (define modstr (object->string (module-name (car v))))
                    (apply string-append "=== module " modstr "===\n" (map (lambda (w)
                      (define symstr (symbol->string (caar w)))
                      (string-append "* [[[(xhtml-examine (module-ref (resolve-interface (quote " modstr ")) (quote " symstr "))) | " symstr "]]]\n")
                    ) (cdr v)))
                  ) matches-by-module))             
                )
                "none\n"
              )

            ))
          )
        )
      )
      ; else no query string
      (make-wiki "{{Proverb:default}}")
    )
  )
))

(define xhtml-help-wrapper* (with-docstring
  "This is a private function used in the expansion of the xhtml-help macro"
  (lambda args
  (if (null? args)
    (xhtml-help* #f)
    (apply make-xhtml (map (lambda (arg)
      (xhtml-help* (cond
        ((string? arg) arg)
        ((symbol? arg) (symbol->string arg))
        (else (object->string arg))
      ))
    ) args))
  )
  )
))

;
; This definition is split into separate syntax and syntax rules so that
; the syntax rules can be rebound into simply 'help' when the user is using
; an XHTML interaction environment.
;
(cond-expand 
 (guile
  (define-syntax xhtml-help (syntax-rules ()
    ((xhtml-help)
      (xhtml-help-wrapper*)
    )
    ((xhtml-help body ...)
      (apply xhtml-help-wrapper* '(body ...))
    )
  ))
  (define-syntax help (syntax-rules ()
    ((help)
      (xhtml-help-wrapper*)
    )
    ((help body ...)
      (apply xhtml-help-wrapper* '(body ...))
    )
  ))

  ; this is a macro so that it can capture your scope with the widget pickler
  (define-syntax make-widgetscript (syntax-rules ()
    ((make-widgetscript str)
      (with-xml-tag `(widgetscript xmlns "proverb" key ,(html-quote-escape-string (number->string (xhtml-widget-pickler)))) (make-raw-text (xml-body-escape-string str)))
    )
    ((make-widgetscript args ...)
      (let* ((l (list args ...)) (ll (length l)) (str (list-ref l (- ll 1))))
        (with-xml-tag `(widgetscript xmlns "proverb" key ,(html-quote-escape-string (number->string (xhtml-widget-pickler))) ,@(list-head l (-  ll 1))) (make-raw-text (xml-body-escape-string str)))
      )
    )
  ))
  ; this is a simpler-to-use macro version of the above
  (define-syntax widgetscript-apply-with-attrs (syntax-rules ()
    ((widgetscript-apply-with-attrs attrs str args)
      (begin
      ; this is tricky. we can't just bake 'args' into the javascript
      ; text because the hygenic macro expansion will rewrite args.
      (with-xml-tag `(widgetscript xmlns "proverb" key ,(html-quote-escape-string (number->string (xhtml-widget-pickler args))) ,@attrs) (make-raw-text (string-append
        "shell.inner_evaluator.eval_to_js(widget.formals,shell,function(v) {("
        (xml-body-escape-string str)
        ").apply(this,v);});"
      )))
      )
    )
  ))
  (define-syntax widgetscript-apply (syntax-rules ()
    ((widgetscript-apply str args)
      (begin
      ; this is tricky. we can't just bake 'args' into the javascript
      ; text because the hygenic macro expansion will rewrite args.
      (with-xml-tag `(widgetscript xmlns "proverb" key ,(html-quote-escape-string (number->string (xhtml-widget-pickler args)))) (make-raw-text (string-append
        "shell.inner_evaluator.eval_to_js(widget.formals,shell,function(v) {("
        (xml-body-escape-string str)
        ").apply(this,v);});"
      )))
      )
    )
  ))
 )
 (else
   #t
 )
)

(define clear (with-docstring
  "clear the output area in the javascript shell"
  (lambda ()
    (dribble (make-widgetscript "shell.clear();"))
    *unspecified*
  )
))

(define xhtml-search-browser (with-docstring
  "command will be called initially with #f, subsequently with a search term as a string"
  (lambda (command)
    (make-widgetscript "
      var current;
      var back_array=Array();
      var forward_array=Array();
      var div=widget.div;

      var painter=function(v) {
        if(back_array.length > 0) {
          current=back_array[0];
          div.removeChild(current);
        }
        back_array.unshift(v);
        div.appendChild(v);
      };

      var subshell=widget.shell.subshell(painter);

      var inp;
      var back,forward;
      inp=document.createElement('input');
      inp.type='text';

      inp.onchange=function() {
        forward_array=Array();
        subshell.string_evaluator(widget.wrapper('(command ' + subshell.string_escape(inp.value) + ')'),false);
        inp.value='';
      }

      back=document.createElement('input');
      back.type='button';
      back.value='back';
      back.onclick=function(e) {
        if(back_array.length > 1) {
          var current=back_array.shift();
          div.removeChild(current);
          forward_array.push(current);
          current=back_array[0];
          div.appendChild(current);
        }
      };

      forward=document.createElement('input');
      forward.type='button';
      forward.value='forward';
      forward.onclick=function(e) {
        if(forward_array.length > 0) {
          var current=back_array[0];
          div.removeChild(current);
          current=forward_array.pop();
          back_array.unshift(current);
          div.appendChild(current);
        }
      };
      div.appendChild(inp);
      div.appendChild(back);
      div.appendChild(forward);
         
      div.appendChild(document.createElement('br'));

      subshell.string_evaluator(widget.wrapper('(command #f)'),false);
    ")
  )
))

(define xhtml-help-browser (with-docstring
  "create an interactive browser widget for searching the help system"
  (lambda ()
    (xhtml-search-browser xhtml-help*)
  )
))

(define xhtml-ls* (with-docstring
  "Do something like shell ls in the browser for <path>"
  (lambda (path)
  (define s (catch #t (lambda () (stat path)) always-false))
  (if s
    (cond
      ((eq? (stat:type s) 'directory)
        (apply make-xhtml (cons
          (with-xml-tag 'h3 (make-text (string-append path ":")))
          (let ()
            (define d (catch #t (lambda () (opendir path)) always-false))
            (if d
              (let helper ((racc '()))
                (define dirent (readdir d))
                (if (eof-object? dirent)
	          (reverse racc)
                  (helper (cons (make-xhtml (make-shortcut-anchor `(ls ,(string-append path "/" dirent)) (make-text dirent)) (with-xml-tag 'br)) racc))
                )
              )
              (list (with-xml-tag 'h3 (make-text "That you cannon open")))
            )
          )
        ))
      )
      ((eq? (stat:type s) 'regular)
        (with-xml-tag 'h3 (make-text (string-append path ": regular file")) (make-shortcut-anchor `(more ,path) (make-text " view ")))
      )
      (else s)
    )
    s
  )
  )
))

(cond-expand
 (guile
(define-macro (xhtml-ls arg)
  `(cond
    ((string? (quote ,arg))
      (xhtml-ls* ,arg)
    )
    ((symbol? (quote ,arg))
      (xhtml-ls* (symbol->string (quote ,arg)))
    )
    ((char? (quote ,arg))
      (xhtml-ls* (string ,arg))
    )
    (else "Try (ls \"foo\") or (ls 'foo)")
  )
)
 )
 (else #f)
)

(define xhtml-more* (with-docstring
  "do something like shell more in the browser"
  (lambda (filename)
  (define p (open-input-file filename))
  (if p
    (let helper ((racc '()))
      (define v (read-line p))
      (if (eof-object? v)
        (apply make-xhtml (reverse racc))
        (helper (cons (make-xhtml (make-text v) (with-xml-tag 'br)) racc))
      )
    )
    (make-text "couldn't open the file")
  )
  )
))

(cond-expand
 (guile
(define-macro (xhtml-more arg)
  `(cond
    ((string? (quote ,arg))
      (xhtml-more* ,arg)
    )
    ((symbol? (quote ,arg))
      (xhtml-more* (symbol->string (quote ,arg)))
    )
    ((char? (quote ,arg))
      (xhtml-more* (string ,arg))
    )
    (else "Try (more \"foo\") or (more 'foo)")
  )
)
 )
 (else #f)
)

(define xhtml-make-color-dictionary (with-docstring
  "make a color dictionary -- an association between values and colors"
  (lambda ()
    ; hash, size, current
    ; we may need nothing more than the hashtable here
    (list (make-eq-hashtable) 1000 0)
  )
))

(define xhtml-cd-get-color (with-docstring
  "return the color associated with sym in cd. manufacture a new color on demand."
  (lambda (cd sym)
  (define hexify-byte (lambda (v)
    (if (> v 15)
      (number->string v 16)
      (string-append "0" (number->string v 16))
    )
  ))
  (or (hashtable-ref (car cd) sym #f) (let ()
    (define new-color (string-append "#" (hexify-byte (random 256))
                                         (hexify-byte (random 256))
                                         (hexify-byte (random 256))))
    (hashtable-set! (car cd) sym new-color)
    (list-set! cd 2 (+ (list-ref cd 2) 1))
    new-color
  ))
  )
))

(define colortable (with-docstring
  "Draw a colortable in the browser. accept a color dictionary <dic>, a table cell size <size> in pixels, a width at which to wrap <wrap>, the boolean flag <dense?> and a list of Scheme values <values>. This function is called by colorspark."
 (lambda (dic size wrap dense? values)
  (define s (number->string size))

  (define rowify (lambda (subset)
    ; inverted results
    (define i (let invert ((l subset) (raccs (map (lambda (v) '()) (car values))))
      (if (null? l)
        (map reverse raccs)
        (invert (cdr l) (map cons (car l) raccs))
      )
    ))
    (map (lambda (v)
      (apply with-xml-tag (cons "<tr>" (map (lambda (w)
        (with-xml-tag (string-append "<td onclick=\"alert('" (html-quote-escape-string (html-escape-string (object->string w))) "');\" width='" s "' height= '" s "' bgcolor='" (xhtml-cd-get-color dic w) "'>"))
      ) v)))
    ) i)
  ))

  (if dense?
    (let helper ((cursor values) (racc '()))
      (if (> (length cursor) wrap)
        (helper (list-tail cursor wrap) (cons (rowify (list-head cursor wrap)) racc))
        (apply with-xml-tag "<table cellspacing='1'>" (apply append (reverse
          (cons (rowify cursor) racc))
        )) 
      )
    )
    (let helper ((cursor values))
      (if (> (length cursor) wrap)
        (apply make-xhtml
          (with-xml-tag "<table cellspacing='1'>" (rowify (list-head cursor wrap)))
          (with-xml-tag "<br>")
          (helper (list-tail cursor wrap))
        )
        (apply with-xml-tag "<table cellspacing='1'>" (rowify cursor))
      )
    )
  )
  )
))

(define colorspark (with-docstring
  "make a kind of sparkline for <values>, a list of the form: ( ( datum0 ... ) ...) with each element of <values> having the same length. Each element in <values> will correspond to a column in the sparkline. Each color in the sparkline will correspond to a unique datum in <values>."
  (lambda (values)
  (if (eq? (length (car values)) 1)
    (colortable (xhtml-make-color-dictionary) 3 100 #t values)
    (colortable (xhtml-make-color-dictionary) 3 100 #f values)
  )
  )
))

(define xhtml-show-stack (with-docstring
  "make a hypertext presentation of the stack-valued object <stack>"
  (lambda-w-types ((stack stack?))
    (define max-frames 50)
    (let helper ((f (stack-ref stack 0)) (foofoo #f) (left max-frames) (racc '()))
      (define prev (frame-previous f))
      (define proc (frame-procedure f))
      (define args (frame-arguments f))
      (define source (frame-source f))

      (if (and prev (> left 0))
        (let ()
          (define entry (if proc
            (list (string-append (object->string (procedure-name proc)) (if source (object->string (source-properties source)) "unknown")))
            (list (if source (object->string (source-properties source)) "unknown"))
          ))
          (helper prev #f (- left 1) (cons (cons entry f) racc))
        )
        (let ()
          (define show-colorized-source (lambda ()
            (define searcher (lambda (form in)
              (let inner-helper ((cursor in) (pre-racc '()))
                (if (null? cursor)
                  #f
                  (if (equal? form (car cursor))
                    (list (reverse pre-racc) (cdr cursor))
                    (inner-helper (cdr cursor) (cons (car cursor) pre-racc))
                  )
                )
              )
            ))
            (define frames-outer-to-inner (map cdr racc))
            (let ()
              (define sources-outer-to-inner (map (lambda (v)
                (let ((s (frame-source v))) (if (memoized? s) (unmemoize-expr s) s))
              ) frames-outer-to-inner))
              (let helper ((cursor sources-outer-to-inner) (fcursor frames-outer-to-inner))
                (if (null? cursor)
                  (dribble "DONE!\n")
                  (begin
                    (if (null? (cdr cursor))
                      (begin
                        (with-xml-tag '(span style "color: red;") (make-text (object->string (car cursor))))
                      )
                      (let ((match (searcher (cadr cursor) (car cursor))))
                        (if match
                          (begin
                            (with-xml-tag 'span
                              (apply with-xml-tag '(span style "color: black;")
                                (make-text "(")
                                (map (lambda (v) (make-text (object->string v))) (car match))
                              )
                              (helper (cdr cursor) (cdr fcursor))
                              (apply with-xml-tag '(span style "color: black;") (append
                                (map (lambda (v) (make-text (object->string v))) (cadr match))
                                (list (make-text ")"))
                              ))
                            )
                          )
                          (with-xml-tag 'div
                            (let ((p (frame-procedure (car fcursor))))
                              (if p
                                (let ((e (procedure-environment p)))
                                  (if e
                                    (let ((m (environment-module e)) (pn (procedure-name p)))
                                      (if m
                                        (with-xml-tag 'span
                                          (make-wiki (string-append "In call to procedure [[Scheme:" (symbol->string pn) "|" (symbol->string pn) "]] from module " (object->string (module-name m))))
                                        )
                                        (make-text (format #f "IN CALL TO PROCEDURE: ~a from environment ~a\n" (procedure-name p) e))
                                      )
                                    )
                                    (make-text (format #f "IN CALL TO PROCEDURE: ~a\n" (procedure-name p)))
                                  )
                                )
                                (make-text (format #f "IN CALL TO UNKNOWN PROCEDURE:\n"))
                              )
                            )
                            (make-text "\nFrame source:\n")
                            (make-text (object->string (car cursor)))
                            (make-text "\nSubordinate stack:\n")
                            (with-xml-tag '(div style "position: relative; left: 20px; border-width: 1px; border-style: dotted;")
                              (helper (cdr cursor) (cdr fcursor))
                            )
                          )
                        )
                      )
                    )
                  )
                )
              )
            )
          ))

          (define frobify (lambda (index)
            (define f (cdr (list-ref racc index)))
            (format #f "Frame real?: ~a\nFrame Procedure: ~a\nFrame Arguments: ~aFrame Source: ~a\n"
              (object->string (frame-real? f))
              (object->string (frame-procedure f))
              (object->string (frame-arguments f)) 
              (let ((s (frame-source f)))
                (if (memoized? s)
                  (unmemoize-expr s)
                  (object->string (frame-source f)) 
                )
              )
            )
          ))
          (widgetscript-apply "function(l) {
            var ul;
            var x;
            ul=document.createElement('ul');
            var sub=widget.shell.subshell(function(v) {
              widget.div.appendChild(v);
            },false);
            var s;
            sub.string_evaluator(widget.wrapper('(show-colorized-source)'),false);
            for(x=0;x<l.length;x++) {
              var n=document.createTextNode(''+l[x]);
              (function() {
                var li=document.createElement('li');
                var foo=x;
                var expanded=false;
                li.addEventListener('click',function(e) {
                  if(!expanded) {
                    expanded=document.createElement('div');
                    widget.shell.inner_evaluator.eval_to_js(
                      widget.wrapper('(frobify ' + foo + ')'),
                      widget.shell,
                      function(v) {
                        expanded.appendChild(document.createTextNode(v));
                      }
                    );
                    li.appendChild(expanded);
                  } else {
                    li.removeChild(expanded);
                    expanded=false;
                  }
                },false);
                li.appendChild(n);
                ul.appendChild(li);
              })();
            }
            /*widget.div.appendChild(ul);*/
          }" (list (map car racc)))
        )
      )
    )
  )
))

(define xrun* (with-docstring
  "accept and run <thunk>. If <thunk> returns without error, return the received value, otherwise return an XHTML stack browser. The macro xrun wraps its arguments with a thunk for xrun*."
  (lambda (thunk)
  ; wacky-runner should only be on the stack once. thunk may be recursive
  (define wacky-runner (lambda ()
    (thunk)
  ))
  (define wacky-stack #f)
  (define wacky-handler (lambda args
    (define stack (make-stack #t wacky-handler wacky-runner 0))
    (set! wacky-stack stack)
  ))
  (catch #t (lambda () (lazy-catch
    #t
    wacky-runner
    wacky-handler
  )) (lambda args (if wacky-stack
    (with-xml-tag "<div>"
      (make-text (format #f "error: ~a\n" (if (and (list? args) (> (length args) 2)) (apply format #f (list-ref args 2) (list-ref args 3)) "unknown type")))
      (xhtml-show-stack wacky-stack))
    (with-xml-tag "<div>"
      (make-text (format #f "error: ~a\n" (if (and (list? args) (> (length args) 2)) (apply format #f (list-ref args 2) (list-ref args 3)) "unknown type")))
  ))))
  )
))

(define-macro (xrun form)
  `(xrun* (lambda () ,form))
)

(define make-image (with-docstring
  "accept an image as a string and return an XHTML inline image. Scheme will attempt to guess a mime type."
  (lambda (image-string . args)
    (define mimetype (or (image-guess-mime-type image-string) "text/octet-stream"))
    (with-xml-tag `(img ,@args src ,(make-data-url image-string mimetype)))
  )
))

(define make-script (with-docstring
  "accept unescaped Javascript text as script-text, correctly escape it, and wrap it in a script tag"
  (lambda (script-text)
    (with-xml-tag 'script (make-raw (xml-body-escape-string script-text)))
  )
))

(define nav (with-docstring
  "a lame hypertext way to cruise around Guile"
  (lambda ()
  (apply with-xml-tag (cons "<div>" `(
    ,(make-text (identify-scheme-type))
    ,(with-xml-tag "<br>")
    ,(make-text (string-append 
      "Current module: " (object->string (module-name (current-module)))
    ))
    ,(with-xml-tag "<br>")
    ,(make-text "Navigate by module:\n")
    ,@(map (lambda (v)
       (define n (module-name v))
       (make-xhtml
         (make-text "  ") (make-shortcut-anchor `(xhtml-examine (resolve-interface (quote ,n))) (make-text (string-append (object->string n))))
         (with-xml-tag "<br>")
       )
    ) (cons (current-module) (module-uses (current-module))))
  )))
  )
))

(define wikify-source (with-docstring
  "render the Scheme source source as a wiki element"
  (lambda (source)
    (make-wiki (apply string-append (map (lambda (v)
      (string-append "%%" v "\n")
    ) (studio-string-split (with-output-to-string (lambda () (pretty-print source))) #\newline))))
  )
))

(define xhtml-examine (with-docstring
  "create a detailed XHTML view of <object>"
  (lambda (object)
  (define wikify-value (lambda (v)
    ;(define l (studio-string-split (call-with-output-string (lambda (p) (pretty-print v p))) #\newline))
    (define l (studio-string-split (call-with-output-string (lambda (p) (display v p))) #\newline))
    (make-wiki (apply string-append (map (lambda (v)
      (string-append "%%" v "\n")
    ) l)))
  ))
  (cond
    ((module? object)
      (make-wiki (string-append "{{Scheme:" (object->string (module-name object)) "}}"))
    )
    ((closure? object) (let ()
      (define ds (object-property object 'documentation)) 
      (if (and ds (string? ds))
        (make-xhtml
          (if (eq? (object-property ds 'format) 'wiki)
            (make-wiki ds)
            (wikify-value ds)
          )
          (wikify-value (procedure-source object))
        )
        (wikify-value (procedure-source object))
      )
    ))
    ((list? object) (let ()
      object
    ))
    ((vector? object) (let ()
      object
    ))
    ((string? object)
      (if (> (string-length object) 2048)
        (make-text "a large string")
        (make-text object)
      )
    )
    (else (make-text (format #f "~a" object)))
  )
  )
))

;
; states used by open widgets
; 
(define xhtml-widget-stores '())

(define xhtml-environment-pickler (with-docstring
  "saves the environment e and the additional values in the list formals for a widget. they persist until the widget closes." 
  (lambda (e formals)
    (define store (vector e formals))
    (let ()
      (define a (object-address store))
      (set! xhtml-widget-stores (cons (cons a store) xhtml-widget-stores))
      a
    )
  )
))

(define-syntax xhtml-widget-pickler (syntax-rules ()
  ((xhtml-widget-pickler)
    (xhtml-environment-pickler (the-environment) '())
  )
  ((xhtml-widget-pickler formals)
    (xhtml-environment-pickler (the-environment) formals)
  )
))

(define xhtml-widget-delete! (with-docstring
  "delete the pickled widget context <pickle> from the global list of stored contexts"
  (lambda (pickle)
    (set! xhtml-widget-stores (remove (assoc pickle xhtml-widget-stores) xhtml-widget-stores))
  )
))

(define xhtml-widget-eval (with-docstring
  "evaluate <form> in the pickled widget context <pickle>"
  (lambda (form pickle)
    (local-eval form (vector-ref (cdr (assoc pickle xhtml-widget-stores)) 0))
  )
))

(define xhtml-widget-formals (with-docstring
  "return the formals jammed into a widget function"
  (lambda (pickle)
    (vector-ref (cdr (assoc pickle xhtml-widget-stores)) 1)
  )
))

(define fork-add! (with-docstring "fork-add and return a clickable link to select the cookie"
  (lambda ()
    (define c (fork-add))
    (make-wiki (format #f "[[[>~a|select ~a]]]" c c))
  )
))

(define push-cookie (with-docstring "command the browser to switch focus to cookie"
  (lambda (cookie)
    (dribble (make-wiki (format #f "Your current cookie is ``!shell.get_eval_cookie()``\n")))
    (with-xml-tag "<shell>" (make-raw-text
      "!shell.set_eval_cookie('"
      (js-quote-escape-string (symbol->string cookie))
      "')"
    ))
  )
))

(define fork-push (with-docstring "fork-push does a fork-add and commands
the browser to switch focus to the child"
  (lambda ()
    (define c (fork-add))
    (push-cookie c)
  )
))

(define retrieve-file (with-docstring
  "(retrieve-file FILENAME) -- make an anchor tag linking to FILENAME encoded as a data url. The MIME type will be text/octet-stream"
  (lambda (filename) (retrieve-file-with-mimetype filename "text/octet-stream"))
))

(define retrieve-file-with-mimetype (with-docstring
  "(retrieve-file FILENAME MIMETYPE) -- make an anchor tag linking to FILENAME encoded as a data url. The MIME type will be MIMETYPE"
  (lambda (filename mimetype)
    (with-xml-tag `(a href ,(make-data-url (filename->string filename) mimetype)) (make-text filename))
  )
))

(define gc-widget (with-docstring
  "create a Proverb browser widget that monitors system garbage collection in Guile"
  (lambda ()
  (define responder (snatch-responder))
  (define hooker (lambda args
    (responder 'i 'dribble (format #f "~a cells\n" (cdr (assoc 'cells-allocated (gc-stats)))))
  ))
  (add-hook! after-gc-hook hooker)
  (responder 'i 'dribble (make-widgetscript "
    widget.finalize=function() {
      alert('gc widget finalize');
      shell.inner_evaluator.eval_to_string('] ' + widget.wrapper('(begin (remove-hook! after-gc-hook hooker) (responder (quote r) 5))'),shell,function(v) {
        alert('gc widget removed hook');
      });
    };
    alert('still here');
  "))
  )
))

(define with-charset (with-docstring
  "create XML tags to cause string <str> to be rendered in the browser which character set <charset>"
  (lambda (str charset)
    (with-xml-tag `(iframe
                    style "border: 0px;"
                    src ,(make-data-url str "text/plain" 'charset charset)))
  )
))

(define xhtml-make-live-updater (with-docstring
  "an XHTML demonstration that runs the thunk <func> <count> times with at least <delay> seconds between runs"
  (lambda (count delay func)
  (define id (make-dom-id))
  (define s (snatch-responder))
  (s 'i 'dribble (with-xml-tag (format #f "<div id='~a'>" id)
    (make-text "FOO")
  ))
  (let helper ((c count))
    (if (eq? c 0)
      'ok
      (begin
        (s 'i 'dribble (with-xml-tag (format #f "<replace target='~a'>" id)
          (func)
        ))
        (usleep (inexact->exact (floor (* delay 1000000))))
        (helper (- c 1))
      )
    )
  )
  (s 'r 'ok)
  )
))

(define pretty-print-filter (with-docstring
  "Accept a Scheme form as a string <str> and pretty-print it in XHTML. The result is an XHTML sub-tree"
  (lambda (str)
  (call-with-input-string str (lambda (ip)
    (define r (read ip))
    (make-text (call-with-output-string (lambda (op)
      (pretty-print r op)
    )))
  ))
  )
))

(define hexbase-filter (with-docstring
  "accept a string with embedded Scheme numbers and rewrite it as an XHTML text fragment with the Scheme numbers in hex"
  (lambda (str)
  (define f (lambda (v)
    (cond
      ((number? v) (string-append "#x" (number->string v 16)))
      ((null? v) v)
      ((pair? v) (cons (f (car v)) (f (cdr v))))
      (else v)
    )
  ))
  (call-with-input-string (string-append "(" str ")") (lambda (ip)
    (define r (read ip))
    (make-text (call-with-output-string (lambda (op)
      (for-each (lambda (rv)
        (display (f rv) op) (display " " op)
      ) r)
    )))
  ))
  )
))

(define decbase-filter (with-docstring
  "accept a string with embedded Scheme numbers and rewrite it as an XHTML text fragment with the Scheme numbers in base 10"
  (lambda (str)
  (define f (lambda (v)
    (cond
      ((number? v) (number->string v 10))
      ((null? v) v)
      ((pair? v) (cons (f (car v)) (f (cdr v))))
      (else v)
    )
  ))
  (call-with-input-string (string-append "(" str ")") (lambda (ip)
    (define r (read ip))
    (make-text (call-with-output-string (lambda (op)
      (for-each (lambda (rv)
        (display (f rv) op) (display " " op)
      ) r)
    )))
  ))
  )
))

(define filter-widget (with-docstring
  "XHTML widget that passes the selection in the browser to the Scheme procedure <formatter> as a string and rewrites the browser selection in-place with the result of <formatter>. Most values of <formatter> will return XHTML subtrees. If <formatter> has object documentation, this will be printed in the widget body."
  (lambda (formatter)
    (widgetscript-apply "function(desc) {
      var subshell;
      subshell=widget.shell.subshell(function(v) {
        portable_replace_selection(window,v);
      });
      widget.div.appendChild(document.createTextNode(desc));
      widget.div.appendChild(document.createElement('br'));
      var button;
      button=document.createElement('input');
      button.type='button';
      button.value='filter the selection';
      widget.div.appendChild(button);
      widget.div.appendChild(document.createElement('br'));
      button.addEventListener('click',function(e) {
        var s=portable_selection_to_string(window);
        if(s) {
          subshell.string_evaluator(
            '] ' + widget.wrapper('(formatter ' + subshell.string_escape(s) + ')'),
              shell);
        }
        e.stopPropagation();
      },false);
    }" (list (or (object-documentation formatter) "a scheme procedure")))
  )
))

(define context-menu-selection-widget (with-docstring
  "sort of a thingy to hang a Scheme selection processor in the context menu"
  (lambda (name formatter)
    (widgetscript-apply-with-attrs '(hidden "true") "function(desc) {
      shell.add_to_context_menu(''+desc,function(e) {
        var s=portable_selection_to_string(window);
        if(s && s!='') {
          shell.string_evaluator(
            '] ' + widget.wrapper('(formatter ' + shell.string_escape(s) + ')'),
              shell);
        } else {
          //e.target.style.backgroundColor='orange';
          alert('context menu with no selection: ');
        }
      });
    }" (list (or name (object-documentation formatter) "a scheme procedure")))
  )
))

(define context-menu-filter-widget (with-docstring
  "sort of a thingy to hang a Scheme selection rewriter in the context menu"
  (lambda (name formatter)
    (widgetscript-apply-with-attrs '(hidden "true") "function(desc) {
      var subshell;
      subshell=widget.shell.subshell(function(v) {
        portable_replace_selection(window,v);
      });
      shell.add_to_context_menu(''+desc,function(e) {
        var s=portable_selection_to_string(window);
        if(s && s!='') {
          subshell.string_evaluator(
            '] ' + widget.wrapper('(formatter ' + subshell.string_escape(s) + ')'),
              shell);
        } else {
          alert('context menu with no selection: ');
        }
      });
    }" (list (or name (object-documentation formatter) "a scheme procedure")))
  )
))

(define make-fullscreen (lambda (x)
  (widgetscript-apply "function(x) {
    with_fullscreen_div(function(d) {
      var subshell=widget.shell.subshell(function(v) {
        d.appendChild(v);
      });
      subshell.string_evaluator(']' + widget.wrapper('x'));
    });
  }" '())
))

;;; this left over fragment is kind of interesting
;  var shwrapfunc=function(scheme_name) {
;    return function() {
;      var cont=false;
;      var s='(' + scheme_name + ' ';
;      var x,v;
;      for(x=0 ; x < arguments.length ; x++) {
;        v=arguments[x];
;        if(typeof(v) == 'string') {
;          s+=''+scheme_stringify(v)+' ';
;        } else if(x==(arguments.length - 1) && typeof(v) == 'function') {
;          cont=v;
;        } else {
;          s+=''+v+' ';
;        }
;      }
;      s+=')';
;      widget.shell.string_evaluator('] ' + widget.wrapper(s));
;    }
;  };

(define xhtml-read-indirect (lambda (url)
  (define key (gensym "yield-key-"))
  (define resumer (lambda (v) (bogo-resume key v) "ok"))

  (dribble (make-xhtml
    (with-xml-tag 'script (make-raw-text "jsexp_data=false;"))
    (widgetscript-apply "function(url) {
      widget.div.appendChild(document.createTextNode('Proverb is using your browser to read from ' + url + ' ... '));
      var s=document.createElement('script');
      s.src=url;
      s.onload=function() {
        widget.div.appendChild(document.createTextNode('success'));
        widget.shell.inner_evaluator.eval_to_string('] ' + widget.wrapper('(resumer (quote ' + jsexp_data + '))'),widget.shell,function(v) { alert(v); });
      };
      document.body.appendChild(s);
    }" (list url))
  ))

  (tire-eval-yield key)
))

(define xhtml-help-index (with-docstring
  "produce a hypertext index of all the help articles available inside Proverb"
  (lambda ()
    (define index-by-wiki (aggregate-alist-for-assoc (help-index)))
    (make-wiki (apply string-append (map (lambda (a)
      (define wiki-name (car a))
      (apply string-append
        "=== Results from the " wiki-name " wiki ===\n" 
        (map (lambda (v)
          (define article-name (symbol->string v))
          (string-append "* [[" wiki-name ":" article-name " | " article-name "]]\n")
        ) (cdr a))
      )
    ) index-by-wiki)))
  )
))

(define saved-attrs '())

(define proverb-jsshell-background-widget (with-docstring
  "return an xhtml object that initializes some persistient client and server relationships. this is typically dribbled or returned as part of the welcome form"
  (lambda args
    (define delete-attr! (lambda (key)
      (define aval (assoc key saved-attrs))
      (if aval
        (begin
          (set! saved-attrs (delete aval saved-attrs))
          (delete-attr! key)
        )
      )
    ))
    (define save-attr (lambda (key value)
      (delete-attr! key)
      (set! saved-attrs (cons (cons key value) saved-attrs))
      (save-prefs)
      #t
    ))
    (make-xhtml
      (context-menu-filter-widget "hex" hexbase-filter)
      (context-menu-filter-widget "prettyprint" pretty-print-filter)
      (context-menu-selection-widget "help" (lambda (term)
        (if (string? term)
          (xhtml-help* term)
          (format #t "unable to get help for ~a\n" term)
        )
      ))
      ; auto-save every ten minutes
      (xhtml-auto-save-history (* 600 1000))
      ; get preference updates
      (widgetscript-apply-with-attrs '(hidden "true") "function() {
        jsprefs.onchange=function(key,oldval,newval) {
          widget.shell.inner_evaluator.eval_to_js(widget.wrapper('(save-attr ' + scheme_stringify(''+key) + ' '  + scheme_stringify(''+newval) + ')'),widget.shell,function(v) { return true; },false);
        };
      }" '())
    )
  )
))

(define welcome (with-docstring
  "create a multi-media welcome message for Proverb. This function can be redefined to customize Proverb when Proverb is embedded in other applications."
  (lambda ()
    (make-xhtml
      (proverb-jsshell-background-widget)
      (with-xml-tag "<table border='0' width='100%'>"
        (with-xml-tag "<tr>"
          (with-xml-tag "<td>"
            (make-wiki "{{Proverb:welcome}}")
          )
          (with-xml-tag "<td align='right'>"
            (with-xml-tag "<sanchor href='http://www.ida.org'>" (make-image (filename->string (string-append support-dir "/artwork/ida.png")) 'width 64))
            (with-xml-tag "<br>")
            (with-xml-tag "<small>" (make-wiki "{{attribution-small}}"))
          )
        )
      )
    )
  )
))
(define welcome-back (with-docstring
  "create a multi-media 'welcome back' message for Proverb. This function can be redefined to customize Proverb when Proverb is embedded in other applications."
  (lambda ()
    (make-xhtml
      (proverb-jsshell-background-widget)
      (make-wiki "{{Proverb:welcome_back}}")
    )
  )
))

(load "svg.scm")

(define make-video-from-url (with-docstring
  "accept a video as an URL and a mime type and return an XHTML inline video."
  (lambda (url mime . rest)
  (define width 640)
  (define height 480)
  (widgetscript-apply "function(url,mime,width,height) {

    if (mime == 'video/quicktime') {
      var movie_width = width || 320;
      var movie_height = height || 240;
 
      var o=document.createElement('object');
      o.setAttribute('type',mime);
      o.setAttribute('classid','clsid:02BF25D5-8C17-4B23-BC80-D3488ABDDC6B');
      o.setAttribute('width',movie_width);
      o.setAttribute('height',movie_height);
      o.setAttribute('codebase',\"http://www.apple.com/qtactivex/qtplugin.cab\");
      o.setAttribute('pluginspage','http://www.apple.com/quicktime/download/');

      function mp(k,v) {
        var p=document.createElement('param');
        p.setAttribute('name',k);
        p.setAttribute('value',v);
        o.appendChild(p);
      }

      mp('src',url);
      mp('enablejavascript',\"True\");
      mp('postdomevents',\"True\");
      //mp('controller',\"false\");
      mp('controller',\"true\");
      mp('autoplay',\"true\");

      /* the 'e' case for when we fail on the plugin */
      {
        var e=document.createElement('embed');
        e.setAttribute('src',url);
        e.setAttribute('width',movie_width);
        e.setAttribute('height',movie_height);
        o.appendChild(e);
      }

      widget.div.appendChild(o);
    } else if (mime == 'application/x-shockwave-flash') {
      var movie_width = width || 320;
      var movie_height = height || 240;
 
      var o=document.createElement('object');
      var e=document.createElement('embed');

      o.setAttribute('width',movie_width);
      o.setAttribute('height',movie_height);
      e.setAttribute('width',movie_width);
      e.setAttribute('height',movie_height);
      e.setAttribute('type',mime);
      e.setAttribute('src',url);

      function mp(k,v) {
        var p=document.createElement('param');
        p.setAttribute('name',k);
        p.setAttribute('value',v);
        o.appendChild(p);
        e.setAttribute(k,v);
      }

      mp('movie',url);
      mp('allowFullScreen',true);

      o.appendChild(e);

      widget.div.appendChild(o);
    } else {
      widget.div.appendChild(document.createTextNode('I do not understand the video mime type ' + mime));
    }
}

  " (list url mime width height))
  )
))

(define make-video (with-docstring
  "accept a video as a string and return an XHTML inline video. Scheme will attempt to guess a mime type."
  (lambda (video-string . rest)
    (define key (symbol->string (gensym "cache-")))
    (define mime (let ()
      (define m (member 'mime rest))
      (if m
        (cadr m)
        (or (image-guess-mime-type video-string) "text/octet-stream")
      )
    ))
    (let ()
      (define bogo (make-http-cache-fragment video-string mime))
      (http-tire-cache-add key (cadr bogo) (caddr bogo))
      (make-video-from-url (string-append "/cache?key=" (url-encode-string key)) mime)
    )
  )
))

(define xhtml-save-history (with-docstring
  "save the non-persistient elements of the browser command line history to ~/.proverb/history"
  (lambda ()
    (define proverb-base (string-append (getenv "HOME") "/.proverb"))
    (define save-history (lambda (elements)
      (if (access? proverb-base (bitwise-ior X_OK W_OK))
        (let ()
          (define hfile (string-append proverb-base "/history"))
          (define writer (lambda (f)
            (for-each (lambda (v)
              (write v f) (newline f)
            ) elements)
            (close f)
            (make-text "OK. history saved.")
          ))
          (if (access? hfile W_OK)
            (let ()
              (define f (open-file hfile "a"))
              (writer f)
            )
            (if (access? hfile F_OK)
              (make-text "history file exists but is not writeable")
              (let ()
                (define f (open-file hfile "w"))
                (writer f)
              )
            )
          ) 
        )
        (let ()
          (and
            (mkdir proverb-base)
            (save-history elements)
          )
        )
      )
    ))
    (widgetscript-apply-with-attrs '(hidden "true") "function() {
      var l='(';
      var x;
      for(x=0;x<shell.input_element.jsshell_history.length;x++) {
        var e=shell.input_element.jsshell_history[x];
        if(e.session == shell.session_id && !e.saved) {
          l+=' ('+scheme_stringify(e.value)+') ';
          e.saved = true;
        }
      }
      l+=')';
      widget.shell.string_evaluator('] ' + widget.wrapper('(save-history (quote ' + l + '))'), false);
    }" '())
  )
))

(define xhtml-auto-save-history (with-docstring
  "automatically save the non-persistient elements of the browser command line history to ~/.proverb/history every MILLISECONDS milliseconds"
  (lambda (milliseconds)
    (define proverb-base (string-append (getenv "HOME") "/.proverb"))
    (define save-history (lambda (elements)
      (if (access? proverb-base (bitwise-ior X_OK W_OK))
        (let ()
          (define hfile (string-append proverb-base "/history"))
          (define writer (lambda (f)
            (for-each (lambda (v)
              (write v f) (newline f)
            ) elements)
            (close f)
            (make-text "OK. history saved.")
          ))
          (if (access? hfile W_OK)
            (let ()
              (define f (open-file hfile "a"))
              (writer f)
            )
            (if (access? hfile F_OK)
              (make-text "history file exists but is not writeable")
              (let ()
                (define f (open-file hfile "w"))
                (writer f)
              )
            )
          ) 
        )
        (let ()
          (and
            (mkdir proverb-base)
            (save-history elements)
          )
        )
      )
    ))
    (widgetscript-apply-with-attrs '(hidden "true") "function(ms) {
      var round=0;
      var saver=function() {
        var l='(';
        var x;
        for(x=0;x<shell.input_element.jsshell_history.length;x++) {
          var e=shell.input_element.jsshell_history[x];
          if(e.session == shell.session_id && !e.saved) {
            l+=' ('+scheme_stringify(e.value)+') ';
            e.saved=true;
          }
        }
        l+=')';
        shell.inner_evaluator.eval_to_string('] ' + widget.wrapper('(begin (save-history (quote ' + l + ')))'), widget.shell,function (v) {
          if(ms) {
            setTimeout(saver,ms);
          }
        });
      }
      saver();
    }" (list milliseconds))
  )
))

; unlike history, preferences changes are posted back to Proverb
; immediately when the background widget is started. No need to fish them
; out of the browser at the time of this call.
(define save-prefs (with-docstring
  "save the non-persistient elements of the browser jsshell preferences to ~/.proverb/jsprefs. returns #t or a string with an error message"
  (lambda ()
    (define proverb-base (string-append (getenv "HOME") "/.proverb"))
    (define inner-save-prefs (lambda (elements)
      (if (access? proverb-base (bitwise-ior X_OK W_OK))
        (let ()
          (define hfile (string-append proverb-base "/jsprefs"))
          (define writer (lambda (f)
            (for-each (lambda (v)
              (write v f) (newline f)
            ) elements)
            (close f)
            #t
          ))
          (if (access? hfile W_OK)
            (let ()
              (define f (open-file hfile "a"))
              (writer f)
            )
            (if (access? hfile F_OK)
              (make-text "history file exists but is not writeable")
              (let ()
                (define f (open-file hfile "w"))
                (writer f)
              )
            )
          ) 
        )
        (and
          (mkdir proverb-base)
          (inner-save-prefs elements)
        )
      )
    ))
    (inner-save-prefs saved-attrs)
  )
))

;
;
;
(define xhtml-save-prefs (with-docstring
  "save the non-persistient elements of the browser jsshell preferences and return a text node"
  (lambda ()
    (define r (save-prefs))
    (if (string? r)
      (make-text r)
      (if r
        (make-text "OK. preferences saved.")
        (make-text "Preferences not saved.")
      )
    )
  )
))

(define xhtml-make-terminal-port (with-docstring
  "you really don't want this function"
  (lambda ()
    #f
  )
))

(define xhtml-popen (with-docstring
  "open the command command-str as in inferior subprocess with standard I/O redirected to a web terminal"
  (lambda (command-str)
  (define p #f)
  (define r #f)
  (define send (lambda (s)
    (if (string? s)
      (if (and p (output-port? (cdr p)))
        (if (port-closed? (cdr p))
          "port-closed"
          (catch #t (lambda ()
            (r 'i 'output s)
            (display s (cdr p))
            (force-output (cdr p))
            #t
          ) (lambda args
            (object->string args)
          ))
        )
        (string-append "not-a-port: " (object->string p))
      )
      #f
    )
  ))
  (define bogobogo (lambda ()
    (define psub (popen-fports command-str))
    (define rsub (snatch-embeddable-responder))
    (set! r rsub)
    (set! p psub)
    (r 'i 'ready)
    (call-with-new-thread (lambda ()
      (let helper ((racc '()))
        (define c (char-ready? (car p)))
        (if (or c (null? racc))
          (let ()
            (define v (read-char (car p)))
            (if (eof-object? v)
              (begin
                (unless (null? racc)
                  (r 'i 'output (list->string (reverse racc)))
                )
                (r 'r "DONE")
              )
              (helper (cons v racc))
            )
          )
          (begin
            (r 'i 'output (list->string (reverse racc)))
            (helper '())
          )
        )
      )
    ))
  ))

  (widgetscript-apply "
    function(send) {
    var foo;
    var bar;
    var drawfunc=function() {
      var ready=false;
      foo.id='vt100';
      bar=new VT100(foo);
      bar.crLfMode=true;
      bar.keysPressed=function(ch) {
        if(ready) {
          send(function(v) {
            alert('bogus send returns: ' + v);
          },ch);
          widget.shell.inner_evaluator.eval_to_js('] ' + widget.wrapper('(send ' + scheme_stringify(ch) + ')'),widget.shell,function(v) {
            if(v != true) {
              alert('bad response: ' + v);
              alert('] ' + widget.wrapper('(send ' + scheme_stringify(ch) + ')'));
            }
          });
        } else {
          alert('other end not ready yet');
        }
      };
      widget.shell.inner_evaluator.eval_to_js('] ' + widget.wrapper('(bogobogo)'),widget.shell,function(v) {
        if(v.type == 'i') {
          var itype=dump_node_to_string(v.rest[0]);
          if(itype == 'output') {
            v.rest.shift();
            bar.vt100(dump_node_to_string(v.rest[0]));
          } else if(itype == 'ready') {
            ready=true;
          } else {
            alert('non-output I message: ' + itype);
          }
        }
      });
    };
    foo=document.createElement('div');
    foo.style.height='25em';
    /*foo.style.width='40em';*/
    foo.style.borderStyle='dotted';
    widget.div.appendChild(foo);
      var drawn=false;
      var handler=function(e) {
        if(!drawn && widget.div.parentNode) {
          drawn=true;
	  document.body.removeEventListener('DOMSubtreeModified',handler,false);
          drawfunc();
        }
      };
      document.body.addEventListener('DOMSubtreeModified',handler,false);
  }
  " (list (with-xml-tag `(ref key ,(xhtml-environment-pickler #f (list send))))))
  )
))

(define wikipoint (with-docstring
  "accept wiki text <article-body> and render it in a split screen view that holds
text with inline examples and the results of evaluations side-by-side"
  (lambda (article-body)
    (widgetscript-apply "
      function(body) {

	var h = document.body.clientHeight-100;

	var sub = document.createElement('div');
	sub.style.position = 'relative';
	sub.style.height = h + 'px';
	widget.div.appendChild(sub);
	
	var wikidiv = document.createElement('div');
	wikidiv.style.position = 'absolute';
	wikidiv.style.width = '40%';
	wikidiv.style.maxWidth = '40%';
	wikidiv.style.top = '0%';
	wikidiv.style.left = '0%';
	wikidiv.style.maxHeight = h + 'px';
	wikidiv.style.overflow = 'scroll';
	sub.appendChild(wikidiv);
	
	var evaldiv=document.createElement('div');
	evaldiv.style.position = 'absolute';
	evaldiv.style.width = '40%';
	evaldiv.style.maxWidth = '60%';
	evaldiv.style.top = '0%';
	evaldiv.style.left = '40%';
	evaldiv.style.maxHeight = h + 'px';
	evaldiv.style.overflow = 'scroll';
	sub.appendChild(evaldiv);

	var subshell = widget.shell.subshell(function(v) {
  	evaldiv.appendChild(v);
	  evaldiv.appendChild(document.createElement('br'));
	});

	wikidiv.appendChild(gin_up_wiki_div(body,subshell));

      }
    " (list article-body))
  )
))

(define make-preformatted-text (with-docstring
  "accept a string of pre-formatted plain text and generate tags that,
when displayed, are recognizable as a single block of text whose formatting
is intentional. Presentation may be controlled ultimately by user preference."
  (lambda (string)
    (with-xml-tag '(div class "wiki_literal") (make-text string))
  )
))

(define make-bookmarklet (with-docstring
  "emit the location string for a bookmarklet that points at the current Proverb"
  (lambda (name . extra)
    (define r (snatch-responder))
    (define func (lambda (url)
      (critmsg "URL: ~a\n" url)
      (r 'r (with-xml-tag `(a href ,(string-append
        "javascript:"
        (url-encode-string (string-append
          "{"
          (if (null? extra)
            ""
            (string-append "jsshell_bookmarklet_extra=["
            (apply string-append (map (lambda (v)
              (string-append "[\"" (car v) "\",\"" (cadr v) "\"],")
            ) extra))
            "];") 
          )
        
          ;"var foo=document.createElement('script'); foo.src='"
          ;(http-server-get-url (current-http-server))
          ;"/jsshell/jsshell-bookmarklet.js"
          ;"'; document.body.appendChild(foo);void(0);}"

          "var foo=document.createElement('script'); foo.src='"
          url "/jsshell-bookmarklet.js"
          "'; document.body.appendChild(foo);void(0);}"

        )))) (if (string? name) (make-text name) name)))
    ))

    (r 'i 'dribble (make-widgetscript 'hidden 'true
      "widget.shell.inner_evaluator.eval_to_string(
        widget.wrapper('(func ' + scheme_stringify(document.URL.replace('/jsshell.html','').replace(/\\/\\/.*:.*@/,'//')) + ')'),
        widget.shell,
        function(v) {
          true;
        }
      )"
    ))
  )
))

)
