#!/bin/bash
exec guile -e client-main -s $0 "$@"
!#

;
; This code is the kernel of next-generation lash
;

(use-modules (ice-9 session))
(use-modules (ice-9 syncase))
(use-modules (ice-9 documentation))
(use-modules (ice-9 rdelim))
(use-modules (ice-9 regex))
(use-modules (ice-9 rw))
(use-modules (ice-9 format))

(debug-enable 'debug)

;(debug-set! stack 2000000)

(load "proverb-web-common.scm")

(trace-args! select)

(define ctr 0)
(define name 'foo)
(define gen-tag (lambda ()
  (define tag (string->symbol (format #f "~a-~a-~a" (getpid) name ctr)))
  (set! ctr (+ ctr 1))
  tag
))

(define bogo-lock #f)
(define-macro (with-lock lock body)
  `(if (not ,lock) (error-preserving-wrap (lambda () (set! ,lock #t)) (lambda () ,body) (lambda ()  (set! ,lock #f))) (display "LOCK HELD!\n"))
)

(define setup-ioforward (lambda (local-port mux remote-cookie)
  (define chan (tire-mux-lookup-cookie mux remote-cookie))
  (define tag (gen-tag))
  (define from (make-tire-procedure-channel (lambda (form)
    (case (car form)
      ((r)
        (critmsg "RESPONSE!\n")
        (close local-port)
      )
      ((i)
        (if (eq? (list-ref form 2) 'stream)
          (begin
            (display (list-ref form 3) local-port)
          )
        )
      )
      (else (error "BAD FORM"))
    )
  )))
  (critmsg "SETUP IOFORWARD TAG: ~a\n" tag)
  (let ()
    (define subform `(t ,tag ,remote-cookie read-stream))
    (tire-mux-add-request! mux (cons (list-ref subform 1) (cons from (lambda ()
      (display "SETUP IOFORWARD CONTINUATION TAG: ") (display tag) (display "\n")
    ))))
    (channel-enqueue-sync chan mux subform)
  )
))

(define foo-tire (lambda (mux cookie form)
  (define chan (tire-mux-lookup-cookie mux cookie))
  (define tag (gen-tag))

  (let ()
    (define responded #f)
    (define bogo-select-post (lambda (state rfds wfds xfds)
      (with-lock bogo-lock (let ()
        (define retval (tire-mux-select-post state rfds wfds xfds))
        ; stop once we got our answer
        (if responded
          #f
          retval
        )
      ))
    ))
    (define rval #f)
    (define subform `(t ,tag ,cookie ,@form))
    (define from (make-tire-procedure-channel (lambda (form)
      (case (car form)
        ((r)
          (format #t "R: ~a\n" form)
          (set! rval (list-ref form 2))
        )
        ((i)
          (if (eq? (list-ref form 2) 'stream)
            (display (list-ref form 3) (current-output-port))
            (format #t "I: ~a\n" form)
          )
        )
        ((t)
          (display "T!\n")
        )
        ((e)
          (display "E!\n")
        )
        (else
          (error "WHAT THE HECK!\n")
        )
      )
    )))

    (tire-mux-add-request! mux (cons (list-ref subform 1) (cons from (lambda () (set! responded #t) (display "FOO!\n")))))
    (channel-enqueue-sync chan mux subform)
    (critmsg "BAR!\n")
    (run-select-loop (list (list mux tire-mux-select-pre bogo-select-post)) (lambda () (not responded) ))

    (critmsg "BAZ!\n")
    rval
  )
))

(define foo-eval (lambda (mux cookie form)
  (foo-tire mux cookie `(eval ,form))
))

;
; call as clash.scm URL
;
(define client-main (lambda (args)
  (define url (cadr args))
  (define mux (tire-mux-setup '()))

  (define bogo-select-pre (lambda args
    ;(display "TIRE MUX SELECT PRE: ") (display args) (newline)
    (apply tire-mux-select-pre args)
  ))

  (let ()
    (define c (make-tire-url-channel url))
    (define p (tire-parse-url url))
    (display "PARSED URL: ") (display p) (newline)
    (if c
      (let ()
        (define cookie (list-ref p 3))
        (tire-mux-add-channel! mux c)
        (tire-mux-hash-cookie! mux cookie c)

	(display (foo-eval mux cookie '(getpid))) (newline)

        ;(let ()
        ;  (define stdout-cookie (foo-eval mux cookie '(port->cookie (open-input-file "/etc/passwd"))))
        ;  (display "STDOUT-COOKIE: ") (display stdout-cookie) (newline)
        ;  (foo-tire mux stdout-cookie '(read-stream))
        ;)

        (let ()
          (define stdout-cookie (foo-eval mux cookie '(port->cookie (open-input-file "/etc/passwd"))))
          (define issue-cookie (foo-eval mux cookie '(port->cookie (open-input-file "/etc/passwd"))))
          (define outstanding-ports 2)
          (display "FORWARDING STDOUT TO /tmp/bogobogo\n")
          (display "FORWARDING ISSUE TO /tmp/barbar\n")
          (setup-ioforward (open-output-file "/tmp/bogobogo") mux stdout-cookie)
          (setup-ioforward (open-output-file "/tmp/barbar") mux issue-cookie)

          ;(run-select-loop (list (list mux tire-mux-select-pre tire-mux-select-post)) (lambda () (display "STILLHERE: ") (display (tire-mux-get-requests mux)) (display "\n") #t))
          (run-select-loop (list (list mux bogo-select-pre tire-mux-select-post)) (lambda () #t))

        )

      )
    )
  )

  (display "SHUTTING DOWN!\n")

))

