
function evaluator_unix_shell() {

  // See section 2.10 of the POSIX report

  // some token names here are derived from that language

  var expr_table = {

    /* operators */

    OR_IF: { pattern: /^(\|\|)(.*)/
    },
    AND_IF: { pattern: /^(&&)(.*)/
    },
    DSEMI: { pattern: /^(;;)(.*)/
    },

    DLESS: { pattern: /^(<<)(.*)/
    },
    DGREAT: { pattern: /^(>>)(.*)/
    },
    LESSAND: { pattern: /^(<&)(.*)/
    },
    GREATAND: { pattern: /^(>&)(.*)/
    },
    LESSGREAT: { pattern: /^(<>)(.*)/
    },
    DLESSDASH: { pattern: /^(<<-)(.*)/
    },

    CLOBBER: { pattern: /^(>\|)(.*)/
    },

    /* reserved words */
    IF: { pattern: /^(if)(.*)/
    },
    THEN: { pattern: /^(then)(.*)/
    },
    ELSE: { pattern: /^(else)(.*)/
    },
    ELIF: { pattern: /^(elif)(.*)/
    },
    FI: { pattern: /^(fi)(.*)/
    },
    DO: { pattern: /^(do)(.*)/
    },
    DONE: { pattern: /^(done)(.*)/
    },

    CASE: { pattern: /^(case)(.*)/
    },
    ESAC: { pattern: /^(esac)(.*)/
    },
    WHILE: { pattern: /^(while)(.*)/
    },
    UNTIL: { pattern: /^(until)(.*)/
    },
    FOR: { pattern: /^(for)(.*)/
    },

    /* these are also 'reserved words' */

    LBRACE: { pattern: /^({)(.*)/
    },
    RBRACE: { pattern: /^(})(.*)/
    },
    BANG: { pattern: /^(!)(.*)/
    },

    IN: { pattern: /^(in)(.*)/
    },

    /* dsridge added */
    PIPE: { pattern: /^(\|)(.*)/
    },
    LPAREN: { pattern: /^(\()(.*)/
    },
    RPAREN: { pattern: /^(\))(.*)/
    },
    AMP: { pattern: /^(&)(.*)/
    },
    SEMI: { pattern: /^(;)(.*)/
    },
    LESS: { pattern: /^(<)(.*)/
    },
    GREAT: { pattern: /^(>)(.*)/
    },
    DOLLAR: { pattern: /^(\$)(.*)/
    },

    simple_WORD: { pattern: /^([-A-Za-z]+)(.*)/,
            consumes: true
    },
    DQUOT: { pattern: /^(")(.*)/, 
    },
    squot_str: { pattern: /^('[^']*')(.*)/,
                 consumes: true
    },
    WORD: {
      rules: [
        [ "simple_WORD" ],
        [ "DOLLAR", "simple_WORD" ],
        [ "DOLLAR", "LBRACE", "simple_WORD", "RBRACE" ],
        [ "DQUOT", "simple_WORD", "DQUOT" ],
        [ "squot_str" ],
      ]
    },

    ASSIGNMENT_WORD: { pattern: /^([A-Za-z]+=[A-Za-z]*)(.*)/,
      consumes: true
    },
    IO_NUMBER: { pattern: /^([0-9]+)(.*)/,
      consumes: true
    },
    NEWLINE: { pattern: /^(\n)(.*)/
    },
    
    NAME: {
      rules: [
        [ "simple_WORD" ]
      ]
    },

    complete_command: {
      rules: [
        [ "list" , "spaces", "separator" ],
        [ "list" ]
      ],
      //debug: true
    },
    list: {
      rules: [
        [ "and_or", "spaces", "separator_op", "spaces", "list" ],
        [ "and_or" ],
      ],
      //debug: true
    },
    and_or: {
      rules: [
        [ "pipeline", "spaces", "AND_IF", "spaces", "linebreak", "spaces", "and_or" ],
        [ "pipeline", "spaces", "OR_IF", "spaces", "linebreak", "spaces", "and_or" ],
        [ "pipeline" ],
      ],
      //debug: true
    },
    pipeline: {
      rules: [
        [ "pipe_sequence" ], 
        [ "BANG" , "spaces", "pipe_sequence" ],
      ],
      //debug: true
    },
    pipe_sequence: {
      rules: [
        [ "command", "spaces", "PIPE", "spaces", "linebreak", "spaces", "pipe_sequence" ],
        [ "command" ],
      ],
      //debug: true
    },
    command: {
      rules: [
        [ "simple_command" ],
        [ "compound_command", "spaces", "redirect_list" ],
        [ "compound_command" ],
        [ "function_definition" ],
      ],
      //debug: true
    },
    compound_command: {
      rules: [
        [ "brace_group" ],
        [ "subshell" ],
        [ "for_clause" ],
        [ "case_clause" ],
        [ "if_clause" ],
        [ "while_clause" ],
        [ "until_clause" ] 
      ]
    },
    subshell: {
      rules: [ 
        [ "LPAREN", "spaces", "compound_list", "spaces", "RPAREN" ]
      ]
    },
    compound_list: {
      rules: [
        [ "term" ],
        [ "newline_list", "spaces", "term", "spaces", "separator" ],
        [ "newline_list", "spaces", "term" ],
        [ "term" , "spaces", "separator" ],
      ]
    },
    term: {
      rules: [
        [ "and_or", "spaces", "separator", "spaces", "term" ],
        [ "and_or" ],
      ]
    },
    for_clause: {
      rules: [
        [ "FOR", "spaces", "NAME", "spaces", "linebreak", "spaces", "do_group" ],
        [ "FOR", "spaces", "NAME", "spaces", "linebreak", "spaces", "IN", "spaces", "sequential_sep", "spaces", "do_group"],
        [ "FOR", "spaces", "NAME", "spaces", "linebreak", "spaces", "IN", "spaces", "wordlist", "spaces", "sequential_sep", "spaces", "do_group"],
      ]
    },
    wordlist: {
      rules: [
        [ "WORD", "spaces", "wordlist" ],
        [ "WORD" ],
      ]
    },
    case_clause: {
      rules: [
        [ "CASE", "spaces", "WORD", "spaces", "linebreak", "spaces", "IN", "spaces", "linebreak", "spaces", "case_list", "spaces", "ESAC" ],
        [ "CASE", "spaces", "WORD", "spaces", "linebreak", "spaces", "IN", "spaces", "linebreak", "spaces", "case_list_ns", "spaces", "ESAC" ],
        [ "CASE", "spaces", "WORD", "spaces", "linebreak", "spaces", "IN", "spaces", "linebreak", "spaces", "ESAC" ]
      ]
    },
    case_list_ns: {
      rules: [
        [ "case_item_ns", "spaces", "case_list" ], 
        [ "case_item_ns" ]
      ]
    }, 
    case_list: {
      rules: [
        [ "case_item", "spaces", "case_list" ], 
        [ "case_item" ]
      ]
    },
    case_item_ns: {
      rules: [
        [ "pattern", "spaces", "RPAREN", "spaces", "linebreak" ],
        [ "pattern", "spaces", "RPAREN", "spaces", "compound_list", "spaces", "linebreak" ],
        [ "LPAREN", "spaces", "pattern", "spaces", "RPAREN", "spaces", "linebreak" ],
        [ "LPAREN", "spaces", "pattern", "spaces", "RPAREN", "spaces", "compound_list", "spaces", "linebreak" ],
      ]
    },
    case_item: {
      rules: [
        [ "pattern", "spaces", "RPAREN", "spaces", "linebreak", "spaces", "DSEMI", "spaces", "linebreak" ],
        [ "pattern", "spaces", "RPAREN", "spaces", "compound_list", "spaces", "DSEMI", "spaces", "linebreak" ],
        [ "LPAREN", "spaces", "pattern", "spaces", "RPAREN", "spaces", "linebreak", "spaces", "DSEMI", "spaces", "linebreak" ],
        [ "LPAREN", "spaces", "pattern", "spaces", "RPAREN", "spaces", "compound_list", "spaces", "DSEMI", "spaces", "linebreak" ],
      ]
    },
    pattern: {
      rules: [
        [ "WORD", "spaces", "PIPE", "spaces", "pattern" ],
        [ "WORD" ],
      ]
    },
    if_clause: {
      rules: [
        [ "IF", "spaces", "compound_list", "spaces", "THEN", "spaces", "compound_list", "spaces", "else_part", "spaces", "FI" ],
        [ "IF", "spaces", "compound_list", "spaces", "THEN", "spaces", "compound_list", "spaces", "FI" ],
      ]
    },
    else_part: {
      rules: [
        [ "ELIF", "spaces", "compound_list", "spaces", "THEN", "spaces", "else_part" ],
        [ "ELSE", "spaces", "compound_list" ],
      ]
    },
    while_clause: {
      rules: [
        [ "WHILE", "spaces", "compound_list", "spaces", "do_group" ]
      ]
    },
    until_clause: {
      rules: [
        [ "UNTIL", "spaces", "compound_list", "spaces", "do_group" ]
      ]
    },
    function_definition: {
      rules: [
        [ "fname", "spaces", "LPAREN", "spaces", "RPAREN", "spaces", "linebreak", "spaces", "function_body" ]
      ]
    }, 
    function_body: {
      rules: [
        [ "compound_command", "spaces", "redirect_list" ],
        [ "compound_command" ],
      ]
    },
    fname: {
      rules: [
        [ "NAME" ]
      ]
    },
    brace_group: {
      rules: [
        [ "LBRACE", "spaces", "compound_list", "spaces", "RBRACE" ]
      ]
    },
    do_group: {
      rules: [
        [ "DO", "spaces", "compound_list", "spaces", "DONE" ]
      ]
    },
    spaces: {
      pattern: /([ ]*)(.*)/,
      silent: true
    },
/*
    spaces: {
      rules: [
        [ "space", "spaces" ],
        []
      ]
    },
*/
    simple_command: {
      rules: [
        [ "cmd_prefix", "spaces", "cmd_word", "spaces", "cmd_suffix" ],
        [ "cmd_prefix", "spaces", "cmd_word" ],
        [ "cmd_prefix" ],
        [ "cmd_name" , "spaces", "cmd_suffix" ],
        [ "cmd_name" ]
      ],
      //debug: true
    },
    cmd_name: {
      rules: [
        [ "WORD" ]
      ],
      //debug: true
    },
    cmd_word: {
      rules: [
        [ "WORD" ]
      ]
    },
    cmd_prefix: {
      rules: [
        [ "io_redirect", "spaces", "cmd_prefix" ],
        [ "io_redirect" ],
        [ "ASSIGNMENT_WORD", "spaces", "cmd_prefix" ],
        [ "ASSIGNMENT_WORD" ] 
      ],
      //debug: true
    },
    cmd_suffix: {
      rules: [
        [ "io_redirect", "spaces", "cmd_suffix" ],
        [ "io_redirect" ],
        [ "WORD", "spaces", "cmd_suffix" ],
        [ "WORD" ]
      ]
    },
    redirect_list: {
      rules: [
        [ "io_redirect", "spaces", "redirect_list" ],
        [ "io_redirect" ]
      ] 
    },
    io_redirect: {
      rules: [
        [ "io_file" ],
        [ "IO_NUMBER", "spaces", "io_file" ],
        [ "io_here" ],
        [ "IO_NUMBER", "spaces", "io_here" ]
      ]
    },
    io_file: {
      rules: [
        [ "LESS", "spaces", "filename" ],
        [ "LESSAND", "spaces", "filename" ],
        [ "GREAT", "spaces", "filename" ],
        [ "GREATAND", "spaces", "filename" ],
        [ "DGREAT", "spaces", "filename" ],
        [ "LESSGREAT", "spaces", "filename" ],
        [ "CLOBBER", "spaces", "filename" ]
      ]
    },
    filename: {
      rules: [
        [ "WORD" ]
      ]
    },
    io_here: {
      rules: [
        [ "DLESS", "spaces", "here_end" ],
        [ "DLESSDASH", "spaces", "here_end" ]
      ]
    },
    here_end: {
      rules: [
        [ "WORD" ] 
      ]
    }, 
    newline_list: {
      rules: [
        [ "NEWLINE", "spaces", "newline_list" ],
        [ "NEWLINE" ]
      ], 
    },
    linebreak: {
      rules: [
        [ "newline_list" ],
        []
      ]
    },
    separator_op: {
      rules: [
        [ "AMP" ],
        [ "SEMI" ]
      ]
    },
    separator: {
      rules: [
        [ "separator_op", "spaces", "linebreak" ],
        [ "newline_list" ]
      ]
    }, 
    sequential_sep: {
      rules: [
        [ "SEMI", "spaces", "linebreak" ],
        [ "newline_list" ]
      ]
    }
  };

  var eat_table=function(state,table) {
    var incomplete=false;
    var chase_rule = function(rule) {

      if(state[rule] && state[rule].chased==true) {
        return;
      }

      //document.body.appendChild(document.createTextNode('chasing ' + rule));
      //document.body.appendChild(document.createElement('br'));

      var r;
      r=table[rule];
      var c;
      c=new Object;
      state[rule]=c;
      c.patterns=new Array;
      c.root_patterns=new Object;
      c.root_rules=new Object;

      if(r.pattern) {
        /* this is a pattern rule, and so is terminal */
        c.patterns.push(r.pattern);
        c.root_rules[rule]=true; // self
        //document.body.appendChild(document.createTextNode('marking chased ' + rule));
        //document.body.appendChild(document.createElement('br'));
        c.chased=true; 
      } else if(r.rules) {
        var bogo=false;
        var y;
        for(y=0;y<r.rules.length;y++) {
          if(r.rules[y].length == 0) {
            /* this rule admits a null match */
            c.nullmatch=true;
          } else {
            var n=r.rules[y][0]; 
            if(!c.root_rules[n]) {
              if(state[n]) {
                if(state[n].chased) {
                } else {
                  /* we are in the middle of chasing this rule already */
                  incomplete=["found", n, "during", rule];
                  bogo=true;
                }
              } else {
                chase_rule(n);
              }
              if(state[n].chased) {
                var z;
                for(z in state[n].root_rules) {
                  c.root_rules[z]=true;
                }
              }
            }
          }    
        }
        if(!bogo) {
          //document.body.appendChild(document.createTextNode('marking chased ' + rule));
          //document.body.appendChild(document.createElement('br'));
          c.chased=true;
        }
      } else {
        alert('bad rule ' + rule);
        c.chased=true;
      }
    };
    var x;
    var pattern_rules;
    var total_rules;
    pattern_rules=0;
    total_rules=0;
    for(x in table) {
      total_rules++;
      if(table[x].pattern) {
        pattern_rules++;
      }
      chase_rule(x);
    }
    //alert('total rules: ' + total_rules + 'pattern rules: ' + pattern_rules + ' incomplete: ' + incomplete);
  }

  var fastcheck=function(state,str,table,rule) {
    var x;
    var f;
    f=state.fastpath[rule];
    if(f.chased) {
      if(f.nullmatch) {
        return true;
      } else {
        var x;
        for(x in f.root_rules) {
          if(str.match(table[x].pattern)) {
            return true;
          }
        }
        return false;    
      }
    } else {
      /* we haven't chased this rule so return a possible match */
      return true;
    }
  }

  var match_rule=function(state,str,table,rule,pass,fail) {
    //alert('matching: ' + str + ' (' + typeof(str) + ') against ' + rule);
    var k;
    var m;
    var real_rule=rule;
    var real_str=str;

    var fastcheck_res=fastcheck(state,str,table,rule);

    if(!fastcheck_res) {
      if(state.use_fastcheck) {
        return fail(state);
      }
    }

    if(state.test_fastcheck) {
      var old_pass=pass;
      var old_fail=fail;

      pass = function(state,production,remaining_str) {
        if(fastcheck_res) {
          if(!state.fastcheck_true_positive) {
            state.fastcheck_true_positive=1;
          } else {
            state.fastcheck_true_positive++;
          }
        } else {
          alert('fastcheck false negative on rule ' + rule + ' with str ' + real_str);
        }
        return old_pass(state,production,remaining_str);
      };

      fail = function(state) {
        if(fastcheck_res) {
          if(!state.fastcheck_false_positive) {
            state.fastcheck_false_positive=1;
          } else {
            state.fastcheck_false_positive++;
          }
        } else {
          if(!state.fastcheck_true_negative) {
            state.fastcheck_true_negative=1;
          } else {
            state.fastcheck_true_negative++;
          }
        }
        return old_fail(state);
      };
    }

    var foo=function(rule,match) {
      if(match.length==3) {
        var prodname;
        if(rule.production) {
          prodname=rule.production;
        } else {
          prodname=k;
        }
/*
        if(rule.next) {
          return rule.next(state,match[2],pass);
        } else {
*/
          if(rule.consumes) {
            return pass(state,[prodname,match[1]],match[2]);
          } else {
            return pass(state,[prodname,true],match[2]);
          }
/*
        }
*/
      } else {
        alert('strange match on ' + rule);
        return;
      }
    };

    if(typeof(rule) == "string") {
      k=rule;
      var e=table[rule];
      if(!e) {
        alert('no entry for rule: ' + rule);
      } else {
        if(e.debug) {
          if(!confirm('entering rule: ' + rule + ' with an input string of length: ' + str.length)) {
            return;
          }
        }
      }

      if(e.rules) {
        var outer_rule,inner_rule;
        var original_str=str;
        var flatten_racc=function(racc) {
          var a=[];
          var cursor=racc;
          while(cursor.length > 0) {
            a.unshift(cursor[0]);
            cursor=cursor[1];
          }
          return a;
        }
        var rule_checker=function(state,str,outer_index,inner_index,racc) {
          if((!e.consolidated) && e.rules.length > 1) {
            //if(e.rules[0][0] == e.rules[1][0]) {
            //  alert('possible consolidation opportunity with ' + rule + ' -- ' + e.rules);
            //}
            e.consolidated=true;  
          }
          if(outer_index < e.rules.length) {
            // we have rules left
            if(inner_index<e.rules[outer_index].length) {
              // we have stages left in this rule ...
              return match_rule(state,str,table,e.rules[outer_index][inner_index],function(state,production,str) {
                // ... and we matched the next one
                // Damned recursion limit...
                if(Math.random() < .1) {
                  setTimeout(function() {
                    rule_checker(state,str,outer_index,inner_index+1,[production,racc]);
                  },0);
                  return false;
                } else {
                  return rule_checker(state,str,outer_index,inner_index+1,[production,racc]);
                }
              },function() {
                // ... and we didn't match the next one. advance to the next rule
                return rule_checker(state,original_str,outer_index+1,0,[]);
              });
            } else {
              // we have no stages left in this rule and so have matched it!
              if(e.silent) {
                return pass(state,flatten_racc(racc)[0],str);
              } else {
                return pass(state,[rule,flatten_racc(racc)],str);
              }
            }
          } else {
            // our list of rules is empty. We obviously didn't match anything.
            return fail(state);
          }
        };
        return rule_checker(state,str,0,0,[]);
      } else if(e.patterns) {
        var l;
        for(l=0;l<e.patterns.length;l++) {
          if(m=str.match(e.patterns[l])) {
            return foo(e,m);
          }
        }
      } else if(m=str.match(e.pattern)) {
        return foo(e,m);
      }
    } else {
      alert('non-string rule!');
    }
    return fail(state);
  };

  this.language='sh';
  this.prompt=function(sh) {
    return document.createTextNode(this.language+'>');
  };
  var bogo_dumper=function(res,preamble) {
    var d=document.createElement('div');
    var x;
    d.appendChild(document.createTextNode(preamble+'Production: ' + res[0]));
    d.appendChild(document.createElement('br'));
    if(typeof(res[1]) == "string") {
      d.appendChild(document.createTextNode(preamble+'produced: ' + res[1]));
    } else {
      if(res.length > 0 && res[1].length > 0) {
        for(x=0;x<res[1].length;x++) {
          //d.appendChild(document.createTextNode(preamble+'something here: ' + res[1][x]));
          //d.appendChild(document.createElement('br'));
          d.appendChild(bogo_dumper(res[1][x],preamble+'..'));
        }
      }
      //if(racc.length > 1) {
      //  d.appendChild(bogo_dumper(racc[1]));
      //}
    }
    return d;
  };
  this.evaluator=function(expression,document_context,shell,continuation) {
    this.eval_to_js(expression,shell,function(r) {
      var flatten_blob_list = function(bl) {
        var rv=[];
        var cursor=bl;
        while(cursor) {
          if(cursor[0] == "blob_list") {
            rv.push(cursor[1][0]);
            cursor=cursor[1][1];
          } else {
            break;
          }
        }
        return ["flat_blob_list",rv];
      };
      var t=document_context.createTextNode(r+'');
      var d=document.createElement('div');
      //d.appendChild(bogo_dumper(r,''));
      //d.appendChild(document.createElement('br'));
      if(r[0] == "complete_command") {
        if(r[1][0][0] == "blob_list") {
         // d.appendChild(document.createTextNode('hey! got a blob list!' + flatten_blob_list(r[1][0])));
          d.appendChild(bogo_dumper(["entry",[flatten_blob_list(r[1][0])]],''));
          d.appendChild(document.createElement('br'));
        } else {
          d.appendChild(bogo_dumper(r,''));
          d.appendChild(document.createElement('br'));
        }
      }
      continuation(d);
    });
  };
  this.tabcompleter=function(string,shell,continuation) {
    return false;
  };
  this.eval_to_string = function(string,shell,continuation) {
    return this.eval_to_js(string,shell,function(v) {
      var dump_production = function(p) {
        return 'production name: ' + p[0] + ' -- kjshfkjhds ' ;
      };
      continuation(dump_production(v));
    });
  };
  var done=false;
  var accelerated_state=false;
  this.eval_to_js = function(string,shell,continuation) {
    if(!done) { 
      done=true;
      accelerated_state=new Object;
      eat_table(accelerated_state,expr_table);
/*
      var foo;
      for(foo in accelerated_state['complete_command']['root_rules']) {
        alert('foo: ' + foo);
      }
*/
    }
  
    var state=new Object;
    state.vars=new Object;
    state.vars["VAR_PATH"]="/bin:/usr/bin";
    state.fastpath=accelerated_state;
    state.use_fastcheck=true;
    state.test_fastcheck=false;
   
    return match_rule(state,string,expr_table,'complete_command',function(state,production,str) {
      //alert('production: ' + production);
      //alert('true_positive: ' + state.fastcheck_true_positive + ' false_positive: ' + state.fastcheck_false_positive + ' true negative: ' + state.fastcheck_true_negative);
      continuation(production);
    },function() {
      continuation('NOT OK: ' + arguments);
    });
  };
};

hybrid_evaluators['='] = new evaluator_unix_shell;
