/*
 * Readline-like functions for the javascript shell
 */

var jsshell_readline = {
  // this array currently gets populated by jsshell-site.js for RPC jsshells.
  default_history: [],

  make_history_element: function(string,session,time) {
    var elem = {
      value: string,
      session: session,
      time: time
    };
    return elem;
  },

  cursor_up: function(object) {
    if(object.jsshell_history.length != 0) {
      if(object.jsshell_history_cursor) {
        if(object.jsshell_history_cursor > 0) {
          object.jsshell_history_cursor--;
        }
      } else {
        if(object.jsshell_history_cursor == 0) {
          //alert('end of history');
        } else {
          object.jsshell_history_cursor=object.jsshell_history.length - 1;
        }
      }
      object.value=(object.jsshell_history[object.jsshell_history_cursor]).value;
    } else {
      // the history is empty
    }
  },

  cursor_up_uniq: function(object) {
    if(object.jsshell_history.length != 0) {
      if(!object.jsshell_history_cursor == null) {
          object.jsshell_history_cursor=object.jsshell_history.length;
      }

      if(object.value == "") {
        jsshell_readline.cursor_up(object);
      } else {
        while(1) {
          if(object.jsshell_history_cursor > 0) {
              object.jsshell_history_cursor--;
          } else { // end of history
              break;
          }
          if(object.value != (object.jsshell_history[object.jsshell_history_cursor]).value) {
            object.value=(object.jsshell_history[object.jsshell_history_cursor]).value;
            break;
          } // else this entry is the same as our present value;
        }
      }
    } else {
      // the history is empty
    }
  }
};

jsshell_readline.cursor_down = function(object) {
  if(object.jsshell_history_cursor != null) {
    if(object.jsshell_history_cursor < (object.jsshell_history.length - 1) ) {
      object.jsshell_history_cursor++;
      object.value=(object.jsshell_history[object.jsshell_history_cursor]).value;
    } else {
      object.value='';
      object.jsshell_history_cursor = null;
    }
  } // else there is no active cursor to go down
}

jsshell_readline.cursor_down_uniq = function(object) {
  if(object.jsshell_history_cursor != null) {
    if(object.value == "") {
      jsshell_readline.cursor_down(object);
    } else {
      while(1) {
        if(object.jsshell_history_cursor < (object.jsshell_history.length - 1) ) {
          object.jsshell_history_cursor++;
          if(object.value != (object.jsshell_history[object.jsshell_history_cursor]).value) {
            object.value=(object.jsshell_history[object.jsshell_history_cursor]).value;
            break;
          } // else this entry is the same as our present value;
        } else {
          object.value='';
          object.jsshell_history_cursor = null;
          break;
        }
      }
    }
  } // else there is no active cursor to go down
}

jsshell_readline.cursor_clear = function(object) {
    //object.value='';
    object.jsshell_history_cursor = null;
}

var try_session_read = function(key) {
  if(window.JSON) {
    // Firefox 3.6 says window.sessionStorage is a protected read from file:// URLs
    try {
      var s=window.sessionStorage;
      if(s) {
        var v=window.sessionStorage[key];  
        if(v) {
          if(v!='') {
            return v;
          } else {
            return false;
          }
        } else {
          return false;
        }
      }
    } catch (e) {
      return false;
    }
  }
};

var try_session_update = function(key,value) {
  if(window.JSON) {
    // Firefox 3.6 says window.sessionStorage is a protected read from file:// URLs
    try {
      var s=window.sessionStorage;
      if(s) {
        window.sessionStorage[key]=value;
        return true;
      } else {
        return false;
      }
    } catch (e) {
      return false;
    }
  }
};

jsshell_readline.cursor_setup = function(object) {
  object.jsshell_history=jsshell_readline.default_history;
  var v=try_session_read('replete_readline_session_history');
  if(v) {
    var p;
    var i;
    try {
      p=JSON.parse('[' + v + ']');
    } catch(e) {
      // the session history was unparseable. kill it with fire!
      try_session_update('replete_readline_session_history','');
      p=false;
    }
    if(p) {
      for(i in p) {
        object.jsshell_history.push(p[i]);
      }
    }
  }
  object.jsshell_history_cursor=null;
}

jsshell_readline.history_append = function(object,entry,session) {
  var e=jsshell_readline.make_history_element(entry,session,false);
  object.jsshell_history[object.jsshell_history.length]=e;
  var v=try_session_read('replete_readline_session_history');
  if(v) {
    if(window.JSON) {
      try_session_update('replete_readline_session_history',v+','+JSON.stringify(e));
    }
  } else {
      if(window.JSON) {
        try_session_update('replete_readline_session_history',JSON.stringify(e));
      }
  }
};

jsshell_readline.expression_search = function(object,c) {
  alert('searching');
}

jsshell_readline.prefix_search = function(object,prefix) {
  var i,x,v;
  var vals,uniqvals;
  var cursor=false;

  i=object.jsshell_history.length;
  vals = [];
  uniqvals = [];
  for(x=0;x<i;x++) {
    v=(object.jsshell_history[x]).value;
    if((v.length >= prefix.length) && (prefix == v.substring(0,prefix.length))) {
      vals[vals.length]=(v);
    } 
  }
  vals=vals.sort();
  for(x=0;x<vals.length;x++) {
    if(vals[x] != cursor) {
      uniqvals[uniqvals.length]=vals[x];
    }
    cursor=vals[x];
  }
  if (vals.length>0) {
    var y;
    var c;
    var common_prefix="";
    if(vals[0].length == 0) {
      common_prefix=""; // clearly there is no common root
    } else {
      for(y=0;y<vals[0].length;y++) {
        c=vals[0][y];
        for(x=1;x<vals.length;x++) {
          if(vals[x].length <= y) {
            break;
          } else if(vals[x][y] != c) {
            break;
          }
        }
        if(x == vals.length) {
          common_prefix+=c;
        } else {
          break;
        }
      }
    }
    return [common_prefix,uniqvals];
  } else {
    return ["",[]];
  }
}

jsshell_readline.search = function (object) {
  var i,x,prefix,v;
  var vals,uniqvals;
  var cursor=false;
  var re;

  i=object.jsshell_history.length;
  vals = [];
  uniqvals = [];
  prefix=object.value;
  // turn a literal string into a regular expression
  // There is probably a cheaper strstr
  re=prefix.replace(/(.)/g,"[$1]").replace(/[[][\^][\]]/g,"[\\^]").replace(/[[][\]][\]]/g,"[\\]]");
  re=RegExp(re);
  
  for(x=0;x<i;x++) {
    v=(object.jsshell_history[x]).value;
    if(v.match(re)) {
      vals[vals.length]=(v);
    } 
  }
  vals=vals.sort();
  for(x=0;x<vals.length;x++) {
    if(vals[x] != cursor) {
      uniqvals[uniqvals.length]=vals[x];
    }
    cursor=vals[x];
  }
  if (vals.length>0) {
    var y;
    var c;
    var common_prefix="";
    if(vals[0].length == 0) {
      common_prefix=""; // clearly there is no common root
    } else {
      for(y=0;y<vals[0].length;y++) {
        c=vals[0][y];
        for(x=1;x<vals.length;x++) {
          if(vals[x].length <= y) {
            break;
          } else if(vals[x][y] != c) {
            break;
          }
        }
        if(x == vals.length) {
          common_prefix+=c;
        } else {
          break;
        }
      }
    }
    return [common_prefix,uniqvals];
  } else {
    return ["",[]];
  }
}
  
