// Functions for using Scheme as an XML RPC language

(function() {

function dump_sexp_number_to_string(v) {
  return v.attributes.getNamedItem('value').value;
}

function dump_sexp_number_to_js(v,shell) {
  return parseFloat(v.attributes.getNamedItem('value').value);
}

var dump_sexp_number = make_text_class_dumper(dump_sexp_number_to_string,"scheme_number");

function dump_sexp_symbol_to_string(v) {
  return v.attributes.getNamedItem('name').value;
}

var dump_sexp_symbol = make_text_class_dumper(dump_sexp_symbol_to_string,"scheme_symbol");

function dump_sexp_boolean_to_js(v,shell) {
  var which=v.attributes.getNamedItem('value').value;
  if(which == 'true') {
    return true;
  } else if(which == 'false') {
    return false;
  } else {
    alert('unknown boolean');
  }
}

function dump_sexp_boolean_to_string(v) {
  var which=v.attributes.getNamedItem('value').value;
  if(which == 'true') {
    return "#t";
  } else if(which == 'false') {
    return "#f";
  } else {
    alert('unknown boolean');
  }
}

var dump_sexp_boolean = make_text_class_dumper(dump_sexp_boolean_to_string,"scheme_boolean");

function dump_sexp_unspecified_to_string(string) {
  return "*unspecified*"
}

var dump_sexp_unspecified = make_text_class_dumper(dump_sexp_unspecified_to_string,"scheme_unspecified");

function dump_sexp_list(v,document_context,namespace,shell,continuation) {
  var i,node;
  
  var span=document_context.createElement('span');

  span.className="scheme_list";
  span.appendChild(document_context.createTextNode("("));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,function(e) {span.appendChild(e); });
    break;
  }
  for(i=i+1;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    span.appendChild(document_context.createTextNode(" "));
    dump_node(node,document_context,namespace,shell,function(e) {span.appendChild(e); });
  }

 span.appendChild(document_context.createTextNode(")"))
 continuation(span);
}

function dump_sexp_list_to_js(v,shell) {
  var i,node;
  var r=[];
  
  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    r.push(dump_node_to_js(node,shell));
    break;
  }
  for(i=i+1;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    r.push(dump_node_to_js(node,shell));
  }

  return r;
}

function dump_sexp_vector_to_js(v,shell) {
  // lists and vectors are equivalent for this purpose
  return dump_sexp_list_to_js(v,shell)
}

function dump_sexp_list_to_string(node) {
  return '(a-list-bogusly-dumped-to-string)';
}

function dump_sexp_quote(v,document_context,namespace,shell,continuation) {
var i,node;
 
  var span=document_context.createElement('span');

  //span.className="scheme_list";
  span.appendChild(document_context.createTextNode("'"));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,function(e) {span.appendChild(e); });
    break;
  }

  continuation(span);
}

function dump_sexp_quote_to_string(node) {
  var first_node=false,i;

  for(i=0;i<v.childNodes.length;i++) {
    first_node=v.childNodes[i];
    if(first_node.nodeType != 1) continue;
    break;
    first_node=false;
  }

  if(first_node) {
    return "'" + dump_node_to_string(first_node);
  } else {
    return "(quote)";
  }
}

function dump_quasiquote(v,document_context,namespace,shell,continuation) {
var i,node;
 
  var span=document_context.createElement('span');

  //span.className="scheme_list";
  span.appendChild(document_context.createTextNode("`"));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,function(e) {span.appendChild(e); });
    break;
  }

  continuation(span);
}

function dump_quasiquote_to_string(node) {
  var first_node=false,i;

  for(i=0;i<v.childNodes.length;i++) {
    first_node=v.childNodes[i];
    if(first_node.nodeType != 1) continue;
    break;
    first_node=false;
  }

  if(first_node) {
    return "`" + dump_node_to_string(first_node);
  } else {
    return "(quasiquote)";
  }
}

function dump_sexp_vector(v,document_context,namespace,shell,continuation) {
  var i,node;
    var span=document_context.createElement('span');
    var pseudo_continuation=function(e) {
      span.appendChild(e);
    }

  span.className="scheme_vector";
  pseudo_continuation(document_context.createTextNode("#("));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
    break;
  }
  for(i=i+1;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    pseudo_continuation(document_context.createTextNode(" "));
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
  }

 pseudo_continuation(document_context.createTextNode(")"));
 continuation(span);
}

function dump_sexp_vector_to_string(node) {
  return '#(a-vector-bogusly-dumped-to-string)';
}

function dump_sexp_hash(v,document_context,namespace,shell,continuation) {
  var i,node;
    var span=document_context.createElement('span');
    var pseudo_continuation=function(e) {
      span.appendChild(e);
    }

  span.className="scheme_hash";
  pseudo_continuation(document_context.createTextNode("#H("));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
    break;
  }
  for(i=i+1;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    pseudo_continuation(document_context.createTextNode(" "));
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
  }

 pseudo_continuation(document_context.createTextNode(")"));
 continuation(span);
}

function dump_sexp_hash_to_string(node) {
  return '#(a-hash-bogusly-dumped-to-string)';
}

function dump_sexp_pair(v,document_context,namespace,shell,continuation) {
  var i,node;

  var s=document.createElement('span');
  var pseudo_continuation=function(e) {
    s.appendChild(e);
  };

  s.className="scheme_pair";
  pseudo_continuation(document_context.createTextNode("("));

  for(i=0;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
    break;
  }
  pseudo_continuation(document_context.createTextNode(" . "));
  for(i=1;i<v.childNodes.length;i++) {
    node=v.childNodes[i];
    if(node.nodeType != 1) continue;
    dump_node(node,document_context,namespace,shell,pseudo_continuation);
    break;
  }

 pseudo_continuation(document_context.createTextNode(")"));
 continuation(s);
}

function dump_sexp_string(v,document_context,namespace,shell,continuation) {
  dump_string_without_quoting(v,document_context,namespace,shell,function(span) {
    span.insertBefore(document_context.createTextNode("\""),span.firstChild);
    span.appendChild(document_context.createTextNode("\""));
    span.className="scheme_string";
    continuation(span);
  },function(s) {
    return s.replace(/\"/g,"\\\"");
  });
}

function dump_sexp_string_to_string(v) {
  var acc="";
  var i;
  var node;

  if(v.firstChild) {
    for(i=0;i<v.childNodes.length;i++) {
      node=v.childNodes[i];
      if(node.nodeType != 3) {
        if(node.nodeName == 'char') {
          var code=String.fromCharCode(parseInt(node.attributes.getNamedItem('code').value));
          acc+=code;
        } else {
           alert('unknown node while decoding string');
        }
        continue;
      }
      acc+=node.nodeValue;
    }
    return acc;
  } else {
    alert('bad string node');
    return false;
  }
}

function dump_sexp_string_to_js(v,shell) {
  // string strings and javascripts strings are the same thing
  return dump_sexp_string_to_string(v);
}

function dump_char_to_string(v) {
  return v.firstChild.nodeValue+"";
}

var dump_char = make_text_class_dumper(dump_char_to_string,"scheme_char");

// Don't add a new tag here without updating the documentation
// in portable/doc.scm

// scheme-specific dumpers
dumpers['number'] = dump_sexp_number;
dumpers['symbol'] = dump_sexp_symbol;
dumpers['boolean'] = dump_sexp_boolean;
dumpers['string'] = dump_sexp_string;
dumpers['char'] = dump_char;
dumpers['list'] = dump_sexp_list;
dumpers['quote'] = dump_sexp_quote;
dumpers['quasiquote'] = dump_quasiquote;
dumpers['vector'] = dump_sexp_vector;
dumpers['hash'] = dump_sexp_hash;
dumpers['pair'] = dump_sexp_pair;
dumpers['unspecified'] = dump_sexp_unspecified;

string_dumpers['number'] = dump_sexp_number_to_string;
string_dumpers['symbol'] = dump_sexp_symbol_to_string;
string_dumpers['boolean'] = dump_sexp_boolean_to_string;
string_dumpers['string'] = dump_sexp_string_to_string;
string_dumpers['char'] = dump_char_to_string;
string_dumpers['list'] = dump_sexp_list_to_string;
string_dumpers['quote'] = dump_sexp_quote_to_string;
string_dumpers['quasiquote'] = dump_quasiquote_to_string;
string_dumpers['vector'] = dump_sexp_vector_to_string;
string_dumpers['hash'] = dump_sexp_hash_to_string;
string_dumpers['unspecified'] = dump_sexp_unspecified_to_string;

js_dumpers['number'] = dump_sexp_number_to_js;
js_dumpers['boolean'] = dump_sexp_boolean_to_js;
js_dumpers['string'] = dump_sexp_string_to_js;
js_dumpers['list'] = dump_sexp_list_to_js;
js_dumpers['vector'] = dump_sexp_vector_to_js;

function dump_sexp(exp,document_context,namespace,shell,continuation) {
  var sexp = exp.getElementsByTagName('sexp'); 
  var i,j;
  var node;

  for (i=0;i<sexp.length;i++) {
    for(j=0;j<sexp[i].childNodes.length;j++) {
      node=sexp[i].childNodes[j];
      if(node.nodeType != 1) continue;
      dump_node(node,document_context,namespace,shell,continuation);
    } 
  }
}

})();

function xml_to_tire(xml) {
  var tire = new Object();
  var sexp=xml.getElementsByTagName('sexp');

  if(!sexp) { 
    alert('xml_to_tire: no sexp!');
  }

  var i,j,k,index;
  var node,subnode;

  tire.rest=new Array();

  if((sexp.length == 1)) { 
    for (i=0;i<sexp.length;i++) {
      for(j=0;j<sexp[i].childNodes.length;j++) {
        node=sexp[i].childNodes[j];
        if(node.nodeType != 1) continue;
        if(node.nodeName == 'list') {
          for(k=0,index=0;k<node.childNodes.length;k++) {
            subnode=node.childNodes[k];
            if(subnode.nodeType != 1) continue;
            if(index == 0) {
              tire.type=dump_node_to_string(subnode);
            } else if(index == 1) {
              tire.tag=dump_node_to_string(subnode);
            } else { 
              tire.rest[tire.rest.length]=subnode;
            }
            index++;
          }
        } else {
          alert('unknown sexp subnode: ' + node.nodeName);
        }
      }  
    }
  }

  return tire;
}

function sexp_to_string(exp) {
  var sexp = exp.getElementsByTagName('sexp'); 
  var i,j;
  var node;
  var result='';

  for (i=0;i<sexp.length;i++) {
    for(j=0;j<sexp[i].childNodes.length;j++) {
      node=sexp[i].childNodes[j];
      if(node.nodeType != 1) continue;
      result += dump_node_to_string(node);
    } 
  }

  return result;
}

