/* Routines for an AJAX to TIRE bridge */

/*
 *   waiter pools are for servers, not for URLs
 *   (browsers may limit simultaneous connections by server)
 *
 *   The pool objects do not themselves make RPC calls
 */
var make_server_waiter_pool = function(server) {
  var counter=0;
  var pool = new Object;
  pool.length=0;
  pool.keys=new Array;
  pool.add_waiter=function(keybase) {
    var realkey=server+'-'+keybase+'-'+counter;
    counter++; 
    pool.keys.unshift(realkey);
    pool[realkey]='yup';
    pool.length++;
    return realkey;
  };
  pool.remove_waiter=function(key) {
    var foo;
    var done=false;
    for(foo in pool.keys) {
      if(pool.keys[foo] == key) {
        delete pool.keys[foo];
        done=true;
      }
    }
    delete pool[key];
    if(done) {
      pool.length--;
      return true;
    } else {
      return false;
    }
  };
  pool.waiter_succeeds=function(key) {
    if(pool.remove_waiter(key)) {
      pool.success_count++;
    }
  };
  pool.waiter_fails=function(key) {
    if(pool.remove_waiter(key)) {
      pool.failure_count++;
    }
  };
  pool.success_count=0;
  pool.failure_count=0;
  pool.server=server;
  pool.deferred_transmit_queue=new Array;
  pool.rpc_count = 0;
  pool.rpc_defer_count = 0;
  pool.rpc_undefer_count = 0;
  return pool;
};

function mux_to_wiki(shell,tire_mux) {
  var num_servers=0;
  var waiters_fragment='';

  for(var server in tire_mux.outstanding_waiters) {
      var this_pool=tire_mux.outstanding_waiters[server];

      waiters_fragment+='* server ' + server + '\n';
      num_servers++;

      waiters_fragment+="* Total RPCs: " + this_pool.rpc_count + '\n';
      waiters_fragment+='** Successful RPCs: ' + this_pool.success_count + ' -- these are HTTP RPCs, not necessarily TIRE forms \n';
      waiters_fragment+='** Failed RPCs: ' + this_pool.failure_count + '\n';
      waiters_fragment+="* Deferred RPCs: " + this_pool.rpc_defer_count + ' -- this is the number of RPCs that have enqueued in the client but which could not issue until outstanding RPC calls completed \n';
      waiters_fragment+="* Un-deferred RPCs: " + this_pool.rpc_undefer_count + ' -- this is the number of RPCs deferred in the client which were subsequently issued\n';
      waiters_fragment+='** Outstanding RPCs: ' + this_pool.length + '\n';
      for(var key in this_pool.keys) {
        waiters_fragment+='*** waiter ' + this_pool.keys[key] + '\n';
      }
  }

  var wikistr =
"=== Jsshell TIRE Mux ===\n" + 
"There are waiters for " + num_servers + " servers\n";

wikistr += waiters_fragment +
"==== RPC statistics ====\n" +
"* Pruned RPCs: " + tire_mux.rpc_prune_count + ' -- these are outstanding HTTP RPC calls that were shut down by the server to minimize open connections\n' +
//"* Missing RPCs: " + (tire_mux.rpc_count - ( tire_mux.outstanding_waiters.success_count + tire_mux.outstanding_waiters.failure_count + tire_mux.outstanding_waiters.length ) ) +
'\n';

  return libreplete.paint(gin_up_wiki_div(wikistr,'Jsshell',shell));
};

function make_tire_ajax_bridge(url,error_handler) {
  var url_get_server = function(url) {
    if(url.match(/^http/)) {
      return url.replace(/^http[^:]*:\/\//,"").replace(/:.*/,"").replace(/\/.*/,"");
    } else {
      return document.URL.replace(/^http[^:]*:\/\//,"").replace(/:.*/,"").replace(/\/.*/,"");
    }
  }

  // feed this to the server
  var session_id=Math.floor(1024*1024*1024*4*Math.random()) + "";

  var mux_bridge = new Object();
  var tag_hash = new Object();
  var connections = new Array();
  var mux_url_default=url;

  var cookie_hash = new Object();

  var outstanding_waiters = new Object();
 
  var needs_waiters=false;

  mux_bridge.set_url=function(url) {
    mux_url_default=url;
  };

  mux_bridge.hash_cookie = function(cookie,url) {
    cookie_hash[cookie]=url;
  };

  mux_bridge.unhash_cookie = function(cookie,url) {
    cookie_hash[cookie]=false;
  };

  mux_bridge.error_handler = error_handler;
  mux_bridge.outstanding_waiters = outstanding_waiters;

  mux_bridge.rpc_success_count = 0;
  mux_bridge.rpc_fail_count = 0;
  mux_bridge.rpc_prune_count = 0;
  mux_bridge.rpc_defer_count = 0;
  mux_bridge.rpc_undefer_count = 0;

  // needs_waiters should be set to true if we have exposed a cookie
  // name to the other end of the mux and should be prepared to accept
  // completely asynchronous forms (not just i and r forms)

  var tally_tags=function() {
    var count=0;
    for(var key in tag_hash) {
      count++;
    }
    return count; 
  }

  var add_waiter=function(url,waiters,label) {

    var key=waiters.add_waiter(label);

    waiters.rpc_count++;
    xmlrpc(url,'(http-wait "' + session_id + '")',function (exp) {
      if(exp.getElementsByTagName('sexp').length == 0) {
        alert('http-wait: there was an error. Logged already? : ' + exp);
      } else {
        catcher(url,waiters,exp,key);
      }
    } , function(e) {
      if(e.status == 204) {
        // RPC terminated quietly with no information
        // shut down by Proverb to force a single waiter
        mux_bridge.rpc_prune_count++;
      } else if(e.status == 200) {
        if(mux_bridge.error_handler) {
          mux_bridge.error_handler(e)
        } else {
          // HTTP seemed OK but didn't return an XML response. Proverb shut us
          // down but should have done it more politely.
          alert('http-wait shut down impolitely with ' + e.status + ' error status');
        }
      } else {
        if(mux_bridge.error_handler) {
          mux_bridge.error_handler(e)
        } else {
          alert('http-wait: serious tire rpc error: ' + e + ' -- status -- ' + e.status);
        }
      }
      waiters.waiter_fails(key);
    });
  }

  var catcher=function(url,waiters,exp,key) {
    waiters.waiter_succeeds(key);
    var t=xml_to_tire(exp);
    if(t) { 
      var m = tag_hash[t.tag];
      if(m) {
        if(m.length>0) {
          m[0](t);
          if(t.type == 'r') {
            m.shift();
            if(m.length == 0) { 
              tag_hash[t.tag]=false; 
              delete tag_hash[t.tag];
            }
          } else if(t.type == 'i') {
            if(t.rest.length > 0) {
              if(dump_node_to_string(t.rest[0]) == 'route') {
                if(t.rest.length > 1) {
                  var cookie=dump_node_to_string(t.rest[1]);
                  mux_bridge.hash_cookie(cookie,url);
                }
                // this is a routing update and can hidden. This mux doesn't have a routing table.
              } else if(dump_node_to_string(t.rest[0]) == 'unroute') {
                if(t.rest.length > 1) {
                  var cookie=dump_node_to_string(t.rest[1]);
                  mux_bridge.unhash_cookie(cookie,url);
                }
              }
            }
          } else {
            //alert('non-r form');
          }
        } else {
          alert('hash bucket for tag is empty');
        }
      } else { 
        //alert('no outstanding transmit for this tag: ' + t.tag + ' -- ' + exp);
      }

      if(waiters.length == 0) {
        if((tally_tags() > 0) || needs_waiters) {
          add_waiter(url,waiters,'catcher');
        }
      }

    } else {
      alert('unable to convert xml response to tire');
    }
  }

  mux_bridge.cancel=function() {
    var l=tag_hash.length; 
    var x;
    for(var key in tag_hash) {
      mux_bridge.transmit(function(tire) {
        if(tire.tire_error) {
          alert('error in cancellation!');
        } else {
          alert('cancellation complete');
        }
      },"s",key,"cancel");
    }
  };

  mux_bridge.max_waiters=2;

  // this is a varargs function
  mux_bridge.transmit=function(continuation,tireform,tag) {

    // our error continuation can be called more
    // than once as the error propagates through
    // the system.
    var error_count=0;
    var cookie=false;
    var this_url=false;
    var server;

    var these_outstanding_waiters;

    // we only let you register continuations for 't' and 's' commands
    // as well.
    if(continuation && (tireform == "t" || tireform == "s")) {
      var th;
      th=tag_hash[tag];
      if(!th) { 
        th=tag_hash[tag]=new Array();
      }
      th[th.length]=continuation;
    }

    if(tireform == "t" && arguments.length > 3) {
      var lookup;
      cookie=arguments[3];
      lookup=cookie_hash[cookie];
      if(lookup) {
        this_url=lookup;
      } else {
        this_url=mux_url_default;
      }
    } else {
      this_url=mux_url_default;
    }

    server=url_get_server(this_url);

    if(outstanding_waiters[server]) {
      these_outstanding_waiters=outstanding_waiters[server];
    } else {
      these_outstanding_waiters=outstanding_waiters[server]=make_server_waiter_pool(server);
    }

  
    var sexp;
    sexp='(http-rpc ';
    var sexp=sexp + '"' + session_id + '" (' + tireform + ' ' + tag;
    var i;
    for(i=3;i<arguments.length;i++) {
      sexp=sexp+' '+arguments[i];
    } 
    sexp = sexp + '))';
    var handle_error=function(e) {
      error_count++;
      //alert('(error count: ' + error_count + ') rpc error in tire transmit (form: ' + sexp +') error is: ' + e + ' type=' + e.type + ' status=' + e.status);
      if(error_count == 1) {
        mux_bridge.rpc_fail_count++;
        var tire_error=new Object;
        tire_error.tire_error=e;
        continuation(tire_error);
      } // otherwise don't squawk about it twice
    }

    // rpcs may issue out of order

    var issuer=function() {
      var waiter_key=these_outstanding_waiters.add_waiter('transmit');
      these_outstanding_waiters.rpc_count++;
      xmlrpc(this_url,sexp,function(e) {
          if(these_outstanding_waiters.length <= mux_bridge.max_waiters && these_outstanding_waiters.deferred_transmit_queue.length > 0) {
            var defer=these_outstanding_waiters.deferred_transmit_queue.shift();
            mux_bridge.rpc_undefer_count++;
            defer();
          }
          catcher(this_url,these_outstanding_waiters,e,waiter_key); 
        },function(e) {
          these_outstanding_waiters.waiter_fails(waiter_key);
          if(e.status == 204) {
            // No content. Proverb shut down a redundant connection
            mux_bridge.rpc_prune_count++;
          } else if(e.status == 200) {
            // HTTP seemed OK but didn't return an XML response. Proverb shut us
            // down but should have done it more politely.
            alert('http-rpc shut down impolitely with 200 error status');
          } else if(e.status == 0) {
            // HTTP probably failed utterly
            handle_error(e);
          } else {
            // 
            handle_error(e);
          }
      });
    };

    if(these_outstanding_waiters.length >= mux_bridge.max_waiters) {
      mux_bridge.rpc_defer_count++;
      these_outstanding_waiters.deferred_transmit_queue.push(issuer);
    } else {
      issuer();
    }
  }

  mux_bridge.show=function() {
    var p=document.createElement('p');
    var x;
    p.appendChild(document.createTextNode('TIRE STATE...'));
    p.appendChild(document.createElement('br'));
    p.appendChild(document.createTextNode('OUTSTANDING TAGS...'));
    p.appendChild(document.createElement('br'));
    for(x in tag_hash) {
      if(tag_hash[x]) {
        p.appendChild(document.createTextNode('tag_hash.' + x + ' == ' +typeof(tag_hash[x]) + '\n'));
        p.appendChild(document.createElement('br'));
      }
    }
    p.appendChild(document.createTextNode('WAITERS...'));
    p.appendChild(document.createElement('br'));
    for(x in outstanding_waiters.keys) {
        p.appendChild(document.createTextNode('waiter[' + outstanding_waiters.keys[x] + '] == ' + outstanding_waiters.keys[x]));
        p.appendChild(document.createElement('br'));
    }
   
    {
    var a;
      a=document.createElement('a');
      a.appendChild(document.createTextNode('launch a waiter'));
      a.onclick=function() {
        alert('not right now...');
        add_waiter('manual');
      }
      p.appendChild(a);
      p.appendChild(document.createElement('br'));
    }

    return libreplete.paint(p);
  };

  mux_bridge.session_id=session_id;

  return mux_bridge;
}

// evaluator support for TIRE-backed languages
the_tire_selector_evaluator=new function() {
  this.language="TIRE cookie selector";
  this.evaluator=function(str,doc,sh,cont) {
    str=str.replace(/^[\s]*/,"");
    sh.set_eval_cookie(str);
  }
  this.tabcompleter=false;
};
hybrid_evaluators[">"]=the_tire_selector_evaluator;

