//
// Code for persistient preference management
//   this is defined first because it is used early
// 
var jsprefs = {
  prefs: new Array,

  mkpref: function(name,description,default_value,type,predicate) {
    var pref = {
      jspref: true,
      name: name,
      desc: description,
      default_value: default_value,
      value: default_value,
      type: type,
      predicate: predicate
    }

    jsprefs.prefs[name]=pref;

    if(window.localStorage) {
      var lspref=localStorage.getItem('jsshell_jspref_' + name);
      if(lspref) {
        jsprefs.setpref_string(name,lspref);
      }
    }

    return pref;
  },

  ispref: function(name) {
    if(jsprefs.prefs[name]) {
      return true;
    } else {
      return false;
    }
  },

  getpref: function(name) {
    if(jsprefs.prefs[name]) {
      return jsprefs.prefs[name].value;
    } else {
      return false;
    }
  },

  setpref: function(name,value) {
    var old;
    if(jsprefs.prefs[name]) {
      old=jsprefs.prefs[name].value;
      jsprefs.prefs[name].value=value;
      if(window.localStorage) {
        localStorage.setItem('jsshell_jspref_' + name,value);
      }
      if(jsprefs.prefs.onchange) {
        jsprefs.prefs.onchange(name,old,value);
      }
      return old;
    } else {
      return false;
    }
  },

  setpref_string: function(name,value) {
    var pref=jsprefs.prefs[name];
    if(pref) {
      if(pref.type == 'number') {
        var v=parseInt(value);
        if(pref.predicate && pref.predicate(v)) {
          jsprefs.setpref(name,v);
        } else if(!pref.predicate) {
          jsprefs.setpref(name,v);
        }
      } else if(pref.type == 'boolean') {
        var v,set=false;
        if (value == 'true') { 
          set=true;
          v=true;
        } else if(value == 'false') {
          set=true;
          v=false;
        }
        if(set) {
          if(pref.predicate && pref.predicate(v)) {
            jsprefs.setpref(name,v);
          } else if(!pref.predicate) { 
            jsprefs.setpref(name,v);
          }
        }
      } else {
        if(pref.predicate && pref.predicate(value)) {
          jsprefs.setpref(name,value);
        } else if(pref.predicate) {
          jsprefs.setpref(name,value);
        }
      }
    }
  },

  loadpref: function(name,value) {
    var old;
    if(jsprefs.prefs[name]) {
      old=jsprefs.prefs[name].value;
      jsprefs.prefs[name].value=value;
      return old;
    } else {
      return false;
    }
  },

  loadpref_string: function(name,value) {
    var pref=jsprefs.prefs[name];
    if(pref) {
      if(pref.type == 'number') {
        var v=parseInt(value);
        if(pref.predicate && pref.predicate(v)) {
          jsprefs.loadpref(name,v);
        } else if(!pref.predicate) {
          jsprefs.loadpref(name,v);
        }
      } else if(pref.type == 'boolean') {
        var v,set=false;
        if (value == 'true') { 
          set=true;
          v=true;
        } else if(value == 'false') {
          set=true;
          v=false;
        }
        if(set) {
          if(pref.predicate && pref.predicate(v)) {
            jsprefs.loadpref(name,v);
          } else if(!pref.predicate) { 
            jsprefs.loadpref(name,v);
          }
        }
      } else {
        if(pref.predicate && pref.predicate(value)) {
          jsprefs.loadpref(name,value);
        } else if(!pref.predicate) {
          jsprefs.loadpref(name,value);
        }
      }
    }
  }
};

var is_positive_number = function(v) {
  if((typeof(v) == 'number') && v>0) {
    return true;
  } else {
    return false;
  }
};

var is_nonnegative_number = function(v) {
  if((typeof(v) == 'number') && v>=0) {
    return true;
  } else {
    return false;
  }
};

var is_string = function(v) {
  if(typeof(v) == 'string') {
    return true;
  } else {
    return false;
  }
};

jsprefs.mkpref('autocomplete_delay',
       'minimum time in milliseconds from the end of user input until a speculative completion event can occur',
        1000,
       'number',
       is_positive_number
       );
jsprefs.mkpref('autocomplete_minsize',
       'minimum number of characters that must be supplied before a speculative completion event can occur',
        4,
       'number',
       is_positive_number
       );
jsprefs.mkpref('spinner_delay',
       'time in milliseconds from the start of evaluation of a form until a visual element appears indicating ongoing work',
        1000,
       'number',
       is_positive_number
       );
jsprefs.mkpref('welcome_form',
       'language form that should run when a new user session starts',
       'jsshell_welcome(shell);',
       'string',
       false);
jsprefs.mkpref('welcome_back_form',
       'language form that should run when a user reloads a session',
       'jsshell_welcome_back(shell);',
       'string',
       false);
jsprefs.mkpref('wikis_echo',
       'non-executable links clicked in wiki text echo their associated command as shell commands',
       true,
       'boolean',
       false);
jsprefs.mkpref('anchors_eval',
       'executable links clicked in hypertext evaluate their associated commands',
       true,
       'boolean',
       false);
jsprefs.mkpref('string_trunc_threshold',
       'how many string characters to print before generating an ellipsis',
       256,
       'number',
       is_positive_number);
jsprefs.mkpref('theme',
       'default jsshell CSS theme',
       'themes/1998/1998.css',
       'string',
       is_string);
jsprefs.mkpref('smart_scroll',
       'use heuristics to decide how far to scroll the screen on output',
       true,
       'boolean',
       false);
jsprefs.mkpref('noprompt',
       'if false, the Replete interface has a prompt at the bottom. true when Repl is dedicated to hosting a widget.',
       false,
       'boolean',
       false);


// Globals assigned during initialization
var can_read_selection=false; // the browser supports reading the selection
var can_delete_selection=false; // the browser supports deleting the current selection



function with_xml_tag(shell,tagdef,inner) {
  var tp=new DOMParser();
  var doc=tp.parseFromString('<xhtml>'+tagdef+'</xhtml>','text/xml');
  var x=doc.getElementsByTagName('xhtml');
  if(x) {
    dump_node(x[0],document,false,shell,function(e) {
      inner(e);
    });
  }
};

jsshell_welcome=function(shell) {
  var d=document.createElement('div');
  with_xml_tag(shell,"<table border='0' width='100%'></table>",function(t) {
    d.appendChild(t);
    with_xml_tag(shell,"<tr></tr>",function(r) {
      t.appendChild(r);
      with_xml_tag(shell,"<td></td>",function(c) {
        r.appendChild(c);
        c.appendChild(gin_up_wiki_div('{{welcome}}','Jsshell',shell));
      });
      with_xml_tag(shell,"<td align='right'></td>",function(c) {
        r.appendChild(c);
        with_xml_tag(shell,"<small></small>",function(s) {
          s.appendChild(gin_up_wiki_div('Replete was written by the:\n\n[http://www.ida.org/aboutus/organization/hpcc.php Center for Computing Sciences]\n\n[http://www.ida.org Institute for Defense Analyses]\n\n\nReplete is a component of Proverb','Jsshell',shell)); 
          c.appendChild(s);
        });
      });
    });
  });
  return libreplete.paint(d);
};

jsshell_welcome_back=function(shell) {
  var d=document.createElement('div');
  with_xml_tag(shell,"<table border='0' width='100%'></table>",function(t) {
    d.appendChild(t);
    with_xml_tag(shell,"<tr></tr>",function(r) {
      t.appendChild(r);
      with_xml_tag(shell,"<td></td>",function(c) {
        r.appendChild(c);
        c.appendChild(gin_up_wiki_div('Welcome back to Replete','Jsshell',shell));
      });
      with_xml_tag(shell,"<td align='right'></td>",function(c) {
        r.appendChild(c);
        with_xml_tag(shell,"<small></small>",function(s) {
          s.appendChild(gin_up_wiki_div('Replete was written by the:\n\n[http://www.ida.org/aboutus/organization/hpcc.php Center for Computing Sciences]\n\n[http://www.ida.org Institute for Defense Analyses]\n\n\nReplete is a component of Proverb','Jsshell',shell)); 
          c.appendChild(s);
        });
      });
    });
  });
  return libreplete.paint(d);
};

function jsshell_error(target,value,document_context) {
  var newp = document_context.createElement('div');
  var tn = document_context.createTextNode(value);
  newp.className='iwta_error';
  newp.appendChild(tn);
  return newp;
}

function text_area_resizer(textarea) {
  var form_updater=function() {
    for(var x in textarea) {
      if(!textarea) continue;
      if(typeof(textarea.rows) != "number") continue;
      textarea.rows=countLines(textarea.value,textarea.cols)+1;
    }
    setTimeout(form_updater,300);
  };
  function countLines(strtocount,cols) {
    var hard_lines=1;
    var last=0;
    while(true) {
      last=strtocount.indexOf("\n",last+1);
      hard_lines++;
      if(last == -1) break;
    }
    var soft_lines=Math.round(strtocount.length / (cols-1));
    var hard = eval("hard_lines " + unescape("%3e") + "soft_lines;");
    if(hard) soft_lines = hard_lines;
    if(soft_lines<3)
      return 3;
    else
      return soft_lines;
  }
  form_updater();
}

var libreplete = new Object;

libreplete.generate_representation = function(value,document_context) {
  var doc;
  var string_trunc_threshold = jsprefs.getpref('string_trunc_threshold');

  function simple_truncate_string(s) {
    if(s.length > string_trunc_threshold) {
      return '"' + s.substr(0,string_trunc_threshold) + '"';
    } else {
      return '"' + s + '"';
    }      
  }

  // a smart string is a string printed partially with an elipsis that expands it fully
  function make_smart_string(s) {
    var tn;
    if(s.length > string_trunc_threshold) {
      var newp = doc.createElement('div');
      tn = doc.createTextNode(simple_truncate_string(s));
      var a = doc.createElement('a');
      a.innerHTML="...";
      a.onclick=function() {
        var newtn = doc.createTextNode('"' + s + '"');
        newp.removeChild(a);
        newp.removeChild(tn);
        newp.appendChild(newtn);
      };
      a.onhover=function() { alert('hover') };
      a.oncontextmenu=function(args) { alert('contextmenu:' + args); };
      newp.appendChild(tn);
      newp.appendChild(a);
      return newp;
    } else {
      tn = doc.createTextNode(simple_truncate_string(s));
      //newp.appendChild(tn);
      return tn;
    }
  }

  doc=document_context; 
  var valtype = typeof(value);

  if(valtype == 'string') {
    var newp=make_smart_string(value);
    newp.className='iwta_scalar';
    return newp;
  } else if(value == null) {
    var newp = doc.createElement('div');
    var tn = doc.createTextNode('null');

    newp.className='iwta_error';
    newp.appendChild(tn);
    return newp;
  } else if(valtype == 'object' && value.jsshell_paintable==true) {
    return value.value;
  } else if(valtype == 'object') {
    var i;
    //alert('dumping an object with length ' + value.length);
    //for(i in value) {
    //  alert('a key in the object: ' + i);
    //}; 
    try {
      var newp = doc.createElement('div');
      var tn = doc.createTextNode(value);
      var newanchor = doc.createElement('a');
      newp.className='iwta_object';
      newanchor.onclick=function() {
        var bogusdiv=document_context.createElement('div');
        var bogusclose=document_context.createElement('a');
        bogusclose.innerHTML='close';
        bogusclose.onclick=function() {
          newp.removeChild(bogusdiv);
        }
        bogusdiv.appendChild(bogusclose);
        newp.appendChild(bogusdiv);
        examine(bogusdiv,value,document_context); 
      }
      newanchor.appendChild(tn);
      newp.appendChild(newanchor);

      return newp;
    } catch (e) {
      var newp = doc.createElement('div');
      var tn = doc.createTextNode('object with no representation');
      newp.className='iwta_error';
      newp.appendChild(tn);
      return newp;
    }
  } else if(valtype == 'function') {
    try {
      var newp = doc.createElement('div');
      var tn = doc.createTextNode(value);
      var newanchor = doc.createElement('a');
      newp.className='iwta_object';
      newanchor.onclick=function() { examine(newp,value+'',document_context); };
      newanchor.appendChild(tn);
      newp.appendChild(newanchor);
      return newp;
    } catch (e) {
      var newp = doc.createElement('div');
      var tn = doc.createTextNode('object with no representation');
      newp.className='iwta_error';
      newp.appendChild(tn);
      return newp;
    }
  } else {
    try {
      var newp = doc.createElement('div');
      newp.className='iwta_scalar';
      newp.innerHTML=value;
      return newp;
    } catch (e) {
      var newp = doc.createElement('div');
      var tn = doc.createTextNode(typeof(value) + ' with no representation');
      newp.className='iwta_error';
      newp.appendChild(tn);
      return newp;
    }
  }
}

// don't really paint, just wrap in a paintable
// returning one of these from the evaluator suppresses
// the normal Javascript object printer and uses value
// as the DOM node to be rendered as the result.
libreplete.paint = function(value) {
  var pv = new Object();
  pv.jsshell_paintable=true;
  pv.value=value;
  return pv;
}

// convert a number to an ordinal string suitable for use
// in the prose of decent, civilized persons.
libreplete.ordinalize=function(n) {
  switch(n) {
    case 0: return "zeroth";
    case 1: return "first";
    case 2: return "second";
    case 3: return "third";
    case 4: return "fourth";
    case 5: return "fifth";
    case 6: return "sixth";
    case 7: return "seventh";
    case 8: return "eighth";
    case 9: return "ninth";
    case 10: return "tenth";
    case 11: return "eleventh";
    case 12: return "twelfth";
    case 13: return "thirteenth";
    case 14: return "fourteenth";
    case 15: return "fifteenth";
    case 16: return "sixteenth";
    case 17: return "seventeenth";
    // my schoolboy style guide said to stop here
    default: {
      var m=n%10;
      switch(m) {
        case 1: return n+'st';
        case 2: return n+'nd';
        case 3: return n+'rd';
        default: return n+'th';
      };
    };
  }
};


hybrid_evaluators=new Array();

var hybrid_default_evaluator=false;

// a hybrid evaluator that switches languages based on a prefix
function evaluator_hybrid() {
  var with_evaluator = function(ex,f) {
    if(typeof(ex) == "string") {
      if(hybrid_default_evaluator.canmatch && hybrid_default_evaluator.canmatch(ex)) {
        f(hybrid_default_evaluator,ex,"");
      } else {
        var e=hybrid_evaluators[ex[0]];
        if(e) {
          f(e,ex.substring(1),ex.substring(0,1)); 
        } else {
          f(hybrid_default_evaluator,ex,"");
        }
      }
    } else {
      alert('non-string expression for hybrid evaluator: ' + typeof(ex) + ' -- ' + ex);
    }
  };
  var with_loader = function(ex,f) {
    // should match on extensions?
    if(typeof(ex) == "string") {
      f(hybrid_default_evaluator);
    } else {
      alert('non-string expression for hybrid loader: ' + typeof(ex) + ' -- ' + ex);
    }
  };

  this.language = "Hybrid prefix evaluator";
  this.prompt = function(sh) {
    if(hybrid_default_evaluator.prompt) {
      return hybrid_default_evaluator.prompt(sh);
    } else {
      return document.createTextNode(hybrid_default_evaluator.language + '@' + sh.private_eval_cookie + '>');
    }
  };

  this.evaluator = function(expression,document_context,shell,continuation) {
    with_evaluator(expression,function(e,expression) {
      if(e.evaluator) {
        e.evaluator(expression,document_context,shell,continuation);
      } else {
        alert('that evaluator has no evaluator method');
      }
    });
  }
  this.loader = function(expression,filename,document_context,shell,continuation) {
    with_loader(filename,function(e) {
      if(e.loader) {
        e.loader(expression,filename,document_context,shell,continuation);
      } else if(e.evaluator) {
        e.evaluator(expression,document_context,shell,continuation);
      } else {
        alert('that evaluator has no loader or evaluator method');
      }
    });
  };

  this.tabcompleter = function(expression,shell,continuation) {
    with_evaluator(expression,function(e,expression,prefix) {
      if(e.tabcompleter) {
        if(prefix == "") {
          e.tabcompleter(expression,shell,continuation);
        } else {
          e.tabcompleter(expression,shell,function(v) {
            if(v) {
              continuation(prefix+v);
            } else {
              continuation(false);
            }
          });
        }
      } else {
        continuation(false);
      }
    });
  }

  this.eval_to_js = function(expression,shell,continuation) {
    with_evaluator(expression,function(e,expression) {
      if(e.eval_to_js) {
        e.eval_to_js(expression,shell,continuation);
      } else {
        alert('that evaluator has no eval_to_js_method');
      }
    });
  };

  this.eval_to_string = function(expression,shell,continuation) {
    with_evaluator(expression,function(e,expression) {
      if(e.eval_to_string) {
        e.eval_to_string(expression,shell,continuation);
      } else {
        alert('that evaluator has to eval_to_string method');
      }
    });
  };

  this.cancel = function(v,shell) {
    if(v) {
    } else {
      if(hybrid_default_evaluator.cancel) {
        hybrid_default_evaluator.cancel(v,shell);
      }
    }
  };

};

the_hybrid_evaluator = new evaluator_hybrid();

// The default evaluator -- a tiny Javascript evaluation environment
function evaluator_js() {
  this.language="Javascript";
  this.prompt=function(sh) {
    return document.createTextNode(this.language+">");
  };

  var handheld_wrap=function(e) {
    return e.replace(/lll/g,"(")
            .replace(/rrr/g,")")
            .replace(/sss/g,";")
            .replace(/uuu/g,"_");
  };
  
  this.evaluator=function(expression,document_context,shell,continuation) {
    var retval;
    try {
      var evalval = eval(handheld_wrap(expression));
      retval = libreplete.generate_representation(evalval,document_context);
    } catch(e) {
      // Neither Safari nor Firefox appear to populate error objects with this
      if(!e.expression) {
        e.expression = expression;
      }
      retval = libreplete.generate_representation(e,document_context);
    }
    continuation(retval);
  };

  // find the longest shared root between two strings
  var find_common_root=function(stringa,stringb) {
    var x,l,a;
    l=min(stringa.length,stringb.length);
    a='';
    for(x=0;x<l;x++) {
      if(stringa[x] == stringb[x]) {
       a+=stringa[x];
      } else {
        break;
      }
    }
    return a;
  };

  // tab complete an identifier in an object
  var single_matcher = function(object,string) {
    var exp,x,root;
    exp=new RegExp('^'+string);
    root=false;
    for(x in object) {
      if(exp.exec(x)) {
        if(root) {
          root=find_common_root(x+'',root);
        } else {
          root=x+'';
        }
      }
    } 
    if(root) {
      return root;
    } else {
      return false;
    }
  };

  var last_matched_object=false;
  var last_matched_prefix=false;
  var last_matched_root=false;

  // an exceptionally lame tabcompleter for Javascript that
  // uses introspection
  var multi_matcher = function(string) {
    var object_cursor=window;
    var string_cursor=string;
    var m;
    var acc;

    if(last_matched_prefix && (string.length >= last_matched_prefix.length) && (last_matched_prefix==string.substring(0,last_matched_prefix.length))) {
      acc=last_matched_prefix;
      string_cursor=string.substring(last_matched_prefix.length);
      object_cursor=last_matched_object;
    } else {
      last_matched_prefix=false;
      last_matched_object=window;
      last_matched_root=false;
      acc='';
    }
    while(1) {
      m=string_cursor.match(/^([A-Za-z][A-Za-z0-9_]*)(.*)/);
      if(m) {
        var identifier=m[1];
        if(m[2] == "") {
          // then this is the terminal match
          var r=single_matcher(object_cursor,identifier);
          if(r) {
            acc+=r;
            return acc;
          } else {
            return false;
          }
        } else {
          var o=object_cursor[identifier];
          if(o) {
            // complete through a dot
            var mnext=m[2].match(/^([ \t]*[.])(.*)/);
            if(mnext) {
              acc+=(identifier+mnext[1]);
              object_cursor=o;
              string_cursor=mnext[2];
              last_matched_object=object_cursor;
              last_matched_prefix=acc;
              last_matched_root=string_cursor;
            } else {
              // complete through a paren
              mnext=m[2].match(/^([ \t]*[\(])(.*)/);
              if(mnext) {
                acc+=(identifier+mnext[1]);
                string_cursor=mnext[2];
                last_matched_prefix=acc;
                last_matched_root=string_cursor;
              } else {
                // complete through a comma
                mnext=m[2].match(/^([ \t]*[,])(.*)/);
                if(mnext) {
                  acc+=(identifier+mnext[1]);
                  string_cursor=mnext[2];
                  last_matched_prefix=acc;
                  last_matched_root=string_cursor;
                } else {
                  return false;
                }
              }
            }
          } else {
            return false;
          }
        }
      } else {
        return false;
      }
    }
  };

  //if( (multi_matcher("documen") != "document") ||
  //    (multi_matcher("document.bod") != "document.body") ||
  //    (multi_matcher("alert(documen") != "alert(document") ||
  //    (multi_matcher("alert(document.bod") != "alert(document.body") ||
  //    //(multi_matcher("alert(document.body,documen") != "alert(document.body,document") ||
  //    0
  //  ) {
  //  alert('tabcomplete testcase failure');
  //}
  
  this.tabcompleter = function(string,shell,continuation) {
    continuation(multi_matcher(string));
    return false;
  };

  this.eval_to_js = function(string,shell,continuation) {
    var retval;
    try {
      var evalval = eval(string);
      retval = evalval;
    } catch(e) {
      retval=e;
    }
    continuation(retval);
  };
  this.eval_to_string = function(string,shell,continuation) {
    var retval;
    try {
      var evalval = eval(string);
      retval = evalval;
    } catch(e) {
      retval = e;
    }
    continuation(''+retval);
  };
  this.expression_wrapper=handheld_wrap;
};

the_js_evaluator=new evaluator_js();
hybrid_evaluators["!"]=the_js_evaluator;

function evaluator_href() {
  this.language = "URLs";
  this.prompt=function(sh) {
    return document.createTextNode("URL>");
  };
  this.evaluator = function(expression,document_context,shell,continuation) {
    var inframe=false;
    expression=expression.replace(/^[\s]*/,"");
    if(expression.length>1 && expression.match(/^:/)) {
      inframe=true;
      expression=expression.substring(1);
    }
    if(!expression.match(/:/) && !expression.match(/^\//)) {
      expression='http://' + expression;
    }
    if(inframe) {
      var i=document.createElement('iframe');
      i.src=expression;
      continuation(i);
    } else {
      var new_window = window.open(expression,'','');
      continuation(document.createTextNode('visiting ' + expression));
    } 
  }
  this.tabcompleter = false;
};
hybrid_evaluators[":"]=new evaluator_href();

function evaluator_notes() {
  this.language = "notes";
  this.prompt=function(sh) {
    return document.createTextNode("notes>");
  };
  this.evaluator = function(expression,document_context,shell,continuation) {
    var d=document.createElement('div');
    d.appendChild(document.createTextNode(Date()));
    d.appendChild(document.createElement('br'));
    d.appendChild(document.createTextNode(expression));
    continuation(d);
  };
  this.tabcompleter = false;
};
hybrid_evaluators["#"]=new evaluator_notes();

function evaluator_wiki() {
  this.language = "wiki";
  this.prompt=function(sh) {
    return document.createTextNode("wiki>");
  };
  this.evaluator = function(expression,document_context,shell,continuation) {
    continuation(gin_up_wiki_div(expression,'Jsshell',shell));
  };
  this.tabcompleter = false;
};
hybrid_evaluators["|"]=new evaluator_wiki();


hybrid_evaluators["^"]=new function() {
  this.language = "hybrid language switcher";
  this.evaluator = function(expression,document_context,shell,continuation) {
    var e;
    e=hybrid_evaluators[expression];
    if(e) {
      hybrid_default_evaluator=e;
    }
  };
  this.tabcompleter = false;
};

function readfile(file,next) {
  var r=new FileReader();
  if(r.readAsText) {
    r.addEventListener('load',function(ep) {
      next(ep.target.result);
    },false);
    r.addEventListener('error',function(ep) {
      next(false);
    },false);
    r.readAsText(file);
  }
}

function evaluator_include() {
  this.language = "host file include";
  this.evaluator = function(expression,document_context,shell,continuation) {
    var d=document_context.createElement('div');
    if(window.FileReader) {
      var i,is;
      i=document_context.createElement('input');
      i.type='file';
      is=document_context.createElement('button');
      is.appendChild(document.createTextNode('evaluate'));
      is.addEventListener('click',function(e) {
        readfile(i.files.item(0),function(text) {
          shell.inner_evaluator.evaluator(text,document_context,shell,continuation);
        });
      },false);
      d.appendChild(document_context.createTextNode('file chooser'));
      d.appendChild(i);
      d.appendChild(is);
    } else {
      d.appendChild(document_context.createTextNode('no local file chooser in this browser'));
    }
    continuation(d);
  };
  this.tabcompleter = false;
};
hybrid_evaluators["<"]=new evaluator_include();

var jsshell_current_evaluator=the_hybrid_evaluator;

var eval_tag_count = 0;
var default_eval_cookie = "local";

/* make a scheme printed string from a javascript string */
function scheme_stringify(s) {
  var ns="",x,c,cc;
  for(x=0;x<s.length;x++) {
    c=s.charAt(x);
    cc=c.charCodeAt(0);
    if(c=="\"") {
      ns=ns+"\\\"";
    } else if(c=="\\") {
      ns=ns+"\\\\";
    } else if(cc == 0) {
      //ns=ns+"\\000";
      ns=ns+"\\0";
    } else if(cc >= 128) {
      ns=ns+"\\x"+cc.toString(16);
    } else if(cc > 255) {
      alert('encounterd wide char in scheme_stringify.');
    } else {
      //ns=ns+c;
      ns=ns+String.fromCharCode(cc);
    }
  }
  return "\"" + ns + "\"";
}

var rpc_url=false;

var interactive_tire_mux_error_handler = function(shell,e) {
  var online_error=function() {
    var error_text='The TIRE mux is lying drunk on the floor. It moans gently as you kick it. You can [[[!mux_to_wiki(shell,shell.tire_mux)|show]]] the state of the mux.';

    shell.output_append_target.appendChild(gin_up_wiki_div(error_text,'Jsshell',shell));
  }
  if(navigator && typeof(navigator.onLine) == 'boolean') {
    if(navigator.onLine) {
      online_error();
    } else {
      var error_text='Dude. Your TIRE mux is the loneliest mux in the world. Your browser is offline.';
      shell.output_append_target.appendChild(gin_up_wiki_div(error_text,'Jsshell',shell));
    }
  } else {
    online_error();
  }
  //return paint(gin_up_wiki_div(error_text,'Jsshell',shell));
};

function boot_tire(shell) {
  if(!shell.tire_mux) {
    shell.tire_mux=make_tire_ajax_bridge(rpc_url,function(e) {
      interactive_tire_mux_error_handler(shell,e);
    });
  }
};

// an alternate evaluator for RPC systems
// calls continuation either with the error object or the tire 'i' results and 'r' result
function jsshell_rpc_tire_evaluator(command,expression,shell,continuation,error_continuation) {

  if(!shell) {
    alert('no shell!');
  }

  if(!shell.tire_mux) {
    document.body.appendChild(libreplete.generate_representation(shell,document));
  }

  try {
    shell.tire_mux.transmit(function(tire) {
      if(tire.tire_error) {
        interactive_tire_mux_error_handler(shell,tire.tire_error);
        if(error_continuation) {
          error_continuation(tire.tire_error);
        }
      } else {
        if(tire.type == 'r') {
          continuation(tire);
        } else if(tire.type == 'i') {
          if(tire.rest.length > 0) {
            if(dump_node_to_string(tire.rest[0]) == 'route') {
              // this is a routing update and can hidden. This mux doesn't have a routing table.
            } else if(dump_node_to_string(tire.rest[0]) == 'unroute') {
              // this is a routing update and can hidden. This mux doesn't have a routing table.
            } else {
              continuation(tire);
            }
          }
        } else {
          error_continuation(tire);
        }
      } 
    },"t","js-"+shell.session_id+"-eval-tag-"+eval_tag_count,shell.private_eval_cookie,command,scheme_stringify(expression));
    eval_tag_count++;
  } catch(e) {
    continuation(e);
  }
}

function evaluator_foo_tire() {
  this.tire_command='seval';
  this.expression_wrapper=function(expression) {
    return expression;
  };
};

last_error=false;

evaluator_foo_tire.prototype['evaluator'] = function (expression,document_context,shell,continuation) {
  var finished=false;
  var spinner=false;
  var spinner_timeout=setTimeout(function() {
    if(!finished) {
      spinner=document.createElement('span');
      spinner.className='waiting';
      spinner.appendChild(document.createTextNode('waiting'));
      continuation(spinner);
      spinner_timeout=false;
    }
  },jsprefs.getpref('jsshell_spinner_delay'));

  continuation(spinner);
  jsshell_rpc_tire_evaluator(this.tire_command,this.expression_wrapper(expression),shell,
    function(v) {
      if(v.type=='r') {
        finished=true;
        if(spinner_timeout) {
          clearTimeout(spinner_timeout);
        }
        if(spinner && spinner.parentNode) { spinner.parentNode.removeChild(spinner); }
        dump_node_array(v.rest,document_context,false,shell,continuation);
      } else if(v.type == 'i') {
        var itype=dump_node_to_string(v.rest[0]);
        // skip past the 'output' or 'error' tag. we should probably consume that and
        if( itype == "output" ) {
          v.rest.shift();
          //s.appendChild(document_context.createTextNode(t));
          dump_string_without_quoting(v.rest[0],document_context,false,shell,function(e) {
            e.className='iwta_output';
            continuation(e);
          },false);
        } else if ( itype == "error") {
          v.rest.shift();
          //s.appendChild(document_context.createTextNode(t));
          dump_string_without_quoting(v.rest[0],document_context,false,shell,function(e) {
            e.className='iwta_error';
            continuation(e);
          },false);
        } else if ( itype == "async") {
          v.rest.shift();
          //s.appendChild(document_context.createTextNode(t));
          dump_string_without_quoting(v.rest[0],document_context,false,shell,function(e) {
            e.className='iwta_async';
            continuation(e);
          },false);
        } else {
          v.rest.shift();
          dump_node_array(v.rest,document_context,false,shell,continuation);
        }
      } else {
        continuation(libreplete.generate_representation(v,document_context));
      }
    },function(v) {
      //continuation('error: ');
      if(spinner_timeout) {
        clearTimeout(spinner_timeout);
      }
      if(spinner && spinner.parentNode) { spinner.parentNode.removeChild(spinner); }
      last_error=v;
      continuation(libreplete.generate_representation(v,document_context));
    }
  );
};


evaluator_foo_tire.prototype.language='generic TIRE';
evaluator_foo_tire.prototype['eval_to_string'] = function(string,shell,continuation) {
  jsshell_rpc_tire_evaluator("seval",string,shell,function(v) {
    if(v.type == 'r') {
      continuation(dump_node_to_string(v.rest[0]));
    } else {
      false;
    }
  },function(v) {
      if(v.type == 'e') {
        continuation(dump_node_to_string(v.rest[0]));
      }
      shell.output_append_target.appendChild(document.createTextNode('tire seval eval_to_string error in ' + string + '\n'));
      shell.output_append_target.appendChild(libreplete.generate_representation(v,document));
  });
};
evaluator_foo_tire.prototype['eval_to_js'] = function(string,shell,continuation) {
  jsshell_rpc_tire_evaluator("seval",string,shell,function(v) {
    if(v.type == 'r') {
      continuation(dump_node_to_js(v.rest[0],shell));
    } else {
      // huh. I don't really know how to accomodate i responses here
      if(v.type == 'i') {
        continuation(v);
      }
    }
  },function(v) {
      if(v.type == 'e') {
        continuation(dump_node_to_js(v.rest[0],shell));
    }
    shell.output_append_target.appendChild(document.createTextNode('eval_to_js error\n'));
    shell.output_append_target.appendChild(libreplete.generate_representation(v,document));
  });
};
evaluator_foo_tire.prototype['cancel'] = function(val,shell) {
  if(!shell.tire_mux) {
    alert('cancelling without mux');
  } else {
    shell.tire_mux.cancel();
  }
};

// allow for history completion
evaluator_foo_tire.prototype['tabcompleter'] = false;

function evaluator_scm_rpc() {
  var that = new evaluator_foo_tire();

  var handheld_wrap=function(e) {
    return e.replace(/lll/g,"(").replace(/rrr/g,")");
  };
  
  that.language = 'Scheme'; 
  that.tabcompleter = function(string,shell,continuation) {
    if(navigator.onLine) {
      jsshell_rpc_tire_evaluator("complete",handheld_wrap(string),shell,function(v) {
        if(v.type == 'r') {
          continuation(dump_node_to_js(v.rest[0]));
        } else {
          false;
        }
      },function(v) {
        if(v.type == 'e') {
          continuation(dump_node_to_string(v.rest[0]));
        }
        shell.output_append_target.appendChild(document.createTextNode('complete error\n'));
        shell.output_append_target.appendChild(libreplete.generate_representation(v,document));
      });
    }
  };
  that.canmatch=function(ex) {
    if(ex.match(/^#[xotf]/)) { // certain collisions with the notes reader
      return true;
    } else {
      return false;
    }
  }
  that.loader=function(expression,filename,document_context,shell,continuation) {
    that.evaluator("(begin " + expression + ")",document_context,shell,continuation);
  };

  // an optimization for doing Scheme on handhelds
  // without direct-entry punctuation
  that.expression_wrapper=handheld_wrap;

  return that;
};

the_scm_rpc_evaluator = new evaluator_scm_rpc();
hybrid_evaluators["]"]=the_scm_rpc_evaluator;

function evaluator_help() {
  var scm_that = new evaluator_scm_rpc;
  var js_that = new evaluator_js;

  var scm_expression_wrapper = function(expression) {
    expression=expression.replace(/^[\s]*/,"");
    if(expression == "") {
      return '(help)';
    } else {
      return '(help ' + scheme_stringify(expression) + ')';
    }
  };

  this.evaluator = function(expression,document_context,shell,continuation) {
    if(expression == "") {
      var w=jsprefs.getpref('default_wiki');
      if(w && jspref_wikis[w] && jspref_wikis[w].get_article && jspref_wikis[w].home) {
        var r;
        r=domwiki.get_default_article(shell,w);
        if(r && r.jsshell_paintable) {
          continuation(r.value);
        } else {
          continuation(r);
        }
/*
        jspref_wikis[w].get_article(jspref_wikis[w].home,shell,function(v) {
          continuation(gin_up_wiki_div(v,w,shell));
        });
*/
      } else {
        continuation(document_context.createTextNode('I have no home page to show you'));
      }
    } else {
      if(rpc_url) {
        scm_that.evaluator(scm_expression_wrapper(expression),document_context,shell,continuation);
      } else {
        jsshell_help_search(expression,function(v) {
          var r='=== Search Results ===\n';
          var l;
          for(l=0;l<v.length;l++) {
            r+='* [[' + v[l] + '|' + v[l] + ']]\n';
          }

          continuation(gin_up_wiki_div(r,'Jsshell',shell)); 
        });
      }
    }
  
  };
  this.language='help';
  this.tabcompleter=false;

  return this;
};

hybrid_evaluators['?'] = new evaluator_help;

function old_evaluator_unix_shell() {
  var that = new evaluator_scm_rpc;
  that.language='UNIX';
  that.expression_wrapper=function(expression) {
    expression=expression.replace(/^[\s]*/,"");
    if(expression == "") {
      return expression;
    } else {
      return '(xhtml-popen ' + scheme_stringify(expression) + ')';
    }
  };
  that.tabcompleter=false;
  return that;
};

hybrid_default_evaluator=the_js_evaluator;

function evaluator_bg_scm_rpc() {
  var that=new evaluator_scm_rpc;

  that.language = "Scheme threads";
  that.tire_command='seval_t';

  return that; 
};

the_bg_scm_rpc_evaluator = new evaluator_bg_scm_rpc();
hybrid_evaluators["&"]=the_bg_scm_rpc_evaluator; // used to be the thread_rpc_evaluator

// an alternate evaluator for RPC systems
// calls continuation either with the error object or the tire 'i' results and 'r' result
function jsshell_raw_tire_driver(expression,shell,continuation,error_continuation) {

  // eat leading and closing parens
  if(expression[0]=="(") {
    expression=expression.substring(1);
    if((expression.length>0) && expression[expression.length-1]==")") {
      expression=expression.substring(0,expression.length-1);
    }
  }

  var action_rest=expression.match(/[ ]*([^ ]{1,})(.*)/);

  if(action_rest.length<2) {
    continuation(document.createTextNode("bad form. You must supply at least an action"));
  } else {

    if(!shell.tire_mux) {
      shell.tire_mux=make_tire_ajax_bridge(rpc_url,function(e) { interactive_tire_mux_error_handler(shell,e); });
    }

    var action=action_rest[1];
    if(action=="t") {
      var tag_rest=action_rest[2].match(/[ ]*([^ ]{1,})(.*)/);
      if(tag_rest.length<2) {
        continuation(document.createTextNode("bad form. Try (t tag cookie command [body])."));
      } else {
        var tag=tag_rest[1];
        var cookie_rest=tag_rest[2].match(/[ ]*([^ ]{1,})(.*)/);
        if(cookie_rest.length<2) {
          continuation(document.createTextNode("bad form. Try (t tag cookie command [body])."));
        } else {
          var cookie=cookie_rest[1];
          var command_rest=cookie_rest[2].match(/[ ]*([^ ]{1,})(.*)/);
          if(command_rest.length<2) {
            continuation(document.createTextNode("bad form. Try (t tag cookie command [body])."));
          } else {
            var command=command_rest[1];
            var rest=command_rest[2]; 

            try {
              shell.tire_mux.transmit(function(tire) {
                if(tire.tire_error) {
                  interactive_tire_mux_error_handler(shell,tire.tire_error);
                  error_continuation(tire.tire_error);
                } else {
                  if(tire.type == 'r') {
                   continuation(tire);
                  } else if(tire.type == 'i') {
                    if(tire.rest.length > 0) {
                      if(dump_node_to_string(tire.rest[0]) == 'route') {
                        // this is a routing update and can hidden. This mux doesn't have a routing table.
                      } else if(dump_node_to_string(tire.rest[0]) == 'unroute') {
                        // this is a routing update and can hidden. This mux doesn't have a routing table.
                      } else {
                        continuation(tire);
                      }
                    }
                  } else {
                    error_continuation(tire);
                  }
                }
              },"t",tag,cookie,command,''+rest);
            } catch(e) {
              alert('wah wah wah wah...');
              continuation(e);
            }
          }
        }
      }
    }
  }
}

// an alternate evaluator for RPC systems
function evaluator_raw_tire() {
  this.language = "TIRE";
  this.evaluator = function(expression,document_context,shell,continuation) {
    jsshell_raw_tire_driver(expression,shell,
      function(v) {
        if(v.type=='r') {
          dump_node_array(v.rest,document_context,false,shell,continuation);
        } else if(v.type == 'i') {
          // skip past the 'output' or 'error' tag. we should probably consume that and
          // use it to apply a style
          v.rest.shift();
          dump_node_array(v.rest,document_context,false,shell,continuation);
        } else {
          continuation(libreplete.generate_representation(v,document_context));
        }
      },function(v) {
        continuation(document.createTextNode('TIRE error: '));
        continuation(libreplete.generate_representation(e,document_context));
      }
    );
  };

  this.eval_to_js = function(string,shell,continuation) {
    jsshell_raw_tire_evaluator(string,shell,function(v) {
      if(v.type == 'r') {
        continuation(dump_node_to_js(v.rest[0],shell));
      } else {
        false;
      }
    },function(v) {
      if(v.type == 'e') {
        continuation(dump_node_to_js(v.rest[0],shell));
      }
      shell.output_append_target.appendChild(document.createTextNode('raw tire eval_to_js error\n'));
      shell.output_append_target.appendChild(libreplete.generate_representation(v,document));
    });
  };

  this.eval_to_string = function(string,shell,continuation) {
    jsshell_raw_tire_evaluator(string,shell,function(v) {
      if(v.type == 'r') {
        continuation(dump_node_to_string(v.rest[0]));
      } else {
        false;
      }
    },function(v) {
      if(v.type == 'e') {
        continuation(dump_node_to_string(v.rest[0]));
      }
      shell.output_append_target.appendChild(document.createTextNode('raw tire eval_to_string error in ' + string + '\n'));
      shell.output_append_target.appendChild(libreplete.generate_representation(v,document));
    });
  };
};

the_raw_tire_evaluator=new evaluator_raw_tire();
hybrid_evaluators[")"]=the_raw_tire_evaluator;

// the jsshell class
function jsshell() {
  this.clear = function() {
    var o=document.getElementById(output_element);
    var i;
    for(i=0;i<o.childNodes.length;) {
      o.removeChild(o.childNodes[0]);
    }
  };

  // temporarily assigned the value of the context menu
  this.context_menu_div = false;

  this.jsprefs = jsprefs;

  this.input_element = false;
  //this.output_element = false;

  this.inner_evaluator = jsshell_current_evaluator;

  this.dump_node_to_string = dump_node_to_string;

  this.generate_representation = libreplete.generate_representation;
  this.string_escape=scheme_stringify;

  // if this is true, shell appends results
  this.jspref_shell_appends=true;
  // a shortcut anchor evaluates the 'form' of the anchor and echos if jspref_shortcuts_echo is
  // non-false
  this.jspref_shortcuts_echo=true;

  this.prompt = function(sh) {
    if(this.inner_evaluator.prompt) {
      return this.inner_evaluator.prompt(sh);
    } else {
      return document.createTextNode('>');
    }
  }

  var shell=this;
  //var string_evaluator;
  var output_window;
  var output_document;
  var output_append_target;

  // this may be a bit RPC specific
  shell.private_eval_cookie=default_eval_cookie;

  shell.get_eval_cookie=function() {
    return shell.private_eval_cookie;
  }

  shell.set_eval_cookie=function(cookie) {
    shell.private_eval_cookie=cookie;
    shell.update_prompt();
  }

  shell.session_id=Math.floor(1024*1024*1024*4*Math.random()) + "";

  // make a subshell for use in widgets
  function _subshell(render_func) {
    this.jsprefs = shell.jsprefs;

    this.session_id=shell.session_id;
    this.show_keybindings=shell.show_keybindings;
    this.inner_evaluator=shell.inner_evaluator;
    this.string_evaluator=function(expression,should_echo) {
      shell.inner_evaluator.evaluator(expression,document,this,function(result) {
        if(result) {
          render_func(result);
        }
      });
    };
    this.file_evaluator=function(expression,filename) {
      shell.inner_evaluator.loader(expression,filename,document,this,function(result) {
        if(result) {
          render_func(result);
        }
      });
    };
    this.string_escape=shell.string_escape;
    this.generate_representation=shell.generate_representation;

    this.prompt=shell.prompt;

    this.subshell = function (render_func) {
      return new _subshell(render_func);
    }

    // this may be a bit RPC specific
    // subshells inherit the parent's cookie of the moment
    this.private_eval_cookie=shell.private_eval_cookie;
    this.tire_mux=shell.tire_mux;

    this.get_eval_cookie=function() {
      return this.private_eval_cookie;
    }


    return this;
  }
  this.subshell = function (render_func) {
    return new _subshell(render_func);
  }

  function echo(target,value,document_context,reissue_handler) {
    var newp = document_context.createElement('div');

    var i;
    newp.className='iwta_echo';

    var s_array=value.split("\n");
    var j;
    for(j=0;j<s_array.length;j++) {
      newp.appendChild(document_context.createTextNode(tty_string_escape(s_array[j])));
      break;
    }
    for(j=1;j<s_array.length;j++) {
      newp.appendChild(document_context.createElement('br'));
      newp.appendChild(document_context.createTextNode(tty_string_escape(s_array[j])));
    }
    // we only need these buttons in the case that we don't
    // support inline editing.
    //if(!can_read_selection) {
    //  if(reissue_handler) {
    //    var reissue = document_context.createElement('a');
    //    reissue.innerHTML='<b>R</b>';
    //    newp.appendChild(document_context.createTextNode('   ...'));
    //    newp.appendChild(reissue);
    //    reissue.onclick=function() { reissue_handler(); };
    //  };
    //}
    if(!can_delete_selection) {
      var nuke = document_context.createElement('a');
      nuke.innerHTML='<b>X</b>';
      newp.appendChild(document_context.createTextNode('...'));
      newp.appendChild(nuke);
      nuke.onclick=function() { target.parentNode.removeChild(target); };
    }
    target.appendChild(newp);
    return newp;
  }

  var combined_evaluator = function(expression,filename,should_echo) {
    if(shell.jspref_shell_appends) {
      var ofbuf = output_window;
      var newdiv;

      // lets capture some parameters of ofbuf before we transform the document
      var old_scrollTop=ofbuf.scrollTop;
      var old_clientHeight=ofbuf.clientHeight;
      var old_scrollHeight=ofbuf.scrollHeight;

      newdiv = ofbuf.document.createElement('div');
      newdiv.className='iwta_expr';

      shell.output_append_target.appendChild(newdiv);
     
      if(!filename) {
        if(should_echo) {
          jsshell_readline.history_append(shell.input_element,expression,shell.session_id);
        }
      }

      if(should_echo) {
        if(filename) {
          echo(newdiv,"Loading " + filename,ofbuf.document,function() {
            return "CAN'T REDO A LOAD RIGHT NOW";
          });
        } else {
          echo(newdiv,expression,ofbuf.document,function() {
            string_evaluator(expression,should_echo);
          });
        }
      }

      var next=function(result) {
        if(result) {
          newdiv.appendChild(result);

          var ofbuf=shell.output_append_target;
          if(jsprefs.getpref('smart_scroll')) {
            var thresh=100;
            var dangle_thresh=50;
            var divisor=4; // we are willing to scroll the ofbuf
                           // by up to 1/divisor if the top of
                           // result is withing thresh pixels of
                           // being visible
            var mknotice=function(message) {
              var notice_div=document.createElement('div');
              notice_div.className='floating notification';
              notice_div.style.left='20px';
              notice_div.style.top='20px';
              if(message) {
                notice_div.appendChild(document.createTextNode('more... click or ctrl-] to warp: ' + message));
              } else {
                notice_div.appendChild(document.createTextNode('more... click or ctrl-] to warp...'));
              }
              document.body.appendChild(notice_div);
              setTimeout(function() {
                if(notice_div.parentNode) {
                  notice_div.parentNode.removeChild(notice_div);
                }
              },2000);
              // if you click this div then you can visit the start
              // of the result
              notice_div.addEventListener('click',function(e) {
                e.stopPropagation();
                //ofbuf.scrollTop = ofbuf.scrollHeight;
                ofbuf.scrollTop = newdiv.offsetTop;
                if(notice_div.parentNode) {
                  notice_div.parentNode.removeChild(notice_div);
                }
              },false);
              notice_div.addEventListener('touchstart',function(e) {
                e.stopPropagation();
                //ofbuf.scrollTop = ofbuf.scrollHeight;
                ofbuf.scrollTop = newdiv.offsetTop;
                if(notice_div.parentNode) {
                  notice_div.parentNode.removeChild(notice_div);
                }
              },false);
            };
            if(ofbuf.clientHeight && ofbuf.scrollHeight) {
              var dangling_ofbuf=(ofbuf.scrollHeight-ofbuf.scrollTop)-ofbuf.clientHeight;
              if(newdiv.clientHeight) {
                if(
                    dangling_ofbuf < ( newdiv.clientHeight + dangle_thresh)
                  ) {
                  // then let's say that the user was scrolled to the bottom
                  // and we'll keep scrolling as we go
                  if((ofbuf.scrollHeight != ofbuf.clientHeight) && newdiv.clientHeight > ( ofbuf.scrollHeight / divisor)) {
                    // but we're already scrolled and this result is just too much
                    //ofbuf.scrollTop += Math.floor(ofbuf.scrollheight / divisor);
                    mknotice('1');
                  } else {
                    ofbuf.scrollTop = ofbuf.scrollHeight;
                  } 
                } else {
                  mknotice();
                }
              } else {
                // the newdiv doesn't have a height yet
                if(
                    dangling_ofbuf < dangle_thresh
                  ) {
                  // then let's say we're scrolled to the bottom
                  ofbuf.scrollTop = ofbuf.scrollHeight;
                } else {
                  mknotice();
                }
              } 
            } else {
              ofbuf.scrollTop = ofbuf.scrollHeight;
            }
          }
        }
      };
      if(filename && shell.inner_evaluator.loader) {
        shell.inner_evaluator.loader(expression,filename,ofbuf.document,shell,next);
      } else {
        shell.inner_evaluator.evaluator(expression,ofbuf.document,shell,next);
      }
    } else {
      //alert('output_element: ' + output_element);
      newdiv=document.getElementById(output_element);

      if(filename && shell.inner_evaluator.loader) {
        shell.inner_evaluator.loader(expression,filename,document,shell,function(result) {
          if(result) {
            // we're not appending
            var x;
            while(newdiv.childNodes.length>0) {
              newdiv.removeChild(newdiv.childNodes[0]);
            }
            newdiv.appendChild(result);
          }
        });
      } else {
        shell.inner_evaluator.evaluator(expression,document,shell,function(result) {
          if(result) {
            // we're not appending
            var x;
            while(newdiv.childNodes.length>0) {
              newdiv.removeChild(newdiv.childNodes[0]);
            }
            newdiv.appendChild(result);
          }
        });
      }
    }

    if(should_echo) {
      shell.update_prompt();
    }
  };
  var string_evaluator = function(expression,should_echo) {
    combined_evaluator(expression,false,should_echo);
  };
  var file_evaluator = function(expression,filename,should_echo) {
    combined_evaluator(expression,filename||"unknown",should_echo);
  };

  this.string_evaluator=string_evaluator;
  this.file_evaluator=file_evaluator;

  this.update_prompt = function () {
    this.prompt_elem.removeChild(this.prompt_elem.childNodes[0]);
    if(this.prompt) {
      if(typeof(this.prompt) == "string") {
        this.prompt_elem.appendChild(document.createTextNode(this.prompt));
      } else if (typeof(this.prompt) == "function") {
        this.prompt_elem.appendChild(this.prompt(this)); 
      } else if(this.prompt.childNodes) {
        this.prompt_elem.appendChild(this.prompt);
      } else {
        this.promt_elem.appendChild(document.createTextNode('Prompt>'));
      }
    } else {
      this.prompt_elem.appendChild(document.createTextNode('Prompt>'));
    }
  }

  this.evaluator = function(object) {

    // at this point, 'this' is the shell
    if(object.value != "") {

      this.string_evaluator(object.value,true);

      // clear the entry widget
      object.value='';

      // update the prompt
      this.update_prompt();
    }
  }

  this.absorb_input_element = function(element) {
    if(typeof(element) == "string") {
      element=document.getElementById(element);
    }

    // i'm not sure that false is the right phase
    element.addEventListener('keypress', function(e) {
      return jsshell_text_keypress_handler(shell,this,e);
    },false);
    element.addEventListener('keydown',function(e) {
      return jsshell_text_keydown_handler(shell,this,e);
    },false);

    // autocomplete is probably not the right thing for a repl input
    element.setAttribute('autocomplete','off');
    element.setAttribute('autocapitalize','off'); // for Mobile Safari
    element.setAttribute('autocorrect','off'); // for Mobile Safari

    var drophandler=function(e) {
      var ep,f;
      if(e.originalEvent) {
        ep=e.originalEvent;
      } else {
        ep=e;
      }
      if(ep.dataTransfer && ep.dataTransfer.files) {
        readfile(ep.dataTransfer.files[0],function(text) {
          //shell.string_evaluator(text,true);
          shell.file_evaluator(text,ep.dataTransfer.files[0].name,true);
        });
      } else {
        if(!ep.dataTransfer) {
          alert('no dataTransfer');
        } else {
          alert('another problem');
        }
      }
      e.stopPropagation();
      e.preventDefault();
      diddrop=true;
      dragclear();
    };

    var oldvalue=false;
    var diddrop=false;
    var dragstart=false;

    var dragenter_count=0;
    var dragleave_count=0;

    // drag events:
    //// on the source element
    // dragstart
    // drag -- fired periodically
    // dragend
    //// on an arbitrary element
    // dragenter -- should be once, but may be more if we change the contents of the element
    // dragover -- periodically
    // dragleave -- should be once, but may be more if we change the contents of the element
    // drop
    // on the target element

    var foodiv=false;

    var quirksmode_findpos=function(obj) {
      var curleft = curtop = 0;
      if (obj.offsetParent) {	
        do {
          curleft += obj.offsetLeft;
          curtop += obj.offsetTop;	
        } while (obj = obj.offsetParent);
      }
      return [curleft,curtop];
    };
 
    var dragsetup=function() {
      if(!dragstart) {  
        var offset=quirksmode_findpos(element);
        foodiv=document.createElement('div');
        foodiv.appendChild(document.createTextNode('drop a file to evaluate its contents'));
        foodiv.className='floating notification';
        foodiv.style.left=offset[0]+'px';
        foodiv.style.top=(offset[1] - 200) +'px';
        document.body.appendChild(foodiv);
        dragstart=true;
        element.className='iwt iwt_droptarget';
        //oldvalue=element.value;
        //element.value='drop your file here to evaluate its contents';
      }
    };

    var dragclear=function() {
      if(foodiv) {
        foodiv.parentNode.removeChild(foodiv);
        foodiv=false;
      }
      dragstart=false;
      element.className='iwt';
    };

    element.addEventListener('dragenter',function(e) {
      dragsetup();
      dragenter_count++;
      e.stopPropagation();
      e.preventDefault();
    },false);

    element.addEventListener('dragleave',function(e) {
      dragclear();
      e.stopPropagation();
      e.preventDefault();
    },false);

    element.addEventListener('dragover',function(e) {
      e.stopPropagation();
      e.preventDefault();
    },false);

   
    // FireFox 3.6, at least on Linux,
    //  appears to have a default behavior
    // for drop on inputs that pastes the URL into
    // the input. That's why this now registers on
    // the way down rather than the way back 
    element.addEventListener('drop',drophandler,true);

    // filter stray keystrokes out of the input area
    if(element.parentNode) {
      element.parentNode.addEventListener('keypress',function(e) {
        e.stopPropagation();
      },false);
    }

    // create a blank history and cursor
    jsshell_readline.cursor_setup(element);

    element.jsshell_input_mode=null;

    // grab the focus
    element.focus();

    shell.input_element=element;

    window.addEventListener('paste',function(e) {
      if(element.selectionStart) {
        // a completion was possibly rejected
        element.value=element.value.substring(0,element.selectionStart);
        completions_rejected++;
      }
      if(e.clipboardData) {
        var t=e.clipboardData.getData('Text');
        if(t) {
          var s=t.split("\n");
          if(s.length > 1) { 
            jsshell_switch_to_multiline(shell,shell.input_element);
          }
        }
        //document.body.appendChild(libreplete.generate_representation(e,document));
        //alert(e.clipboardData.getData('Text'),document);
      }
      //alert('window paste: ' + e.target + ' data: ' + e.data + ' -- value: ' + e.value + ' clipboard: ' + e.clipboardData);
    },false);
  }
  this.absorb_prompt_element = function(element) {
    if(typeof(element) == "string") {
      element=document.getElementById(element);
    }

    var prompt_elem=document.createElement('span');
    element.appendChild(prompt_elem);
    this.prompt_elem=prompt_elem;

    if(this.prompt) {
      if(typeof(this.prompt) == "string") {
        prompt_elem.appendChild(document.createTextNode(this.prompt));
      } else if (typeof(this.prompt) == "function") {
        prompt_elem.appendChild(this.prompt(this)); 
      } else if(this.prompt.childNodes) {
        prompt_elem.appendChild(this.prompt);
      } else {
        promt_elem.appendChild(document.createTextNode('Prompt>'));
      }
    } else {
      prompt_elem.appendChild(document.createTextNode('Prompt>'));
    }
  }

  var current_bogo = false;
  var current_div = false;
  var output_element = false;
  var output_document = false;
  var output_append_target = false;

  var absorb_output_element_common = function(output_append_target) {

    shell.output_append_target = output_append_target;

    // see if we need to provide a reissue button
    if(output_window && output_window.getSelection) {
        var s=output_window.getSelection();
        can_read_selection=true;
        if(s && s.deleteFromDocument)
          can_delete_selection=true;
    }

    //var context_menu_div=document.createElement('div');
    shell.context_menu_div=document.createElement('div');
    shell.context_menu_div.style.position='absolute';
    shell.context_menu_div.className='floating context_menu';
    
    var menu_selected = false;
    var last_styled=false;
    var menu_adder = function(name,fn) {
      var n=document.createElement('span');
      var t=document.createTextNode(name);
      n.appendChild(t);
      n.addEventListener('mouseover',function(e) {
        n.style.backgroundColor='green';
        last_styled=n;
        n.addEventListener('mouseout',function(e) {
          menu_selected=false;
          n.style.backgroundColor='';
          last_styled=false;
        },false);
        menu_selected=fn;
      },false);
      shell.context_menu_div.appendChild(n);
      shell.context_menu_div.appendChild(document.createElement('br'));
    }
    menu_adder('evaluate',function(e) {
      var sel=portable.selection_to_string(output_window);  
      if(sel) {
        string_evaluator(sel,true);
      }
    });
    menu_adder('edit',function(e) {
      var sel=portable.selection_to_string(output_window);  
      if(sel) {
        jsshell_switch_to_multiline_with_data(shell,shell.input_element,sel);
      }
    });
    menu_adder('copy',function(e) {
      portable.copy();
      return true;
    });
    menu_adder('paste',function(e) {
      if(e.target) {
        portable.paste(e.target);
      }
      return true;
    });
    menu_adder('delete',function(e) {
      portable.delete_selection(output_window); 
    });
    menu_adder('spin off',function(e) {
      portable.spinoff_selection(output_window);
    });
    //menu_adder('menu diag',function(e) {
    //  alert('diag: e.pageY: ' + e.pageY + ' -- e.clientY: ' + e.clientY + ' menu height: ' + shell.context_menu_div.clientHeight + ' menu top: ' + shell.context_menu_div.offsetTop + ' -- scrollTop: ' + document.body.scrollTop);
    //});
    menu_adder('shim',function(e) {
      var d=document.createElement('div');
      d.style.height='200px';
      d.style.width='200px';
      d.style.overflow='scroll';
      portable.shim_selection(window,d);
    });
    menu_adder('shout',function(e) {
      with_fullscreen_div(function(d) {
        var x,s;
        s=window.getSelection();
        for(x=0;x<s.rangeCount;x++) {
          d.appendChild(document.createTextNode(s.getRangeAt(x)+''));
        }
      });
    });
    menu_adder('select column',function(e) {
      var selector=function(table,column_index) {
        var ranges=[];
        var cursor=table;
        var i=0,j=0;
        for(i=0;i<cursor.childNodes.length;i++) {
          if(cursor.childNodes[i].nodeName == 'TBODY') {
            return selector(cursor.childNodes[i],column_index); 
          } else if(cursor.childNodes[i].nodeName == 'TR') {
            var column_count=0;
            for(j=0;j<cursor.childNodes[i].childNodes.length;j++) {
              if(cursor.childNodes[i].childNodes[j].nodeName == 'TD') {
                var td=cursor.childNodes[i].childNodes[j];
                var cell_width;
                cell_width=td.getAttribute('colspan');
                if(cell_width) {
                  cell_width=parseInt(cell_width);
                }
                if(!cell_width) { 
                  cell_width=1;
                }
                if((column_count <= column_index) && (column_count+cell_width > column_index)) {
                  //td.style.backgroundColor='blue';
                  var thisrange;
                  thisrange=document.createRange();
                  ranges[ranges.length]=thisrange;
                  thisrange.selectNodeContents(td);
                  window.getSelection().addRange(thisrange);
                }
                column_count+=cell_width;
              }
            }
          }
        }
        return true;
      };
      var find_table = function(tr,column_index) {
        var cursor=tr;
        while(cursor) {
          if(cursor.nodeName == 'TABLE') {
            return selector(cursor,column_index);
          }
          cursor=cursor.parentNode;
        }
      }
      var find_tr = function(td) {
        var cursor=td;
        while(cursor) {
          if(cursor.nodeName == 'TR') {
            var i=0;
            var column_count=0;
            for(i=0;i<cursor.childNodes.length;i++) {
              if( cursor.childNodes[i].nodeName == 'TD') {
                var cell_width;
                cell_width=cursor.childNodes[i].getAttribute('colspan');
                if(cell_width) {
                  cell_width=parseInt(cell_width);
                }
                if(!cell_width) { 
                  cell_width=1;
                }
     
                if( cursor.childNodes[i] == td) {
                  return find_table(cursor,column_count);
                }
                column_count+=cell_width;
              }
            }
          }
          cursor=cursor.parentNode;
        }
      }
      var find_td = function(root) {
        var cursor=root;
        while(cursor) {
          if(cursor.nodeName == 'TD') {
            return find_tr(cursor);
          }
          cursor=cursor.parentNode;
        }
      }
      find_td(e.target);
    });

    shell.add_to_context_menu=menu_adder;

    var lastheight=false;
    var lastwidth=false;
    var adjust_menu=function(d,downe) {
      var page_client_delta_y,page_client_delta_x;
      page_client_delta_x=downe.pageX - downe.clientX;
      page_client_delta_y=downe.pageY - downe.clientY;
      if(d.clientHeight) {
        // the context menu is mapped
        lastheight=d.clientHeight;
        lastwidth=d.clientWidth;
        if(d.offsetTop + d.clientHeight > window.innerHeight + page_client_delta_y) {
          // this case is only true if the menu doesn't fit
          // on-screen and the height of the context
          // menu has changed since last being mapped or has
          // never been mapped.
          d.style.top=page_client_delta_y + (window.innerHeight-lastheight) + -5 + 'px';
        }
        if(d.offsetLeft + d.clientWidth > window.innerWidth + page_client_delta_x) {
          // this case is only true if the menu doesn't fit
          // on-screen and the height of the context
          // menu has changed since last being mapped or has
          // never been mapped.
          d.style.left=page_client_delta_x + (window.innerWidth-lastwidth) + -5 + 'px';
        }
      } else {
        if(lastheight && ((downe.clientY + lastheight)) > window.innerHeight) {
          d.style.top=page_client_delta_y + (window.innerHeight-lastheight) + -5 + 'px';
        } else {
          d.style.top=downe.pageY+'px';
        }
        if(lastwidth && ((downe.clientX + lastwidth)) > window.innerWidth) {
          d.style.left=page_client_delta_x + (window.innerWidth-lastwidth) + -5 + 'px';
        } else {
          d.style.left=downe.pageX+'px';
        }
      }
    }
    var context_handler=function(downe) {
      var d=shell.context_menu_div;

      adjust_menu(d,downe); // use the last known height,width

      document.body.appendChild(d);
      // update to reflect possibly new height,width
      adjust_menu(d,downe);

      {
        var newf=function(e) {
          document.body.removeChild(d);
          output_window.removeEventListener('mouseup',newf,true);
          if(last_styled) {
            last_styled.style.backgroundColor='';
          }
          if(menu_selected) {
            var m=menu_selected;
            menu_selected=false;
            m(downe);
          }
        }  
        output_window.addEventListener('mouseup',newf,true);
      }
    }
   
    // support executing the selection
    // via meta-click or meta-j or meta-enter
    // Firefox 2.0/Mac doesn't want a clicker on Window.document
    // without an event listener on the Window
    // click clears the selection in Safari
    var mousedown_handler=function(e) {
      // mozilla on linux allows the ctrl key
      // safari eats ctrl click
      // command-click on the apple in firefox
      // this handler fires in firefox on mac;
      if(portable.meta_key_p(e) || (e.button == 2)) {
        e.stopPropagation();
        e.preventDefault();
        /* firefox has a context handler event */
        /* prevent context menu event from firing */
        var cfunc=function(e) {
          e.stopPropagation();
          e.preventDefault();
          output_window.removeEventListener('contextmenu',cfunc,true);
        };
        output_window.addEventListener('contextmenu',cfunc,true);
        context_handler(e);
      }
    }

/*
    // firefox on linux has contextmenu
    output_window.addEventListener('contextmenu',function(e) {
      e.stopPropagation();
      e.preventDefault();
      context_handler(e);
    },false);
*/

    // firefox on mac uses this path
    output_window.addEventListener('mousedown',mousedown_handler,false);
    //output_document.addEventListener('mousedown',mousedown_handler,false);

    var keybindings=new Object();
    shell.keybindings=keybindings;

    shell.show_keybindings = function() {
      var wikistr="";
      var k;
      for(k in keybindings) {
        wikistr+="* meta-" + k + " -- " + keybindings[k].desc + "\n";
      }
      return gin_up_wiki(wikistr,'Jsshell',shell);
    };

    var make_standard_keybinding=function(name,description,handler) {
      keybindings[name]=function(e) {
        handler(e);
        e.stopPropagation();
        e.preventDefault();
        return false;
      };
      keybindings[name].desc=description;
    };
    make_standard_keybinding('k','show keybindings',function(e) {
      var n=output_document.createElement('div');
      for(j in keybindings) {
        var t=output_document.createTextNode('key: ' + j + ' -- ' + keybindings[j].desc);
        n.appendChild(t);
        n.appendChild(output_document.createElement('br'));
      }
      output_append_target.appendChild(n);
    });
    make_standard_keybinding('s','show the source of the selection',function(e) {
      alert(portable.selection_to_source(output_window));      
    });
    make_standard_keybinding('j','evaluate the selection',function(e) {
      var sel=portable.selection_to_string(output_window);
      string_evaluator(sel,true);
    });
    make_standard_keybinding('h','delete the selection',function(e) {
      portable.delete_selection(output_window); 
    });
    make_standard_keybinding('p','hide the selection',function(e) {
      portable.hide_selection(output_window); 
    });
    make_standard_keybinding('n','spin the selection off into a new window',function(e) {
      portable.spinoff_selection(output_window);
    });
    make_standard_keybinding('backspace','delete the selection',function(e) {
      if(!in_window_editing) {
        portable.delete_selection(output_window); 
      }
    });
    make_standard_keybinding('[','scroll to mark or beginning of buffer',function(e) {
      if(shell.mark) {
        shell.output_append_target.scrollTop=shell.mark;
        shell.mark=false;
      } else {
        shell.output_append_target.scrollTop=0;
      }
    });
    make_standard_keybinding(']','scroll to the end of the buffer',function(e) {
      var ofbuf=shell.output_append_target;
      shell.mark=ofbuf.scrollTop;
      if(ofbuf) {
        ofbuf.scrollTop = ofbuf.scrollHeight;
      }
    });
    make_standard_keybinding('c','cancel outstanding commands',function(e) {
      if(shell.inner_evaluator.cancel) {
        // this is a cancel without a handle to a specific outstanding command
        shell.inner_evaluator.cancel(false,shell);
      }
    });
    

    var window_keypress_handler=function(e) {
      if(shell.input_element == e.target) {
        alert('a keypress event on the main input element leaked through');
      } else {
      var str;
      if(portable.meta_key_p(e)) {
        str=portable.keypress_to_string(e);
        var handler=keybindings[str];
        if (handler) {
          return handler(e);
        } else if(str=='i') {
          // this should toggle the immediate mode entry form
          var bogoinput;
          var bogotext;
          bogoinput=document.createElement('input');
          if(e.shiftKey) {
            alert('shift key is on');
            bogoinput.type='textarea';
          } else { 
            bogoinput.type='text';
          }
          bogoinput.className='iwti';
          bogotext=document.createTextNode('eval:'); 
          framediv.appendChild(bogotext);
          framediv.appendChild(bogoinput);
          shell.absorb_input_element(bogoinput);
          e.stopPropagation();
        } else if(str=='g') {
          //alert('should show you the tree of this event: ' + e.target + ' -- ' + e.target.innerHTML +  ' -- and -- ' + e.currentTarget);
          e.stopPropagation();
          return false;
        } else {
          if(e.charCode == 27) {
            var h=keybindings['['];
            if(h) {
              h(e);
            }
          } else if(e.charCode == 29) {
            var h=keybindings[']'];
            if(h) {
              h(e);
            }
          }// else {
           // alert('unknown meta key: ' + str + ' -- ' + e.keyCode + ' -- ' + e.charCode);
          //}
        }
      } else {
        var w;
        w=window.getSelection();
        if(w.isCollapsed) {
          if(e.target && e.target.type=='text') {
            // this event was sent to a text element. leave it alone
          } else if(e.target && e.target.type=='textarea') {
            // this event was sent to a textarea element. leave it alone
          } else {
            // send it to the default input element
            if(shell.input_element) { 
              shell.input_element.focus();
              portable.bounce_keypress(e,shell.input_element);
            }
          }
        } else {
          var handler;
          if(e.keyCode == e.DOM_VK_BACKSPACE || e.keyCode == 8) {
            handler=keybindings['backspace'];
            if(handler) {
              return handler(e);
            }
          } else {
                //alert('another keypress with a non-null selection');
          }
        }
      }
    }
    };

    output_window.addEventListener('keypress',window_keypress_handler,false);
 
    init_widgets(shell);

    boot_tire(shell);

    true;
  };

  //  this is when we output to a div in the parent document
  this.absorb_output_element_div = function(output_div_name) {
    output_element=output_div_name;
    output_window=window;
    output_document=document;
    if(typeof(output_div_name) == 'string') {
      output_append_target=document.getElementById(output_div_name);
    } else {
      output_append_target=output_div_name;
    }

    absorb_output_element_common(output_append_target);

  }

  // absorb_output_element should only be called for appending shells
  this.absorb_output_element_frame = function(output_div_name) {
    // refreshing. 'this' is the shell
    // do this onload
    var newframe=document.createElement('iframe');
    newframe.name='outputframe2';
    newframe.id='outputframe2';

    var framediv=document.getElementById(output_div_name);
    //alert(framediv + ' ' + newframe);
    framediv.appendChild(newframe);
    //alert(window.frames.outputframe2);
    output_window=window.frames.outputframe2;
    newframe.style.width='100%';
    newframe.style.height='100%';

    var iframe_inline=true;

    output_append_target=output_window.document.body;

    output_document=output_window.document;

    absorb_output_element_common(output_append_target);

    newframe.type="application/xhtml+xml";
 
    iframe_inline=false; 
    if(!iframe_inline) {
      if(navigator.vendor.match(/Apple/)) {
        newframe.src='sub.html';
      } else {
        newframe.src='sub.xhtml';
      }
    }

    // is paste supported anywhere but safari?
    output_document.onPaste = function(args) {
      alert('paste!');
    }
    output_document.addEventListener('paste',function(e) {
      alert('paste foo!');
    },false);
    if(newframe.contentWindow && newframe.contentWindow.document) {
      newframe.contentWindow.document.addEventListener('paste',function(e) {
        alert('got a paste! should write a useful handler!');
      },false);
    }

    if(in_window_editing) {
      output_window;
    }
  }
 
  shell.defaults=function() {
    var d=document.createElement('div');
    d.innerHTML="Defaults editor<hr/>";
    shell.echo(d);
    return d;
  }  

}

function create_jsshell_with_input_and_output(input_name,output_name) {
  var shell = new jsshell();
  if(input_name) {
    shell.absorb_input_element(input_name);
  }
  //shell.absorb_output_element_frame(output_name);
  shell.absorb_output_element_div(output_name);
  return shell;
}

function create_jsshell_with_input_and_output_and_prompt(input_name,output_name,prompt_name) {
  var shell = new jsshell();
  if(input_name) {
    shell.absorb_input_element(input_name);
  }
  if(prompt_name) {
    shell.absorb_prompt_element(prompt_name);
  }
  //shell.absorb_output_element_frame(output_name);
  shell.absorb_output_element_div(output_name);
  return shell;
}

function create_jsshell_with_output(output_name) {
  var shell = new jsshell(false);
  shell.absorb_output_element_frame(output_name);
  return shell;
}

function jsshell_switch_to_multiline_with_data(shell,input_element,data) {
  if(input_element && input_element.parentNode) { 
    var parent=input_element.parentNode;
    var old=input_element;
    var tempdiv=document.createElement('div');
    var ta=document.createElement('textarea');
    tempdiv.className='iwia';
    if(data) {
      if(old.value!="") {
        ta.value=old.value+' '+data;
      }
    } else {
      if(old.value!="") {
        ta.value=old.value+"\n";
      }
    }
    var cleanup=function() {
      parent.removeChild(tempdiv);
      parent.appendChild(old);          
      if(input_element && input_element.jsshell_on_resize) {
        input_element.jsshell_on_resize();
      } 
    };
    ta.setAttribute('rows','8');
    ta.setAttribute('cols','65');
    ta.className='iwt';
    var button=document.createElement('input');
    button.type='button';
    button.value='submit';
    button.addEventListener('click',function(e) {
      shell.evaluator(ta);
      cleanup();
    },false);
    var cancel_button=document.createElement('input');
    cancel_button.type='button';
    cancel_button.value='cancel';
    cancel_button.addEventListener('click',function(e) {
      cleanup();
    },false);
    parent.removeChild(old);
    old.value='';
    tempdiv.appendChild(ta);
    tempdiv.appendChild(document.createElement('br'));
    tempdiv.appendChild(button);
    tempdiv.appendChild(cancel_button);
    parent.appendChild(tempdiv);
    ta.focus();
    text_area_resizer(ta);
    if(input_element && input_element.jsshell_on_resize) {
      input_element.jsshell_on_resize();
    }
  }
}

function jsshell_switch_to_multiline(shell,input_element) {
  return jsshell_switch_to_multiline_with_data(shell,input_element,false);
}

var stroke_count=0;
var last_timeout=false;

// for key events that do not add a character to the input area
function jsshell_text_keydown_handler(shell,object,event) {

  function up_handler() {
    if(portable.shift_key_p(event)) {
      jsshell_readline.cursor_up_uniq(object);
    } else {
      jsshell_readline.cursor_up(object);
    }
    event.stopPropagation();
    event.preventDefault();
  }

  function down_handler() {
    if(portable.shift_key_p(event)) {
      jsshell_readline.cursor_down_uniq(object);
    } else {
      jsshell_readline.cursor_down(object);
    }
    event.stopPropagation();
    event.preventDefault();
  }

  function tab_handler(shell,object) {
    if(object.value != '') {
      get_completion(shell,object,object.value,function(v) {
        if(v) {
          object.value=v;
        }
      });
    }
    event.preventDefault();
    event.stopPropagation();
  }

  if(event.DOM_VK_DOWN) {
  } else {
    event.DOM_VK_TAB=9;
    event.DOM_VK_LEFT=37;
    event.DOM_VK_UP=38;
    event.DOM_VK_RIGHT=39;
    event.DOM_VK_DOWN=40;
  }

  if(event.keyCode == event.DOM_VK_UP) {
    stroke_count++;
    up_handler();
  } else if(event.keyCode == event.DOM_VK_DOWN) {
    stroke_count++;
    down_handler();
  } else if(event.keyCode == event.DOM_VK_TAB) {
    stroke_count++;
    tab_handler(shell,object);
  }
};

function get_completion_from_history(object,exp) {
  var sres;
  sres=jsshell_readline.prefix_search(object,exp);
  if(sres[0] != "") {
    return sres[0];
  } else {
    return false;
  }
};

function get_completion(shell,object,exp,cont) {
  if(exp != '') {
    if(shell.inner_evaluator.tabcompleter) { 
      shell.inner_evaluator.tabcompleter(exp,shell,function (string) {
        if(string) {
          cont(string);
        } else {
          cont(get_completion_from_history(object,exp));
        }
      });
    } else {
      cont(get_completion_from_history(object,exp));
    }
  } else {
    cont(false);
  }
}

var completions_attempted=0;   // number of calls for completions
var completions_failed=0;      // number of those calls that returned invalid completions
var completions_successful=0;  // number of those calls that returned successful completions
var completions_toolate=0;     // number of those calls that did not complete before the
                               // user typed another character
var completions_rejected=0;    // number of completions that the user typed through
var completions_accepted=0;    // number of completions that the user appeared to accept

function jsshell_text_keypress_handler(shell,object,event) {
  var k;
  
  stroke_count++;

  if(last_timeout) {
    clearTimeout(last_timeout);
    last_timeout=false;
  }

  if(event.charCode) {
    k=event.charCode;
  } else
    k=event.keyCode;

  function enter_handler() {
    // clear the cursor
    jsshell_readline.cursor_clear(object);
    object.jsshell_input_mode=null;
    // evaluate
    shell.evaluator(object);
    event.stopPropagation();
    event.preventDefault();
  }

  function manage_key_timers(object) {
    // capture the current stroke count
    var now=stroke_count;
    var delay=jsprefs.getpref('autocomplete_delay');
    var minsize=jsprefs.getpref('autocomplete_minsize');
    if(delay) {
      if(minsize && object.value.length >= minsize) {
        last_timeout=setTimeout(function() {
          if(stroke_count == now) {
            if(object.selectionStart == object.value.length) {
              completions_attempted++;
              get_completion(shell,object,object.value,function(v) {
                if(typeof(v) == "string") {
                  // get completion is async. check again
                  if(stroke_count == now) {
                    if(v.length > 0) { 
                      var old_value=object.value;
                      object.value=v;
                      object.selectionStart=old_value.length;
                      object.selectionEnd=v.length;
                      completions_successful++;
                    } else {
                      completions_failed++;
                    }
                  } else {
                    completions_toolate++;
                  }                
                } else {
                  completions_failed++;
                }
              });
            }
          }
        },delay);
      }
    }
  }

  function fallthrough_handler() {
    // this key is editing the value. scrub the cursor
    if(object.jsshell_input_mode == 'readline-search') {
      jsshell_readline.expression_search(object,e);
      event.stopPropagation();
      event.preventDefault();
      return false;
    } else {
      var now=stroke_count;   
      if(object.selectionStart && ( object.selectionStart != object.selectionEnd ) ) {
        completions_rejected++;
      }
      //jsshell_readline.cursor_clear(object);
      manage_key_timers(object);
      return true;
    }
  }

  if(event.DOM_VK_DOWN) { 
    if(event.keyCode == event.DOM_VK_ENTER) {
      return enter_handler();
    } else if(event.keyCode == event.DOM_VK_RETURN) {
      return enter_handler();
    } else if(event.keyCode == event.DOM_VK_LEFT) {
      if(object.selectionStart) {
        if(object.selectionStart != object.selectionEnd && typeof(object.selectionStart) == 'number' && object.selectionStart>0) {
          // a completion was possibly rejected
          object.value=object.value.substring(0,object.selectionStart);
          completions_rejected++;
          //object.selectionEnd=object.selectionStart=object.selectionStart-1;
          event.stopPropagation();
          event.preventDefault();
          return false;
        }
      } else {
        return fallthrough_handler();
      }
    } else if(event.keyCode == event.DOM_VK_RIGHT) {
      if(object.selectionStart) {
        if(object.selectionStart == object.selectionEnd && typeof(object.selectionStart) == 'number') {
          // a completion was possibly accepted
          object.selectionEnd=object.selectionStart=object.selectionStart+1;
          completions_accepted++;
          event.stopPropagation();
          event.preventDefault();
          return false;
        }
      } else {
        return fallthrough_handler();
      }
    } else if(portable.meta_key_p(event)) {
      if(k == 117) { /* ctrl-U */
        //jsshell_readline.cursor_clear(object);
        object.value='';
        event.stopPropagation();
        event.preventDefault();
        return false;
      } else if(k == 109) { /* ctrl-M */
        jsshell_switch_to_multiline(shell,event.target);
        event.stopPropagation();
        event.preventDefault();
        return false;
      } else if(k == 114) { /* ctrl-R */
	var sres,x,vals;
        sres=jsshell_readline.search(object);
        vals=sres[1];
        object.value=sres[0];
        for(x=0;x<vals.length;x++) {
          shell.output_append_target.appendChild(document.createTextNode(vals[x]));
          shell.output_append_target.appendChild(document.createElement('br'));
        }
        event.stopPropagation();
        event.preventDefault();
      } else if(k == 99) { /* ctrl-C */
        var h=keybindings['c'];
        if(h) {
          h(event);
        }
      } else if(k == 27) { /* ctrl-[ */
        var h=shell.keybindings['['];
        if(h) {
          h(event);
        }
      } else if(k == 29) { /* ctrl-] */
        var h=shell.keybindings[']'];
        if(h) {
          h(event);
        }
      } else {
        // continue propagation
        return true;
      }
    } else {
      return fallthrough_handler();
    }
  } else {
    // WebKit (Safari) uses this path
    if(event.keyCode == 13) {
      // desktop Safari uses this path (at least on the Mac)
      return enter_handler();
    } else if(event.keyCode == 10) {
      // iPhone Safari uses this path
      return enter_handler();
    } else if(event.keyCode == 18 && event.ctrlKey == true) { // ctrl R
      alert('ctrlR -- searching through ');
      object.jsshell_input_mode='readline-search';
      event.stopPropagation();
      return false;
    } else if(event.keyCode == 3 && event.ctrlKey == true) { // ctrl C
      alert('ctrlC -- clearing ');
      object.jsshell_input_mode=null;
      event.stopPropagation();
      return false;
    } else if(event.keyCode == 21 && event.ctrlKey == true) { // ctrl U
      object.value='';
      object.jsshell_input_mode=null;
      event.stopPropagation();
      return false;
    } else if(event.keyCode == 29 && event.ctrlKey == true) { // ctrl ]
      var h=keybindings['['];
      if(h) {
        h(event);
      }
      return false;
    } else if(event.keyCode == 27 && event.ctrlKey == true) { // ctrl [
      var h=keybindings['['];
      if(h) {
        h(event);
      }
      return false;
    } else if(event.altKey || event.ctrlKey || event.metaKey) {
      // this is a meta key we don't understand
      return true;
    } else {
      //alert('foo: ' + event.keyCode);
      return fallthrough_handler();
    }
  }
}
// some convenience functions for calling inside the shell
function examine(target,object,document_context) {

  var newp = document_context.createElement('div');

  newp.innerHTML='This value is a ' + typeof(object);
  target.appendChild(newp);

  var newt = document_context.createElement('table');
  var newtb = document_context.createElement('tbody');

  // peek at all the keys
  for(i in object) {
    var newtr = document_context.createElement('tr');
    var nametd = document_context.createElement('td');
    var valtd = document_context.createElement('td');
    var nametext = document_context.createTextNode(i);

    nametd.insertBefore(nametext,null);
    try {
      valtd.insertBefore(libreplete.generate_representation(object[i],document_context),null);
    } catch(e) {
      valtd.insertBefore(document_context.createTextNode('protected'),null);
    }
    newtr.insertBefore(nametd,null);
    newtr.insertBefore(valtd,null);
    newtb.insertBefore(newtr,null);
  }
  newt.insertBefore(newtb,null);

  newp.appendChild(newt);
}

function mw_json_get_text(json_obj) {
  if(json_obj.query && json_obj.query.pages) {
    for(var k in json_obj.query.pages) {
      for(var j in json_obj.query.pages[k].revisions) {
        for(var i in json_obj.query.pages[k].revisions[j]) {
          if (i=='*') {
            return json_obj.query.pages[k].revisions[j][i];
          }
        }
      }
    }
  }
  return "object wasn't fully baked!";
}

var sleazy_global_continuation=false;

function jsshell_mw_callback(json_obj) {
  var d=document.createElement('div');
  d.innerHTML=wiki2html(mw_json_get_text(json_obj));
  sleazy_global_continuation(d);
}

function with_fullscreen_div(next) {
  var scroll_cursor=0;
  var d=document.createElement('div');
  var d_content=document.createElement('div');
  var padding=false;

  var viewport_d=document.createElement('div');

  // viewport_d should be the only element 
  // with a style related to actual window sizes in pixels
  /* viewport_d should consume exactly 100% of the visible viewport */
  viewport_d.style.position='absolute';

  var rightsize=function() {
    viewport_d.style.width=window.innerWidth+'px';
    //viewport_d.style.height=window.innerHeight+'px';
    viewport_d.style.height='100%';
    viewport_d.style.left='0px';
    // this works for me on the iPad. The iPad
    // maintains that scrollTop is 0 for documents
    // that are taller than the screen. iPhoneOS
    // maintains a virtual viewport that may be
    // larger than the real screen
    viewport_d.style.top=(window.pageYOffset)+'px';
  };

  var goaway=function() {
    if(viewport_d.parentNode) {
      window.removeEventListener('resize',bogus_resizer,false);
      viewport_d.parentNode.removeChild(viewport_d);
    }
  };
  var bogus_resizer=function(e) {
    rightsize();  
  };
  window.addEventListener('resize',bogus_resizer,false);
  rightsize();

  var viewport_background_d=document.createElement('div');
  viewport_background_d.className='fullscreen_overscan';
  viewport_d.appendChild(viewport_background_d);


  // this is the div onto which we will install touch events.
  // since we're in a full-screen mode, this div should
  // be as large as possible
  var touch_d=document.createElement('div');
  touch_d.position='relative';
  touch_d.style.height='100%';
  touch_d.style.width='100%';
 
  d.className='fullscreen_container';
  d.style.position='relative';
  d.style.height='75%';
  d.style.minHeight='75%';
  d.style.margin='12.5%';

  d_content.className='fullscreen_content';

  next(d_content);

  d.appendChild(d_content);
  viewport_d.addEventListener('click',function(e) {
    goaway();
  },false);

  touch_d.addEventListener('touchstart',function(e_start) {
    var left_slide=0;
    var t0_start=e_start.changedTouches.item(0);
    var t0_start_x=t0_start.clientX;
    var t0_start_y=t0_start.clientY;

    var move=function(e_move) {
      var t0_move=e_move.changedTouches.item(0);
      var t0_move_x=t0_move.clientX;
      var t0_move_y=t0_move.clientY;
      if(Math.abs(t0_move_x - t0_start_x) > Math.abs(t0_move_y - t0_start_y)) {
        left_slide=(t0_move_x - t0_start_x);
        if(t0_move_x < t0_start_x) {
          // we may be advancing to the next page
          if(padding) {
            if(left_slide < -150) {
              left_slide = -150;
            }
          }
        } else {
          // we may be returning to the previous page
          if(d_content.scrollTop == 0) {
            if(left_slide > 150) {
              left_slide = 150;
            }
          }
        }
        d.style.left=left_slide+'px';
        e_move.preventDefault();
      } else {
        // the event is predominantly in Y
      }
    };
    var end=function(e_end) {
      d.removeEventListener('touchmove',move,true);
      d.removeEventListener('touchend',end,true);
      var t0_end=e_end.changedTouches.item(0);
      var t0_end_x=t0_end.clientX;
      var t0_end_y=t0_end.clientY;
      d.style.left='0px';
      if(Math.abs(t0_end_x - t0_start_x) > Math.abs(t0_end_y - t0_start_y)) {
        // then this is predominantly an X event
        if(t0_end_x < t0_start_x) {
          // we advance to the next page
          if(!padding) {
            scroll_cursor+=d.clientHeight;
            if( ( scroll_cursor + d_content.clientHeight ) > d_content.scrollHeight) { 
              padding=document.createElement('div');
              padding.style.height=d.clientHeight+'px';
              padding.appendChild(document.createTextNode('this area intentionally left blank'));
              d_content.appendChild(padding);
            }
            d_content.scrollTop=scroll_cursor;
          }
        } else {
          // we return to the previous page
          var h=d.clientHeight;
          if(scroll_cursor>=h) {
            scroll_cursor-=h;
          } else {
            scroll_cursor=0;
          }
          if(padding) {
            padding.parentNode.removeChild(padding);
            padding=false;
          }
          d_content.scrollTop=scroll_cursor;
        }
        e_end.preventDefault();
        //alert('d.clientWidth: ' + d.clientWidth + ' -- under_d.clientWidth: ' + under_d.clientWidth + ' under_under_d.clientWidth: ' + under_under_d.clientWidth);
      } else {
        // the event is predominantly in Y
        if(Math.abs(t0_end_y - t0_start_y) > 50) {
          goaway();
        }
      }
    };
    touch_d.addEventListener('touchend',end,true);
    touch_d.addEventListener('touchmove',move,true);
  },false);

  //d.style.minHeight='100px';
  //d.style.minWidth='100px';

  //under_under_d.appendChild(under_d);
  touch_d.appendChild(d);
  viewport_d.appendChild(touch_d);
  document.body.appendChild(viewport_d);
}

function showprefs(shell) {
  var k;
  var d=document.createElement('div');
  d.appendChild(document.createTextNode('jsshell preferences:'));
  d.appendChild(document.createElement('br'));
  for(k in jsprefs.prefs) {
    if(jsprefs.prefs[k].jspref) {
    (function(k) { 
      var f;
      f=document.createElement('input');
      f.type='text';
      f.value=jsprefs.prefs[k].value;
      f.addEventListener('change',function(e) {
        if(jsprefs.prefs[k].type == 'number') {
          var v=parseInt(f.value);
          if(jsprefs.prefs[k].predicate && jsprefs.prefs[k].predicate(v)) {
            shell.string_evaluator('!jsprefs.setpref_string("' + k + '",' + v + ');',true)
          }
          f.value=jsprefs.prefs[k].value;
        } else if(jsprefs.prefs[k].type == 'boolean') {
          var v,set=false;
          if (f.value == 'true') { 
            set=true;
            v=true;
          } else if(f.value == 'false') {
            set=true;
            v=false;
          }
          if(set) {
            if(jsprefs.prefs[k].predicate && jsprefs.prefs[k].predicate(v)) {
              shell.string_evaluator('!jsprefs.setpref("' + k + '",' + v + ');',true)
            } else if(!jsprefs.prefs[k].predicate) { 
              shell.string_evaluator('!jsprefs.setpref("' + k + '",' + v + ');',true)
            }
          }
          f.value=jsprefs.prefs[k].value;
        } else {
          if(jsprefs.prefs[k].predicate && jsprefs.prefs[k].predicate(f.value)) {
            shell.string_evaluator('!jsprefs.setpref_string("' + k + '","' + f.value + '");',true)
          } else if(!jsprefs.prefs[k].predicate) {
            shell.string_evaluator('!jsprefs.setpref_string("' + k + '","' + f.value + '");',true)
          }
          f.value=jsprefs.prefs[k].value;
        }
      },false);
      d.appendChild(document.createTextNode(k+' -- ' + jsprefs.prefs[k].desc));
      d.appendChild(document.createElement('br'));
      d.appendChild(document.createTextNode('current value: '));
      d.appendChild(f);
      d.appendChild(document.createElement('br'));
      d.appendChild(document.createTextNode('default value: ' + jsprefs.prefs[k].default_value));
      d.appendChild(document.createElement('br'));
    })(k);
    }
  }

  return libreplete.paint(d);
};

var make_bookmarklet = function(shell,name) {
  var s=document.createElement('span');
  with_xml_tag(shell,"<a/>",function(v) {
    v.setAttribute('href',
      'javascript:{var foo=document.createElement("script"); foo.src=\"' + 'http://ub235.super.org:51533/jsshell/jsshell-bookmarklet.js' + '\"; document.body.appendChild(foo);void(0);}'
    );
    var t=document.createTextNode(name);
    v.appendChild(t);
    s.appendChild(v);
  });
  return libreplete.paint(s);
};

var jsshell_help_search = function(querystring,continuation) {
 jspref_wikis['Jsshell'].search(querystring,function(r) {
   continuation(r);
 });
};

var jsshell_help_index = function(continuation) {
 jspref_wikis['Jsshell'].index(function(r) {
   continuation(r);
 });
};

function dav_parse_index(x) {

  var i;
  var r = new Array;

  var responses=x.getElementsByTagName('response');
  if(responses.length == 0) {
    responses=x.getElementsByTagNameNS('DAV:','D:response');
  }
  if(responses.length == 0) {
    responses=x.getElementsByTagName('D:response');
  }

  for(i=0;i<responses.length;i++) {
    try {
      var n=responses[i].getElementsByTagName('href');
      if(n.length == 0) {
        n=responses[i].getElementsByTagNameNS('DAV:','href');
      }
      if(n.length == 0) {
        n=responses[i].getElementsByTagName('D:href');
      }
      var t=node_text(n[0]);
      //if(t && t.match(/^wiki\/.*[.]wiki$/)) {
      if(t && t.match(/^[^\/]*\/.*$/)) {
        r.push(t.replace(/^[^\/]*\//,''));
      }
    } catch(e) {
      //alert('ehh? ' + e);
    }
  }

  return r;
};

function dav_fetch_index(url,success,failure) {
  var concluded=false;
  if(0) {
  } else {
    if(window.XMLHttpRequest) {
      var req=new XMLHttpRequest();
      if(!req.DONE) {
        req.DONE=4;
      }
      req.open("GET",url + "/pseudodav.xml",true);
      req.withCredentials = "true";
      req.onreadystatechange = function(e) {
        if(req.readyState == req.DONE && req.status == 200) {
          if(req.responseXML) {
            if(!concluded) {
              concluded=true;
              if(success) {
                success(dav_parse_index(req.responseXML));
              }
            }
          } else {
            if(!concluded) {
              concluded=true;
              if(failure) failure();
            }
          }
        } else if((req.readyState == req.DONE) && (req.status == 404)) {
          if(!concluded) {
            concluded=true;
            if(failure) failure();
          }
        }
      }
      req.send('');
    }
    false;
  }
};

// construct the DOM scaffolding for a widget
libreplete.make_dom_widget = function(document_context,shell) {
  var i,node;
  var d=document_context.createElement('div');
  var decorations=document_context.createElement('div');
  //var d=document_context.createElement('span');

  //  var widget_finalize = false;

  var widget=d;

  d.appendChild(decorations);

  //d.className='widget';

  decorations.className='bad_widget_decoration';
  
  d.addEventListener('mouseover',function(e) {
    decorations.className='widget_decoration';
  },false);
  d.addEventListener('mouseout',function(e) {
    decorations.className='bad_widget_decoration';
  },false);

  widget.finalize=false;

  function make_button(label) {
    var b=document.createElement('input');
    b.type='button';
    b.value=label;
    //b.className='widget_decoration';
    return b;
  }

  if(shell.widget_window) {
    alert('make_widget: widget_window');
  } else {
    var float_button=make_button('float');
    float_button.onclick=function(e) {
      d.style.float='right';
    };
    decorations.appendChild(float_button);

    var close_button=make_button('close');
    var orig_delete_target=function() { widget.parentNode.removeChild(widget); };
    var delete_target=orig_delete_target;
    var widget_closer = function() {
      if(widget.finalize) {
        widget.finalize(); 
      }
      if(delete_target) {
        delete_target();
      }
      widget=false;
      delete_target=false;
    };
    widget.close=widget_closer;
    close_button.onclick=function(e) {
      widget_closer();
    };
    var orig_parent;

    decorations.appendChild(close_button);
    var unstick_button=make_button('unstick');
    var orig_unstick_function=function(e) {
      var width=widget.clientWidth;
      var height=widget.clientHeight;
      var l=document.createElement('div');
      l.style.backgroundColor='blue';
      document.body.appendChild(l);
      orig_parent=widget.parentNode;
      orig_parent.removeChild(widget);
      portable.make_subwindow_with_element_and_dimensions(document,l,widget,'iwta_body',width,height);
      delete_target=function() { l.parentNode.removeChild(l); };
      unstick_function=function(e) {
        var new_window=window.open('/jsshell/sub.xhtml','','');

        new_window.onload=function() {
          if(delete_target) {
            delete_target();
          }
          delete_target=function() { new_window.close(); };
          new_window.addEventListener('unload',function(e) {
              alert('widget window finalize!');
              //if(widget.finalize) widget.finalize();
	  }, false);

          new_window.document.body.appendChild(d);
        }

      }
      restick_function=function(e) {
        var p;
        unstick_function=orig_unstick_function;
        p=widget.parentNode;
        p.removeChild(widget);
        orig_parent.appendChild(widget); 
        document.body.removeChild(p.parentNode);
        restick_function=orig_restick_function;
        delete_target=orig_delete_target;
      }
    };
    var unstick_function=orig_unstick_function;
    unstick_button.onclick=function(e) {
      unstick_function(e);
    }
    decorations.appendChild(unstick_button);
    var restick_button=make_button('restick');
    var orig_restick_function=function(e) {
      alert('already stuck!'); 
    }
    var restick_function=orig_restick_function;
    restick_button.onclick=function(e) {
      restick_function(e);
    }
    decorations.appendChild(restick_button);
   d.appendChild(document_context.createElement('br'));
  }

  return d;
};

libreplete.make_widget=function(document_context,shell,script_body,key,hidden) {
  var bogo_widget=new Object();

  bogo_widget.shell=shell;

  // Why must these widgets know about Scheme?
  if(key) {
    bogo_widget.wrapper=function (s) {
      return '(xhtml-widget-eval (quote ' + s + ') ' + key + ')';
    };
    bogo_widget.formals='](xhtml-widget-formals ' + key + ')';
    bogo_widget.finalize=function() {
      shell.inner_evaluator.evaluator('](xhtml-widget-delete! ' + key + ')',
                                      document_context,
                                      shell,
                                      function(v) { return false; }
      );
    }
  }

  return bogo_widget;
};
