#!/bin/bash
# -- find proverb
PROVERB=`which proverb`
if [ -z "$PROVERB" ] ; then
  PROVERB=`dirname "$0"`/../proverb
fi
# -- exec
exec "${PROVERB}" --noweb --load "$0" --eval '(promagnon-main)' -- "$0" "$@"
# -- or die
exit 1
!#

; We have now switched gears into Scheme

;
; Keeper-aliver for proverb
;

(define make-socket (lambda (host portno)
  (define s (socket AF_INET SOCK_STREAM 0))
  (define ip (if host
    (if (string? host)
      (car (hostent:addr-list (gethostbyname host)))
      host
    )
    INADDR_ANY
  ))
  (setsockopt s (vector-ref (getproto "IP") 2) SO_REUSEADDR 1)
  (let helper ()
    (catch #t (lambda ()
      (bind s AF_INET ip portno)
    ) (lambda args
      (critmsg "Spinning...\n")
      (sleep 1)
      (helper)
    ))
  )
  (listen s 1)
  (cons s (vector-ref (getsockname s) 2))
))

(define promagnon-main (lambda ()
  (define promagnon-options (let helper ((cursor (cdr (command-line))) (racc '()))
    (if (null? cursor)
      (reverse racc)
      (if (string=? (car cursor) "--")
        (reverse racc)
        (helper (cdr cursor) (cons (car cursor) racc))
      )
    )
  ))
  (define proverb-options (let helper ((cursor (cdr (command-line))))
    (if (null? cursor)
      '()
      (if (string=? (car cursor) "--")
        (cdr cursor)
        (helper (cdr cursor))
      )
    )
  ))
  (define port #f)
  (define host #f)
  (define uid #f)
  (define gid #f)
  (define user #f)
  (define accepted-options `(
    ("--port" ,(lambda (p) (set! port (string->number p))) "use specified port for HTTP")
    ("--host" ,(lambda (h) (set! host h)) "listen on the interface with the address of HOST")
    ("--local" ,(lambda () (set! host INADDR_LOOPBACK)) "listen on the loopback interface")
    ("--uid" ,(lambda (p) (set! uid (string->number p))) "run as user UID")
    ("--gid" ,(lambda (p) (set! gid (string->number p))) "run as group GID")
    ("--user" ,(lambda (p) (set! user p)) "run as user USERNAME")
  ))

  (if (null? proverb-options)
    (begin
      (display "Usage: promagnon [promagnon-option] ... -- <path-to-proverb> [proverb-option] ...\n" (current-error-port))
      (display "----\n" (current-error-port))
      (display "promagnon accepts the following options:\n" (current-error-port))
      (for-each (lambda (v)
        (display (car v) (current-error-port))
        (display " -- " (current-error-port))
        (display (caddr v) (current-error-port))
        (newline (current-error-port)) 
      ) accepted-options)
      (exit 1)
    )
  )

  (let helper ((cursor promagnon-options))
    (if (null? cursor)
      #t
      (cond
        ((assoc (car cursor) accepted-options) => (lambda (a)
          (define n (cadr (assoc 'arity (procedure-properties (cadr a)))))
          (apply (cadr a) (cdr (list-head cursor (+ n 1))))
          (helper (list-tail cursor (+ n 1)))
        ))
        (else #t)
      )
    )
  )

  (let ()
    (define s (make-socket host (or port 0)))
 
    (if port
      (format (current-error-port) "promagnon launching proverb on port ~a\n" port)
      (format (current-error-port) "promagnon launching proverb\n")
    )
    (if user
      (let ((n (getpwnam user)))
        (if n 
          (begin
            (if (not uid) (set! uid (vector-ref n 2)))
            (if (not gid) (set! gid (vector-ref n 3)))
          )
          (begin
            (format (current-error-port) "promagnon couldn't resolve user ~a\n" user)
            (exit 1)
          )
        )
      )
    )
    (if gid
      (begin
        (setgid gid)
        (setegid gid)
      )
    )
    (if uid
      (begin
        (setuid uid)
        (seteuid uid)
      )
    )
    (let helper ()
      (define argv (append
        proverb-options
        `("--http-socket" ,(number->string (fileno (car s))))
      ))
      (system (string-join (map (lambda (v) (string-append "\"" v "\"")) argv) " "))
      (sleep 2)
      (format (current-error-port) "promagnon relaunching proverb\n")
      (helper)
    )
  )
))

; relaunch is run inside of Proverb
(define relaunch (lambda ()
  (define s (current-http-server))
  (define p (getpid))
  (define proverb-dir (dirname (car (command-line))))
  (let ()
    (define l (http-server-get-listener s))
    ((snatch-responder) 'r "OK")
    (concurrent-apply (lambda (v)
      (define f (fileno v))
      (let ()
        (define argv (cons
          (string-append proverb-dir "/proverb")
          (append
            (cdr (command-line))
           `("--http-socket" ,(number->string f))
          )
        ))
        ; set this fd to no close on exec
        (fcntl f F_SETFD (logand (fcntl f F_GETFD) (lognot FD_CLOEXEC)))
        ; map the child's stdout to stderr
        (dup2 2 1)
        ; kill our dad
        (kill p 9)
        ; and away we go
        (apply execlp (car argv) argv)
        (critmsg "Uh Oh. system returned\n" f)
      )
    ) (list l)) 
  )
))
