#!/usr/bin/env python

#
# A sample TIRE server for Python.
#

import string, sys, subprocess, lispio, re, os, thread,code, time, rlcompleter
import StringIO

# We're going to use stdio for TIRE. Hide it from the process.
moved_stdin_fd=os.dup(0);
moved_stdout_fd=os.dup(1);
moved_stdin=os.fdopen(moved_stdin_fd,'r');
moved_stdout=os.fdopen(moved_stdout_fd,'w');
os.dup2(2,1);
new_stdin=open("/dev/null");
new_stdin_fd=new_stdin.fileno();
os.dup2(new_stdin_fd,0);

unquoter = re.compile('(^")|("$)')

def send_tire_response(channel,action,tag,*value):
  x=0
  y=len(value)
  acc=[]
  while x < y:
    acc.append(str(value[x]))
    acc.append(" ")
    x=x+1
  channel.writelines(["(" + action + " " + tag + " " + "".join(acc) + " )"])
  channel.flush()

def make_responder(channel,tag):
  return lambda value: send_tire_response(channel,"r",tag,value)

def make_raw_responder(channel,tag):
  return lambda *value: send_tire_response(channel,"i",tag,*value)

interpreter=code.InteractiveInterpreter()

# Jam some bindings into the interpreter
interpreter.runcode(code.compile_command("def dribble(val):\n  raw_responder('dribble',val)\n"))
interpreter.runcode(code.compile_command("def make_raw(your_str):\n  return '(#.(make-passval xml-raw (lambda (v p u) (display #b#7#xml-raw p))) #b#'+str(len(your_str))+'#' + your_str + ')'\n"))

cookie_dict = dict([])

interpreter_methods_dict = dict([])

def seval(responder,raw_responder,form):
  printval=[]
  bogo_printval=[]
  soft_stdout=StringIO.StringIO()
  soft_stderr=StringIO.StringIO()
  def printer(v):
    printval.append(v)
  def bogo_printer(v):
    bogo_printval.append(v)
  old_hook = sys.displayhook
  thiscode=code.compile_command(form)
  sys.displayhook = printer
  old_stdout=sys.stdout
  old_stderr=sys.stderr
  sys.stdout=soft_stdout
  sys.stderr=soft_stderr
  interpreter.locals['raw_responder']=raw_responder
  if thiscode:
    try:
      result=interpreter.runcode(thiscode)
    except:
      result="You lose in eval"
  else:
    result="You lose in compile"
  sys.stdout=old_stdout
  sys.stderr=old_stderr
  sys.displayhook = old_hook
  iprint=soft_stdout.getvalue()
  if not iprint == "" :
    raw_responder('output' , '"' + iprint + '"') 
  iprint=soft_stderr.getvalue()
  if not iprint == "" :
    raw_responder('error' , '"' + iprint + '"') 
  if len(printval) > 0:
    responder('"' + str(printval[0]) + '"')
  elif type(result) == type(""):
    responder('"' + result + '"')
  else:
    responder(result)

def make_threaded_method(method):
  def newfunc(responder,raw_responder,form):
    thread.start_new_thread(method,(responder,raw_responder,form))
  return newfunc

def bogo_eval(responder,raw_responder,form):
  generate_local_cookie(responder)

interpreter_methods_dict['seval']=make_threaded_method(seval)

interpreter_methods_dict['eval']=bogo_eval

my_cookie="python-"+str(os.getpid())

mux_methods_dict = dict([])

def enum(responder,raw_responder,form):
  responder(my_cookie)

mux_methods_dict['enum']=enum

def generate_local_cookie(responder):
  responder(my_cookie)

cookie_dict['local']=interpreter_methods_dict
cookie_dict[my_cookie]=interpreter_methods_dict
cookie_dict['-']=mux_methods_dict

completer = rlcompleter.Completer()

#
# This tab-completes in the wrong context
# and also does a terrible job on arguments
# and syntax
def tire_tab_complete(str,responder):
  responder('"' + completer.complete(str,0) + '"')

#
# Run a thread to manage a tag queue
#
def run_tag_thread(tag,entry,in_channel,out_channel,finalize):
  queue_lock=entry[0]
  queue=entry[1]
  sleep_lock=entry[2]
  queue_lock.acquire()
  while len(queue) > 0:
    readval=queue.pop(0)
    queue_lock.release()

    if readval[0] == "t":
      cookie=readval[2]
      command=readval[3]
      methods=cookie_dict[cookie]
      method=methods[command]
      if method:
        if len(readval) > 4:
          form=readval[4]
        else:
          form=0
	thread.start_new_thread(method,(make_responder(out_channel,tag),make_raw_responder(out_channel,tag),form))
      elif command == "complete":
        tire_tab_complete(readval[4],make_responder(out_channel,tag))
    elif readval[0] == "i":
      send_tire_response(out_channel,"e",tag,"i-form-not-supported")
    elif readval[0] == "r":
      send_tire_response(out_channel,"e",tag,"r-form-not-supported")
    elif readval[0] == "e":
      send_tire_response(out_channel,"e",tag,"e-form-not-supported")
    else:
      print "TAG THREAD: ERROR!"

    queue_lock.acquire()

  queue_lock.release()
  finalize()

def run_mux(in_channel,out_channel):
  tag_dict = dict([])
  while 1:
    readval=lispio.getSexp(in_channel)
    tag=readval[1]
    if not tag in tag_dict:
      # a lock for the queue, the queue, a lock for sleep, a lock for good measure, a thread object
      tag_dict[tag]=[thread.allocate_lock(),[],thread.allocate_lock(),thread.allocate_lock(),False]
    (tag_dict[tag])[0].acquire()
    (tag_dict[tag])[1].append(readval)
    (tag_dict[tag])[0].release()
    (tag_dict[tag])[3].acquire()
    if (tag_dict[tag])[2].locked():
      (tag_dict[tag])[2].release()
    (tag_dict[tag])[3].release()
    if not (tag_dict[tag])[4]:
      def closer():
        (tag_dict[tag])[4]=False;
      thread.start_new_thread(run_tag_thread,(tag,tag_dict[tag],in_channel,out_channel,closer))

run_mux(moved_stdin,moved_stdout)

sys.exit(0);
