#!/bin/bash

#
# Rover is the trivial shell interface to proverb-web's RPC service
# This exists to demonstrate the possible, not the attractive
#

# A hack to make rover into a TIRE<->HTTP bridge
if [ "$1" == "raw" ] ; then
  IS_RAW=true
  shift
else
  IS_RAW=""
fi

ADDR=$1
if [ -z "${ADDR}" ] ; then
  ADDR="file:///tmp/proverb-${USER}.html"
fi
COMMAND=$2

get_url_from_filename() {
  FILENAME=$1
  cat ${FILENAME} | grep '<a href=' | sed "s/<a href=[\"\']\([^\"\']*\)[\"\'].*>/\1/"
}

# Were we handed a file url with a redirection
if echo "${ADDR}" | grep "file://" > /dev/null ; then
  echo "Rover starting with a file redirection URL" 1>&2
  FILENAME=`echo "${ADDR}" | sed 's/^file:\/\///'`
  BASE_URL=`get_url_from_filename "${FILENAME}"`
elif echo "${ADDR}" | grep '^/' > /dev/null ; then
  BASE_URL=`get_url_from_filename "${ADDR}"`
else
  if echo "${ADDR}" | grep "^http://" > /dev/null ; then
    BASE_URL="${ADDR}"
  else
    BASE_URL="http://${ADDR}/"
  fi
fi
echo "Rover using base URL ${BASE_URL}" 1>&2
XSL_URL=${BASE_URL}/sexp.xsl
RPC_URL=${BASE_URL}/rpc

XSLT=/tmp/xslt.$$
curl ${XSL_URL} 2> /dev/null > ${XSLT} || (echo "Couldn't get xslt from ${XSL_URL}" ; exit 1)
echo "Rover got the XSLT" 1>&2

SESSION=${RANDOM}
COOKIE="local"
TAG=0

process_command() {
  COMMAND_TEXT="$1"

  if [ -n "${IS_RAW}" ] ; then
    FORM="(http-rpc ${SESSION} ${COMMAND_TEXT})"
    RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    while echo "${RES}" | grep '(i ' > /dev/null 2> /dev/null ; do
      echo "AN I: ${RES}" 1>&2
      FORM="(http-wait ${SESSION})"
      RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    done 
    echo "${RES}"
  elif echo "${COMMAND_TEXT}" | grep '^@' > /dev/null 2> /dev/null ; then
    COOKIE=`echo "${COMMAND_TEXT}" | sed 's/^@//'`
  elif echo "${COMMAND_TEXT}" | grep '^)' > /dev/null 2> /dev/null ; then
    ESC_REPLY=`echo "${COMMAND_TEXT}" | sed 's/^)//'`
    FORM="(http-rpc ${SESSION} (${ESC_REPLY}))"
    RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    while echo "${RES}" | grep '(i ' > /dev/null 2> /dev/null ; do
      echo "AN I: ${RES}" 1>&2
      FORM="(http-wait ${SESSION})"
      RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    done 
    echo "${RES}"
  else
    ESC_REPLY=`echo "${COMMAND_TEXT}" | sed 's/"/\\\"/g'`
    FORM="(http-rpc ${SESSION} (t rover-${SESSION}-${TAG} ${COOKIE} seval \"${ESC_REPLY}\"))"
    TAG=$[ ${TAG} + 1 ]
    RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    while echo "${RES}" | grep '(i ' > /dev/null 2> /dev/null ; do
      echo "AN I: ${RES}" 1>&2
      FORM="(http-wait ${SESSION})"
      RES=`curl -s -d "${FORM}" ${RPC_URL} | xsltproc ${XSLT} -`
    done 
    echo "${RES}"
  fi
}

if `tty > /dev/null` ; then
  WANTS_PROMPT=yes
else
  WANTS_PROMPT=""
fi

if [ -z "${COMMAND}" ] ; then
  if [ -n "${WANTS_PROMPT}" ] ; then
    echo -n "${COOKIE}> "
  fi
  while read ; do
    process_command "${REPLY}"
    if [ -n "${WANTS_PROMPT}" ] ; then
      echo -e -n "\n${COOKIE}> "
    fi
  done  
else
  process_command "${COMMAND}"
fi

rm -f ${XSLT}
