; tire-shim.scm
;
; Code to demonstrate interoperating with a running Proverb
; from a client.
;
; See the article on tire-shim at the bottom
;

(use-modules (studio util))

(define mux (tire-mux-setup '()))
(define cookie #f)

(define setup-forwarding (lambda (url)
  (define p (tire-parse-url url))
  (if (and p (list-ref p 3))
    (let ()
      (define c ((car p)))
      (tire-mux-add-channel! mux c)
      (tire-mux-hash-cookie! mux (list-ref p 3) c)
      (set! cookie (list-ref p 3))
    )
  )
))

(define tire-wrap (lambda (name)
  (define ctr 0)
  (define gen-tag (lambda ()
    (define tag (string->symbol (format #f "~a-~a-~a" (getpid) name ctr)))
    (set! ctr (+ ctr 1))
    tag
  ))
  (lambda args
    (define responded #f)
    (define rval #f)
    (define chan (tire-mux-lookup-cookie mux cookie))
    (define tag (gen-tag))
    (define bogo-select-post (lambda (state rfds wfds xfds)
      (define retval (tire-mux-select-post state rfds wfds xfds))
      ; stop once we got our answer
      (if responded
        #f
        retval
      )
    ))
    (define rewritten-args (map (lambda (v)
      (if (and (closure? v) (not (procedure-name v)))
        `(lambda args
           ;(display "BOGO CALLBACK!\n")
           (tire-do (quote (t bogobogo ,(getpid) eval (display "FOO!"))))
         )
        v
      )
    ) args))
    (let ()
      (define form `(t ,tag ,cookie eval (,name ,@rewritten-args))) 
      (define from (make-tire-procedure-channel (lambda (form)
        (if (and form (list? form) (not (null? form)))
          (case (car form)
            ((r) (if (and (not (null? (cdr form))) (eq? (cadr form) tag))
              (set! rval (list-ref form 2))
              (error "bad-tag")
            ))
            ((i) 
              (write form (current-output-port))
            )
            ((t)
              (error "got a t!") 
            )
            ((e)
              (error "got a e!") 
            )
            (else 
              (error (string-append "bad-command: " form))
            )
          )
          (error "bad-form")
        )
      )))
      (tire-mux-add-request! mux (cons (list-ref form 1) (cons from (lambda ()
        (set! responded #t)
        ;(display "GOT A RESPONSE!\n") (display rval) (newline)
      ))))
      (channel-enqueue chan mux form)
      (run-select-loop (list (list mux tire-mux-select-pre bogo-select-post)) (lambda ()
        ;(display ".")
        (not responded)
      ))
      rval
    )
  )
))

(make-wiki-article 'tire-shim (html-escape-string "
== TIRE shimming ==

The Proverb distribution includes the file support/tire-shim.scm
that can be used to construct a call forwarding interface to a
remote Proverb.

Example:
 ./proverb-web.scm --noweb --tireport 5050
 guile -l proverb-web.scm -l support/tire-shim.scm
 guile> (setup-forwarding \"tire://localhost:5050/local\")
 guile> (define remote-getpid (tire-wrap 'getpid))
 guile> (getpid)
 123
 guile> (remote-getpid)
 guile> 122

"))
